/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: bookmarks module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * This bookmarks module defines a "bookmark manager", which essentially
 * encapsulates the bookmark database and provides bookmark-specific
 * routines to access it.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __BOOKMARKS_H__
#define __BOOKMARKS_H__

#include <stdio.h>
#include <windows.h>
#include "linked-list.h"

/* the maximum length of a bookmark title, in characters */
#define  BOOKMARK_TITLE_LEN  ( 32 )

/* the opaque BookmarkManager type */
typedef void* BookmarkManager;

/* the data relevant to a bookmark.  Everything except the 'dbid' field gets stored persistantly; the
 * dbid field is used internally as the record's primary key to update or delete the record later. */
typedef struct
{
  int record_id;    /* the record-id of the plucker record that contains this bookmark */

  int char_pos;     /* the starting character position in the record of the bookmark */
  int delta_y;      /* the delta-y from the top of the screen to the starting character position of the bookmark */

  TCHAR name[ BOOKMARK_TITLE_LEN ]; /* the title of the bookmark */

  int dbid;         /* "database id" -- the index of the record on disk.  READ-ONLY! */
} Bookmark;


/**
 * Creates a new bookmark manager.  The manager uses the named file as its persistant storage
 * location.  If this function fails, it will return NULL.
 */
BookmarkManager BookmarkManager_create( char *filename );

/**
 * Destroys the given bookmark manager and closes its associated file.  The bookmark manager should not be
 * referenced after this is called.  This function does NOT destroy any LinkedList objects that were created
 * by the getBookmarks routine.
 */
void BookmarkManager_destroy( BookmarkManager bm );

/**
 * Enumerates all bookmarks in the bookmark manager in a linked list.  It is the calling routine's responsibility
 * to destroy the list when it is finished with it.
 */
LinkedList* BookmarkManager_getBookmarks( BookmarkManager bm );

/**
 * Adds the given bookmark both to the linked list of bookmarks (in sorted order, by title), and to the persistant
 * storage of the bookmark manager.  All three entities must be valid.  A copy is made of the bookmark, so the calling
 * routine may safely destroy or reuse the bookmark structure after calling this routine.
 */
void BookmarkManager_addBookmark( BookmarkManager bm,
                                  LinkedList *bookmarks,
                                  Bookmark *bookmark );

/**
 * Deletes the given bookmark from the given linked list and from the persistant storage of the manager.  This
 * will also deallocate the bookmark, assuming it to the data area of one of the LinkedListNode objects.
 */
void BookmarkManager_deleteBookmark( BookmarkManager hm,
                                     LinkedList *bookmarks,
                                     Bookmark *bookmark );

/**
 * Updates the given bookmark object with the given values, and stores the updated object to the persistant
 * storage.  All parameters must be valid.  The 'name' must not be longer than BOOKMARK_TITLE_LEN characters.
 */
void BookmarkManager_modifyBookmark( BookmarkManager hm,
                                     Bookmark *bookmark,
                                     int record_id,
                                     int char_pos,
                                     int delta_y,
                                     TCHAR* name );

#endif /* __BOOKMARKS_H__ */
