/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: colorbox control
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include "colorbox-ctl.h"


#define COLORBOX_CONTROL_CLASS   L"VadeMecumColorBoxControl"
#define BORDER_SIZE              ( 3 )

typedef struct
{
  COLORREF color;
  BOOL selected;
} COLORBOXDATA;


static BOOL CALLBACK ColorBoxControlProc( HWND, UINT, WPARAM, LPARAM );

static UINT InitColorBoxControl( HWND );
static UINT DestroyColorBoxControl( HWND );
static UINT DrawColorBoxControl( HWND );
static UINT DrawFocus( HWND );
static UINT DoLButtonUp( HWND, UINT, UINT );
static UINT SetColorBoxColor( HWND, WPARAM );
static UINT GetColorBoxColor( HWND );
static UINT SetColorBoxSelected( HWND, WPARAM );
static UINT GetColorBoxSelected( HWND );


ATOM RegisterColorBoxControl( HINSTANCE hInst )
{
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = ColorBoxControlProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = sizeof( COLORBOXDATA* );
  wc.hInstance = hInst;
  wc.hIcon = (HICON)NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH );
  wc.hCursor = (HCURSOR)NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = COLORBOX_CONTROL_CLASS;

  return RegisterClass( &wc );
}


HWND CreateColorBox( HWND hParent,
                     HINSTANCE hInst,
                     DWORD style,
                     int x, int y,
                     int width, int height,
                     DWORD id )
{
  return CreateWindow( COLORBOX_CONTROL_CLASS, L"", style, x, y, width, height, hParent, (HMENU)id, hInst, NULL );
}



static BOOL CALLBACK ColorBoxControlProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
  switch( uMsg )
  {
    case WM_CREATE:
      return InitColorBoxControl( hWnd );

    case WM_DESTROY:
      return DestroyColorBoxControl( hWnd );

    case WM_PAINT:
      return DrawColorBoxControl( hWnd );

    case WM_SETFOCUS:
      return DrawFocus( hWnd );

    case WM_KILLFOCUS:
      return DrawFocus( hWnd );

    case WM_LBUTTONUP:
      return DoLButtonUp( hWnd, LOWORD( lParam ), HIWORD( lParam ) );

    case CLBXM_SET_COLOR:
      return SetColorBoxColor( hWnd, wParam );

    case CLBXM_GET_COLOR:
      return GetColorBoxColor( hWnd );

    case CLBXM_SELECTED:
      return SetColorBoxSelected( hWnd, wParam );

    case CLBXM_IS_SELECTED:
      return GetColorBoxSelected( hWnd );
  }

  return DefWindowProc( hWnd, uMsg, wParam, lParam );
}


static UINT InitColorBoxControl( HWND hWnd )
{
  COLORBOXDATA *data;

  data = (COLORBOXDATA*)malloc( sizeof( COLORBOXDATA ) );
  data->color = RGB( 0, 0, 0 );
  data->selected = FALSE;

  SetWindowLong( hWnd, 0, (LONG)data );

  return 0;
}


static UINT DestroyColorBoxControl( HWND hWnd )
{
  COLORBOXDATA *data;

  data = (COLORBOXDATA*)GetWindowLong( hWnd, 0 );
  free( data );

  return 0;
}


static UINT DrawColorBoxControl( HWND hWnd )
{
  PAINTSTRUCT paint;
  HDC dc;
  HBRUSH brush;
  COLORBOXDATA *data;
  RECT rect;

  data = (COLORBOXDATA*)GetWindowLong( hWnd, 0 );
  BeginPaint( hWnd, &paint );
  dc = GetDC( hWnd );
  
  GetClientRect( hWnd, &rect );
  rect.left += BORDER_SIZE;
  rect.right -= BORDER_SIZE - 1;
  rect.top += BORDER_SIZE;
  rect.bottom -= BORDER_SIZE - 1;

  brush = CreateSolidBrush( data->color );
  FillRect( dc, &rect, brush );
  DeleteObject( brush );

  DrawEdge( dc, &rect, ( data->selected ? EDGE_RAISED : EDGE_SUNKEN ), BF_RECT );

  ReleaseDC( hWnd, dc );
  EndPaint( hWnd, &paint );

  if( GetFocus() == hWnd )
  {
    DrawFocus( hWnd );
  }

  return 0;
}


static UINT DrawFocus( HWND hWnd )
{
  RECT rect;
  HDC dc;

  GetClientRect( hWnd, &rect );
  dc = GetDC( hWnd );

  DrawFocusRect( dc, &rect );

  ReleaseDC( hWnd, dc );

  return 0;
}


static UINT DoLButtonUp( HWND hWnd, UINT x, UINT y )
{
  SetFocus( hWnd );
  SendMessage( GetParent( hWnd ), WM_COMMAND,
               MAKELONG( GetDlgCtrlID( hWnd ), CLBXN_CLICKED ),
               (LONG)hWnd );
  return 0;
}


static UINT SetColorBoxColor( HWND hWnd, WPARAM wParam )
{
  COLORBOXDATA *data;

  data = (COLORBOXDATA*)GetWindowLong( hWnd, 0 );
  data->color = wParam;

  InvalidateRect( hWnd, NULL, TRUE );

  return 0;
}


static UINT GetColorBoxColor( HWND hWnd )
{
  COLORBOXDATA *data;

  data = (COLORBOXDATA*)GetWindowLong( hWnd, 0 );
  return data->color;
}


static UINT SetColorBoxSelected( HWND hWnd, WPARAM wParam )
{
  COLORBOXDATA *data;

  data = (COLORBOXDATA*)GetWindowLong( hWnd, 0 );
  data->selected = wParam;

  InvalidateRect( hWnd, NULL, TRUE );

  return 0;
}


static UINT GetColorBoxSelected( HWND hWnd )
{
  COLORBOXDATA *data;

  data = (COLORBOXDATA*)GetWindowLong( hWnd, 0 );
  return data->selected;
}


