/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: colormap control
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include "colormap-ctl.h"
#include "resrc1.h"

#define EPSILON ( 0.001 )
#define EQWEPS(a,b) (fabs(a-b) < EPSILON)

#define COLORMAP_CONTROL_CLASS   L"VadeMecumColormapControl"

#define BITMAP_HEIGHT 100
#define BITMAP_WIDTH  200
#define INTERSPACE      5
#define BAR_HEIGHT     20

typedef struct
{
  COLORREF color;
  UINT x;
  UINT y;
  UINT saturation;
  HBITMAP colormap;
  HDC hidden;   /* for drawing the fade bar */
  HBITMAP bmp;
} COLORMAP_CONTROL_DATA;


static BOOL CALLBACK ColormapControlProc( HWND, UINT, WPARAM, LPARAM );

static UINT InitColormapControl( HWND );
static UINT DestroyColormapControl( HWND );
static UINT PaintColormapControl( HWND );
static UINT ColormapTapped( HWND, LPARAM );
static UINT ColormapGetColor( HWND );
static UINT ColormapSetColor( HWND, WPARAM );

static UINT DrawFadeBar( HWND, HDC );
static UINT DrawColormapSelection( HWND, HDC );
static UINT DrawFadeBarSelection( HWND, HDC );
static void DrawLine( HDC, POINT*, COLORREF );


ATOM RegisterColormapControl( HINSTANCE hInst )
{
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = ColormapControlProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = sizeof( COLORMAP_CONTROL_DATA* );
  wc.hInstance = hInst;
  wc.hIcon = (HICON)NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH );
  wc.hCursor = (HCURSOR)NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = COLORMAP_CONTROL_CLASS;

  return RegisterClass( &wc );
}


HWND CreateColormapControl( HWND hParent,
                            HINSTANCE hInstance,
                            DWORD style,
                            int x,
                            int y,
                            DWORD id )
{
  return CreateWindow( COLORMAP_CONTROL_CLASS, L"", style,
                       x, y, COLORMAP_WIDTH, COLORMAP_HEIGHT, hParent,
                       (HMENU)id, hInstance, NULL );
}


static BOOL CALLBACK ColormapControlProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
  switch( uMsg )
  {
    case WM_CREATE:
      return InitColormapControl( hWnd );

    case WM_DESTROY:
      return DestroyColormapControl( hWnd );

    case WM_PAINT:
      return PaintColormapControl( hWnd );

    case WM_LBUTTONDOWN:
      return ColormapTapped( hWnd, lParam );

    case CCM_GET_COLOR:
      return ColormapGetColor( hWnd );

    case CCM_SET_COLOR:
      return ColormapSetColor( hWnd, wParam );
  }

  return DefWindowProc( hWnd, uMsg, wParam, lParam );
}


static UINT InitColormapControl( HWND hWnd )
{
  COLORMAP_CONTROL_DATA *data;

  data = (COLORMAP_CONTROL_DATA*)malloc( sizeof( COLORMAP_CONTROL_DATA ) );
  data->color = 0x000000;
  data->x = 0;
  data->y = BITMAP_HEIGHT;
  data->saturation = BITMAP_WIDTH-1;
  data->colormap = LoadBitmap( GetModuleHandle( NULL ), MAKEINTRESOURCE( IDB_COLOR_MAP ) );

  data->hidden = NULL;
  data->bmp = NULL;
  
  SetWindowLong( hWnd, 0, (LONG)data );

  return 0;
}


static UINT DestroyColormapControl( HWND hWnd )
{
  COLORMAP_CONTROL_DATA *data;

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );
  DeleteObject( data->colormap );

  if( data->hidden != NULL )
  {
    DeleteObject( data->bmp );
    DeleteDC( data->hidden );
  }

  free( data );

  return 0;
}


static UINT PaintColormapControl( HWND hWnd )
{
  HDC cdc;
  HDC wdc;
  HBITMAP oldbmp;
  COLORMAP_CONTROL_DATA *data;
  PAINTSTRUCT paint;

  BeginPaint( hWnd, &paint );

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );

  wdc = GetDC( hWnd );

  cdc = CreateCompatibleDC( wdc );
  oldbmp = SelectObject( cdc, data->colormap );
  BitBlt( wdc, 1, 1, BITMAP_WIDTH, BITMAP_HEIGHT, cdc, 0, 0, SRCCOPY );
  SelectObject( cdc, oldbmp );
  DeleteDC( cdc );

  DrawFadeBar( hWnd, wdc );
  DrawColormapSelection( hWnd, wdc );
  DrawFadeBarSelection( hWnd, wdc );

  ReleaseDC( hWnd, wdc );

  EndPaint( hWnd, &paint );

  return 0;
}


static UINT ColormapTapped( HWND hWnd, LPARAM lParam )
{
  int x;
  int y;
  COLORMAP_CONTROL_DATA *data;
  HDC dc;

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );

  x = LOWORD( lParam );
  y = HIWORD( lParam );

  x--;
  y--;

  /* ignore the "null" areas */
  if( x < 0 || y < 0 || x > BITMAP_WIDTH || y > BITMAP_HEIGHT + INTERSPACE + BAR_HEIGHT ) return 0;
  if( y > BITMAP_HEIGHT && y < BITMAP_HEIGHT + INTERSPACE ) return 0;

  dc = GetDC( hWnd );

  if( y < BITMAP_HEIGHT ) /* color area clicked */
  {
    DrawColormapSelection( hWnd, dc );
    DrawFadeBarSelection( hWnd, dc );

    data->x = x+1;
    data->y = y+1;

    DrawColormapSelection( hWnd, dc );
    DrawFadeBar( hWnd, dc );
    DrawFadeBarSelection( hWnd, dc );

    SendMessage( GetParent( hWnd ), WM_COMMAND, MAKELONG( GetDlgCtrlID( hWnd ), CCN_COLOR_CHANGED ), (LPARAM)hWnd );
  }
  else if( y > BITMAP_HEIGHT + INTERSPACE ) /* fade bar clicked */
  {
    DrawFadeBarSelection( hWnd, dc );
    data->saturation = x;
    DrawFadeBarSelection( hWnd, dc );

    SendMessage( GetParent( hWnd ), WM_COMMAND, MAKELONG( GetDlgCtrlID( hWnd ), CCN_COLOR_CHANGED ), (LPARAM)hWnd );
  }

  ReleaseDC( hWnd, dc );

  return 0;
}

static UINT ColormapGetColor( HWND hWnd )
{
  HDC dc;
  COLORREF color = RGB(0,0,0);
  COLORMAP_CONTROL_DATA *data;

  if( IsWindowVisible( hWnd ) )
  {
    data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );

    dc = GetDC( hWnd );
    color = GetPixel( dc, data->saturation, BITMAP_HEIGHT + INTERSPACE + BAR_HEIGHT/2 );
    ReleaseDC( hWnd, dc );
  }

  return (UINT)color;
}


static UINT ColormapSetColor( HWND hWnd, WPARAM wParam )
{
  COLORREF color = (COLORREF)wParam;
  COLORMAP_CONTROL_DATA *data;
  double r, g, b;
  double h, s, v;
  double smallest, delta;
  HDC dc;

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );

  r = GetRValue( color ) / 255.0;
  g = GetGValue( color ) / 255.0;
  b = GetBValue( color ) / 255.0;

  v = max( r, max( g, b ) );
  smallest = min( r, min( g, b ) );

  if( v == smallest ) /* shade of gray */
  {
    s = ( 1 - v ); /* saturation in [0..1], where 0 is white and 1 is pure 'v' */
    v = 0; /* v == 0 is a shade of black */
    h = 0; /* h is irrelevant here, since black is black.  h is also impossible to determine */
  }
  else
  {
    delta = v - smallest;
    s = delta / v;

    if( r == v )
    {
      h = ( g - b ) / delta;
    }
    else if( g == v )
    {
      h = 2.0 + ( b - r );
    }
    else if( b == v )
    {
      h = 4.0 + ( r - g );
    }
    h *= 60.0;
    if( h < 0.0 ) h += 360.0;
    h /= 360.0;
  }

  if( IsWindowVisible( hWnd ) )
  {
    dc = GetDC( hWnd );

    DrawColormapSelection( hWnd, dc );
    DrawFadeBarSelection( hWnd, dc );
  }

  data->x = 1 + (int)( h * ( BITMAP_WIDTH - 1 ) );
  data->y = 1 + (int)( ( 1 - v ) * ( BITMAP_HEIGHT - 1 ) );
  data->saturation = 1 + (int)( s * ( BITMAP_WIDTH - 1 ) );

  if( IsWindowVisible( hWnd ) )
  {
    DrawFadeBar( hWnd, dc );
    DrawColormapSelection( hWnd, dc );
    DrawFadeBarSelection( hWnd, dc );

    ReleaseDC( hWnd, dc );
  }

  return 0;
}


static UINT DrawFadeBar( HWND hWnd, HDC dc )
{
  int x;
  BYTE red;
  BYTE green;
  BYTE blue;
  POINT line[2];
  COLORMAP_CONTROL_DATA *data;
  RECT rect;

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );

  if( data->hidden == NULL )
  {
    data->hidden = CreateCompatibleDC( dc );
    data->bmp = CreateCompatibleBitmap( dc, BITMAP_WIDTH+2, BAR_HEIGHT+2 );

    SelectObject( data->hidden, data->bmp );
  }

  data->color = GetPixel( dc, data->x, data->y );

  rect.left = 0;
  rect.right = rect.left + BITMAP_WIDTH + 2;
  rect.top = 0;
  rect.bottom = rect.top + BAR_HEIGHT + 2;

  Rectangle( data->hidden, rect.left, rect.top, rect.right, rect.bottom );

  line[0].y = 1;
  line[1].y = line[0].y + BAR_HEIGHT;

  red = GetRValue( data->color );
  green = GetGValue( data->color );
  blue = GetBValue( data->color );

  for( x = 0; x < BITMAP_WIDTH; x++ )
  {
    BYTE r;
    BYTE g;
    BYTE b;

    r = 255 - x * ( 255 - red ) / BITMAP_WIDTH;
    g = 255 - x * ( 255 - green ) / BITMAP_WIDTH;
    b = 255 - x * ( 255 - blue ) / BITMAP_WIDTH;

    line[0].x = line[1].x = x+1;
    DrawLine( data->hidden, line, RGB( r, g, b ) );
  }

  BitBlt( dc, 0, BITMAP_HEIGHT + INTERSPACE, BITMAP_WIDTH+2, BAR_HEIGHT+2,
          data->hidden, 0, 0, SRCCOPY );

  return 0;
}


static UINT DrawColormapSelection( HWND hWnd, HDC dc )
{
  RECT r;
  COLORMAP_CONTROL_DATA *data;

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );
  r.left = data->x-2;
  r.top = data->y-2;
  r.right = data->x+2;
  r.bottom = data->y+2;

  DrawFocusRect( dc, &r );

  return 0;
}


static UINT DrawFadeBarSelection( HWND hWnd, HDC dc )
{
  RECT r;
  COLORMAP_CONTROL_DATA *data;

  data = (COLORMAP_CONTROL_DATA*)GetWindowLong( hWnd, 0 );
  r.left = data->saturation-2;
  r.top = 1 + BITMAP_HEIGHT + INTERSPACE - 1;;
  r.right = data->saturation+2;
  r.bottom = r.top + BAR_HEIGHT + 2;

  DrawFocusRect( dc, &r );

  return 0;
}


static void DrawLine( HDC dc, POINT* points, COLORREF color )
{
  HPEN oldpen;

  oldpen = SelectObject( dc, CreatePen( PS_SOLID, 1, color ) );
  Polyline( dc, points, 2 );
  DeleteObject( SelectObject( dc, oldpen ) );
}
