/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: colorpicker dialog
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <winbase.h>

#include "colormap-ctl.h"
#include "colorbox-ctl.h"
#include "colorpicker.h"
#include "utilities.h"


typedef struct
{
  COLORPICKERDATA *parms;
  HWND active_foreground;
  HWND active_background;
  HWND active_colorbox;    /* which colorbox currently has focus */
  BOOL customize_open;
} COLORPICKER_WINDOW_DATA;

#define COLORPICKER_CLASS    L"VadeMecumColorPicker"

#define ID_COLORMAP          ( 40001 )
#define ID_CUSTOMIZE         ( 40002 )
#define ID_FOREGROUND_FIRST  ( 40011 )
#define ID_BACKGROUND_FIRST  ( 40021 )

#define BORDER_SIZE          5
#define TEXT_HEIGHT          16
#define COLORBOX_HEIGHT      20
#define COLORBOX_WIDTH       25
#define BUTTON_HEIGHT        20
#define BUTTON_WIDTH         COLORMAP_WIDTH

#define CLIENT_WIDTH         ( BORDER_SIZE * 2 + COLORMAP_WIDTH )
#define CLIENT_HEIGHT        ( BORDER_SIZE * 2 + TEXT_HEIGHT*2 + COLORBOX_HEIGHT*2 + 4 + BUTTON_HEIGHT )
#define EXTRA_HEIGHT         ( 2 + COLORMAP_HEIGHT )

#define SHOW_CUSTOMIZE       L"Show >>"
#define HIDE_CUSTOMIZE       L"<< Hide"

static BOOL CALLBACK ColorPickerProc( HWND, UINT, WPARAM, LPARAM );

static UINT InitColorPicker( HWND, LPCREATESTRUCT );
static UINT DestroyColorPicker( HWND );
static UINT BackgroundColorSlotChanged( HWND, UINT );
static UINT ForegroundColorSlotChanged( HWND, UINT );
static UINT SelectedColorChanged( HWND );
static UINT AcceptSettings( HWND );
static UINT ToggleCustomize( HWND );

ATOM RegisterColorPicker( HINSTANCE hInst )
{
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = ColorPickerProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = sizeof( COLORPICKER_WINDOW_DATA* );
  wc.hInstance = hInst;
  wc.hIcon = (HICON)NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH );
  wc.hCursor = (HCURSOR)NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = COLORPICKER_CLASS;

  return RegisterClass( &wc );
}


BOOL ColorPicker( COLORPICKERDATA *data )
{
  HWND hPicker;
  MSG  msg;
  int  x;
  int  y;
  int  width;
  int  height;
  HDC  dc;

  width = GetSystemMetrics( SM_CXBORDER ) * 2 + CLIENT_WIDTH;
  height = GetSystemMetrics( SM_CYBORDER ) * 2 + GetSystemMetrics( SM_CYCAPTION ) + CLIENT_HEIGHT + EXTRA_HEIGHT;

  dc = GetDC( data->hWnd );
  x = ( GetDeviceCaps( dc, HORZRES ) - width ) / 2;
  y = ( GetDeviceCaps( dc, VERTRES ) - height ) / 3 + 5;
  ReleaseDC( data->hWnd, dc );

  height -= EXTRA_HEIGHT;

  hPicker = CreateWindowEx( WS_EX_CAPTIONOKBTN, COLORPICKER_CLASS, data->title, WS_POPUP | WS_CAPTION | WS_SYSMENU,
                            x, y, width, height, data->hWnd, (HMENU)NULL, GetModuleHandle(NULL), (LPVOID)data );

  if( hPicker == NULL ) return FALSE;

  EnableWindow( data->hWnd, FALSE );

  ShowWindow( hPicker, SW_SHOW );
  UpdateWindow( hPicker );

  while( GetMessage( &msg, NULL, 0, 0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  EnableWindow( data->hWnd, TRUE );

  return msg.wParam;
}


static BOOL CALLBACK ColorPickerProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
  switch( uMsg )
  {
    case WM_CREATE:
      return InitColorPicker( hWnd, (LPCREATESTRUCT)lParam );

    case WM_DESTROY:
      return DestroyColorPicker( hWnd );

    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case IDOK:
          AcceptSettings( hWnd );
          PostQuitMessage( TRUE );
          DestroyWindow( hWnd );
          return 0;

        case ID_COLORMAP:
          if( HIWORD( wParam ) == CCN_COLOR_CHANGED )
          {
            return SelectedColorChanged( hWnd );
          }
          break;

        case ID_CUSTOMIZE:
          return ToggleCustomize( hWnd );

        default:
          if( LOWORD( wParam ) >= ID_BACKGROUND_FIRST )
          {
            return BackgroundColorSlotChanged( hWnd, LOWORD( wParam ) );
          }
          else if( LOWORD( wParam ) >= ID_FOREGROUND_FIRST )
          {
            return ForegroundColorSlotChanged( hWnd, LOWORD( wParam ) );
          }
          break;
      }
      break;

    case WM_CLOSE:
      PostQuitMessage( FALSE );
      DestroyWindow( hWnd );
      return 0;
  }

  return DefWindowProc( hWnd, uMsg, wParam, lParam );
}


static UINT InitColorPicker( HWND hWnd, LPCREATESTRUCT create )
{
  HWND ctl;
  COLORPICKERDATA *parms;
  COLORPICKER_WINDOW_DATA *data;
  int i;
  int y;

  parms = (COLORPICKERDATA*)create->lpCreateParams;
  data = (COLORPICKER_WINDOW_DATA*)malloc( sizeof( COLORPICKER_WINDOW_DATA ) );

  data->parms = parms;
  data->active_background = NULL;
  data->active_foreground = NULL;
  data->customize_open = FALSE;

  SetWindowLong( hWnd, 0, (LONG)data );

  y = BORDER_SIZE;
  CreateStaticControl( hWnd, GetModuleHandle(NULL), 0, L"Foreground Colors:",
                       BORDER_SIZE, y, CLIENT_WIDTH, TEXT_HEIGHT,
                       SS_LEFTNOWORDWRAP | WS_VISIBLE );

  y += TEXT_HEIGHT;
  for( i = 0; i < 8; i++ )
  {
    ctl = CreateColorBox( hWnd, GetModuleHandle(NULL), WS_CHILD | WS_VISIBLE,
                          BORDER_SIZE + i * COLORBOX_WIDTH, y,
                          COLORBOX_WIDTH, COLORBOX_HEIGHT, ID_FOREGROUND_FIRST+i );
    SendMessage( ctl, CLBXM_SET_COLOR, parms->fgcolors[i], 0 );

    if( parms->foreground == parms->fgcolors[i] )
    {
      if( parms->selected_foreground )
      {
        SetFocus( ctl );
        data->active_foreground = ctl;
        data->active_colorbox = ctl;
        SendMessage( ctl, CLBXM_SELECTED, TRUE, 0 );
      }
    }
  }

  y += COLORBOX_HEIGHT + 2;
  CreateStaticControl( hWnd, GetModuleHandle(NULL), 0, L"Background Colors:",
                       BORDER_SIZE, y, CLIENT_WIDTH, TEXT_HEIGHT,
                       SS_LEFTNOWORDWRAP | WS_VISIBLE );

  y += TEXT_HEIGHT;
  for( i = 0; i < 8; i++ )
  {
    ctl = CreateColorBox( hWnd, GetModuleHandle(NULL), WS_CHILD | WS_VISIBLE,
                          BORDER_SIZE + i * COLORBOX_WIDTH, y,
                          COLORBOX_WIDTH, COLORBOX_HEIGHT, ID_BACKGROUND_FIRST+i );
    SendMessage( ctl, CLBXM_SET_COLOR, parms->bgcolors[i], 0 );

    if( parms->background == parms->bgcolors[i] )
    {
      if( parms->selected_background )
      {
        if( !parms->selected_foreground ) SetFocus( ctl );
        data->active_background = ctl;
        data->active_colorbox = ctl;
        SendMessage( ctl, CLBXM_SELECTED, TRUE, 0 );
      }
    }
  }

  y += COLORBOX_HEIGHT + 2;
  CreateCheckBox( hWnd, GetModuleHandle(NULL), ID_CUSTOMIZE, SHOW_CUSTOMIZE,
                  BORDER_SIZE, y,
                  BUTTON_WIDTH, BUTTON_HEIGHT, WS_VISIBLE | BS_PUSHLIKE );

  y += BUTTON_HEIGHT + 2;
  ctl = CreateColormapControl( hWnd, GetModuleHandle(NULL), WS_CHILD,
                               BORDER_SIZE, y, ID_COLORMAP );

  if( parms->selected_foreground )
  {
    SendMessage( ctl, CCM_SET_COLOR, parms->foreground, 0 );
  }
  else if( parms->selected_background )
  {
    SendMessage( ctl, CCM_SET_COLOR, parms->background, 0 );
  }

  return 0;
}


static UINT DestroyColorPicker( HWND hWnd )
{
  COLORPICKER_WINDOW_DATA *data;

  data = (COLORPICKER_WINDOW_DATA*)GetWindowLong( hWnd, 0 );
  free( data );

  return 0;
}


static UINT BackgroundColorSlotChanged( HWND hWnd, UINT id )
{
  COLORPICKER_WINDOW_DATA *data;

  data = (COLORPICKER_WINDOW_DATA*)GetWindowLong( hWnd, 0 );

  if( data->active_background != NULL && id == (UINT)GetDlgCtrlID( data->active_background ) )
  {
    SendMessage( data->active_background, CLBXM_SELECTED, FALSE, 0 );
    SetFocus( hWnd );
    data->active_background = NULL;
    data->active_colorbox = NULL;
    return 0;
  }

  if( data->active_background )
  {
    SendMessage( data->active_background, CLBXM_SELECTED, FALSE, 0 );
  }

  data->active_background = GetDlgItem( hWnd, id );
  data->active_colorbox = data->active_background;

  SendMessage( data->active_background, CLBXM_SELECTED, TRUE, 0 );

  SendMessage( GetDlgItem( hWnd, ID_COLORMAP ), CCM_SET_COLOR,
     SendMessage( data->active_background, CLBXM_GET_COLOR, 0, 0 ), 0 );

  return 0;
}


static UINT ForegroundColorSlotChanged( HWND hWnd, UINT id )
{
  COLORPICKER_WINDOW_DATA *data;

  data = (COLORPICKER_WINDOW_DATA*)GetWindowLong( hWnd, 0 );

  if( data->active_foreground != NULL && id == (UINT)GetDlgCtrlID( data->active_foreground ) )
  {
    SendMessage( data->active_foreground, CLBXM_SELECTED, FALSE, 0 );
    SetFocus( hWnd );
    data->active_foreground = NULL;
    data->active_colorbox = NULL;
    return 0;
  }

  if( data->active_foreground )
  {
    SendMessage( data->active_foreground, CLBXM_SELECTED, FALSE, 0 );
  }

  data->active_foreground = GetDlgItem( hWnd, id );
  data->active_colorbox = data->active_foreground;

  SendMessage( data->active_foreground, CLBXM_SELECTED, TRUE, 0 );

  SendMessage( GetDlgItem( hWnd, ID_COLORMAP ), CCM_SET_COLOR,
     SendMessage( data->active_foreground, CLBXM_GET_COLOR, 0, 0 ), 0 );

  return 0;
}


static UINT SelectedColorChanged( HWND hWnd )
{
  HWND hMap;
  COLORREF color;
  COLORPICKER_WINDOW_DATA *data;

  data = (COLORPICKER_WINDOW_DATA*)GetWindowLong( hWnd, 0 );

  if( data->active_colorbox )
  {
    hMap = GetDlgItem( hWnd, ID_COLORMAP );
    color = SendMessage( hMap, CCM_GET_COLOR, 0, 0 );

    SendMessage( data->active_colorbox, CLBXM_SET_COLOR, color, 0 );
  }

  return 0;
}


static UINT AcceptSettings( HWND hWnd )
{
  COLORPICKER_WINDOW_DATA *data;
  HWND ctl;
  UINT id;

  data = (COLORPICKER_WINDOW_DATA*)GetWindowLong( hWnd, 0 );

  for( id = 0; id < 8; id++ )
  {
    ctl = GetDlgItem( hWnd, ID_FOREGROUND_FIRST + id );
    data->parms->fgcolors[id] = SendMessage( ctl, CLBXM_GET_COLOR, 0, 0 );
    ctl = GetDlgItem( hWnd, ID_BACKGROUND_FIRST + id );
    data->parms->bgcolors[id] = SendMessage( ctl, CLBXM_GET_COLOR, 0, 0 );
  }

  data->parms->selected_foreground = ( data->active_foreground != NULL );
  data->parms->selected_background = ( data->active_background != NULL );

  if( data->parms->selected_foreground )
  {
    data->parms->foreground = SendMessage( data->active_foreground, CLBXM_GET_COLOR, 0, 0 );
  }

  if( data->parms->selected_background )
  {
    data->parms->background = SendMessage( data->active_background, CLBXM_GET_COLOR, 0, 0 );
  }

  return 0;
}


static UINT ToggleCustomize( HWND hWnd )
{
  COLORPICKER_WINDOW_DATA *data;
  HWND ctl;
  RECT rect;

  data = (COLORPICKER_WINDOW_DATA*)GetWindowLong( hWnd, 0 );
  ctl = GetDlgItem( hWnd, ID_CUSTOMIZE );

  data->customize_open = !data->customize_open;
  GetWindowRect( hWnd, &rect );

  if( data->customize_open )
  {
    SetWindowText( ctl, HIDE_CUSTOMIZE );
    rect.bottom += EXTRA_HEIGHT;
    ShowWindow( GetDlgItem( hWnd, ID_COLORMAP ), SW_SHOW );
  }
  else
  {
    SetWindowText( ctl, SHOW_CUSTOMIZE );
    rect.bottom -= EXTRA_HEIGHT;
    ShowWindow( GetDlgItem( hWnd, ID_COLORMAP ), SW_HIDE );
  }

  MoveWindow( hWnd, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, TRUE );

  /* ensure that a colorbox with focus retains that focus after the button is clicked */
  if( data->active_colorbox ) SetFocus( data->active_colorbox );

  return 0;
}
