/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: edit categories dialog
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <windows.h>

#include "edit-categories.h"
#include "get-text.h"
#include "utilities.h"
#include "library.h"

#define EDIT_CATEGORIES_CLASS     L"VadeMecumEditCategories"
#define INTERNAL_BORDER           ( 5 )

#define ID_CATEGORY_LIST          ( 40001 )
#define ID_ADD_CATEGORY           ( 40002 )
#define ID_DELETE_CATEGORY        ( 40003 )
#define ID_RENAME_CATEGORY        ( 40004 )
#define ID_SELECT_ALL             ( 40005 )
#define ID_SELECT_NONE            ( 40006 )


static BOOL CALLBACK EditCategoriesWindowProc( HWND, UINT, WPARAM, LPARAM );

static BOOL CreateEditCategoriesWindow( HWND, LPCREATESTRUCT );
static void AddCategory( HWND );
static void DeleteCategory( HWND );
static void SaveCategorySelections( HWND );
static void RenameCategory( HWND );
static void SelectAllCategories( HWND );
static void SelectNoCategory( HWND );


static Library *s_library = NULL;
static Document *s_document = NULL;
static EditCategoriesMode s_mode;


ATOM RegisterEditCategoriesDialog( HINSTANCE hInst )
{
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = EditCategoriesWindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInst;
  wc.hIcon = (HICON)NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH );
  wc.hCursor = (HCURSOR)NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = EDIT_CATEGORIES_CLASS;

  return RegisterClass( &wc );
}


BOOL ShowEditCategoriesDialog( HWND hWnd,
                               EditCategoriesMode mode,
                               Library *library,
                               Document *document )
{
  HWND hDlg;
  MSG msg;
  RECT work_area;

  GetWindowRect( hWnd, &work_area );

  s_mode = mode;
  s_library = library;
  s_document = document;

  hDlg = CreateWindowEx( WS_EX_CAPTIONOKBTN,
                         EDIT_CATEGORIES_CLASS,
                         ( mode == ECD_MODE_DOCUMENT ? L"Document Categories" : L"Category Filter" ),
                         WS_POPUP | WS_CAPTION | WS_SYSMENU,
                         work_area.left, work_area.top,
                         work_area.right - work_area.left,
                         work_area.bottom - work_area.top,
                         hWnd, (HMENU)NULL, library->hInstance, NULL );

  if( hDlg == NULL )
    return FALSE;

  ShowWindow( hDlg, SW_SHOW );
  UpdateWindow( hDlg );

  while( GetMessage( &msg, NULL, 0, 0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  return msg.wParam;
}




static BOOL CALLBACK EditCategoriesWindowProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
  MSG msg_info;

  switch( msg )
  {
    case WM_CREATE:
      return CreateEditCategoriesWindow( hWnd, (LPCREATESTRUCT)lParam );

    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case IDOK:
          SaveCategorySelections( hWnd );
          PostQuitMessage( TRUE );
          DestroyWindow( hWnd );
          return 0;
        case ID_ADD_CATEGORY:
          AddCategory( hWnd );
          return 0;
        case ID_RENAME_CATEGORY:
          RenameCategory( hWnd );
          return 0;
        case ID_DELETE_CATEGORY:
          DeleteCategory( hWnd );
          return 0;
        case ID_SELECT_ALL:
          SelectAllCategories( hWnd );
          return 0;
        case ID_SELECT_NONE:
          SelectNoCategory( hWnd );
          return 0;
      }
      break;

    case WM_CLOSE:
      /* if the window is not being destroyed because of an OK button click, then
       * post a quit message.  Otherwise, leave the quit message that was previously
       * posted there where it is. */
      if( !PeekMessage( &msg_info, hWnd, WM_QUIT, WM_QUIT, PM_NOREMOVE ) )
      {
        PostQuitMessage( FALSE );
      }

      DestroyWindow( hWnd );
      return 0;
  }

  return DefWindowProc( hWnd, msg, wParam, lParam );
}


static BOOL CreateEditCategoriesWindow( HWND hWnd, LPCREATESTRUCT lpcs )
{
  HWND ctl;
  int  category;
  int  index;
  int  y;
  int  height;
  int  width;
  RECT rect;
  int  last_category;

  y = INTERNAL_BORDER;
  height = 16;

  GetClientRect( hWnd, &rect );

  CreateStaticControl( hWnd, s_library->hInstance, 0, L"Categories:",
                       INTERNAL_BORDER, y, rect.right - 2*INTERNAL_BORDER, height,
                       SS_LEFTNOWORDWRAP | WS_VISIBLE );

  y += height + 4;
  height = rect.bottom - y - INTERNAL_BORDER - 30;

  if( s_mode == ECD_MODE_FILTER )
  {
    height -= 30; /* for the select all/none buttons */
  }

  ctl = CreateListBox( hWnd, s_library->hInstance, ID_CATEGORY_LIST,
                       INTERNAL_BORDER, y,
                       rect.right - 2*INTERNAL_BORDER, height,
                       LBS_MULTIPLESEL | LBS_NOINTEGRALHEIGHT | LBS_SORT | WS_VISIBLE );

  /* do not include the "Unfiled" category in this list if mode == ECD_MODE_DOCUMENT */
  last_category = MAX_CATEGORIES;
  if( s_mode == ECD_MODE_DOCUMENT ) last_category -= 1;

  for( category = 0; category < last_category; category += 1 )
  {
    CategoryDefinition *def = &s_library->categories[ category ];
    BOOL select;

    if( def->name[0] != '\0' )
    {
      index = SendMessage( ctl, LB_ADDSTRING, 0, (LPARAM)def->name );
      SendMessage( ctl, LB_SETITEMDATA, index, (LPARAM)def );

      select = FALSE;
      switch( s_mode )
      {
        case ECD_MODE_DOCUMENT:
          select = ( ( s_document->categories & ( 1 << def->id ) ) == ( 1 << def->id ) );
          break;

        case ECD_MODE_FILTER:
          select = ( ( s_library->current_categories & ( 1 << def->id ) ) == ( 1 << def->id ) );
          break;
      }

      SendMessage( ctl, LB_SETSEL, select, index );
    }
  }

  y += height + 5;
  height = 25;
  width = ( rect.right - 2 * INTERNAL_BORDER - 10 ) / 3;

  CreateButtonControl( hWnd, s_library->hInstance, ID_ADD_CATEGORY, L"Add",
                       INTERNAL_BORDER, y, width, height, WS_VISIBLE, FALSE );

  CreateButtonControl( hWnd, s_library->hInstance, ID_DELETE_CATEGORY, L"Delete",
                       INTERNAL_BORDER + width + 5, y, width, height, WS_VISIBLE, FALSE );

  CreateButtonControl( hWnd, s_library->hInstance, ID_RENAME_CATEGORY, L"Rename",
                       INTERNAL_BORDER + width*2 + 10, y, width, height, WS_VISIBLE, FALSE );

  if( s_mode == ECD_MODE_FILTER )
  {
    y += height + 5;
    width = ( rect.right - 2 * INTERNAL_BORDER - 5 ) / 2;

    CreateButtonControl( hWnd, s_library->hInstance, ID_SELECT_ALL, L"Select All",
                         INTERNAL_BORDER, y, width, height, WS_VISIBLE, FALSE );

    CreateButtonControl( hWnd, s_library->hInstance, ID_SELECT_NONE, L"Select None",
                         INTERNAL_BORDER + width + 5, y, width, height, WS_VISIBLE, FALSE );
  }

  return 0;
}


static void AddCategory( HWND hWnd )
{
  GETUSERTEXTOPTS opts;
  TCHAR name[ MAX_CATEGORY_NAME_LEN+1 ];
  CategoryDefinition *def;

  name[0] = '\0';

  opts.flags = ES_LEFT | ES_AUTOHSCROLL;
  opts.hInstance = s_library->hInstance;
  opts.hWnd = hWnd;
  opts.prompt = L"Enter the category name:";
  opts.title = L"New Category";
  opts.width = 180;
  opts.height = 20;
  opts.text = name;
  opts.text_len = sizeof( name ) / sizeof( name[0] );

  if( GetUserText( &opts ) && wcslen( name ) > 0 )
  {
    HWND ctl;
    int idx;

    ctl = GetDlgItem( hWnd, ID_CATEGORY_LIST );
    idx = SendMessage( ctl, LB_FINDSTRING, 0, (LPARAM)name );

    if( idx == LB_ERR )
    {
      def = Library_findCategoryByName( s_library, name );
      if( def != NULL )
      {
        idx = SendMessage( ctl, LB_ADDSTRING, 0, (LPARAM)def->name );
        SendMessage( ctl, LB_SETITEMDATA, idx, (LPARAM)def );
      }
      /* TODO: report when the add failed */
    }
  }
}


static void DeleteCategory( HWND hWnd )
{
  CategoryDefinition *def;
  HWND ctl;
  int idx;
  int count;
  int items[ MAX_CATEGORIES ];
  TCHAR message[ 512 ];

  ctl = GetDlgItem( hWnd, ID_CATEGORY_LIST );
  count = SendMessage( ctl, LB_GETSELCOUNT, 0, 0 );

  if( count < 1 )
  {
    MessageBox( hWnd, L"Please select a category to delete.", L"Error", MB_OK | MB_ICONWARNING );
    return;
  }

  SendMessage( ctl, LB_GETSELITEMS, count, (LPARAM)items );

  wsprintf( message, L"Do you really want to delete these %d categories?", count );
  if( MessageBox( hWnd, message, L"Are you sure?", MB_YESNO | MB_ICONQUESTION ) == IDYES )
  {
    for( idx = count-1; idx >= 0; idx -= 1 )
    {
      if( items[ idx ] == MAX_CATEGORIES-1 ) continue;

      def = (CategoryDefinition*)SendMessage( ctl, LB_GETITEMDATA, items[ idx ], 0 );
      SendMessage( ctl, LB_DELETESTRING, items[ idx ], 0 );
      Library_deleteCategory( s_library, def );
    }
  }
}


static void SaveCategorySelections( HWND hWnd )
{
  HWND ctl;
  int count;
  int item;
  CategoryDefinition *def;
  int selected_categories = 0;

  ctl = GetDlgItem( hWnd, ID_CATEGORY_LIST );
  count = SendMessage( ctl, LB_GETCOUNT, 0, 0 );

  for( item = 0; item < count; item += 1 )
  {
    if( SendMessage( ctl, LB_GETSEL, item, 0 ) )
    {
      def = (CategoryDefinition*)SendMessage( ctl, LB_GETITEMDATA, item, 0 );
      selected_categories |= ( 1 << def->id );
    }
  }

  if( s_mode == ECD_MODE_DOCUMENT )
  {
    s_document->categories = selected_categories;

    /* if the document has been assigned a category, remove it from the "unfiled" category */
    if( ( s_document->categories & ~( 1 << ( MAX_CATEGORIES-1 ) ) ) != 0 )
    {
      s_document->categories &= ~( 1 << ( MAX_CATEGORIES-1 ) );
    }
    else if( s_document->categories == 0 )
    {
      s_document->categories = ( 1 << ( MAX_CATEGORIES-1 ) );
    }
  }
  else
  {
    s_library->current_categories = selected_categories;
    if( selected_categories == 0 )
    {
      s_library->current_categories = 0xFFFF;
    }
  }
}


static void RenameCategory( HWND hWnd )
{
  GETUSERTEXTOPTS opts;
  CategoryDefinition *def;
  HWND ctl;
  int idx;
  TCHAR name[ MAX_CATEGORY_NAME_LEN+1 ];

  ctl = GetDlgItem( hWnd, ID_CATEGORY_LIST );
  idx = SendMessage( ctl, LB_GETSELCOUNT, 0, 0 );

  if( idx > 1 )
  {
    MessageBox( hWnd, L"You can only rename one category at a time.", L"Error", MB_OK | MB_ICONWARNING );
    return;
  }
  else if( idx < 1 )
  {
    MessageBox( hWnd, L"Please select a category to rename.", L"Error", MB_OK | MB_ICONWARNING );
    return;
  }

  SendMessage( ctl, LB_GETSELITEMS, 1, (LPARAM)(LPINT)&idx );

  def = (CategoryDefinition*)SendMessage( ctl, LB_GETITEMDATA, idx, 0 );
  if( def->id == MAX_CATEGORIES-1 )
  {
    MessageBox( hWnd, L"You cannot rename the 'Unfiled' category.", L"Error", MB_OK );
    return;
  }

  wcscpy( name, def->name );

  opts.flags = ES_LEFT | ES_AUTOHSCROLL;
  opts.hInstance = s_library->hInstance;
  opts.hWnd = hWnd;
  opts.prompt = L"Enter the new category name:";
  opts.title = L"Rename Category";
  opts.width = 180;
  opts.height = 20;
  opts.text = name;
  opts.text_len = sizeof( name ) / sizeof( name[0] );

  if( GetUserText( &opts ) && wcslen( name ) > 0 )
  {
    if( Library_findCategoryByName( s_library, name ) != NULL )
    {
      MessageBox( hWnd, L"A category already exists by that name!", L"Duplicate Name", MB_OK | MB_ICONWARNING );
    }
    else
    {
      wcscpy( def->name, name );
      SendMessage( ctl, LB_DELETESTRING, idx, 0 );
      idx = SendMessage( ctl, LB_ADDSTRING, 0, (LPARAM)name );
      SendMessage( ctl, LB_SETITEMDATA, idx, (LPARAM)def );
    }
  }
}


static void SelectAllCategories( HWND hWnd )
{
  int count;
  HWND ctl;

  ctl = GetDlgItem( hWnd, ID_CATEGORY_LIST );
  count = SendMessage( ctl, LB_GETCOUNT, 0, 0 );
  SendMessage( ctl, LB_SELITEMRANGE, TRUE, MAKELPARAM( 0, count-1 ) );
}


static void SelectNoCategory( HWND hWnd )
{
  int count;
  HWND ctl;

  ctl = GetDlgItem( hWnd, ID_CATEGORY_LIST );
  count = SendMessage( ctl, LB_GETCOUNT, 0, 0 );
  SendMessage( ctl, LB_SELITEMRANGE, FALSE, MAKELPARAM( 0, count-1 ) );
}

