/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: font cache module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <windows.h>
#include <stdlib.h>
#include <string.h>

#include "font_cache.h"


typedef struct _font_cache_item_   FontCacheItem;
typedef struct _font_cache_record_ FontCacheRecord;

struct _font_cache_item_
{
  FontCacheItem *next;

  HFONT   font;
  LOGFONT def;
};

struct _font_cache_record_
{
  FontCacheItem *items;

  int cache_size;
  int max_cache_size;

  GlyphCache glyph_cache;
};


static void static_font_cache_item_delete( FontCacheItem *item );



FontCache FontCache_new( int size, int glyph_cache_size )
{
  FontCacheRecord *record;

  record = (FontCacheRecord*)malloc( sizeof( FontCacheRecord ) );
  if( record == NULL )
  {
    return NULL;
  }

  memset( record, 0, sizeof( *record ) );

  record->max_cache_size = size;

  record->glyph_cache = GlyphCache_new( glyph_cache_size );
  if( record->glyph_cache == NULL )
  {
    free( record );
    return NULL;
  }

  return (FontCache)record;
}


void FontCache_delete( FontCache cache )
{
  FontCacheRecord *record = (FontCacheRecord*)cache;
  FontCacheItem *item;
  FontCacheItem *next;

  item = record->items;
  while( item != NULL )
  {
    next = item->next;
    static_font_cache_item_delete( item );
    item = next;
  }

  GlyphCache_delete( record->glyph_cache );

  memset( record, 0, sizeof( *record ) );
  free( record );
}


HFONT FontCache_get( FontCache  cache,
                     LOGFONT   *font_def )
{
  FontCacheRecord *record = (FontCacheRecord*)cache;
  FontCacheItem   *item;
  FontCacheItem   *prev;

  for( item = record->items, prev = NULL; item != NULL; prev = item, item = item->next )
  {
    if( memcmp( font_def, &item->def, sizeof( LOGFONT ) ) == 0 )
    {
      if( prev != NULL )
      {
        /* move this item to the head of the list, to indicate it as the MRU item */
        prev->next = item->next;
        item->next = record->items;
        record->items = item;
        return item->font;
      }
    }
  }

  /* if we get here, then the cache missed.
   *   1. create a new FontCacheItem record.
   *   2. create a new HFONT object
   *   3. add the new item to the head of the list.
   *   4. if necessary, delete the last item in the list.
   */

  item = (FontCacheItem*)malloc( sizeof( FontCacheItem ) );
  if( item == NULL )
  {
    /* out of memory */
    return NULL;
  }
  memset( item, 0, sizeof( *item ) );

  memcpy( &item->def, font_def, sizeof( LOGFONT ) );
  item->font = CreateFontIndirect( &item->def );

  if( item->font == NULL )
  {
    /* could not create requested font entity */
    free( item );
    return NULL;
  }

  item->next = record->items;
  record->items = item;
  record->cache_size++;

  if( record->cache_size > record->max_cache_size )
  {
    prev = item;
    while( prev->next->next != NULL ) prev = prev->next;
    static_font_cache_item_delete( prev->next );
    prev->next = NULL;
    record->cache_size--;
  }

  return item->font;
}


static void static_font_cache_item_delete( FontCacheItem *item )
{
  DeleteObject( item->font );
  memset( item, 0, sizeof( *item ) );
  free( item );
}


GlyphInfo* FontCache_getGlyph( FontCache cache,
                               HDC       hdc,
                               HFONT     font,
                               TCHAR     c )
{
  FontCacheRecord *record = (FontCacheRecord*)cache;

  return GlyphCache_get( record->glyph_cache, hdc, font, c );
}
