/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: font cache module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * This defines an object for caching font references, primarily for use
 * during pagination and rendering of a page, but may be used elsewhere when
 * a HFONT handle is needed.  This allows multiple font handles to be open
 * at once, and manages creating and destroying them.  The objects are
 * stored in MRU order; when the cache runs out of room, the items that are
 * least-recently used are destroyed first.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __FONT_CACHE_H__
#define __FONT_CACHE_H__

#include <wingdi.h>       /* LOGFONT, HFONT, etc. */
#include "glyph_cache.h"  /* defines a cache of glyph information */

/* the opaque FontCache object */
typedef void *FontCache;


/**
 * Creates a new FontCache object that can hold 'size' HFONT handles at a time.
 * The FontCache will also have a reference to a GlyphCache object, which
 * can contain 'glyph_cache_size' records at a time.  Returns NULL if the cache
 * cannot be created.
 */
FontCache FontCache_new( int size, int glyph_cache_size );

/**
 * Destroys the given font cache, which also destroys all font handles and the
 * associated glyph cache.
 */
void FontCache_delete( FontCache cache );

/**
 * Returns a HFONT handle for the font matching the given LOGFONT definition.
 * If the handle exists in the cache, it is moved to the front of the cache (making
 * it the most-recently-used item) and then returned.  If no such font exists in
 * the cache, a new HFONT object is created, added to the front of the cache, and
 * returned.  If the cache is not big enough for a new object, the least-recently
 * used handle in the cache is destroyed to make room.
 */
HFONT FontCache_get( FontCache  cache,
                     LOGFONT   *font_def );

/**
 * Returns a reference to a GlyphInfo object, which contains cached metric information
 * for the given character in a given font.  (See glyph_cache.h)
 */
GlyphInfo* FontCache_getGlyph( FontCache cache,
                               HDC       hdc,
                               HFONT     font,
                               TCHAR     c );

#endif /* __FONT_CACHE_H__ */
