/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: get text dialog
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include "get-text.h"
#include "utilities.h"

#include <aygshell.h>


#define ID_USER_TEXT            ( 40001 )


#define GET_USER_TEXT_CLASS     L"VadeMecumGetUserText"
#define INTERNAL_BORDER         ( 5 )
#define PROMPT_HEIGHT           ( 16 )

static BOOL CALLBACK GetUserTextWindowProc( HWND, UINT, WPARAM, LPARAM );

static BOOL CreateGetUserTextWindow( HWND );


static GETUSERTEXTOPTS *static_opts = NULL;



ATOM RegisterGetTextWindow( HINSTANCE hInst )
{
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = GetUserTextWindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInst;
  wc.hIcon = (HICON)NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH );
  wc.hCursor = (HCURSOR)NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = GET_USER_TEXT_CLASS;

  return RegisterClass( &wc );
}


BOOL GetUserText( GETUSERTEXTOPTS *opts )
{
  int cx;
  int cy;
  int width;
  int height;
  int caption_height;
  int border_width;
  int border_height;
  RECT work_area;
  HWND hWnd;
  MSG msg;
  SHMENUBARINFO mbi;

  static_opts = opts;

  caption_height = GetSystemMetrics( SM_CYCAPTION );
  border_width   = GetSystemMetrics( SM_CXBORDER );
  border_height  = GetSystemMetrics( SM_CXBORDER );

  SystemParametersInfo( SPI_GETWORKAREA, 0, &work_area, 0 );

  /* make sure the requested width and height will fit on the screen */

  if( (int)( opts->width + border_width * 2 + INTERNAL_BORDER * 2 ) > (int)( work_area.right - work_area.left ) )
  {
    opts->width = ( work_area.right - work_area.left - border_width * 2 - INTERNAL_BORDER * 2 );
  }

  if( (int)( opts->height + border_height * 2 + INTERNAL_BORDER * 2 + caption_height + PROMPT_HEIGHT ) > (int)( work_area.bottom - work_area.top ) )
  {
    opts->height = ( work_area.bottom - work_area.top - border_height * 2 - INTERNAL_BORDER * 2 - caption_height - PROMPT_HEIGHT );
  }

  width = border_width * 2 + INTERNAL_BORDER * 2 + opts->width;
  height = border_height * 2 + INTERNAL_BORDER * 2 + caption_height + PROMPT_HEIGHT + opts->height;

  cx = work_area.left + ( work_area.right - work_area.left - width ) / 2;
  cy = work_area.top; /*+ ( work_area.bottom - work_area.top - height ) / 4;*/

  hWnd = CreateWindowEx( WS_EX_CAPTIONOKBTN,
                         GET_USER_TEXT_CLASS,
                         opts->title,
                         WS_POPUP | WS_CAPTION | WS_SYSMENU,
                         cx, cy, width, height,
                         opts->hWnd, (HMENU)NULL, opts->hInstance, NULL );

  if( hWnd == NULL )
  {
    return FALSE;
  }

  mbi.cbSize = sizeof( mbi );
  mbi.hwndParent = hWnd;
  mbi.dwFlags = SHCMBF_EMPTYBAR;
  mbi.hInstRes = opts->hInstance;

  SHFullScreen( hWnd, SHFS_SHOWSIPBUTTON );
  SHCreateMenuBar( &mbi );

  EnableWindow( opts->hWnd, FALSE );

  ShowWindow( hWnd, SW_SHOW );
  UpdateWindow( hWnd );

  while( GetMessage( &msg, NULL, 0, 0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  EnableWindow( opts->hWnd, TRUE );

  return msg.wParam;
}




static BOOL CALLBACK GetUserTextWindowProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
  MSG msg_info;
  HWND ctl;

  SHInputDialog( hWnd, msg, wParam );

  switch( msg )
  {
    case WM_CREATE:
      return CreateGetUserTextWindow( hWnd );

    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case IDOK:
          SHFullScreen( hWnd, SHFS_HIDESIPBUTTON );
          ctl = GetDlgItem( hWnd, ID_USER_TEXT );
          GetWindowText( ctl, static_opts->text, static_opts->text_len );
          PostQuitMessage( TRUE );
          DestroyWindow( hWnd );
          return 0;
      }
      break;

    case WM_CLOSE:
      SHFullScreen( hWnd, SHFS_HIDESIPBUTTON );

      /* if the window is not being destroyed because of an OK button click, then
       * post a quit message.  Otherwise, leave the quit message that was previously
       * posted there where it is. */
      if( !PeekMessage( &msg_info, hWnd, WM_QUIT, WM_QUIT, PM_NOREMOVE ) )
      {
        PostQuitMessage( FALSE );
      }

      DestroyWindow( hWnd );
      return 0;
  }

  return DefWindowProc( hWnd, msg, wParam, lParam );
}


static BOOL CreateGetUserTextWindow( HWND hWnd )
{
  HWND ctl;

  CreateStaticControl( hWnd, static_opts->hInstance, 0, static_opts->prompt,
                       INTERNAL_BORDER, INTERNAL_BORDER, static_opts->width, PROMPT_HEIGHT,
                       SS_LEFTNOWORDWRAP | WS_VISIBLE );

  ctl = CreateEditControl( hWnd, static_opts->hInstance, ID_USER_TEXT,
                           INTERNAL_BORDER, INTERNAL_BORDER + PROMPT_HEIGHT,
                           static_opts->width, static_opts->height,
                           static_opts->flags | WS_VISIBLE );

  SetWindowText( ctl, static_opts->text );
  SetFocus( ctl );
  SendMessage( ctl, EM_SETSEL, 0, -1 );

  return 0;
}
