/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: glyph cache module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <windows.h>

#include "glyph_cache.h"

#include <stdlib.h>
#include <string.h>


typedef struct _glyph_cache_record_ GlyphCacheRecord;
typedef struct _glyph_cache_item_   GlyphCacheItem;

struct _glyph_cache_record_
{
  GlyphCacheItem *items;

  int cache_size;
  int max_cache_size;
};

struct _glyph_cache_item_
{
  GlyphCacheItem *next;
  
  TCHAR c;
  HFONT font;

  GlyphInfo info;
};


static void static_glyph_cache_item_delete( GlyphCacheItem *item );



GlyphCache GlyphCache_new( int size )
{
  GlyphCacheRecord *record = NULL;

  record = (GlyphCacheRecord*)malloc( sizeof( GlyphCacheRecord ) );
  if( record == NULL )
  {
    return NULL;
  }

  memset( record, 0, sizeof( *record ) );
  record->max_cache_size = size;

  return (GlyphCache)record;
}


void GlyphCache_delete( GlyphCache cache )
{
  GlyphCacheRecord *record = (GlyphCacheRecord*)cache;
  GlyphCacheItem *item;
  GlyphCacheItem *next;

  item = record->items;
  while( item != NULL )
  {
    next = item->next;
    static_glyph_cache_item_delete( item );
    item = next;
  }

  memset( record, 0, sizeof( *record ) );
  free( record );
}


void GlyphCache_invalidate( GlyphCache cache,
                            HFONT font )
{
  GlyphCacheRecord *record = (GlyphCacheRecord*)cache;
  GlyphCacheItem   *prev;
  GlyphCacheItem   *item;
  GlyphCacheItem   *next;

  prev = NULL;
  item = record->items;

  while( item != NULL )
  {
    if( font == NULL || item->font == font )
    {
      next = item->next;

      if( prev != NULL )
      {
        prev->next = next;
      }
      else
      {
        record->items = next;
      }

      static_glyph_cache_item_delete( item );
      item = next;
    }
    else
    {
      prev = item;
      item = item->next;
    }
  }
}


GlyphInfo* GlyphCache_get( GlyphCache cache,
                           HDC hdc,
                           HFONT font,
                           TCHAR c )
{
  GlyphCacheRecord *record = (GlyphCacheRecord*)cache;
  GlyphCacheItem *item;
  GlyphCacheItem *prev;
  SIZE size;

  switch( c )
  {
    case 160: /* non-breaking space */
      c = ' ';
      break;
  }

  prev = NULL;
  item = record->items;
  
  while( item != NULL )
  {
    if( item->font == font && item->c == c )
    {
      if( prev != NULL )
      {
        prev->next = item->next;
        item->next = record->items;
        record->items = item;
      }

      return &item->info;
    }

    prev = item;
    item = item->next;
  }

  /* if we get here, then the cache missed.  We need to create a new item
   * and add it to the cache. */

  item = (GlyphCacheItem*)malloc( sizeof( GlyphCacheItem ) );
  if( item == NULL )
  {
    return NULL;
  }
  memset( item, 0, sizeof( *item ) );

  item->next = record->items;
  item->font = font;
  item->c = c;
  
  SelectObject( hdc, font );

  if( !GetTextExtentPoint32( hdc, (LPCTSTR)&c, 1, &size ) )
  {
    free( item );
    return NULL;
  }

  item->info.width = (unsigned char)size.cx;
  item->info.height = (unsigned char)size.cy;

  record->items = item;
  record->cache_size++;

  if( record->cache_size > record->max_cache_size )
  {
    prev = item;
    while( prev->next->next != NULL ) prev = prev->next;
    static_glyph_cache_item_delete( prev->next );
    prev->next = NULL;
    record->cache_size--;
  }

  return &item->info;
}


static void static_glyph_cache_item_delete( GlyphCacheItem *item )
{
  memset( item, 0, sizeof( *item ) );
  free( item );
}
