/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: highlights module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * This module defines an interface whereby the application can query, add,
 * and delete highlight records.  It abstracts the lowerlevel "simple-db"
 * API calls.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __HIGHLIGHTS_H__
#define __HIGHLIGHTS_H__

#include <stdio.h>
#include "linked-list.h"

typedef void* HighlightManager;
typedef struct highlight_t Highlight;

struct highlight_t
{
  int record_id;    /* the record-id of the plucker record that contains this highlight */

  int start_pos;    /* the starting character position in the record of the highlight */
  int end_pos;      /* the ending character position in the record of the highlight */

  int fg;           /* the foreground color of the highlight (-1 means no change) */
  int bg;           /* the background color of the highlight (-1 means no change) */

  int annotation;   /* index into the annotation db (-1 for none) */

  int dbid;         /* "database id" -- the index of the record on disk.  READ-ONLY! */
};


/**
 * Instantiate a new highlight manager that uses the named file as its database
 * file.  Returns NULL if the manager cannot be created.
 */
HighlightManager HighlightManager_create( char *filename );

/**
 * Destroys the given highlight manager.  This will close the associated file handle, but
 * will not free any LinkedList objects returned by the getHighlights call.
 */
void HighlightManager_destroy( HighlightManager hm );

/**
 * Returns a LinkedList object that enumerates all of the highlights in the database for
 * the given document record id.  It is the calling routine's responsibility to destroy
 * the linked list when done with it.  The highlights in the linked list are in sorted order,
 * by starting and ending position of the highlights.
 */
LinkedList* HighlightManager_getHighlights( HighlightManager hm,
                                            int record_id );

/**
 * Adds a copy of the given highlight record to both the peristant storage of the highlight
 * manager, and to the linked list of highlights (in sorted order).
 */
void HighlightManager_addHighlight( HighlightManager hm,
                                    LinkedList *highlights,
                                    Highlight *highlight );

/**
 * Deletes the given highlight both from the manager's persistant storage, and from the linked
 * list.  The highlight record is also freed, so be sure to pass the data area of a LinkedListNode
 * as the third parameter.
 */
void HighlightManager_deleteHighlight( HighlightManager hm,
                                       LinkedList *highlights,
                                       Highlight *highlight );

/**
 * Modifies the foreground, background, and annotation values of the given highlight, storing
 * them to the database of the manager.  The values are written whether they differ from the
 * original or not.
 */
void HighlightManager_modifyHighlight( HighlightManager hm,
                                       Highlight *highlight,
                                       int fg,
                                       int bg,
                                       int annotation );

#endif /* __HIGHLIGHTS_H__ */
