/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: history module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * Represents a list of reader history values, per document.  This is used
 * implement the forward/backward interface.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __HISTORY_H__
#define __HISTORY_H__


#include <stdio.h> /* FILE */

#include <windows.h>
#include "PalmTypes.h"

#ifdef PDAV_SHOWGUIDE

typedef enum
{
  HISTORY_ITEM_TYPE_PLUCKER,
  HISTORY_ITEM_TYPE_MAP,
  HISTORY_ITEM_TYPE_EXHIBITOR,
  HISTORY_ITEM_TYPE_EXHIBITOR_LIST,
  HISTORY_ITEM_TYPE_SESSION,
  HISTORY_ITEM_TYPE_SESSION_LIST,
  HISTORY_ITEM_TYPE_SESSION_SEARCH,

} HISTORY_ITEM_TYPE;


typedef struct
{
  int record;       /* the record index */
  int char_index;   /* the index of the first character on the first full line on the screen */
  int y_offset;     /* the yoffset from the top of the screen to the first full line on the screen */

}HISTORY_ITEM_PLUCKER;

typedef struct
{
  float zoomLevel;
  Int32 xCentre;
  Int32 yCentre;
  RECT highlightRect;

} HISTORY_ITEM_MAP;

typedef struct
{
  UInt16 id;

} HISTORY_ITEM_SESSION;

typedef struct
{
  int year;
  int month;
  int day;
  char keyword[256+1];
  char track[256+1];
  char cme[256+1];
} HISTORY_ITEM_SESSION_LIST;

typedef struct
{
  int year;
  int month;
  int day;
  char keyword[256+1];
  char track[256+1];
  char cme[256+1];
} HISTORY_ITEM_SESSION_SEARCH;

typedef struct
{
  UInt16 id;

} HISTORY_ITEM_EXHIBITOR;

typedef struct
{
  UInt16 startId;
  char limitChar;

} HISTORY_ITEM_EXHIBITOR_LIST;

#endif

typedef struct _history_item HISTORY_ITEM;

/* a single item in a list of history items, representing an historical position
 * in a document. */

typedef struct _history_item
{
  HISTORY_ITEM *next;
  HISTORY_ITEM *previous;

#ifndef PDAV_SHOWGUIDE
  int record;       /* the record index */
  int char_index;   /* the index of the first character on the first full line on the screen */
  int y_offset;     /* the yoffset from the top of the screen to the first full line on the screen */

#else

  TCHAR current_document[MAX_PATH]; /* the name of the current document, or "" if there is no current one */
  HISTORY_ITEM_TYPE type;

  union
  {
    HISTORY_ITEM_PLUCKER plucker;
    HISTORY_ITEM_MAP map;
    HISTORY_ITEM_SESSION session;
    HISTORY_ITEM_SESSION_LIST sessionList;
    HISTORY_ITEM_SESSION_SEARCH sessionSearch;
    HISTORY_ITEM_EXHIBITOR exhibitor;
    HISTORY_ITEM_EXHIBITOR_LIST exhibitorList;
  } item;
#endif
};

/* a history list for a document. */

typedef struct
{
  HISTORY_ITEM *head;     /* the first element of the list (the oldest position in the history) */
  HISTORY_ITEM *tail;     /* the last element of the list (the newest position in the history) */
  HISTORY_ITEM *current;  /* the current element of the list (the current position in the history) */

  int capacity;           /* the maximum number of history items that may be stored */
  int length;             /* the # of history items being stored */
} HISTORY_LIST;


/**
 * Creates a new (empty) history list, capable of holding 'capacity' history items.
 * Returns NULL if the list cannot be created for any reason.
 */
HISTORY_LIST *HistoryList_new( int capacity );

/**
 * Deletes the given history list and frees all associated memory.
 */
void HistoryList_delete( HISTORY_LIST *list );

#ifndef PDAV_SHOWGUIDE

/**
 * Adds the given position after the 'current' element.  If the current element is not the tail
 * element, then all elements after the current element are destroyed (ie, the list is truncated
 * at the current element), and then the new element is added.  The new element becomes the current
 * element, which is then returned.
 */
HISTORY_ITEM* HistoryList_add( HISTORY_LIST *list,
                               int record,
                               int char_index,
                               int y_offset );

#else

void HistoryList_setHistoryItemPlucker(HISTORY_ITEM *hi,
                                       LPCTSTR filename,
                                       int record,
                                       int char_index,
                                       int y_offset );

void HistoryList_setHistoryItemMap(HISTORY_ITEM *hi,
                                   float zoomLevel,
                                   Int32 xCentre,
                                   Int32 yCentre,
                                   RECT *highlightRect);

void HistoryList_setHistoryItemSession(HISTORY_ITEM *hi,
                                       LPCTSTR filename,
                                       UInt16 id);

void HistoryList_setHistoryItemSessionList(HISTORY_ITEM *hi,
                                           LPCTSTR filename,
                                           int year, int month, int day, const char *keyword, const char *track, const char *cme);

void HistoryList_setHistoryItemSessionSearch(HISTORY_ITEM *hi, int year, int month, int day, const char *keyword, const char *track, const char *cme);

void HistoryList_setHistoryItemExhibitor(HISTORY_ITEM *hi,
                                         LPCTSTR filename,
                                         UInt16 id);

void HistoryList_setHistoryItemExhibitorList(HISTORY_ITEM *hi,
                                             LPCTSTR filename,
                                             UInt16 startId,
                                             char limitChar);

HISTORY_ITEM* HistoryList_add( HISTORY_LIST *list );


#endif


/**
 * Returns the current element in the list, or NULL if the list is empty.  Unless the list is
 * empty, there will ALWAYS be a current element in the list.
 */
HISTORY_ITEM *HistoryList_current( HISTORY_LIST *list );

/**
 * If the current element is not the head of the list, then the current element becomes the
 * element immediately before the current element.  The current element is returned, unless
 * the current element was already the head of the list, in which case NULL is returned.
 */
HISTORY_ITEM *HistoryList_goBack( HISTORY_LIST *list );

/**
 * If the current element is not the tail of the list, then the current element becomes the
 * element immediately after the current element.  The current element is returned, unless
 * the current element was already the tail of the list, in which case NULL is returned.
 */
HISTORY_ITEM *HistoryList_goForward( HISTORY_LIST *list );

/**
 * Returns non-zero if there are history items in the list before the current item (ie,
 * the current item is not the head of the list).
 */
int HistoryList_priorElementsExist( HISTORY_LIST *list );

/**
 * Returns non-zero if there are history items in the list after the current item (ie,
 * the current item is not the tail of the list).
 */
int HistoryList_subsequentElementsExist( HISTORY_LIST *list );

#ifndef PDAV_SHOWGUIDE

/**
 * Serializes the given history list to the given FILE, at the current position in the FILE.
 */
int HistoryList_serialize( HISTORY_LIST *list,
                           FILE *file );

/**
 * Unserializes a history list from the given FILE, at the current position in the FILE.
 */
HISTORY_LIST*  HistoryList_unserialize( FILE *file,
                                        int capacity );
#endif

#endif /* __HISTORY_H__ */