/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: image viewer module
 * Copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <windows.h>
#include <aygshell.h>

#include "image-viewer.h"
#include "resrc1.h"
#include "plkfmt.h"

#define ID_25        ( 25 )
#define ID_50        ( 50 )
#define ID_100       ( 100 )
#define ID_200       ( 200 )
#define ID_400       ( 400 )
#define ID_FIT       ( 1000 )


static Library      *g_library;
static HWND          g_vscroll;
static HWND          g_hscroll;
static int           g_screenWidth;
static int           g_screenHeight;
static int           g_record;
static PluckerDB     g_db;
static int           g_hPos;
static int           g_vPos;
static float         g_fmagn;
static int           g_imagn;


static HWND CreateImageViewerDialog( HWND hParent );
static void OnInit( HWND hWnd );
static void OnPaint( HWND hWnd );
static void OnScroll( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );
static BOOL RecognizeContextMenuGesture( HWND hWnd, LPARAM lParam );
static void DoPopupMenu( HWND hWnd, LPARAM lParam );
static void ResizeScrollbars( HWND hWnd );


UINT32 OpenImageViewer( HWND hWnd, Library *library, PluckerDB db, int record )
{
  HWND hViewer;
  MSG msg;

  g_library = library;
  g_record = record;
  g_db = db;
  g_fmagn = 1.0f;
  g_imagn = ID_100;

  SetCursor( LoadCursor( NULL, IDC_WAIT ) );

  hViewer = CreateImageViewerDialog( hWnd );


  ShowWindow( hViewer, SW_SHOW );
  UpdateWindow( hViewer );

  while( GetMessage( &msg, hViewer, 0, 0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  return 0;
}

BOOL CALLBACK ImageViewerDialogProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
  RECT rect;
  float tmpX, tmpY;
  PluckerImage *image;

  switch( uMsg )
  {
  case WM_CREATE:
    OnInit( hWnd );
    break;
  case WM_COMMAND:
    switch( LOWORD( wParam ) )
    {
    case IDOK:
      DestroyWindow( hWnd );
      break;

    case ID_25:
    case ID_50:
    case ID_100:
    case ID_200:
    case ID_400:
      if( g_imagn != (int)LOWORD( wParam ) )
      {
        g_hPos = g_vPos = 0;
        g_imagn = LOWORD( wParam );
        g_fmagn = (float)g_imagn / 100.0f;

        GetClientRect( hWnd, &rect );
        InvalidateRect( hWnd, &rect, FALSE );
        ResizeScrollbars( hWnd );
      }
      break;
    case ID_FIT:

      if( g_imagn != (int)ID_FIT )
      {
        image = PluckerDBGetImage( g_db, g_record );
        GetClientRect( hWnd, &rect );
        g_imagn = ID_FIT;
        tmpX = (float)(rect.right - rect.left) / (float)image->width;
        tmpY = (float)(rect.bottom - rect.top) / (float)image->height;
        g_fmagn = tmpX < tmpY ? tmpX : tmpY;

        GetClientRect( hWnd, &rect );
        InvalidateRect( hWnd, &rect, FALSE );
        ResizeScrollbars( hWnd );
      }
      break;

    default:
      return FALSE;
    }
    break;

  case WM_HSCROLL:
  case WM_VSCROLL:
    OnScroll( hWnd, uMsg, wParam, lParam );
    break;

  case WM_LBUTTONDOWN:
    if( RecognizeContextMenuGesture( hWnd, lParam ) )
    {
      DoPopupMenu( hWnd, lParam );
    }
    break;

  case WM_CLOSE:
    DestroyWindow( hWnd );
    break;

  case WM_PAINT:
    OnPaint( hWnd );
    break;

  default:
    return DefWindowProc( hWnd, uMsg, wParam, lParam );
  }
  return TRUE;
}

static HWND CreateImageViewerDialog( HWND hParent )
{
  HWND hWnd;
  TCHAR szWindowClass[ 64 ];
  


  LoadString( g_library->hInstance, IDS_IMAGE_VIEWER_CLASS, szWindowClass, 64 );
  hWnd = CreateWindow( szWindowClass, L"Image viewer", WS_VISIBLE | WS_POPUP,
                       CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
                       hParent, NULL, g_library->hInstance, NULL );
  return hWnd;
}

static void OnInit( HWND hWnd )
{
  SHINITDLGINFO idi;
  RECT rect;
  PluckerImage *image;
  int ver, hor;
  HDC dc;

  SHFullScreen( hWnd, SHFS_SHOWTASKBAR );

  idi.dwMask = SHIDIM_FLAGS;
  idi.hDlg = hWnd;
  idi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_FULLSCREENNOMENUBAR;
  SHInitDialog( &idi );  

  /* Resize the window to use the entire screen, except the start menu */
  GetClientRect( hWnd, &rect );
  dc = GetDC( NULL );

  ver = GetDeviceCaps( dc, VERTRES ) - (rect.bottom - rect.top);
  hor = GetDeviceCaps( dc, HORZRES ) - (rect.right - rect.left);
  MoveWindow( hWnd, hor+rect.left, ver+rect.top, rect.right -rect.left, rect.bottom-rect.top, TRUE );

  ReleaseDC( hWnd, dc );

  g_vscroll = g_vscroll = NULL;
  g_vPos = g_hPos = 0;

  image = PluckerDBGetImage( g_db, g_record );

  if( !image )
    return;


  ResizeScrollbars( hWnd );

  SetCursor( NULL );
}

static void OnPaint( HWND hWnd )
{
  PAINTSTRUCT ps;
  HDC hdc=0;
  int cx, cy;
  PluckerImage *image;
  int width, height;

  image = PluckerDBGetImage( g_db, g_record );
  if( !image )
    return;

  BeginPaint( hWnd, &ps );

  hdc = CreateCompatibleDC( ps.hdc );
  SelectObject( hdc, image->image );

  width = (int)(image->width * g_fmagn);
  height = (int)(image->height * g_fmagn);

  cx = width > g_screenWidth ? g_screenWidth : width;
  cy = height > g_screenHeight ? g_screenHeight : height;

  if( g_imagn == ID_100 )
    BitBlt( ps.hdc, 0, 0, cx, cy, hdc, g_hPos, g_vPos, SRCCOPY );
  else
    StretchBlt( ps.hdc, 0, 0, cx, cy, hdc, (int)(g_hPos/g_fmagn), (int)(g_vPos/g_fmagn), 
                (int)(cx/g_fmagn), (int)(cy/g_fmagn), SRCCOPY );

  DeleteDC( hdc );

  EndPaint( hWnd, &ps );
}

static void OnScroll( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
  HWND scroll;
  int pos;
  SCROLLINFO si;
  RECT rect;
  int maxPos, screen;

  if( msg == WM_HSCROLL )
  {
    scroll = g_hscroll;
    pos = g_hPos;
    screen = g_screenWidth;
  }
  else /* msg == WM_VSCROLL */
  {
    scroll = g_vscroll;
    pos = g_vPos;
    screen = g_screenHeight;
  }

  si.cbSize = sizeof( si );
  si.fMask = SIF_RANGE;
  GetScrollInfo( scroll, SB_CTL, &si );
  maxPos = si.nMax-screen;

  switch( LOWORD( wParam ) )
  {
  case SB_LINEUP:
    pos -= 10;
    break;
  case SB_LINEDOWN:
    pos += 10;
    break;
  case SB_PAGEUP:
    pos -= (msg == WM_HSCROLL) ? g_screenWidth : g_screenHeight;
    break;
  case SB_PAGEDOWN:
    pos += (msg == WM_HSCROLL) ? g_screenWidth : g_screenHeight;
    break;
  case SB_THUMBTRACK:
    si.fMask = SIF_TRACKPOS;
    GetScrollInfo( scroll, SB_CTL, &si );
    pos = si.nTrackPos;
    break;
  case SB_THUMBPOSITION:
    si.fMask = SIF_POS;
    GetScrollInfo( scroll, SB_CTL, &si );
    pos = si.nPos;
    break;
  default:
    return;
  }

  if( pos < 0 ) pos = 0;
  if( pos > maxPos ) pos = maxPos;

  si.fMask = SIF_POS;
  si.nPos = pos;
  SetScrollInfo( scroll, SB_CTL, &si, TRUE );

  if( msg == WM_HSCROLL )
  {
    g_hPos = pos;
  }
  else /* msg == WM_VSCROLL */
  {
    g_vPos = pos;
  }

  GetClientRect( hWnd, &rect );
  InvalidateRect( hWnd, &rect, FALSE );
  UpdateWindow( hWnd );
}

static BOOL RecognizeContextMenuGesture( HWND hWnd, LPARAM lParam )
{
#ifndef PDAV_SHOWGUIDE
  SHRGINFO shrg;
  DWORD rc;

  shrg.cbSize = sizeof( shrg );
  shrg.hwndClient = hWnd;
  shrg.ptDown.x = LOWORD( lParam );
  shrg.ptDown.y = HIWORD( lParam );
  shrg.dwFlags = SHRG_RETURNCMD;

  rc = SHRecognizeGesture( &shrg );

  if( rc == GN_CONTEXTMENU )
  {
    return TRUE;
  }
#endif
  return FALSE;
}

static void DoPopupMenu( HWND hWnd, LPARAM lParam )
{
  HMENU menu;
  POINT pt;

  menu = CreatePopupMenu( );

  AppendMenu( menu, MF_STRING | (g_imagn == ID_25 ? MF_CHECKED : 0), ID_25, L"25%" );
  AppendMenu( menu, MF_STRING | (g_imagn == ID_50 ? MF_CHECKED : 0), ID_50, L"50%" );
  AppendMenu( menu, MF_STRING | (g_imagn == ID_100 ? MF_CHECKED : 0), ID_100, L"100%" );
  AppendMenu( menu, MF_STRING | (g_imagn == ID_200 ? MF_CHECKED : 0), ID_200, L"200%" );
  AppendMenu( menu, MF_STRING | (g_imagn == ID_400 ? MF_CHECKED : 0), ID_400, L"400%" );
  AppendMenu( menu, MF_STRING | (g_imagn == ID_FIT ? MF_CHECKED : 0), ID_FIT, L"Fit to screen" );

  pt.x = LOWORD( lParam );
  pt.y = HIWORD( lParam );
  ClientToScreen( hWnd, &pt );

  TrackPopupMenu( menu, TPM_LEFTALIGN | TPM_TOPALIGN, pt.x, pt.y,
                  0, hWnd, NULL );

  DestroyMenu( menu );
}

static void ResizeScrollbars( HWND hWnd )
{
  SCROLLINFO si;
  PluckerImage *image;
  BOOL useVscroll = FALSE;
  BOOL useHscroll = FALSE;
  int vscrollWidth, hscrollHeight;
  RECT rect;
  unsigned int width, height;

  image = PluckerDBGetImage( g_db, g_record );

  vscrollWidth  = GetSystemMetrics( SM_CXVSCROLL );
  hscrollHeight = GetSystemMetrics( SM_CYHSCROLL );

  GetClientRect( hWnd, &rect );

  width = (unsigned int)(image->width * g_fmagn);
  height = (unsigned int)(image->height * g_fmagn);

  if( height > (unsigned int)(rect.bottom - rect.top) )
  {
    useVscroll = TRUE;
  }

  if( width > (unsigned int)(rect.right - rect.left) || 
      (useVscroll && width > (unsigned int)(rect.right - rect.left - vscrollWidth)) )
  {
    useHscroll = TRUE;
  }

  if( useHscroll && height > (unsigned int)(rect.bottom - rect.top - hscrollHeight) )
  {
    useVscroll = TRUE;
  }

  g_screenWidth = rect.right - rect.left;
  g_screenWidth -= useVscroll ? vscrollWidth : 0;

  g_screenHeight = rect.bottom - rect.top;
  g_screenHeight -= useHscroll ? hscrollHeight : 0;


  si.cbSize = sizeof( si );
  si.fMask = SIF_PAGE | SIF_RANGE | SIF_POS;
  si.nMin = 0;
  si.nPos = 0;
 
  if( g_vscroll )
  {
    DestroyWindow( g_vscroll );
    g_vscroll = NULL;
  }

  if( useVscroll )
  {
    g_vscroll = CreateWindow(L"scrollbar", L"", SBS_VERT | WS_CHILD,
                             rect.right - vscrollWidth, rect.top,
                             vscrollWidth, g_screenHeight, hWnd,
                             (HMENU)NULL, g_library->hInstance, NULL );
    
    si.nMax = (int)(image->height*g_fmagn);
    si.nPage = g_screenHeight;
    SetScrollInfo( g_vscroll, SB_CTL, &si, TRUE );
    ShowWindow( g_vscroll, SW_SHOW );
  }

  if( g_hscroll )
  {
    DestroyWindow( g_hscroll );
    g_hscroll = NULL;
  }

  if( useHscroll )
  {
    g_hscroll = CreateWindow(L"scrollbar", L"", SBS_HORZ | WS_CHILD,
                             rect.left, rect.bottom - hscrollHeight,
                             g_screenWidth, hscrollHeight, hWnd,
                            (HMENU)NULL, g_library->hInstance, NULL );
    
    si.nMax = (int)(image->width*g_fmagn);
    si.nPage = g_screenWidth;
    SetScrollInfo( g_hscroll, SB_CTL, &si, TRUE );
    ShowWindow( g_hscroll, SW_SHOW );
  }
  
}