/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: library interface
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <windows.h>
#include <commctrl.h>
#include <aygshell.h>
#include <stdlib.h>   /* abs() */
#include <string.h>   /* _wcsicmp() */
#include <stdlib.h>   /* qsort */

#include "vade_mecum.h"
#include "library-ui.h"
#include "resrc1.h"
#include "library.h"
#include "reader-ui.h"
#include "reader-settings.h"
#include "utilities.h"
#include "about.h"
#include "edit-categories.h"

#ifdef PDAV_SHOWGUIDE
#include "PDAVShowGuide.h"
#endif

#define ID_BOOK_LIST       ( 40001 )
#define ID_EXIT_READER     ( 40002 )
#define ID_READ_BOOK       ( 40003 )
#define ID_RENAME_BOOK     ( 40004 )
#define ID_DELETE_BOOK     ( 40005 )
#define ID_CATEGORIZE      ( 40006 )
#define ID_REFRESH_LIBRARY ( 40007 )
#define ID_CHANGE_CATEGORY ( 40008 )
#define ID_SETTINGS        ( 40009 )
#define ID_LIB_SHOW_ABOUT  ( 40010 )
#define ID_CATEGORY_LIST   ( 40011 )
#define ID_DELETE_METADATA ( 40012 )


/* reader column definitions */
#define COL_TITLE          (     1 )
#define COL_SIZE           (     2 )


#define LIBRARY_MARGIN        ( 1 )


static Library g_lib_data;
static HWND    g_category_list;
static HFONT   g_category_font;

static int CreateLibraryWindow( HWND hWnd );
static int DestroyLibraryWindow( HWND hWnd );
static void RepopulateListControl( HWND hWnd );
static void DoPopupMenu( HWND hWnd, WPARAM wParam, LPARAM lParam );
static void EditBookTitle( HWND hWnd );
static void OpenBook( HWND hWnd );
void OpenDocument( HWND hWnd, Document *document );
static void DoEditSettings( HWND hWnd );
static void DoRefreshLibrary( HWND hWnd );
static void ChangeLibrarySort( HWND hWnd, int column );
static void SortList( HWND hWnd );
static int ListSorter( LPARAM item1, LPARAM item2, LPARAM sortData );
static void CategorizeDocument( HWND hWnd );
static void ChangeCategory( HWND hWnd );
static void BuildSelectedCategoryList( TCHAR *list_text );
static int CategoryComparer( const void *arg1, const void *arg2 );
static void DeleteSelectedBook( HWND hWnd );
static void DeleteSelectedMetadata( HWND hWnd );

LRESULT CALLBACK LibraryWindowProc( HWND hWnd,
                                    UINT uMsg,
                                    WPARAM wParam,
                                    LPARAM lParam )
{
  LPNMHDR nmhdr;
  LV_DISPINFO *pdi;
  NMLISTVIEW *pnmv;
  HWND tmp;
  TCHAR *fname;
  PNMRGINFO mrgi;

  switch( uMsg )
  {
    case WM_CREATE:
      CreateLibraryWindow( hWnd );
      break;
    case WM_DESTROY:
      DestroyLibraryWindow( hWnd );
      break;
    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case ID_EXIT_READER:
          DestroyWindow( hWnd );
          break;
        case ID_RENAME_BOOK:
          EditBookTitle( hWnd );
          break;
        case ID_READ_BOOK:
          OpenBook( hWnd );
          break;
        case ID_SETTINGS:
          DoEditSettings( hWnd );
          break;
        case ID_LIB_SHOW_ABOUT:
          ShowAboutDialog( hWnd );
          break;
        case ID_REFRESH_LIBRARY:
          DoRefreshLibrary( hWnd );
          break;
        case ID_CATEGORIZE:
          CategorizeDocument( hWnd );
          break;
        case ID_CHANGE_CATEGORY:
        case ID_CATEGORY_LIST:
          ChangeCategory( hWnd );
          break;
        case ID_DELETE_BOOK:
          DeleteSelectedBook( hWnd );
          break;
        case ID_DELETE_METADATA:
          DeleteSelectedMetadata( hWnd );
          break;
      }
      break;
    case WM_REFRESH_LIBRARY:
      DoRefreshLibrary( hWnd );
      break;
    case WM_DO_POPUP_MENU:
      DoPopupMenu( hWnd, wParam, lParam );
      break;
    case WM_CTLCOLORSTATIC:
      /* make sure the category listing is colored blue */
      if( (HWND)lParam == g_category_list )
      {
        HDC dc = (HDC)wParam;

        SetTextColor( dc, RGB( 0, 0, 255 ) );

        return (LRESULT)GetStockObject( WHITE_BRUSH );
      }
      break;
    case WM_EXIT_VADE_MECUM:
      fname = (TCHAR*)lParam;
      if( fname != NULL )
        wcscpy( g_lib_data.current_document, fname );
      DestroyWindow( hWnd );
      break;
    case WM_NOTIFY:
      nmhdr = (LPNMHDR)lParam;
      switch( nmhdr->code )
      {
        case LVN_ENDLABELEDIT:
          pdi = (LV_DISPINFO*)lParam;
          wcscpy( g_lib_data.document_cache[ pdi->item.lParam ].title, pdi->item.pszText );
          ListView_SetItemText( nmhdr->hwndFrom, pdi->item.iItem, 0, g_lib_data.document_cache[ pdi->item.lParam ].title );
          break;
        case LVN_COLUMNCLICK:
          pnmv = (NMLISTVIEW*)lParam;
          ChangeLibrarySort( hWnd, pnmv->iSubItem+1 );
          break;
        case GN_CONTEXTMENU:
          if( nmhdr->idFrom == ID_BOOK_LIST )
          {
            mrgi = (PNMRGINFO)lParam;
            PostMessage( hWnd,
                         WM_DO_POPUP_MENU, 
                         ListView_GetSelectedCount( mrgi->hdr.hwndFrom ),
                         MAKELONG( mrgi->ptAction.x, mrgi->ptAction.y ) );
            break;
          }
          return DefWindowProc( hWnd, uMsg, wParam, lParam );
        case NM_CLICK:
          if( nmhdr->idFrom == ID_BOOK_LIST )
          {
            tmp = GetDlgItem( hWnd, ID_READ_BOOK );
            EnableWindow( tmp, ( ListView_GetSelectedCount( nmhdr->hwndFrom ) > 0 ) );
            break;
          }
          return DefWindowProc( hWnd, uMsg, wParam, lParam );
        default:
          return DefWindowProc( hWnd, uMsg, wParam, lParam );
      }
      break;
    default:
      return DefWindowProc( hWnd, uMsg, wParam, lParam );
  }

  return 0;
}


HWND LibraryWindowNew( HINSTANCE hInst )
{
  TCHAR window_class[ 64 ];
  TCHAR window_title[ 64 ];

  LoadString( hInst, IDS_LIBRARY_CLASS, window_class, 64 );
  LoadString( hInst, IDS_LIBRARY_TITLE, window_title, 64 );

  if( Library_open( &g_lib_data ) != 0 )
  {
    MessageBox( NULL, L"The library could not be opened.", L"Error", MB_OK | MB_ICONERROR );
    return NULL;
  }

#ifdef PDAV_SHOWGUIDE
  _tcscpy(g_lib_data.current_document,HOME_PLUCKER_DB_NAME);
  _tcscpy(g_lib_data.home_document,g_lib_data.current_document);
#endif

  if( Library_scanBooks( &g_lib_data ) != 0 )
  {
    MessageBox( NULL, L"There was an error scanning for e-books.", L"Error", MB_OK | MB_ICONERROR );
    Library_close( &g_lib_data );
    return NULL;
  }

  g_lib_data.hInstance = hInst;
  g_lib_data.hWnd = CreateWindow( window_class, window_title, WS_VISIBLE,
                                  CW_USEDEFAULT, CW_USEDEFAULT,
                                  CW_USEDEFAULT, CW_USEDEFAULT,
                                  NULL, (HMENU)NULL, hInst, 0 );
                       
  if( g_lib_data.hWnd == NULL )
  {
    Library_close( &g_lib_data );
  }
  else
  {
    if( g_lib_data.current_document[0] != '\0' )
    {
      Document *document;

      document = Library_findDocument( &g_lib_data, g_lib_data.current_document );
      if( document != NULL )
      {
        OpenDocument( g_lib_data.hWnd, document );
      }
#ifndef PDAV_SHOWGUIDE
      g_lib_data.current_document[0] = '\0';
#endif
    }
  }

  return g_lib_data.hWnd;
}


static int CreateLibraryWindow( HWND hWnd )
{
  HWND hList = NULL;
  RECT rect;
  int width;
  int height;
  DWORD lstyle;
  LV_COLUMN column;
  HWND tmp;
  LOGFONT font_def;

  GetClientRect( hWnd, &rect );
  width = rect.right - rect.left;
  height = rect.bottom - rect.top;

  CreateStaticControl( hWnd, g_lib_data.hInstance, 0, L"Displaying: ",
                       LIBRARY_MARGIN, LIBRARY_MARGIN, 60, 18,
                       WS_VISIBLE | SS_LEFTNOWORDWRAP );

  g_category_list = CreateStaticControl( hWnd, g_lib_data.hInstance, ID_CATEGORY_LIST, L"",
                                         LIBRARY_MARGIN + 60, LIBRARY_MARGIN, width - LIBRARY_MARGIN*2 - 60, 18,
                                         WS_VISIBLE | SS_LEFTNOWORDWRAP | SS_NOTIFY );

  /* we want the font for the category list to be blue, and underlined, reminiscent of a hyperlink on a
   * web page, so that the users will know they can tap it to change it. */

  g_category_font = (HFONT)SendMessage( g_category_list, WM_GETFONT, 0, 0 );
  if( g_category_font == NULL )
  {
    g_category_font = (HFONT)GetStockObject( SYSTEM_FONT );
  }

  GetObject( g_category_font, sizeof( font_def ), &font_def );
  font_def.lfUnderline = TRUE;
  g_category_font = CreateFontIndirect( &font_def );
  SendMessage( g_category_list, WM_SETFONT, (WPARAM)g_category_font, MAKELPARAM(TRUE,0) );

  hList = CreateWindowEx( 0, WC_LISTVIEW, NULL,
                          WS_VISIBLE | WS_BORDER | WS_CHILD |
                            LVS_REPORT | LVS_SINGLESEL | LVS_SHOWSELALWAYS | LVS_NOLABELWRAP | LVS_EDITLABELS,
                          LIBRARY_MARGIN, LIBRARY_MARGIN + 18, width-LIBRARY_MARGIN*2, height-75,
                          hWnd, (HMENU)ID_BOOK_LIST, g_lib_data.hInstance, NULL );

  column.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT;

  /* first column -- book title */
  column.fmt = LVCFMT_LEFT;
  column.cx = (width-(LIBRARY_MARGIN+1)*2-GetSystemMetrics( SM_CXVSCROLL )) * 2 / 3;
  column.pszText = L"Title";
  ListView_InsertColumn( hList, 0, &column );

  /* second column -- book size */
  column.fmt = LVCFMT_RIGHT;
  column.cx = (width-(LIBRARY_MARGIN+1)*2-GetSystemMetrics( SM_CXVSCROLL )) * 1 / 3;
  column.pszText = L"Size";
  ListView_InsertColumn( hList, 1, &column );

  /* set full-row selection */
  lstyle = ListView_GetExtendedListViewStyle( hList );
  lstyle |= LVS_EX_FULLROWSELECT;
  ListView_SetExtendedListViewStyle( hList, lstyle );

#define library_btn_width 58
#define library_btn_height 25

  CreateWindowEx( 0, L"button", L"E&xit", WS_VISIBLE | WS_CHILD, LIBRARY_MARGIN, LIBRARY_MARGIN+18+height-75+5, library_btn_width, library_btn_height,
                  hWnd, (HMENU)ID_EXIT_READER, g_lib_data.hInstance, NULL );

  tmp = CreateWindowEx( 0, L"button", L"&Open", WS_VISIBLE | WS_CHILD, LIBRARY_MARGIN+library_btn_width+1, LIBRARY_MARGIN+18+height-75+5, library_btn_width, library_btn_height,
                        hWnd, (HMENU)ID_READ_BOOK, g_lib_data.hInstance, NULL );
  EnableWindow( tmp, FALSE );

  CreateWindow( L"button", L"&About", WS_VISIBLE | WS_CHILD, LIBRARY_MARGIN+((library_btn_width+1)*2), LIBRARY_MARGIN+18+height-75+5, library_btn_width, library_btn_height,
                hWnd, (HMENU)ID_LIB_SHOW_ABOUT, g_lib_data.hInstance, NULL );

  CreateWindow( L"button", L"&Settings", WS_VISIBLE | WS_CHILD, LIBRARY_MARGIN+((library_btn_width+1)*3), LIBRARY_MARGIN+18+height-75+5, library_btn_width, library_btn_height,
                hWnd, (HMENU)ID_SETTINGS, g_lib_data.hInstance, NULL );

  RepopulateListControl( hWnd );

  return 0;
}


static int DestroyLibraryWindow( HWND hWnd )
{
  if( g_category_font != NULL ) DeleteObject( g_category_font );
  Library_close( &g_lib_data );
  PostQuitMessage( 0 );
  return 0;
}


static void RepopulateListControl( HWND hWnd )
{
  HWND hList;
  LV_ITEM item;
  int i;
  int idx = 0;
  TCHAR list_text[ 512 ];

  BuildSelectedCategoryList( list_text );
  SetWindowText( g_category_list, list_text );

  hList = GetDlgItem( hWnd, ID_BOOK_LIST );

  /* freeze redraw while we populate the booklist */
  SendMessage( hList, WM_SETREDRAW, FALSE, 0 );

  item.mask = LVIF_TEXT | LVIF_PARAM;

  ListView_DeleteAllItems( hList );
  for( i = 0; i < g_lib_data.document_cache_size; i++ )
  {
    if( ( g_lib_data.document_cache[ i ].verified ) &&
        ( ( g_lib_data.document_cache[ i ].categories & g_lib_data.current_categories ) != 0 ) )
    {
      TCHAR text[ 32 ];

      /*  insert the new row */
      item.iItem = 0;
      item.iSubItem = COL_TITLE-1;

      item.lParam = i;
      item.pszText = g_lib_data.document_cache[ i ].title;

      idx = ListView_InsertItem( hList, &item );

      /* set the text describing the size of the file */

      if( g_lib_data.document_cache[ i ].size < 1024 )
      {
        wsprintf( text, L"%dB", g_lib_data.document_cache[ i ].size );
      }
      else if( g_lib_data.document_cache[ i ].size < 1024*1024 )
      {
        wsprintf( text, L"%.1fK", g_lib_data.document_cache[ i ].size / ( 1024.0 ) );
      }
      else
      {
        wsprintf( text, L"%.1fM", g_lib_data.document_cache[ i ].size / ( 1024.0 * 1024.0 ) );
      }

      ListView_SetItemText( hList, idx, COL_SIZE-1, text );
    }
  }

  SortList( hWnd );

  /* re-enable redraw once we're done */
  SendMessage( hList, WM_SETREDRAW, TRUE, 0 );
}


static void DoPopupMenu( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
  HMENU menu;
  HWND tmp;

  menu = CreatePopupMenu();
  tmp = GetDlgItem( hWnd, ID_READ_BOOK );

  if( wParam > 0 )
  {
    AppendMenu( menu, MF_STRING, ID_READ_BOOK, L"&Read" );
    AppendMenu( menu, MF_STRING, ID_RENAME_BOOK, L"R&ename" );
    AppendMenu( menu, MF_STRING, ID_DELETE_BOOK, L"&Delete" );
    AppendMenu( menu, MF_STRING, ID_CATEGORIZE, L"&Categorize" );
    AppendMenu( menu, MF_STRING, ID_DELETE_METADATA, L"Delete &metadata" );
    AppendMenu( menu, MF_SEPARATOR, 0, NULL );
    EnableWindow( tmp, TRUE );
  }
  else
  {
    EnableWindow( tmp, FALSE );
  }

  AppendMenu( menu, MF_STRING, ID_REFRESH_LIBRARY, L"Re&fresh List" );
  AppendMenu( menu, MF_STRING, ID_CHANGE_CATEGORY, L"C&hange Category" );
  AppendMenu( menu, MF_STRING, ID_SETTINGS, L"&Edit Settings" );
  AppendMenu( menu, MF_SEPARATOR, 0, NULL );
  AppendMenu( menu, MF_STRING, ID_EXIT_READER, L"E&xit" );

  TrackPopupMenu( menu, TPM_LEFTALIGN | TPM_TOPALIGN, LOWORD( lParam ), HIWORD( lParam ),
                  0, hWnd, NULL );

  DestroyMenu( menu );
}


static void EditBookTitle( HWND hWnd )
{
  HWND hList;
  int  item;

  hList = GetDlgItem( hWnd, ID_BOOK_LIST );
  item = ListView_GetSelectionMark( hList );

  if( item >= 0 )
  {
    ListView_EditLabel( hList, item );
  }
}


static void OpenBook( HWND hWnd )
{
  HWND hList;
  int item;
  LV_ITEM data;

  hList = GetDlgItem( hWnd, ID_BOOK_LIST );
  item = ListView_GetSelectionMark( hList );

  if( item >= 0 )
  {
    data.mask = LVIF_PARAM;
    data.iItem = item;
    data.iSubItem = 0;

    ListView_GetItem( hList, &data );

    OpenDocument( hWnd, &g_lib_data.document_cache[ data.lParam ] );
  }
}


void OpenDocument( HWND hWnd, Document *document )
{
  RECT reader_area;
  HWND hReader;

  _tcscpy(g_lib_data.current_document, document->filename);
  GetClientRect( hWnd, &reader_area );
  hReader = OpenReaderWindow( &g_lib_data, document, &reader_area );

  ShowWindow( hReader, SW_SHOW );
  UpdateWindow( hReader );
}


static void DoEditSettings( HWND hWnd )
{
  UINT32 rc;

  rc = OpenReaderSettings( hWnd, &g_lib_data );

  if( ( rc & RS_BOOKPATH_CHANGED ) != 0 )
  {
    PostMessage( hWnd, WM_REFRESH_LIBRARY, 0, 0 );
  }
}


static void DoRefreshLibrary( HWND hWnd )
{
  Library_scanBooks( &g_lib_data );
  RepopulateListControl( hWnd );
}


static void ChangeLibrarySort( HWND hWnd, int column )
{
  if( abs( g_lib_data.reader_sort_column ) == column )
  {
    g_lib_data.reader_sort_column = -g_lib_data.reader_sort_column;
  }
  else
  {
    g_lib_data.reader_sort_column = column;
  }

  SortList( hWnd );
}


static void SortList( HWND hWnd )
{
  HWND hList;

  if( g_lib_data.reader_sort_column != 0 )
  {
    hList = GetDlgItem( hWnd, ID_BOOK_LIST );
    ListView_SortItems( hList, ListSorter, 0 );
  }
}


static int ListSorter( LPARAM item1, LPARAM item2, LPARAM sortData )
{
  int column;
  int ascending;
  int rc;

  column = abs( g_lib_data.reader_sort_column );
  ascending = ( g_lib_data.reader_sort_column > 0 );

  switch( column )
  {
    case COL_TITLE:
      rc = _wcsicmp( g_lib_data.document_cache[ item1 ].title,
                     g_lib_data.document_cache[ item2 ].title );
      break;
    case COL_SIZE:
      if( g_lib_data.document_cache[ item1 ].size < g_lib_data.document_cache[ item2 ].size )
        rc = -1;
      else if( g_lib_data.document_cache[ item1 ].size > g_lib_data.document_cache[ item2 ].size )
        rc = 1;
      else
        rc = 0;
      break;
    default:
      rc = 0;
  }

  if( !ascending ) rc = -rc;

  return rc;
}


static void CategorizeDocument( HWND hWnd )
{
  HWND hList;
  int item;
  LV_ITEM data;

  hList = GetDlgItem( hWnd, ID_BOOK_LIST );
  item = ListView_GetSelectionMark( hList );

  if( item >= 0 )
  {
    data.mask = LVIF_PARAM;
    data.iItem = item;
    data.iSubItem = 0;

    ListView_GetItem( hList, &data );

    if( ShowEditCategoriesDialog( hWnd, ECD_MODE_DOCUMENT, &g_lib_data, &g_lib_data.document_cache[ data.lParam ] ) )
    {
      RepopulateListControl( hWnd );
    }
  }
}


static void ChangeCategory( HWND hWnd )
{
  if( ShowEditCategoriesDialog( hWnd, ECD_MODE_FILTER, &g_lib_data, NULL ) )
  {
    RepopulateListControl( hWnd );
  }
}


static int CategoryComparer( const void *arg1, const void *arg2 )
{
  return _wcsicmp( *(TCHAR**)arg1, *(TCHAR**) arg2 );
}


static void BuildSelectedCategoryList( TCHAR *list_text )
{
  TCHAR *list[ MAX_CATEGORIES ];
  int idx;
  int element;
  int unselected_count;

  /* first, build the list of selected category names */
  element = 0;
  unselected_count = 0;
  for( idx = 0; idx < MAX_CATEGORIES; idx += 1 )
  {
    if( g_lib_data.categories[idx].name[0] != '\0' )
    {
      if( ( g_lib_data.current_categories & ( 1 << g_lib_data.categories[idx].id ) ) != 0 )
      {
        list[ element++ ] = g_lib_data.categories[idx].name;
      }
      else unselected_count += 1;
    }
  }

  if( unselected_count == 0 )
  {
    wcscpy( list_text, L"All categories" );
  }
  else
  {
    /* sort the list of selected categories */
    qsort( list, element, sizeof( TCHAR* ), CategoryComparer );

    /* build the comma-delimited list of names */
    wcscpy( list_text, L"" );
    for( idx = 0; idx < element; idx += 1 )
    {
      if( idx > 0 ) wcscat( list_text, L", " );
      wcscat( list_text, list[ idx ] );
    }
  }
}


static void DeleteSelectedBook( HWND hWnd )
{
  HWND hList;
  int item;
  LV_ITEM data;
  TCHAR msg[ 512 ];

  hList = GetDlgItem( hWnd, ID_BOOK_LIST );
  item = ListView_GetSelectionMark( hList );

  if( item >= 0 )
  {
    data.mask = LVIF_PARAM;
    data.iItem = item;
    data.iSubItem = 0;

    ListView_GetItem( hList, &data );

    wsprintf( msg, L"Are you sure you want to delete \"%s\"?", g_lib_data.document_cache[ data.lParam ].title );
    if( MessageBox( hWnd, msg, L"Delete Document?", MB_YESNO | MB_ICONQUESTION ) == IDYES )
    {
      int rc;

      rc = MessageBox( hWnd, L"Do you want to delete the metadata for this document as well?", L"Delete Metadata?", 
                       MB_YESNOCANCEL | MB_ICONQUESTION );

      if( rc == IDCANCEL ) return;

      Library_deleteDocument( &g_lib_data, &g_lib_data.document_cache[ data.lParam ], ( rc == IDYES ) );
      RepopulateListControl( hWnd );
    }
  }
}

static void DeleteSelectedMetadata( HWND hWnd )
{
  HWND hList;
  int item;
  LV_ITEM data;
  TCHAR filename[ 512 ];
  char  c_filename[ 512 ];

  hList = GetDlgItem( hWnd, ID_BOOK_LIST );
  item = ListView_GetSelectionMark( hList );

  if( item >= 0 )
  {
    data.mask = LVIF_PARAM;
    data.iItem = item;
    data.iSubItem = 0;

    ListView_GetItem( hList, &data );

    if( MessageBox( hWnd, L"Are you sure that you want to delete the metadata?", L"Delete Metadata?", 
                    MB_YESNO | MB_ICONQUESTION ) == IDYES )
    {
      Library_getMetadataFilename( &g_lib_data, &g_lib_data.document_cache[ data.lParam ], MF_CACHE, c_filename );
      mbstowcs( filename, c_filename, sizeof( c_filename ) );
      DeleteFile( filename );
      Library_getMetadataFilename( &g_lib_data, &g_lib_data.document_cache[ data.lParam ], MF_BOOKMARKS, c_filename );
      mbstowcs( filename, c_filename, sizeof( c_filename ) );
      DeleteFile( filename );
      Library_getMetadataFilename( &g_lib_data, &g_lib_data.document_cache[ data.lParam ], MF_HIGHLIGHTS, c_filename );
      mbstowcs( filename, c_filename, sizeof( c_filename ) );
      DeleteFile( filename );
      Library_getMetadataFilename( &g_lib_data, &g_lib_data.document_cache[ data.lParam ], MF_NOTES, c_filename );
      mbstowcs( filename, c_filename, sizeof( c_filename ) );
      DeleteFile( filename );
      Library_getMetadataFilename( &g_lib_data, &g_lib_data.document_cache[ data.lParam ], MF_HISTORY, c_filename );
      mbstowcs( filename, c_filename, sizeof( c_filename ) );
      DeleteFile( filename );
    }
  }
}