/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: library module
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <stdio.h>     /* FILE, etc. */
#include <string.h>    /* wcscpy */
#include "library.h"
#include "pagination.h"


/* rate at which the document cache will grow to accomodate new entries.  This
 * is the number of new entries that will be added to the cache each time the
 * cache has to grow. */
#define CACHE_GROW_RATE  ( 10 )

/* the default location of the vade mecum metadata files.  If this directory does
 * not exist, it will be automatically created. */
#define VM_DEFAULT_METADATA_DIR (L"\\My Documents\\AAPA2006_Data")

/* the vade mecum settings file */
static TCHAR vmSettingsFile[MAX_PATH];

/* reads the document cache */
static int static_read_document_cache( Library *library );

/* writes the document cache */
static int static_write_document_cache( Library *library );

/* scans each directory in the library->ebook_paths member */
static int static_scan_ebook_directories( Library *library );

/* scan a single directory for ebooks and subdirectories */
static int static_scan_directory( Library *library, TCHAR *directory );

/* initialize the given document with defaults loaded from the Plucker file itself.
 * Returns -1 if the plucker file could not be loaded, 0 otherwise. */
static int static_initialize_document( Library *library, Document *document );

/* add the document to the categories defined by default within it, adding the categories
 * that do not yet exist to the library (if room for them exists). */
static void static_populate_default_categories( Library *library,
                                                PluckerDB plkr,
                                                Document *document );

/* return the bit field that represents the category with the given name.  If the category
 * has not yet been defined, add it to the library and return its new bitfield.  If the
 * category cannot be added (not enough room, for instance), return 0. */
static UINT16 static_find_category_by_name( Library *library, TCHAR *name );

/* builds the name of a perfixed metadata filename for the given document.  Given a filename
 * of 'my ebook.pdb', this creates a filename of 'prefix-my ebook.vdm'. */
static void static_build_metadata_filename( Library *library, Document *document, TCHAR *prefix, TCHAR *cachename );

static void static_get_settings_file_path( TCHAR *path, int max_path_size );


int Library_open( Library *library )
{
  FILE *f;
  int i;

  /* always initialize the library to the default values, so that if the database format changes (ie, new
   * options are added, they will be initialized to their default values regardless of whether the value
   * exists in the file or not. */

  wcscpy( library->ebook_paths, VM_DEFAULT_METADATA_DIR );
  wcscpy( library->metadata_path, VM_DEFAULT_METADATA_DIR );
  library->version = VADE_MECUM_VERSION;
  memset( library->categories, 0, sizeof( library->categories ) );
  library->categories[ MAX_CATEGORIES-1 ].id = MAX_CATEGORIES-1;
  wcscpy( library->categories[ MAX_CATEGORIES-1 ].name, L"Unfiled" );
  library->current_categories = 0xFFFF; /* by default, show all categories */
#ifndef PDAV_SHOWGUIDE
  library->toolbar_pos = TOOLBAR_TOP;
#endif
  library->screen_division = SCREEN_DIV_NONE;
  memset( library->tap_actions, TAP_ACTION_NONE, sizeof( library->tap_actions ) );

  /* the default highlight color is black on yellow */
  library->foreground_selected = FALSE;
  library->foreground = RGB( 0, 0, 0 );
  library->background_selected = TRUE;
  library->background = RGB( 255, 255, 0 );

  /* set up the default color selections */
  library->fg_colors[0] = RGB(0,0,0);
  library->fg_colors[1] = RGB(128,0,0);
  library->fg_colors[2] = RGB(255,0,0);
  library->fg_colors[3] = RGB(0,0,128);
  library->fg_colors[4] = RGB(0,0,255);
  library->fg_colors[5] = RGB(100,0,100);
  library->fg_colors[6] = RGB(200,0,200);
  library->fg_colors[7] = RGB(0,128,0);

  library->bg_colors[0] = RGB(255,255,255);
  library->bg_colors[1] = RGB(255,255,128);
  library->bg_colors[2] = RGB(255,255,0);
  library->bg_colors[3] = RGB(255,128,128);
  library->bg_colors[4] = RGB(128,255,128);
  library->bg_colors[5] = RGB(128,255,255);
  library->bg_colors[6] = RGB(128,128,128);
  library->bg_colors[7] = RGB(255,128,255);

  Library_initializeDefaults( &library->options );

  library->use_hscroll = FALSE;
  library->use_vscroll = TRUE;
  library->pagination_cache_size = 4;
  library->autoscroll_delay = 5; /* 50 milliseconds */
  library->autoscroll_amount = 2;

  library->history_size = 10;
#ifdef PDAV_SHOWGUIDE
  library->history = HistoryList_new( library->history_size );
  library->home_document[0] = '\0';
#endif
  library->full_screen = FALSE;
  library->reader_sort_column = 1; /* sort on the name, ascending, by default */
  library->auto_follow_first_link = FALSE; /* don't follow the top link */
  library->auto_follow_last_link = FALSE; /* don't follow the last link */

  library->current_document[0] = '\0';

  library->progress_meter = PROGRESS_METER_NONE;
#ifndef PDAV_SHOWGUIDE
  library->toolbar_hidden = FALSE;
  library->toolbar_hotspot = HOTSPOT_NONE;
#endif
  library->first_paragraph_is_title = TRUE;

  for ( i = 0; i < NUMBER_RECENT_SEARCHES; i++ )
    *library->recent_searches[ i ].pattern = 0;


#ifndef PDAV_SHOWGUIDE
  static_get_settings_file_path( vmSettingsFile, MAX_PATH );

  f = _wfopen( vmSettingsFile, L"rb" );

  if( f != NULL )
  {
    UINT16 size;

    /* make sure we only read as much of the library as was originally written to disk,
     * in case additional DB fields are added. */

    fread( &size, sizeof( size ), 1, f );
    fseek( f, 0, SEEK_SET );

    if( fread( library, size, 1, f ) != 1 )
    {
      fclose( f );
      return -1;
    }
    
    fclose( f );
  }  
#endif


  library->size = sizeof( *library ) - sizeof( library->document_cache )
                                     - sizeof( library->document_cache_size )
                                     - sizeof( library->font_cache )
                                     - sizeof( library->hInstance )
                                     - sizeof( library->hWnd );

  library->document_cache = NULL;
  library->document_cache_size = 0;

  library->hWnd = NULL;
  library->hInstance = NULL;

  /* hold 10 fonts in memory, and up to 40 glyphs */
  library->font_cache = FontCache_new( 10, 40 );

  /* Check if metadata dir exists, if not create it */
  /* NB code below has no error checking */
  if ( 0xFFFFFFFF == GetFileAttributes( library->metadata_path ) )
    if ( CreateDirectory( library->metadata_path, NULL ) == 0)
      return -1; 

  return 0;
}


int Library_close( Library *library )
{
  int rc = 0;

  rc = Library_flush( library );

  if( library->document_cache != NULL ) free( library->document_cache );
  FontCache_delete( library->font_cache );

  memset( library, 0, sizeof( *library ) );

  return rc;
}


int Library_flush( Library *library )
{
  FILE *f;
  int rc = 0;

  f = _wfopen( vmSettingsFile, L"wb" );
  if( f != NULL )
  {
    if( fwrite( library, library->size, 1, f ) != 1 )
    {
      rc = -1;
    }

    fclose( f );
  }
  else
  {
    rc = -1;
  }

  rc = ( rc != 0 ? rc : static_write_document_cache( library ) );

  return rc;
}


int Library_scanBooks( Library *library )
{
  if( static_read_document_cache( library ) != 0 )
  {
    return -1;
  }

  return static_scan_ebook_directories( library );
}


void Library_getMetadataFilename( Library *library, Document *document, int type, char *filename )
{
  TCHAR temp[ MAX_PATH ];
  TCHAR *suffix;

  *filename = '\0';

  switch( type )
  {
    case MF_CACHE:      suffix = L"cache";  break;
    case MF_BOOKMARKS:  suffix = L"marks";  break;
    case MF_HIGHLIGHTS: suffix = L"hilite"; break;
    case MF_NOTES:      suffix = L"notes";  break;
    case MF_HISTORY:    suffix = L"hist";   break;
    default:
      return;
  }

  static_build_metadata_filename( library, document, suffix, temp );

  wcstombs( filename, temp, MAX_PATH );
}


FILE *Library_openMetadataFile( Library *library, Document *document, int type, int mode )
{
  char filename[ MAX_PATH ];
  char mode_str[10];
  FILE  *f;

  Library_getMetadataFilename( library, document, type, filename );

  mode_str[0] = '\0';

  if( ( mode & MF_READ ) && ( mode & MF_WRITE ) && ( mode & MF_TRUNCATE ) )
  {
    strcat( mode_str, "w+" );
  }
  else if( ( mode & MF_READ ) && ( mode & MF_WRITE ) )
  {
    strcat( mode_str, "r+" );
  }
  else if( mode & MF_WRITE )
  {
    strcat( mode_str, "w" );
  }
  else if( mode & MF_READ )
  {
    strcat( mode_str, "r" );
  }
  strcat( mode_str, "b" ); /* binary mode, always */

  f = fopen( filename, mode_str );

  if( f == NULL && ( mode & MF_CREATE ) ) /* perhaps the file didn't exist? */
  {
    strcpy( mode_str, "w+b" );
    f = fopen( filename, mode_str );
  }

  return f;
}


void Library_initializeDefaults( FormattingDefaults *options )
{
  wcscpy( options->normal_face,   L"Tahoma" );
  wcscpy( options->monotype_face, L"Courier New" );

  options->font_size = 11;
  options->interline_space = 0;
  options->interparagraph_space = 0;
  options->left_margin = 5;
  options->right_margin = 5;
  options->text_color = 0x000000;
  options->text_quality = DEFAULT_QUALITY;
  options->indent_size = 0;
  options->justification_rule = JUSTIFY_NORMAL;
  options->enable_soft_hyphens = TRUE;
}


Document *Library_findDocument( Library *library, TCHAR *filename )
{
  int i;

  for( i = 0; i < library->document_cache_size; i++ )
  {
    if( library->document_cache[i].verified )
    {
      if( wcscmp( library->document_cache[i].filename, filename ) == 0 )
      {
        return &library->document_cache[i];
      }
    }
  }

  return NULL;
}


CategoryDefinition* Library_findCategoryByName( Library *library, TCHAR *name )
{
  int i;
  int blank_pos = -1;

  for( i = 0; i < MAX_CATEGORIES; i++ )
  {
    if( library->categories[ i ].name[ 0 ] == 0 )
    {
      if( blank_pos < 0 ) blank_pos = i;
    }
    else if( wcscmp( library->categories[ i ].name, name ) == 0 )
    {
      return &library->categories[ i ];
    }
  }

  if( blank_pos < 0 ) return NULL;

  wcscpy( library->categories[ blank_pos ].name, name );
  library->categories[ blank_pos ].id = blank_pos;

  return &library->categories[ blank_pos ];
}


int Library_deleteDocument( Library *library, Document *document, int delete_metadata )
{
  TCHAR filename[ 512 ];
  char  c_filename[ 512 ];

  if( !DeleteFile( document->filename ) )
  {
    return 1;
  }

  if( delete_metadata )
  {
    Library_getMetadataFilename( library, document, MF_CACHE, c_filename );
    mbstowcs( filename, c_filename, sizeof( c_filename ) );
    DeleteFile( filename );
    Library_getMetadataFilename( library, document, MF_BOOKMARKS, c_filename );
    mbstowcs( filename, c_filename, sizeof( c_filename ) );
    DeleteFile( filename );
    Library_getMetadataFilename( library, document, MF_HIGHLIGHTS, c_filename );
    mbstowcs( filename, c_filename, sizeof( c_filename ) );
    DeleteFile( filename );
    Library_getMetadataFilename( library, document, MF_NOTES, c_filename );
    mbstowcs( filename, c_filename, sizeof( c_filename ) );
    DeleteFile( filename );
    Library_getMetadataFilename( library, document, MF_HISTORY, c_filename );
    mbstowcs( filename, c_filename, sizeof( c_filename ) );
    DeleteFile( filename );
  }

  document->verified = FALSE;
  document->categories = 0;
  document->filename[0] = '\0';
  document->title[0] = '\0';
  document->size = 0;

  return 0;
}


void Library_deleteCategory( Library *library, CategoryDefinition *def )
{
  int bitfld;
  int idx;

  /* don't allow "unfiled" to be deleted */
  if( def->id == MAX_CATEGORIES-1 ) return;

  bitfld = ( 1 << def->id );
  
  def->id = 0;
  def->name[0] = '\0'; /* delete the category */

  for( idx = 0; idx < library->document_cache_size; idx++ )
  {
    if( library->document_cache[ idx ].verified )
    {
      library->document_cache[ idx ].categories &= ~bitfld;
    }
  }
}


/* scanning ebook directories:
 * 
 * for each '*.vdm' in meta-data directory:
 *   add entry in cache
 *   make entry unverified
 *
 * for each directory in the ebook path:
 *   for each file in the directory
 *     if file is directory, scan directory recursively
 *     if file is ebook (.pdb)
 *       if file does not exist in cache
 *          create new meta-data entry in cache
 *       mark cache item verified
 *
 * for each unverified cache item
 *   delete meta-data files
 *   delete cache item
 */

static int static_read_document_cache( Library *library )
{
  FILE *f;
  TCHAR metadata[ MAX_PATH ];
  UINT16 version;
  UINT16 documents;
  UINT16 i;

  wcscpy( metadata, library->metadata_path );
  wcscat( metadata, L"\\metadata.vdm" );

  library->document_cache_size = 0;

  f = _wfopen( metadata, L"rb" );
  if( f == NULL ) return 0;  /* no document cache to read (first time through?) */

  if( fread( &version, sizeof( version ), 1, f ) != 1 )
  {
    fclose( f );
    return -1;
  }

  if( version != library->version ) /* different versions... what to do? */
  {
    fclose( f );
    return 0; /* pretend all is well */
  }

  if( fread( &documents, sizeof( documents ), 1, f ) != 1 )
  {
    fclose( f );
    return -1;
  }

  if( library->document_cache != NULL )
  {
    free( library->document_cache );
  }

  library->document_cache_size = documents;
  library->document_cache = (Document*)malloc( documents * sizeof( Document ) );

  if( fread( library->document_cache, sizeof( Document ), documents, f ) != documents )
  {
    free( library->document_cache );
    library->document_cache = NULL;
    library->document_cache_size = 0;
    fclose( f );
    return -1;
  }

  fclose( f );

  /* mark all documents invalid */
  for( i = 0; i < documents; i++ )
  {
    /* FIXME: delete the metadata for this document */
    library->document_cache[ i ].verified = FALSE;
  }

  return 0;
}


static int static_write_document_cache( Library *library )
{
  TCHAR filename[ MAX_PATH ];
  int rc = 0;
  FILE *f;

  if( library->document_cache != NULL )
  {
    wcscpy( filename, library->metadata_path );
    wcscat( filename, L"\\metadata.vdm" );

    f = _wfopen( filename, L"wb" );
    if( f != NULL )
    {
      if( fwrite( &library->version, sizeof( library->version ), 1, f ) != 1 )
      {
        rc = -1;
      }
      else if( fwrite( &library->document_cache_size, sizeof( library->document_cache_size ), 1, f ) != 1 )
      {
        rc = -1;
      }
      else if( fwrite( library->document_cache, sizeof( Document ), library->document_cache_size, f ) != library->document_cache_size )
      {
        rc = -1;
      }
      fclose( f );
    }
    else
    {
      rc = -1;
    }
  }

  return rc;
}


static int static_scan_ebook_directories( Library *library )
{
  TCHAR paths_copy[ MAX_EBOOK_PATHS_LEN ];
  TCHAR *path;
  int i;

  wcscpy( paths_copy, library->ebook_paths );
  path = wcstok( paths_copy, L";" );

  while( path != NULL )
  {
    static_scan_directory( library, path );
    path = wcstok( NULL, L";" );
  }

  /* eliminate all document-cache items that are still unverified */
  for( i = 0; i < library->document_cache_size; i++ )
  {
    if( ( !library->document_cache[ i ].verified ) && ( library->document_cache[ i ].filename[ 0 ] != 0 ) )
    {
      /* TODO: delete auxiliary metadata files */
      library->document_cache[ i ].filename[ 0 ] = 0;
    }
  }

  return 0;
}


static int static_scan_directory( Library *library, TCHAR *directory )
{
  HANDLE context;
  WIN32_FIND_DATA data;
  TCHAR pattern[ MAX_PATH ];
  TCHAR filename[ MAX_PATH ];
  int len;
  int i;
  int first_blank;
  BOOL hit;

  wcscpy( pattern, directory );
  wcscat( pattern, L"\\*.*" );

  context = FindFirstFile( pattern, &data );
  if( context == INVALID_HANDLE_VALUE )
  {
    /* I'm assuming that if there are NO files in the directory, or if the directory does
     * not exist, FindFirstFile will return INVALID_HANDLE_VALUE... */
    return 0;
  }

  do
  {
    wcscpy( filename, directory );
    wcscat( filename, L"\\" );
    wcscat( filename, data.cFileName );

    if( ( data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
    {
      static_scan_directory( library, filename );
    }
    else
    {
      len = wcslen( data.cFileName );
      if( wcsicmp( &(data.cFileName[len-4]), L".pdb" ) == 0 )
      {
        /* it's a pdb file, anyway... but is it a Plucker file? */
        /* look in the document cache to see if it already exists there.  If it doesn't,
         * add it. */

        hit = FALSE;
        first_blank = -1;
        for( i = 0; i < library->document_cache_size; i++ )
        {
          if( ( !library->document_cache[ i ].verified ) && 
              ( wcscmp( filename, library->document_cache[ i ].filename ) == 0 ) )
          {
            /* names match, but do the creation times? */
            if( memcmp( &library->document_cache[ i ].created, &data.ftCreationTime, sizeof( FILETIME ) ) == 0 )
            {
              hit = library->document_cache[ i ].verified = TRUE;
            }
            else
            {
              char  c_filename[ MAX_PATH ];
              TCHAR wcs_filename[ MAX_PATH ];

              /* delete the cache and history for this document.
               * NOTE: this may still cause problems if the new version of the document is significantly
               * different than the old, since the history, bookmarks, and highlights will all be off and
               * may cause rendering problems. */
              Library_getMetadataFilename( library, &(library->document_cache[i]), MF_CACHE, c_filename );
              mbstowcs( wcs_filename, c_filename, sizeof( c_filename ) );
              DeleteFile( wcs_filename );
              Library_getMetadataFilename( library, &(library->document_cache[i]), MF_HISTORY, c_filename );
              mbstowcs( wcs_filename, c_filename, sizeof( c_filename ) );
              DeleteFile( wcs_filename );

              /* force this slot to be reread for this document */
              first_blank = i;
            }
            break;
          }
          
          if( ( first_blank < 0 ) && ( library->document_cache[ i ].filename[ 0 ] == 0 ) )
          {
            first_blank = i;
          }
        }

        if( !hit ) /* cache missed -- add it to the cache */
        {
          if( first_blank < 0 ) /* no empty slots in cache */
          {
            library->document_cache_size += CACHE_GROW_RATE; /* grow the cache */
            library->document_cache = (Document*)realloc( library->document_cache,
                                                          sizeof( Document ) * library->document_cache_size );
            memset( &library->document_cache[ library->document_cache_size - CACHE_GROW_RATE ],
                    0, sizeof( Document ) * CACHE_GROW_RATE );
            first_blank = library->document_cache_size - CACHE_GROW_RATE;
          }

          library->document_cache[ first_blank ].verified = TRUE;
          wcscpy( library->document_cache[ first_blank ].filename, filename );
          library->document_cache[ first_blank ].categories = 0;
          library->document_cache[ first_blank ].created = data.ftCreationTime;

          if( static_initialize_document( library, &library->document_cache[ first_blank ] ) != 0 )
          {
            /* oops -- this isn't a plucker file, or the file couldn't be opened */
            library->document_cache[ first_blank ].verified = FALSE;
          }
        }
      }
    }
  }
  while( FindNextFile( context, &data ) );

  FindClose( context );

  return 0;
}


static int static_initialize_document( Library *library, Document *document )
{
  PluckerDB plkr;
  char flat_filename[ MAX_PATH ];
  char *title;
  FILE *f;

  wcstombs( flat_filename, document->filename, sizeof( flat_filename ) );
  plkr = OpenPluckerDBFile( flat_filename );
  if( plkr == NULL )
  {
    return -1; /* if the file cannot open, then we can't do anything */
  }

  static_populate_default_categories( library, plkr, document );
  
  /* get the title */

  title = PluckerDBGetMetadataItem( plkr, METADATA_ID_TITLE );
  if( title == NULL || *title == '\0' )
  {
    title = PluckerDBName( plkr );
    if( title == NULL || *title == '\0' )
    {
      title = flat_filename;
    }
  }

  mbstowcs( document->title, title, MAX_DOCUMENT_TITLE_LEN-1 );

  ClosePluckerDB( plkr );

  /* get the size of the file */
  f = _wfopen( document->filename, L"rb" );
  if( f == NULL )
  {
    return -1; /* this should never happen, but the file couldn't be opened. */
  }

  fseek( f, 0, SEEK_END );
  document->size = ftell( f );
  fclose( f );

  return 0;
}


static void static_populate_default_categories( Library *library,
                                                PluckerDB plkr,
                                                Document *document )
{
  int id;

  id = PluckerUIDOfReservedRecord( plkr, INDEX_ID_DEFAULT_CATEGORY );

  if( id >= 0 )
  {
    char *buffer;
    int   type;
    int   size;
    char *p;
    TCHAR temp_name[ MAX_CATEGORY_NAME_LEN ];

    buffer = PluckerDBGetRecord( plkr, id, &type );
    size = PluckerGetRecordSize( buffer ) - 8;

    p = buffer + 8; /* skip the record header */
    while( size > 0 )
    {
      int len;

      mbstowcs( temp_name, p, MAX_CATEGORY_NAME_LEN );
      document->categories |= static_find_category_by_name( library, temp_name );

      len = strlen( p ) + 1;
      size -= len;
      p += len;
    }
  }

  if( document->categories == 0 ) document->categories = 1 << ( MAX_CATEGORIES-1 );
}


static UINT16 static_find_category_by_name( Library *library, TCHAR *name )
{
  CategoryDefinition *def;

  def = Library_findCategoryByName( library, name );
  if( def == NULL ) return 0;

  return ( 1 << def->id );
}


static void static_build_metadata_filename( Library *library, Document *document, TCHAR *suffix, TCHAR *cachename )
{
  TCHAR *cache_pos;
  TCHAR *fname_pos;

  wcscpy( cachename, library->metadata_path );
  
  cache_pos = ( cachename + wcslen( cachename ) );

  if( wcslen( cachename ) > 0 )
  {
    if( *(cache_pos-1) != '\\' )
    {
      *cache_pos = '\\'; cache_pos++;
    }
  }

  fname_pos = wcsrchr( document->filename, '\\' );
  if( fname_pos != NULL ) fname_pos++;
  else fname_pos = document->filename;

  while( *fname_pos != '\0' && _wcsicmp( fname_pos, L".pdb" ) != 0 )
  {
    *cache_pos = *fname_pos;
    cache_pos++;
    fname_pos++;
  }

  wcscpy( cache_pos, L"-" );
  wcscat( cache_pos, suffix );
  wcscat( cache_pos, L".vdm" );
}

static void static_get_settings_file_path( TCHAR *path, int max_path_size )
{
  size_t i, len;

  GetModuleFileName( NULL, path, max_path_size );

  len = wcslen( path );
  for( i = len-1; i > 0 && path[i] != '\\'; --i);

  if( i == 0 )
    path[0] = '\\';

  path[i+1] = 0;
  wcscat( path, L"vade-mecum.dat");

}