/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: library object module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * The library object is the set of routines that define how the library
 * metadata and document lists are loaded, manipulated, and saved.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __LIBRARY_H__
#define __LIBRARY_H__

#include <stdio.h>
#include "vade_mecum.h"  /* vade mecum types */


/* metadata-file type constants */
#define MF_CACHE      ( 0x01 )
#define MF_BOOKMARKS  ( 0x02 )
#define MF_HIGHLIGHTS ( 0x03 )
#define MF_NOTES      ( 0x04 )
#define MF_HISTORY    ( 0x05 )

/* metadata-file mode constants */
#define MF_READ       ( 0x01 )
#define MF_WRITE      ( 0x02 )
#define MF_TRUNCATE   ( 0x04 )
#define MF_CREATE     ( 0x08 )


#ifdef __cplusplus
extern "C" {
#endif


/* Initializes the given library structure from the global meta-data stored on disk.
 * This global meta-data info is stored in \My Documents\vade-mecum.dat.  If the
 * file does not exist, the structure is initialized to acceptable defaults. If anything
 * fails, the function returns non-zero.
 *
 * This function always sets the hWnd and hInstance fields to NULL.
 */
int Library_open( Library *library );

/* Closes the given library structure by writing the information to the global
 * meta-data file (in \My Documents\vade-mecum.dat).  If the file cannot be
 * written to, the function fails by returning non-zero.
 */
int Library_close( Library *library );

/* Writes the library to disk, essentially saving the state of the library. */
int Library_flush( Library *library );

/* Scans the ebook paths for ebooks and populates the document cache. */
int Library_scanBooks( Library *library );

/* Opens the appropriate metadata file type for the given document.  If the file does not
 * exist and 'writeable' is false, the function returns NULL.  Also returns NULL if there
 * was an error opening the file, such as if the metadata directory does not exist. */
FILE *Library_openMetadataFile( Library *library, Document *document, int type, int mode );

/* Populates 'filename' with the path and name of the metadata file of the given type for the
 * given document. */
void Library_getMetadataFilename( Library *library, Document *document, int type, char *filename );

/* Initialize 'options' structure to the default formatting options. */
void Library_initializeDefaults( FormattingDefaults *options );

/* Returns the document object for the named file, or NULL if the file is not in the library */
Document *Library_findDocument( Library *library, TCHAR *filename );

/* return the CategoryDefinition that represents the category with the given name.  If the category
 * has not yet been defined, add it to the library and return its definition.  If the
 * category cannot be added (not enough room, for instance), return NULL. */
CategoryDefinition* Library_findCategoryByName( Library *library, TCHAR *name );

/* deletes the given document from the library, and from disk.  If 'delete_metadata' is non-zero,
 * all associated metadata for the document is deleted as well. If the file does not exist, returns
 * non-zero, otherwise returns 0. */
int Library_deleteDocument( Library *library, Document *document, int delete_metadata );

/* removes the given category from the list of categories, and removes it from the list of categories
 * for any document in the library. */
void Library_deleteCategory( Library *library, CategoryDefinition *def );

#ifdef __cplusplus
}
#endif


#endif /* __LIBRARY_H__ */
