/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: linked list module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * This defines a simple bi-directional linked list object.  The linked
 * list nodes contain a copy of the information, which should not contain
 * any data that requires special handling on destruction.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __LINKED_LIST_H__
#define __LINKED_LIST_H__

/* forward define the structure types */
typedef struct linked_list_t LinkedList;
typedef struct linked_list_node_t LinkedListNode;
typedef struct linked_list_iterator_t LinkedListIterator;


/* this is the function signature of the functions that the object uses to
 * compare two nodes.  The data area (NODEDATA, see below) is passed to the
 * function, which should compare the two and return and int (like strcmp).
 */
typedef int (*LinkedListCompareFn)( void*, void* );


struct linked_list_t
{
  LinkedListNode *list;          /* the list of nodes, or NULL if the list is empty */
  int data_size;                 /* # of bytes of extra data per node */
  LinkedListCompareFn compare;   /* comparison function, or NULL if the list is unsorted */
};

struct linked_list_node_t
{
  LinkedListNode *prev;          /* previous node in the list, or NULL for the head of the list */
  LinkedListNode *next;          /* next node in the list, or NULL for the tail of the list */
};


/* Returns the data area of the given node.  It assumes that 'node' is
 * of type LinkedListNode*.  If it is not, the results are undefined (and likely dangerous and
 * meaningless).
 */
#define NODEDATA(node, type)  ((type*)( node ? node+1 : NULL ))


/**
 * Creates a new linked list, where each node can hold data_size extra bytes of data and
 * the given 'compare' function is used to sort the list.  If 'compare' is NULL, sorting is
 * not done.
 */
LinkedList* LinkedList_new( int data_size, LinkedListCompareFn compare );

/**
 * Destroys the given linked list and all of its elements.
 */
void LinkedList_delete( LinkedList* list );

/**
 * Inserts a new LinkedListNode containing a copy of 'data' ('data_size' bytes worth of it).
 * The new node is inserted in sorted order (based on the 'compare' function), or at the end
 * if the compare function is NULL.
 */
void LinkedList_insert( LinkedList* list, void* data );

/**
 * Deletes (and destroys) the given node from the list.  The node should not be referenced
 * thereafter.
 */
void LinkedList_deleteNode( LinkedList* list, LinkedListNode *node );

/**
 * Finds the first node that contains the given data.  The data is compared using memcmp.
 */
LinkedListNode* LinkedList_findNode( LinkedList *list, void *data );

#endif /* __LINKED_LIST_H__ */
