/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: notes module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * This notes module defines a "note manager", which essentially
 * encapsulates the note database and provides note-specific
 * routines to access it.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __NOTES_H__
#define __NOTES_H__

#include <stdio.h>
#include <windows.h>

/* the opaque NoteManager type */
typedef void* NoteManager;

/* the data relevant to a note.  The dbid field is used internally as the record's primary key to update
 * or delete the record later. */
typedef struct
{
  TCHAR *note;         /* the text of the note */
  int    note_length;  /* the length of the note */

  int dbid;            /* "database id" -- the index of the record on disk.  READ-ONLY! */
} Note;


/**
 * Creates a new note manager.  The manager uses the named file as its persistant storage
 * location. If this function fails, it will return NULL.
 */
NoteManager NoteManager_create( char *filename );

/**
 * Destroys the given note manager and closes its associated file.  The note manager should not be
 * referenced after this is called.  This function does NOT destroy any Note objects that were created
 * by the getNote routine.
 */
void NoteManager_destroy( NoteManager nm );

/**
 * Retrives the given note, or NULL if the note does not exist.
 */
Note *NoteManager_getNote( NoteManager nm,
                           int id );

/**
 * Releases any memory held by the note object.
 */
void NoteManager_releaseNote( Note* note );

/**
 * Adds the given note to the persistant storage of the note manager.  The note must be a valid, null-terminated
 * string.  A copy is made of the note, so the calling routine may safely destroy or reuse the note buffer after
 * calling this routine.  This function returns the id of the new note object.
 */
int NoteManager_addNote( NoteManager nm,
                         TCHAR *note );

/**
 * Deletes the given note from the persistant storage of the manager.  This will also deallocate the note object,
 * as if NoteManager_releaseNote were called.
 */
void NoteManager_deleteNote( NoteManager nm,
                             Note *note );

/**
 * Deletes the given note from the persistant storage of the manager.
 */
void NoteManager_deleteNoteByID( NoteManager nm,
                                 int id );

/**
 * Updates the given note with the given text, and stores the updated object to the persistant
 * storage.
 */
void NoteManager_modifyNote( NoteManager nm,
                             Note *note,
                             TCHAR *text );

#endif /* __NOTES_H__ */
