/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: pagination module
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

/**
 * TODO: the whole "italic correction" thing needs more work, particularly since
 * the Win32 API's don't support any accurate way of measuring an italic character's
 * width. :(  I'm currently just estimating the italic correction as 1/6 of the width
 * of a lower-case 'm'.
 *
 * What needs to happen now, is: as each space is detected in pagination mode, look to
 * see if the font is italic.  If it is, see if there is room on the line for an italic
 * correction.  If not, break the line at the last available spot.  Then, when a line is
 * broken, see if the font is italic.  If it is, add the italic correction.  This should
 * (hopefully) make sure that italics at the end of a line don't bleed off into the
 * right margin.
 */

#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <dbgapi.h>

#include "pagination.h"

/**
 * For each value in a plucker database representing some pixel offset (like the margins, or
 * the distance between paragraphs), multiply it by PIXEL_SCALE.  This compensates for the
 * increased resolution on a Pocket PC, vs. a Palm device.  (TODO: this should probably be
 * rethought to support variable resolutions.)
 */
#define PIXEL_SCALE   ( 2 )

/**
 * This is the minimum distance between paragraphs when the alignment of a paragraph or the
 * paragraph before it is CENTER or RIGHT, or when the left-margin distance changes between
 * paragraphs.
 */
#define MINIMUM_SPECIAL_INTERPARA_SPACE  ( 6 )

#define SUPERSCRIPT   ( 0 )
#define SUBSCRIPT     ( 1 )

#define NBSP          ( 160 )

#define ROW_HEIGHT( row )        ( (row)->ascent + /*(row)->descent + */(row)->pre_padding )
#define TO_BASELINE( row )       ( (row)->ascent + (row)->pre_padding )
#define PADDING( row )           ( (row)->justified ? (row)->row_padding : 0 )

#define IS_SPACE( c )   ( ( c == ' ' ) || ( c == NBSP ) )

#define FONT_WEIGHT( opts, bold )  ( bold ? FW_BOLD : FW_NORMAL )

#define WRITE_OUT( x )  NKDbgPrintfW x

/* technically, InvertRect is not supported by the API (though it really does seem to
 * be).  Thus, this #define should do the same thing that InvertRect does. */

#define InvertRect( hdc, lprc ) \
    BitBlt( hdc, (lprc)->left, (lprc)->top, \
            (lprc)->right - (lprc)->left, \
            (lprc)->bottom - (lprc)->top, \
            NULL, 0, 0, DSTINVERT )



#define MINUS       (unsigned char)'-'
#define LSQUOTE     (unsigned char)'\x91'
#define RSQUOTE     (unsigned char)'\x92'
#define LQUOTE      (unsigned char)'\x93'
#define RQUOTE      (unsigned char)'\x94'
#define BULLET      (unsigned char)'\x95'
#define N_DASH      (unsigned char)'\x96'
#define M_DASH      (unsigned char)'\x97'
#define SOFT_HYPHEN (unsigned char)'\xAD'

#define IsSoftHyphen(ch, pinforecord) ((ch) == SOFT_HYPHEN && (pinforecord)->defaults.enable_soft_hyphens)

#define UNICODE_N_DASH     0x2013
#define UNICODE_M_DASH     0x2014
#define UNICODE_LS_QUOTE   0x2018
#define UNICODE_RS_QUOTE   0x2019
#define UNICODE_L_QUOTE    0x201C
#define UNICODE_R_QUOTE    0x201D
#define UNICODE_BULLET     0x2022

#define INITIAL_ATTRIBUTES_CAPACITY  ( 10 )

#define BREAK_ON_EOL         0
#define BREAK_ON_SPACE       1
#define BREAK_ON_DASH        2
#define BREAK_ON_SOFT_HYPHEN 3

#define ALIGN_LEFT    0
#define ALIGN_RIGHT   1
#define ALIGN_CENTER  2
#define ALIGN_JUSTIFY 3

#define REGULAR_FONT  0
#define H1_FONT       1
#define H2_FONT       2
#define H3_FONT       3
#define H4_FONT       4
#define H5_FONT       5
#define H6_FONT       6
#define BOLD_FONT     7
#define TT_FONT       8
#define SMALL_FONT    9
#define SUB_FONT     10
#define SUPER_FONT   11

#define IT_NONE       0
#define IT_CHAR       1
#define IT_IMAGE      2
#define IT_RULE       3
#define IT_NEWLINE    4
#define IT_ALIGN      5
#define IT_MARGINS    6
#define IT_DONEITAL   7  /* italics have just finished -- look for need for italic-correction */
#define IT_LINKSTART  8  /* link started */
#define IT_LINKEND    9  /* link ended */


typedef struct _pagination_info_record_ PaginationInfoRecord;
typedef struct _paginated_row_record_   PaginatedRowRecord;
typedef struct _attribute_info_         AttributeInfo;
typedef struct _pagination_state_       PaginationState;
typedef struct _processed_item_         ProcessedItem;

typedef unsigned long int  my_uint32;
typedef unsigned short int my_uint16;
typedef unsigned char      my_uint8;

#define PIR_WRITEABLE_DATA_LEN( pir )   (int)( (char*)&((pir)->defaults) - (char*)(pir) )

struct _pagination_info_record_
{
  /* this is the information in the info record that gets written to disk */
  my_uint32 height;
  my_uint32 row_count;
  my_uint16 page_width;
  my_uint16 attributes_size;

  /* the rest is either not written, or is written separately */
  FormattingDefaults defaults;

  PaginatedRowRecord *rows;
  PaginatedRowRecord *tail;

  FontCache font_cache;

  PluckerDB      document;
  my_uint16      doc_record;

  AttributeInfo *attributes;
  my_uint16 attributes_capacity;

  BOOL pagination_info_exists;

  RangeRecord *links;
  RangeRecord *links_tail;

  my_uint16 nr_large_insets;
  LargeInset* large_insets;
};


#define PRR_WRITEABLE_DATA_LEN( prr )  (int)( (char*)&((prr)->next) - (char*)(prr) )

#pragma pack(1)
struct _paginated_row_record_
{
  my_uint16 attributes  : 10;  /* index into the pagination-info record's attributes[] array */

  my_uint16 descent     :  5;  /* maximum descender value */
  my_uint16 ascent      : 11;  /* maximum ascender value of the row: descent + ascent + pre_padding = height */
  my_uint16 pre_padding :  5;  /* padding added to the top of the row */
  my_uint16 xofs        : 10;  /* x-offset of the row from the left side of canvas, in pixels */
  my_uint32 text_index  : 32;  /* text-index into the buffer of the record */
  my_uint16 text_length : 10;  /* # of bytes that comprise this line */
  my_uint16 length      : 11;  /* length (in pixels) of this row (excluding xofs and right margin) */
  my_uint16 row_padding : 10;  /* # of pixels needed to expand the row to the original row length */
  my_uint16 spaces      : 6;   /* # of space characters in the line */
  my_uint16 paragraph   : 1;   /* whether or not this is the start of a paragraph */
  my_uint16 justified   : 1;   /* whether or not this row was full-justified */

  PaginatedRowRecord *next;
};

struct _attribute_info_
{
  LOGFONT font_def;
  my_uint32 text_color  : 24;  /* 0xbbggrr -- text color */
  my_uint16 has_link    : 1;   /* whether or not the "link" fields are valid */
  my_uint16 link_record : 16;  /* the record in the current document to link to */
  my_uint16 link_para   : 16;
  signed int bline_ofs  : 5;   /* # of pixels offset from the baseline */
  my_uint16 fixed_width : 1;   /* whether or not the font is current monotype or not */
};

#pragma pack(8)
struct _pagination_state_
{
  HFONT current_font;
  AttributeInfo current_attributes;
  PaginationInfoRecord *info;
  LinkedList *highlights;
  HDC hdc;

  PluckerTextRecord *buffer;

  unsigned int position;
  BOOL rendering;
  int last_paragraph_alignment;
  int last_paragraph_left_margin;
  int last_paragraph_newline;
  unsigned short int *attribute_mru_cache;

  unsigned char space_width;
  unsigned char minus_width;
  unsigned char mdash_width;
  unsigned char ndash_width;

  TCHAR last_c;
  TCHAR current_c;

  LONG original_font_size; /* font size before baseline shifting, etc. */

  RangeRecord *active_link;

  HWND hWnd;
  HWND hProgressMeter;

  LinkedList *large_insets;
  int current_large_inset;
};

struct _processed_item_
{
  unsigned char item_type;
  union
  {
    TCHAR c;
    PluckerImage *image;
    struct
    {
      unsigned char height;
      unsigned short int width;
    } rule;
    unsigned char alignment;
    struct
    {
      signed char left;
      signed char right;
    } margins;
    signed char italic_correction;
  };
};


static void static_configure_font( PaginationState *state,
                                   int which_font );

static void static_load_font( PaginationState *state );

static void static_paginate_paragraph( PaginationState *state,
                                       PluckerParagraphDescriptor *paragraph );

static void static_process_item( PaginationState *state,
                                 int             *position,
                                 int              line_width,
                                 BOOL             processing,
                                 ProcessedItem   *item );

static void static_initialize_attributes( PaginationState *state,
                                          PaginatedRowRecord *row );

static unsigned short int static_get_attribute_info( PaginationState *state );

static PaginatedRowRecord *static_paginate_row( PaginationState *state,
                                                int             *position,
                                                int              end_of_paragraph,
                                                int              max_line_width,
                                                int             *break_type );

static void static_horizontal_rule( PaginationState *state,
                                    int x,
                                    int y,
                                    ProcessedItem *item,
                                    PaginatedRowRecord *row );


static void static_render_row( PaginationState *state,
                               PaginatedRowRecord *row,
                               int y );

static void static_dump_pagination( PaginationState *state );

static void static_clear_links( PaginationInfoRecord *record );

static void static_initialize_progress_meter( PaginationState *state,
                                              int max_size );

static void static_update_progress_meter( PaginationState *state,
                                          int pos );

static void static_destroy_progress_meter( PaginationState *state );

static void static_start_link( PaginationState *state,
                               PaginatedRowRecord *row,
                               int start_x, int start_y );

static void static_end_link( PaginationState *state,
                             PaginatedRowRecord *row,
                             int end_x, int end_y );

static void static_correct_link_offsets( PaginationState *state,
                                         RangeRecord *last_processed,
                                         PaginatedRowRecord *row,
                                         int dy );

static int  static_compute_extra_width( PaginationState *state,
                                        int *idx,
                                        int  leeway_per_space,
                                        int *extra_leeway,
                                        int *seen_nonblank,
                                        TCHAR *last_c );

static int static_compute_italic_correction( PaginationState *state,
                                             TCHAR c );


/* finds the LAST active highlight in the list, where an active highlight
 * is one that contains the given character index.  The *list is assumed to
 * contain no "expired" highlights that are not themselves contained in an
 * active highlight. */
static LinkedListNode *static_find_active_highlight( LinkedListNode *list,
                                                     int idx );

static void static_draw_hilite( PaginationState *state,
                                Highlight *h,
                                int x,
                                int w,
                                int y,
                                PaginatedRowRecord *row );


PaginationInfo PaginationInfo_new( FontCache font_cache )
{
  PaginationInfoRecord *record;

  record = (PaginationInfoRecord*)malloc( sizeof( PaginationInfoRecord ) );
  if( record == NULL )
  {
    return NULL;
  }
  memset( record, 0, sizeof( *record ) );

  record->font_cache = font_cache;
  record->attributes_capacity = INITIAL_ATTRIBUTES_CAPACITY;

  record->attributes = (AttributeInfo*)malloc( sizeof( AttributeInfo ) * record->attributes_capacity );
  if( record->attributes == NULL )
  {
    free( record );
    return NULL;
  }

  return (PaginationInfo)record;
}


void PaginationInfo_delete( PaginationInfo info )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;

  PaginationInfo_invalidate( info );
  static_clear_links( record );

  free( record->attributes );

  free( record->large_insets );

  memset( record, 0, sizeof( *record ) );
  free( record );
}


void PaginationInfo_invalidate( PaginationInfo info )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  PaginatedRowRecord *next;

  row = record->rows;
  while( row != NULL )
  {
    next = row->next;
    memset( row, 0, sizeof( *row ) );
    row = next;
  }

  record->rows = record->tail = NULL;
  record->height = record->row_count = 0;

  record->attributes_size = 0;
  record->links = NULL;
  record->links_tail = NULL;
  record->pagination_info_exists = FALSE;
}


unsigned int PaginationInfo_getHeight( PaginationInfo info )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;

  return record->height;
}


unsigned int PaginationInfo_getWidth( PaginationInfo info )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;

  return record->page_width;
}


void PaginationInfo_paginate( PaginationInfo info,
                              PluckerDB document,
                              int which_record,
                              HWND hWnd,
                              HDC hdc,
                              int width,
                              FormattingDefaults *defaults )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PluckerParagraphDescriptor paragraph_descriptor;
  PaginationState state;
  int paragraph;
  LinkedListNode *inset_node;
  int node_nr;

  if( record->pagination_info_exists )
  {
    /* if pagination has been requested, invalidate any prior
     * pagination information */

    PaginationInfo_invalidate( info );
  }

  state.buffer = PluckerDBGetTextRecord( document, which_record );

  if( state.buffer == NULL )
  {
    return;
  }

  state.hWnd = hWnd;
  state.attribute_mru_cache = (unsigned short*)malloc( sizeof( unsigned short ) * record->attributes_capacity );

  record->page_width = width;
  memcpy( &record->defaults, defaults, sizeof( FormattingDefaults ) );

  state.info = record;
  state.hdc = hdc;
  state.rendering = FALSE;
  state.info->document = document;
  state.info->doc_record = which_record;
  state.last_c = state.current_c = 0;
  state.active_link = NULL;

  state.large_insets = LinkedList_new( sizeof( LargeInset ), NULL );

  state.position = 0;

  static_initialize_progress_meter( &state, state.buffer->nr_paragraphs/3 );

  /* set up the "default" attribute set */
  memset( &state.current_attributes, 0, sizeof( state.current_attributes ) );

  state.current_attributes.text_color = defaults->text_color;
  static_configure_font( &state, REGULAR_FONT );                            
  static_load_font( &state );                                              

  /* clear the list of links */
  static_clear_links( record );

  state.last_paragraph_alignment = -1;
  state.last_paragraph_left_margin = state.info->defaults.left_margin;
  state.last_paragraph_newline = FALSE;

  /* paginate each paragraph */
  for( paragraph = 0; paragraph < state.buffer->nr_paragraphs; paragraph++ )
  {
    if( paragraph % 3 == 0 )
    {
      static_update_progress_meter( &state, paragraph/3 );                  
    }

    GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, state.buffer->paragraphs,
                              paragraph * sizeof( PluckerParagraphDescriptor ) );


    /* paginate the paragraph */
    static_paginate_paragraph( &state, &paragraph_descriptor );            

    state.position += paragraph_descriptor.size;
  }

  static_update_progress_meter( &state, state.buffer->nr_paragraphs/3 );                

  if( state.attribute_mru_cache != NULL ) free( state.attribute_mru_cache );
  record->pagination_info_exists = TRUE;

  static_destroy_progress_meter( &state );     
  

  /* Copy the inset data to a more manageable buffer */


  inset_node = state.large_insets->list;
  record->nr_large_insets = 0;

  while( inset_node != NULL )
  {
    inset_node = inset_node->next;
    record->nr_large_insets++;
  }
 
  if( record->nr_large_insets )
  {
    record->large_insets = (LargeInset*)malloc( sizeof( LargeInset ) * record->nr_large_insets );
    
    node_nr = 0;
    for( inset_node = state.large_insets->list; inset_node != NULL; inset_node = inset_node->next )
    {
      memcpy( &record->large_insets[ node_nr ], NODEDATA( inset_node, void* ), sizeof( LargeInset ) );
      node_nr++;
    }
  }
  else
  {
    record->large_insets = NULL;
  }

  LinkedList_delete( state.large_insets );

  /*if( state.info->doc_record == 293 ) static_dump_pagination( &state );*/
}


void PaginationInfo_render( PaginationInfo info,
                            HDC hdc,
                            int dy,
                            int page_ofs,   /* y-offset of top of the screen displaying the text */
                            int canvas_top, /* y-offset of top of the canvas to be drawn on (portion of screen) */
                            int height,     /* height of the portion of the screen to be drawn */
                            LinkedList *highlights,
                            HWND hWnd )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  int y;
  PaginationState state;

  if( !record->pagination_info_exists )
  {
    /* if a document has not yet been paginated, we cannot render it */
    return;
  }

  state.buffer = PluckerDBGetTextRecord( record->document, record->doc_record );

  state.info = record;
  state.hdc = hdc;
  state.active_link = NULL;
  state.last_c = state.current_c = 0;
  state.highlights = highlights;
  
  state.hWnd = hWnd;
  state.large_insets = NULL;
  state.current_large_inset = 0;

  /* find the first row to be rendered */
  y = page_ofs;

  row = state.info->rows;
  while( row != NULL && ROW_HEIGHT( row ) + y < dy+canvas_top )
  {
    y += ROW_HEIGHT( row );
    row = row->next;
  }

  while( state.current_large_inset <  state.info->nr_large_insets &&
         state.info->large_insets[ state.current_large_inset ].y_offs < y-page_ofs )
  {

    state.current_large_inset++;
  }

  if( row ) /* if we haven't scrolled past the last row */
  {
    /* normalize to the canvas */
    y -= dy;

    /* set the text drawing mode to TRANSPARENT */
    SetBkMode( state.hdc, TRANSPARENT );

    state.rendering = TRUE;

    static_initialize_attributes( &state, row );

    while( ( row != NULL ) && ( y < canvas_top+height ) )
    {
      static_render_row( &state, row, y );
      y += ROW_HEIGHT( row );
      row = row->next;
    }
  }
}

int PaginationInfo_lookupPoint( PaginationInfo info,
                                HDC hdc,
                                int page_ofs,
                                int delta_y,
                                int delta_x,
                                int x, int y,
                                LinkedList *highlights,
                                DocumentPoint *point )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  PaginationState state;
  int dy;
  int dx;
  int idx;
  int eol;
  GlyphInfo *glyph;
  int leeway_per_space = 0;
  int extra_leeway = 0;
  int line_width;
  int tmp;
  BOOL seen_non_space;
  TCHAR last_c = 0;
  ProcessedItem item;
  RangeRecord *active_link = NULL;
  RangeRecord *next_link = NULL;
  LinkedListNode *oldest_highlight;
  LinkedListNode *temp_highlight;

  point->has_link = FALSE;
  point->link_range = NULL;
  point->highlight = NULL;

  if( record == NULL )
    return -1;

  /* first, find the row that was clicked */
  dy = page_ofs;
  row = record->rows;

  next_link = record->links;
  if( ( next_link != NULL ) && ( row == (PaginatedRowRecord*)next_link->start_row ) )
  {
    active_link = next_link;

    /* find the link active at the start of the next row; active_link always points to the first
     * link on the current row, next_link always indicates the first link active on the next row. */
    
    while( next_link && next_link->end_row == (PaginatedRow)row ) next_link = next_link->next;
  }

  while( ( row != NULL ) && ( ( y + delta_y ) > ( dy + ROW_HEIGHT( row ) ) ) )
  {
    /* if the active-link (the one that started on or before this row) ends on this row (the row we're
     * done considering, and are about to skip), set the active-link to be NULL, meaning that there might
     * not be a link active on the next row. */

    if( active_link && active_link->end_row == (PaginatedRow)row )
    {
      active_link = NULL;
    }

    /* make sure that if next-link also ends on this row, we skip to the first link that does NOT
     * end on this row... */

    while( next_link && next_link->end_row == (PaginatedRow)row ) next_link = next_link->next;
    
    /* increment dy and move to the next row */

    dy += ROW_HEIGHT( row );
    row = row->next;

    /* if next_link is valid, and it starts (or ends) on this row, AND active-link is NULL, then set active-link
     * to be that link, and move next_link to the first link ending after this row. */

    if( active_link == NULL && next_link && ( next_link->start_row == (PaginatedRow)row || next_link->end_row == (PaginatedRow)row ) )
    {
      active_link = next_link;
      while( next_link && next_link->end_row == (PaginatedRow)row ) next_link = next_link->next;
    }
  }

  /* at this point, row is either NULL (the point clicked is after the last row of text), or it
   * is the row that was clicked.  Also, active_link is either NULL (meaning that there is no link
   * active on this row), or it represents a link that either started before this row and has spanned
   * to include this one, or it is a link that starts somewhere on this row. */

  point->yofs = dy;

  if( row == NULL )
  {
    /* the point requested is after the last row of text. */

    point->row = (PaginatedRow)record->tail;
    point->xofs = record->tail->xofs + record->tail->length + record->tail->spaces * PADDING( record->tail );
    point->text_index = record->tail->text_index + record->tail->text_length;
    point->char_width = 0;
    point->character = 0;

    return CLICK_AT_EOF;
  }

  point->row = (PaginatedRow)row;
  dx = row->xofs;

  /* if the row->xofs is already greater than the requested x-coordinate, then the requested point
   * is somewhere in the left margin. */

  if( dx > x )
  {
    point->xofs = row->xofs;
    point->text_index = row->text_index;
    point->char_width = 0;
    point->character = 0;

    return CLICK_IN_LEFT_MARGIN;
  }

  if( row->spaces > 0 )
  {
    leeway_per_space = PADDING( row ) / row->spaces;
    extra_leeway     = PADDING( row ) % row->spaces;
  }

  idx = row->text_index;
  eol = idx + row->text_length;
  line_width = row->length + row->row_padding;

  state.buffer = PluckerDBGetTextRecord( record->document, record->doc_record );

  state.hdc = hdc;
  state.info = record;
  state.position = row->text_index;
  state.rendering = FALSE;
  state.last_c = state.current_c = 0;

  state.large_insets = NULL;

  static_initialize_attributes( &state, row );
  static_load_font( &state );

  /* now we back up a minute and reconsider the link issue.  Set next_link to be active-link,
   * and set active-link to be NULL.  Then:
   * 
   * (a) if next-link is not NULL, then there is a link active somewhere on the current row.
   * (b) if that link's start row is not the current row, then it must have started on a previous
   *     row (since next-link is now what active-link was---the currently active link).  In this
   *     case, next_link is immediately active.
   * (c) if the start_x of the link is the start of the current row (dx was set earlier to row->xofs),
   *     then next_link is immediately active.
   *
   * if next-link is immediately active, set active-link to next-link and next-link to the next
   * link in the list.  In this way, active-link is ALWAYS the currently active link, and next-link
   * is ALWAYS the next link that could become active.  If active-link is NULL, then there is no link
   * active at the moment. */

  next_link = active_link;
  active_link = NULL;

  if( next_link )
  {
    /* a link is active if it started on a previous row, or if it starts at the beginning of this row */
    if( next_link->start_row != (PaginatedRow)row || next_link->start_x <= dx )
    {
      active_link = next_link;
      next_link = next_link->next;
    }
  }

  /* read the text, measuring offsets until we reach the requested x-coordinate. */

  oldest_highlight = highlights->list;
  seen_non_space = FALSE;

  while( idx < eol )
  {
    /* look to see if there are any active highlights at the current index */
    while( oldest_highlight != NULL && NODEDATA(oldest_highlight,Highlight)->end_pos < idx )
      oldest_highlight = oldest_highlight->next;

    temp_highlight = static_find_active_highlight( oldest_highlight, idx );
    point->highlight = ( temp_highlight == NULL ? NULL : NODEDATA(temp_highlight,Highlight) );

    /* tmp is the index in the text before the item at this index is processed (since
     * static_process_item increments idx after processing the item). */

    tmp = idx;
    static_process_item( &state, &idx, line_width, TRUE, &item );

    switch( item.item_type )
    {
      case IT_DONEITAL:
        if( dx + item.italic_correction > x )
        {
          point->xofs = dx;
          point->char_width = 0;
          point->text_index = tmp;
          point->has_link = state.current_attributes.has_link;
          point->link_record = state.current_attributes.link_record;
          point->link_paragraph = state.current_attributes.link_para;
          point->link_range = (RangeInfo)active_link;
          point->character = 0;
          return CLICK_ON_TEXT;
        }

        dx += item.italic_correction;
        break;

      case IT_IMAGE:
        if( dx + (signed)item.image->width > x )
        {
          point->xofs = dx;
          point->char_width = item.image->width;
          point->text_index = tmp;
          point->has_link = state.current_attributes.has_link;
          point->link_record = state.current_attributes.link_record;
          point->link_paragraph = state.current_attributes.link_para;
          point->link_range = (RangeInfo)active_link;
          point->character = 0;
          return CLICK_ON_TEXT;
        }

        dx += item.image->width;
        break;

      case IT_RULE:
        if( dx + item.rule.width > x )
        {
          point->xofs = dx;
          point->char_width = item.rule.width;
          point->text_index = tmp;
          point->has_link = state.current_attributes.has_link;
          point->link_record = state.current_attributes.link_record;
          point->link_paragraph = state.current_attributes.link_para;
          point->link_range = (RangeInfo)active_link;
          point->character = 0;
          return CLICK_ON_TEXT;
        }
        break;

      case IT_CHAR:
        /* skip initial spaces, just like the paginator and the renderer do */
        if( IS_SPACE( item.c ) && !seen_non_space && item.c != NBSP && !state.current_attributes.fixed_width )
        {
          continue;
        }

        seen_non_space = TRUE;

        if( IsSoftHyphen(item.c, record)/*item.c == SOFT_HYPHEN*/ && idx == eol ) /* end of line, and a soft hyphen */
        {
          item.c = MINUS;
        }
        else if( IsSoftHyphen(item.c, record)/*item.c == SOFT_HYPHEN*/ )
        {
          /* ignore discretionary hyphens that aren't at the line-break */
          continue;
        }

        /* m-dashes are padded on the right if there is any text following them before the
         * line break. */

        if( last_c == UNICODE_M_DASH )
        {
          dx += leeway_per_space;
          if( extra_leeway > 0 )
          {
            dx++;
            extra_leeway--;
          }
        }

        glyph = FontCache_getGlyph( record->font_cache,
                                    hdc,
                                    state.current_font,
                                    item.c );

        if( dx + glyph->width > x )
        {
          point->xofs = dx;
          point->char_width = glyph->width;
          point->text_index = tmp;
          point->has_link = state.current_attributes.has_link;
          point->link_record = state.current_attributes.link_record;
          point->link_paragraph = state.current_attributes.link_para;
          point->link_range = (RangeInfo)active_link;
          point->character = item.c;
          return CLICK_ON_TEXT;
        }

        dx += glyph->width;
        if( IS_SPACE( item.c ) || item.c == UNICODE_M_DASH )
        {
          dx += leeway_per_space;
          if( extra_leeway > 0 )
          {
            dx++;
            extra_leeway--;
          }
        }

        last_c = item.c;
        break;
    }

    /* check to see if the active-link has ended */
    if( active_link )
    {
      /* the active-link has ended if the ending row of the link is the current row, AND
       * the end-x coordinate of the link is less than or equal to the current x coordinate. */
      if( ( active_link->end_row == (PaginatedRow)row ) && ( active_link->end_x <= dx ) )
      {
        active_link = NULL;
      }
    }

    /* check to see if a new link has become active */
    if( ( active_link == NULL ) && ( next_link != NULL ) )
    {
      if( ( next_link->start_row == (PaginatedRow)row ) && ( next_link->start_x <= dx ) )
      {
        active_link = next_link;
        next_link = next_link->next;
      }
    }
  }

  point->highlight = NULL;
  point->xofs = dx;
  point->text_index = idx;
  point->char_width = 0;
  point->character = 0;

  return CLICK_IN_RIGHT_MARGIN;
}


void PaginationInfo_invertRange( PaginationInfo info,
                                 HDC hdc,
                                 int page_ofs,
                                 int yofs,
                                 RangeInfo range )
{
  RangeRecord *range_rec = (RangeRecord*)range;
  PaginatedRowRecord *current;
  RECT rect;
  int  y;

  if( range == NULL )
    return;

  current = (PaginatedRowRecord*)range_rec->start_row;

  if( current == NULL )
    return;

  y = range_rec->start_y - yofs + page_ofs;

  rect.left = range_rec->start_x;
  rect.top  = y + current->pre_padding;

  while( current != (PaginatedRowRecord*)range_rec->end_row )
  {
    rect.right = current->xofs + current->length + PADDING( current );
    rect.bottom = rect.top + current->ascent + current->descent;

    InvertRect( hdc, &rect );

    y += ROW_HEIGHT( current );
    current = current->next;

    rect.top = y + current->pre_padding;
    rect.left = current->xofs;
  }

  rect.right = range_rec->end_x;
  rect.bottom = rect.top + current->ascent + current->descent;

  InvertRect( hdc, &rect );
}


long int PaginationInfo_findParagraphY( PaginationInfo info,
                                        unsigned short int paragraph )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  PluckerTextRecord *buffer;
  PluckerParagraphDescriptor paragraph_descriptor;
  unsigned int position;
  unsigned int para;
  unsigned int para_start;
  long int y;

  buffer = PluckerDBGetTextRecord( record->document, record->doc_record );

  position = 0;

  GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, buffer->paragraphs, 0 );

  y = 0;
  para = 0;
  para_start = position;
  row = record->rows;
  while( row != NULL )
  {
    while( row->text_index >= para_start + paragraph_descriptor.size )
    {
      para_start += paragraph_descriptor.size;
      para++;
      GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, buffer->paragraphs,
                                para * sizeof( PluckerParagraphDescriptor ) );
    }

    if( para == paragraph ) return y;

    y += ROW_HEIGHT( row );
    row = row->next;
  }

  return -1;
}


long int PaginationInfo_deltaYToCharacterIndex( PaginationInfo info,
                                                int delta_y,
                                                int *y_offset )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  long int y;

  y = 0;
  row = record->rows;
  while( row != NULL )
  {
    if( y >= delta_y )
    {
      *y_offset = y - delta_y;
      return row->text_index;
    }

    y += ROW_HEIGHT( row );
    row = row->next;
  }

  /* return the beginning of the last row in the record */
  return record->tail->text_index;
}


long int PaginationInfo_characterIndexToDeltaY( PaginationInfo info,
                                                int index )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  long int y;

  y = 0;
  row = record->rows;
  while( row != NULL )
  {
    if( (int)row->text_index <= index && (int)row->text_index + row->text_length > index )
    {
      return y;
    }

    y += ROW_HEIGHT( row );
    row = row->next;
  }

  /* return y-value of the end of the record */
  return y;
}


void PaginationInfo_createRange( PaginationInfo info,
                                 DocumentPoint *start,
                                 DocumentPoint *end,
                                 RangeRecord *range )
{
  /* make sure the start & end points are in the right order.
   * NOTE: this may change the parameters as a side effect! */

  PaginationInfo_reorderPoints( info, start, end );

  range->next = NULL;
  range->start_row = start->row;
  range->start_x = start->xofs;
  range->start_y = start->yofs;
  range->end_row = end->row;
  range->end_x = end->xofs;
  range->end_y = end->yofs + ROW_HEIGHT( (PaginatedRowRecord*)end->row );
}


void PaginationInfo_reorderPoints( PaginationInfo info,
                                   DocumentPoint *start,
                                   DocumentPoint *end )
{
  DocumentPoint temp;

  if( PaginationInfo_comparePoints( info, start, end ) > 0 )
  {
    temp = *start;
    *start = *end;
    *end = temp;
  }
}


int PaginationInfo_comparePoints( PaginationInfo info,
                                  DocumentPoint *p1,
                                  DocumentPoint *p2 )
{
  if( p1->yofs < p2->yofs ) return -1;
  if( p1->yofs > p2->yofs ) return  1;

  if( p1->xofs < p2->xofs ) return -1;
  if( p1->xofs > p2->xofs ) return  1;

  return 0;
}


void PaginationInfo_selectionInit( PaginationInfo  info,
                                   SelectionState* selection,
                                   DocumentPoint*  anchor )
{
  selection->anchor = selection->start = selection->end = *anchor;
  memset( &selection->range, 0, sizeof( selection->range ) );
}


void PaginationInfo_extendSelection( PaginationInfo  info,
                                     HDC             hdc,
                                     int             yofs,
                                     SelectionState* selection,
                                     DocumentPoint*  point )
{
  if( selection->range.start_row == NULL )
  {
    /* if the selection is not yet defined, define it:
     *
     * if the point is before the anchor, make the point the selection's start.
     * if the point is after the anchor, make the point the selection's end.
     */

    if( PaginationInfo_comparePoints( info, point, &selection->anchor ) < 0 )
    {
      selection->start = *point;
    }
    else
    {
      selection->end = *point;
    }

    PaginationInfo_createRange( info, &selection->start, &selection->end, &selection->range );
    PaginationInfo_invertRange( info, hdc, 0, yofs, &selection->range );
  }
  else
  {
    RangeRecord temp_range;
    DocumentPoint p1, p2;

    /* if the selection has been defined, find how it has been extended (or contracted), invert that
     * portion, and then extend the selection end points */

    /* if the point is BEFORE selection_anchor:
     *   if selection_end is AFTER selection_anchor:
     *      - invert existing selection
     *      - selection_end = selection_anchor
     *   - invert from point to selection_start
     *   - create new selection from point -> selection_anchor
     * else # point is AFTER selection_anchor
     *   if selection_end is BEFORE selection_anchor
     *      - invert existing selection
     *      - selection_start = selection_anchor
     *    - invert from selection_end to point
     *    - create new selection from selection_anchor -> point
     */

    if( PaginationInfo_comparePoints( info, point, &selection->anchor ) < 0 )
    {
      if( PaginationInfo_comparePoints( info, &selection->anchor, &selection->end ) < 0 )
      {
        PaginationInfo_invertRange( info, hdc, 0, yofs, &selection->range );
        selection->end = selection->start = selection->anchor;
      }
      p1 = *point;
      p2 = selection->start;
      PaginationInfo_createRange( info, &p1, &p2, &temp_range );
      PaginationInfo_invertRange( info, hdc, 0, yofs, &temp_range );
      selection->start = *point;
      PaginationInfo_createRange( info, &selection->start, &selection->end, &selection->range );
    }
    else
    {
      if( PaginationInfo_comparePoints( info, &selection->anchor, &selection->start ) > 0 )
      {
        PaginationInfo_invertRange( info, hdc, 0, yofs, &selection->range );
        selection->end = selection->start = selection->anchor;
      }
      p1 = *point;
      p2 = selection->end;
      PaginationInfo_createRange( info, &p1, &p2, &temp_range );
      PaginationInfo_invertRange( info, hdc, 0, yofs, &temp_range );
      selection->end = *point;
      PaginationInfo_createRange( info, &selection->start, &selection->end, &selection->range );
    }
  }
}


void PaginationInfo_selectionDone( PaginationInfo  info,
                                   HDC             hdc,
                                   int             yofs,
                                   SelectionState *selection )
{
  if( selection->range.start_row != NULL )
  {
    PaginationInfo_invertRange( info, hdc, 0, yofs, &selection->range );
  }
  memset( selection, 0, sizeof( *selection ) );
}


int PaginationInfo_emptySelection( PaginationInfo info,
                                   SelectionState *selection )
{
  return ( PaginationInfo_comparePoints( info, &selection->start, &selection->end ) == 0 );
}


int PaginationInfo_serialize( PaginationInfo info,
                              FILE* f )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  RangeRecord *link;
  int link_count = 0;
  int i;

  /* write the pagination info record */
  if( fwrite( record, PIR_WRITEABLE_DATA_LEN( record ), 1, f ) != 1 )
  {
    return -1;
  }

  /* write the rows */
  row = record->rows;
  while( row != NULL )
  {
    if( fwrite( row, PRR_WRITEABLE_DATA_LEN( row ), 1, f ) != 1 )
    {
      return -1;
    }
    row = row->next;
  }

  /* write the attributes */
  if( fwrite( record->attributes, sizeof( AttributeInfo ), record->attributes_size, f ) != record->attributes_size )
  {
    return -1;
  }

  /* write the link list */
  link_count = 0;
  link = record->links;
  while( link != NULL )
  {
    link_count++;
    link = link->next;
  }

  if( fwrite( &link_count, sizeof( link_count ), 1, f ) != 1 )
  {
    return -1;
  }

  link = record->links;
  while( link != NULL )
  {
    fwrite( &link->start_x, sizeof( link->start_x ), 1, f );
    fwrite( &link->start_y, sizeof( link->start_y ), 1, f );
    fwrite( &link->end_x, sizeof( link->end_x ), 1, f );
    fwrite( &link->end_y, sizeof( link->end_y ), 1, f );
    link = link->next;
  }

  fwrite( &record->nr_large_insets, sizeof( record->nr_large_insets), 1, f );
  for( i = 0; i < record->nr_large_insets; i++ )
  {
    fwrite( &record->large_insets[ i ].type, sizeof( record->large_insets[ i ].type ), 1, f );
    fwrite( &record->large_insets[ i ].scrollpos, sizeof( record->large_insets[ i ].scrollpos ), 1, f );
    fwrite( &record->large_insets[ i ].maxpos, sizeof( record->large_insets[ i ].maxpos ), 1, f );
    fwrite( &record->large_insets[ i ].y_offs, sizeof( record->large_insets[ i ].y_offs ), 1, f );
  }
  return 0;
}


int PaginationInfo_unserialize( PaginationInfo  info,
                                PluckerDB       document,
                                int             record,
                                FormattingDefaults *defaults,
                                FILE*           f )
{
  PaginationInfoRecord *precord = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  PaginationState state;
  unsigned short start_x;
  unsigned short start_y;
  unsigned short end_x;
  unsigned short end_y;
  int cy;
  int height;
  int link_count;
  unsigned int i;

  PaginationInfo_invalidate( info );
  precord->document = document;
  precord->doc_record = record;

  memcpy( &precord->defaults, defaults, sizeof( *defaults ) );

  if( fread( precord, PIR_WRITEABLE_DATA_LEN( precord ), 1, f ) != 1 )
  {
    PaginationInfo_invalidate( info );
    return -1;
  }

  /* make sure the attributes array is big enough */
  if( precord->attributes_size > precord->attributes_capacity )
  {
    precord->attributes_capacity = precord->attributes_size;
    precord->attributes = (AttributeInfo*)realloc( precord->attributes,
                                  sizeof( AttributeInfo ) * precord->attributes_capacity );
  }

  /* read rows */
  for( i = 0; i < precord->row_count; i++ )
  {
    row = (PaginatedRowRecord*)malloc( sizeof( PaginatedRowRecord ) );
    if( fread( row, PRR_WRITEABLE_DATA_LEN( row ), 1, f ) != 1 )
    {
      free( row );
      PaginationInfo_invalidate( info );
      return -1;
    }
    row->next = NULL;

    if( precord->rows == NULL )
    {
      precord->rows = row;
    }
    else
    {
      precord->tail->next = row;
    }
    precord->tail = row;
  }

  /* read attributes */
  if( fread( precord->attributes, sizeof( AttributeInfo ), precord->attributes_size, f ) != precord->attributes_size )
  {
    PaginationInfo_invalidate( info );
    return -1;
  }

  /* read the links */
  if( fread( &link_count, sizeof( link_count ), 1, f ) != 1 )
  {
    PaginationInfo_invalidate( info );
    return -1;
  }

  /* the pagination state variable is really not used, except in the start/end link routines, and
   * even there, it is only used to keep track of the "active-link". */

  memset( &state, 0, sizeof( state ) );
  state.info = info;

  row = precord->rows;
  cy = 0;
  height = ROW_HEIGHT( row );

  while( link_count > 0 )
  {
    fread( &start_x, sizeof( start_x ), 1, f );
    fread( &start_y, sizeof( start_y ), 1, f );
    fread( &end_x, sizeof( end_x ), 1, f );
    fread( &end_y, sizeof( end_y ), 1, f );

    /* look for the row on which this link starts.  Because the links are listed in order, this
     * will also be on or after the current row. */
    while( start_y > cy )
    {
      cy += height;
      row = row->next;
      height = ROW_HEIGHT( row );
    }

    static_start_link( &state, row, start_x, cy );

    /* look for the row on which this link ends */
    while( end_y - height > cy )
    {
      cy += height;
      row = row->next;
      height = ROW_HEIGHT( row );
    }

    static_end_link( &state, row, end_x, cy );

    /* that's one less link to account for... */
    link_count--;
  }

  if( fread( &precord->nr_large_insets, sizeof( precord->nr_large_insets ), 1, f ) != 1 )
  {
    PaginationInfo_invalidate( info );
    return -1;
  }

  if( precord->nr_large_insets )
  {
    precord->large_insets = (LargeInset*)malloc( sizeof( LargeInset ) * precord->nr_large_insets );
    for( i = 0; i < (signed)precord->nr_large_insets; i++ )
    {
      fread( &precord->large_insets[ i ].type, sizeof( precord->large_insets[ i ].type ), 1, f );
      fread( &precord->large_insets[ i ].scrollpos, sizeof( precord->large_insets[ i ].scrollpos ), 1, f );
      fread( &precord->large_insets[ i ].maxpos, sizeof( precord->large_insets[ i ].maxpos ), 1, f );
      fread( &precord->large_insets[ i ].y_offs, sizeof( precord->large_insets[ i ].y_offs ), 1, f );
    }
  }
  else
  {
    precord->large_insets = NULL;
  }

  precord->pagination_info_exists = TRUE;

  return 0;
}


int PaginationInfo_copySelectionTo( PaginationInfo info,
                                    SelectionState *selection,
                                    TCHAR* buffer )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  PaginationState state;
  unsigned int idx;
  ProcessedItem item;
  TCHAR *ptr = buffer;
  TCHAR last_c = 0;

  idx = selection->start.text_index;
  row = (PaginatedRowRecord*)selection->start.row;

  state.buffer = PluckerDBGetTextRecord( record->document, record->doc_record );

  state.info = record;
  state.position = idx;
  state.rendering = FALSE;

  while( idx < selection->end.text_index )
  {
    static_process_item( &state, &idx, 0, FALSE, &item );

    switch( item.item_type )
    {
      case IT_RULE:
        *ptr = (UNICODE_M_DASH); ptr++;
        *ptr = (UNICODE_M_DASH); ptr++;
        break;

      case IT_NEWLINE:
        last_c = IT_NEWLINE;
        *ptr = (TCHAR)'\r'; ptr++;
        *ptr = (TCHAR)'\n'; ptr++;
        break;

      case IT_CHAR:
        last_c = item.c;
        if( IsSoftHyphen(item.c, record)/*item.c == SOFT_HYPHEN*/ ) break;
        *ptr = item.c;
        ptr++;
        break;
    }

    if( idx >= row->text_index + row->text_length )
    {
      row = row->next;
      idx = row->text_index;
      if( row->paragraph )
      {
        *ptr = (TCHAR)'\r'; ptr++;
        *ptr = (TCHAR)'\n'; ptr++;
        *ptr = (TCHAR)'\r'; ptr++;
        *ptr = (TCHAR)'\n'; ptr++;
      }
      else if( ( !IsSoftHyphen(last_c, record)/*last_c != SOFT_HYPHEN*/ ) && ( last_c != IT_NEWLINE ) && ( idx < selection->end.text_index ) )
      {
        *ptr = (TCHAR)' '; ptr++;
      }
    }
  }

  *ptr = 0;
  return 0;
}


int PaginationInfo_startsWithLink( PaginationInfo  info,
                                   int            *record,
                                   int            *paragraph )
{
  PaginationInfoRecord *infop = (PaginationInfoRecord*)info;
  PaginationState state;
  unsigned int idx;
  ProcessedItem item;

  state.buffer = PluckerDBGetTextRecord( infop->document, infop->doc_record );

  idx = infop->rows->text_index;

  state.info = infop;
  state.position = idx;
  state.rendering = FALSE;

  while( idx < state.info->rows->text_index + state.info->rows->text_length )
  {
    static_process_item( &state, &idx, 0, TRUE, &item );

    switch( item.item_type )
    {
      case IT_CHAR:
        if( item.c == ' ' || item.c == NBSP ) break;

      case IT_IMAGE:
      case IT_RULE:
        return 0;

      case IT_LINKSTART:
        *record = state.current_attributes.link_record;
        *paragraph = state.current_attributes.link_para;

        if( PluckerDBGetRecordType( infop->document, *record ) == -1 )
          return 0;

        return 1;
    }
  }

  return 0;
}


int PaginationInfo_endsWithLink( PaginationInfo  info,
                                 int            *record,
                                 int            *paragraph )
{
  PaginationInfoRecord *infop = (PaginationInfoRecord*)info;
  PaginationState state;
  unsigned int idx;
  ProcessedItem item;
  int in_link;
  int last_saw_link;

  state.buffer = PluckerDBGetTextRecord( infop->document, infop->doc_record );

  idx = infop->tail->text_index;

  state.info = infop;
  state.position = idx;
  state.rendering = FALSE;

  in_link = state.info->attributes[ state.info->tail->attributes ].has_link;
  *record = state.info->attributes[ state.info->tail->attributes ].link_record;
  *paragraph = state.info->attributes[ state.info->tail->attributes ].link_para;

  last_saw_link = in_link;

  while( idx < state.info->tail->text_index + state.info->tail->text_length )
  {
    static_process_item( &state, &idx, 0, TRUE, &item );

    switch( item.item_type )
    {
      case IT_CHAR:
        if( item.c == ' ' || item.c == NBSP ) break;

      case IT_IMAGE:
      case IT_RULE:
        if( !in_link ) last_saw_link = 0;
        break;
      
      case IT_LINKSTART:
        last_saw_link = 1;
        in_link = 1;
        *record = state.current_attributes.link_record;
        *paragraph = state.current_attributes.link_para;
        break;

      case IT_LINKEND:
        in_link = 0;
        break;
    }
  }

  if( last_saw_link )
    if( PluckerDBGetRecordType( infop->document, *record ) == -1 )
      return FALSE;

  return last_saw_link;
}

int PaginationInfo_getNrLargeInsets( PaginationInfo info )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;

  return record->nr_large_insets;
}

void PaginationInfo_updateLargeInsetScrolls( PaginationInfo info,
                                             HWND hWnd,
                                             int y_offset,
                                             const RECT *reader_bounds )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;
  PaginatedRowRecord *row;
  int y;
  int inset;
  int reader_height;
  int reader_width;
  HWND scroll;
  RECT *rc;
  int top;
  int bottom;
  int cyhscroll;

  reader_height = reader_bounds->bottom - reader_bounds->top;
  reader_width = reader_bounds->right - reader_bounds->left;

  cyhscroll = GetSystemMetrics( SM_CYHSCROLL );

  y = 0;
  row = record->rows;
  while( row != NULL && ROW_HEIGHT( row ) + y < y_offset )
  {
    y += ROW_HEIGHT( row );
    row = row->next;
  }

  inset = 0;
  while( inset < record->nr_large_insets && record->large_insets[ inset ].y_offs < y )
  {
    ShowWindow( GetDlgItem( hWnd, ID_LARGE_INSET_SCROLL+inset ), SW_HIDE );
    inset++;
  }

  while( inset < record->nr_large_insets && record->large_insets[ inset ].y_offs < y_offset+reader_height )
  {
    while( row != NULL && ROW_HEIGHT( row ) + y <= record->large_insets[ inset ].y_offs )
    {
      y += ROW_HEIGHT( row );
      row = row->next;
    }

    if( row == NULL )
      break;

    scroll = GetDlgItem( hWnd, ID_LARGE_INSET_SCROLL+inset );
    rc = (RECT*)GetWindowLong( scroll, GWL_USERDATA );

    top = y-y_offset+reader_bounds->top;
    bottom = top+ROW_HEIGHT( row );

    rc->top = ( top < reader_bounds->top ) ? reader_bounds->top : top;
    rc->bottom = ( bottom > reader_bounds->bottom) ? reader_bounds->bottom : bottom;    
    
    if( bottom > reader_bounds->bottom )
      if( rc->bottom - rc->top < cyhscroll )
        break;

    MoveWindow( scroll, 0, rc->bottom-cyhscroll, 
                reader_width, cyhscroll, FALSE );

    ShowWindow( scroll, SW_SHOW );
    inset++;
  }

  while( inset < record->nr_large_insets )
  {
    ShowWindow( GetDlgItem( hWnd, ID_LARGE_INSET_SCROLL+inset ), SW_HIDE );
    inset++;
  }
}

void PaginationInfo_updateLargeInset( PaginationInfo info, int id, int delta, BOOL rel )
{
  PaginationInfoRecord *record = (PaginationInfoRecord*)info;

  if( rel ) record->large_insets[ id ].scrollpos += delta;
  else record->large_insets[ id ].scrollpos = delta;

  if( record->large_insets[ id ].scrollpos < 0 ) record->large_insets[ id ].scrollpos = 0;
  if( record->large_insets[ id ].scrollpos > record->large_insets[ id ].maxpos ) 
    record->large_insets[ id ].scrollpos = record->large_insets[ id ].maxpos;
}

static void static_configure_font( PaginationState *state,
                                   int which_font )
{
  double  size_multiplier = 1.0;
  TCHAR   *font_face = state->info->defaults.normal_face;
  BOOL    bold = FALSE;
  BOOL    italic = FALSE;
  BOOL    underline;
  BOOL    strikeout;
  BOOL    fixed_width = FALSE;
  int     offset = 0;

  underline = state->current_attributes.font_def.lfUnderline;
  strikeout = state->current_attributes.font_def.lfStrikeOut;

  memset( &state->current_attributes.font_def, 0, sizeof( LOGFONT ) );

  switch( which_font )
  {
    case REGULAR_FONT: break;
    case H1_FONT: size_multiplier = 1.25; bold = TRUE; break;
    case H2_FONT: size_multiplier = 1.2; bold = TRUE; break;
    case H3_FONT: size_multiplier = 1.1; bold = TRUE; break;
    case H4_FONT: size_multiplier = 1.0; bold = TRUE; break;
    case H5_FONT: size_multiplier = 1.0; italic = TRUE; break;
    case H6_FONT: size_multiplier = 1.0; italic = TRUE; break;
    case BOLD_FONT: size_multiplier = 1.0; bold = TRUE; break;
    case TT_FONT: font_face = state->info->defaults.monotype_face; fixed_width = TRUE; break;
    case SMALL_FONT: size_multiplier = 0.85; break;
    case SUB_FONT: offset = -state->info->defaults.font_size/3; size_multiplier = 0.85; break;
    case SUPER_FONT: offset = state->info->defaults.font_size/3; size_multiplier = 0.85; break;
  }

  state->current_attributes.font_def.lfHeight = (int)( -state->info->defaults.font_size * size_multiplier );
  state->current_attributes.font_def.lfWeight = FONT_WEIGHT( &state->info->defaults, bold );
  state->current_attributes.font_def.lfItalic = italic;
  state->current_attributes.font_def.lfCharSet = ANSI_CHARSET; /* FIXME: is this right? */
  state->current_attributes.font_def.lfOutPrecision = OUT_DEFAULT_PRECIS;
  state->current_attributes.font_def.lfClipPrecision = CLIP_DEFAULT_PRECIS;
  state->current_attributes.font_def.lfQuality = state->info->defaults.text_quality;
  state->current_attributes.font_def.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
  wcscpy( state->current_attributes.font_def.lfFaceName, font_face );

  state->current_attributes.font_def.lfUnderline = underline;
  state->current_attributes.font_def.lfStrikeOut = strikeout;
  state->current_attributes.bline_ofs = offset;

  state->current_attributes.fixed_width = fixed_width;

  state->original_font_size = state->current_attributes.font_def.lfHeight;
}


static void static_load_font( PaginationState *state )
{
  SIZE extent;
  TCHAR c;

  state->current_font = FontCache_get( state->info->font_cache, &state->current_attributes.font_def );
  SelectObject( state->hdc, state->current_font );

  /* get the widths of some common characters */
  c = ' ';
  GetTextExtentPoint32( state->hdc, &c, 1, &extent );
  state->space_width = (unsigned char)extent.cx;

  c = '-';
  GetTextExtentPoint32( state->hdc, &c, 1, &extent );
  state->minus_width = (unsigned char)extent.cx;

  c = UNICODE_M_DASH;
  GetTextExtentPoint32( state->hdc, &c, 1, &extent );
  state->mdash_width = (unsigned char)extent.cx;

  c = UNICODE_N_DASH;
  GetTextExtentPoint32( state->hdc, &c, 1, &extent );
  state->ndash_width = (unsigned char)extent.cx;
}


static void static_paginate_paragraph( PaginationState *state,
                                       PluckerParagraphDescriptor *paragraph )
{
  PaginatedRowRecord *row;
  int           eop;
  int           newlines = 0;
  int           alignment = ALIGN_LEFT;
  BOOL          margins_set = FALSE;
  BOOL          alignment_set = FALSE;
  BOOL          paragraph_start = TRUE;
  BOOL          contains_hrule = FALSE;
  BOOL          saw_nonspace_after_newline = FALSE;
  int           left_margin = state->info->defaults.left_margin;
  int           right_margin = state->info->defaults.right_margin;
  int           actual_width;
  int           start;
  int           position;
  int           minimum_interpara_space;
  int           break_type;
  int           rows_created;
  ProcessedItem item;
  RangeRecord   *last_link_processed;

  state->current_attributes.text_color = state->info->defaults.text_color;
  state->current_attributes.has_link = FALSE;

  static_configure_font( state, REGULAR_FONT );
  static_load_font( state );

  position = state->position;
  start = position;
  eop = position + paragraph->size;

  /* force full justification, if requested */
  if( state->info->defaults.justification_rule == JUSTIFY_ALWAYS ) alignment = ALIGN_JUSTIFY;

  /* look for paragraph-level formatting items that occur during the paragraph */
  while( position < eop )
  {
    static_process_item( state, &position, 0, FALSE, &item );              
    switch( item.item_type )
    {
      case IT_MARGINS:
        if( !margins_set )
        {
          left_margin += PIXEL_SCALE * item.margins.left;
          right_margin += PIXEL_SCALE * item.margins.right;
          margins_set = TRUE;
        }
        break;

      case IT_ALIGN:
        if( !alignment_set )
        {
          alignment = item.alignment;

          if( alignment == ALIGN_LEFT && state->info->defaults.justification_rule == JUSTIFY_ALWAYS )
          {
            alignment = ALIGN_JUSTIFY;
          }
          else if( alignment == ALIGN_JUSTIFY && state->info->defaults.justification_rule == JUSTIFY_NEVER )
          {
            alignment = ALIGN_LEFT;
          }

          alignment_set = TRUE;
        }
        break;

      case IT_NEWLINE:
        newlines += 1;
        saw_nonspace_after_newline = FALSE;
        break;

      case IT_RULE:
        contains_hrule = TRUE;
        break;

      case IT_CHAR:
        if( !IS_SPACE( item.c ) )
        {
          saw_nonspace_after_newline = TRUE;
        }
        break;
    }
  }

  /* ignore any newlines that occur at the end of the paragraph */
  if( !saw_nonspace_after_newline && newlines > 0 )
  {
    newlines -= 1;
  }

  /* now, consider whether the situation is right to increase the pre_padding...  This will be the case
   * if any of the following conditions are true:
   *
   *   1) state->last_paragraph_alignment is ( LEFT or JUSTIFY ) AND alignment is ( RIGHT or CENTER )
   *   2) alignment is ( LEFT or JUSTIFY ) AND state->last_paragraph_alignment is ( RIGHT )
   *   3) state->last_paragraph_left_margin != left_margin
   *   4) state->last_paragraph_alignment is ( CENTER )
   *   5) state->last_paragraph_newline is ( TRUE )
   */

  minimum_interpara_space = 0;
  if( ( ( state->last_paragraph_alignment == ALIGN_LEFT ||
          state->last_paragraph_alignment == ALIGN_JUSTIFY ) &&
        ( alignment == ALIGN_RIGHT || alignment == ALIGN_CENTER ) ) ||
      ( state->last_paragraph_alignment == ALIGN_RIGHT &&
        ( alignment == ALIGN_LEFT || alignment == ALIGN_JUSTIFY ) ) ||
      state->last_paragraph_left_margin != left_margin ||
      state->last_paragraph_alignment == ALIGN_CENTER ||
      state->last_paragraph_newline == TRUE )
  {
    minimum_interpara_space = MINIMUM_SPECIAL_INTERPARA_SPACE;
  }

  /* now, format the paragraph */
  position = start;
  rows_created = 0;

  while( position < eop )
  {
    int temp_left_margin = left_margin;

    /* should we indent this line? let's check... */
    if( paragraph_start && ( newlines == 0 ) && !contains_hrule && alignment != ALIGN_CENTER && alignment != ALIGN_RIGHT )
    {
      temp_left_margin += state->info->defaults.indent_size;
    }

    actual_width = state->info->page_width - temp_left_margin - right_margin;
    last_link_processed = state->info->links_tail;

    row = static_paginate_row( state, &position, eop, actual_width, &break_type );  
    if( row == NULL ) continue;

    rows_created += 1;

    row->xofs += temp_left_margin;
    row->pre_padding = 0;

    if( paragraph_start )
    {
      row->pre_padding += ( paragraph->attributes & 0x07 ) * PIXEL_SCALE;

      if( (signed)row->pre_padding + state->info->defaults.interparagraph_space * PIXEL_SCALE > 0 )
      {
        row->pre_padding += state->info->defaults.interparagraph_space * PIXEL_SCALE;
      }
      else
      {
        row->pre_padding = 0;
      }

      row->paragraph = TRUE;
      paragraph_start = FALSE;
    }
    else
    {
      row->paragraph = FALSE;
    }

    if( (signed)row->pre_padding + state->info->defaults.interline_space * PIXEL_SCALE > 0 )
    {
      row->pre_padding += state->info->defaults.interline_space * PIXEL_SCALE;
    }
    else
    {
      row->pre_padding = 0;
    }

    if( row->pre_padding < minimum_interpara_space )
    {
      row->pre_padding = minimum_interpara_space;
    }

    minimum_interpara_space = 0;

    row->justified = 0;

    switch( alignment )
    {
      case ALIGN_LEFT:
        row->spaces = 0;
        break;

      case ALIGN_RIGHT:
        row->xofs += row->row_padding;
        row->spaces = 0;
        break;

      case ALIGN_CENTER:
        row->xofs += row->row_padding / 2;
        row->spaces = 0;
        break;

      case ALIGN_JUSTIFY:
        row->justified = 1;
        if( break_type == BREAK_ON_EOL )
        {
          row->spaces = 0;
          row->justified = 0;
        }
        break;
    }

    /* fix the link offsets for this row, now that we know the true row spacing, etc. */

    static_correct_link_offsets( state, last_link_processed, row, state->info->height );      

    /* append the row to the pagination info */

    state->info->row_count++;
    state->info->height += ROW_HEIGHT( row );

    if( state->info->rows == NULL )
    {
      state->info->rows = row;
    }
    else
    {
      state->info->tail->next = row;
    }

    state->info->tail = row;
    state->info->tail = row;
  }

  if( rows_created > 0 )
  {
    state->last_paragraph_alignment = alignment;
    state->last_paragraph_left_margin = left_margin;
    state->last_paragraph_newline = ( newlines > 0 );
  }
}


static void static_process_item( PaginationState *state,
                                 int             *position,
                                 int              line_width,
                                 BOOL             processing,
                                 ProcessedItem   *item )
{
  int i;
  int j;
  int k;
  unsigned char code;
  unsigned char size;
  unsigned short int s;
  AttributeInfo *attributes = &(state->current_attributes);

  if( item != NULL )
  {
    item->item_type = IT_NONE;
  }

  if( state->buffer->text[ *position ] != '\0' )
  {
    item->item_type = IT_CHAR;
    item->c = state->buffer->text[ *position ];
    *position += 1;

    switch( item->c )
    {
      case M_DASH:  item->c = UNICODE_M_DASH; break;
      case N_DASH:  item->c = UNICODE_N_DASH; break;
      case BULLET:  item->c = UNICODE_BULLET; break;
      case LQUOTE:  item->c = UNICODE_L_QUOTE; break;
      case RQUOTE:  item->c = UNICODE_R_QUOTE; break;
      case LSQUOTE: item->c = UNICODE_LS_QUOTE; break;
      case RSQUOTE: item->c = UNICODE_RS_QUOTE; break;
    }

    if( processing )
    {
      state->last_c = state->current_c;
      state->current_c = item->c;
    }
  }
  else
  {
    *position += 1;

    code = PLKR_FC_CODE( state->buffer->text[ *position ] );
    size = ( state->buffer->text[ *position ] & 0x07 );

    *position += 1;

    switch( code )
    {
      case PLKR_FC_CODE( 0x0A ): /* page-link begins */
        if( size > 0 )
        {
          item->item_type = IT_LINKSTART;

          if( processing )
          {
            attributes->has_link = TRUE;
            attributes->font_def.lfUnderline = TRUE;
            attributes->text_color = 0x00FF0000;
            attributes->link_record = ( (unsigned short)state->buffer->text[ *position ] << 8 ) + state->buffer->text[ *position+1 ];
            attributes->link_para = 0;
          }
          *position += 2;
          if( size > 2 )
          {
            if( processing )
            {
              attributes->link_para = ( (unsigned short)state->buffer->text[ *position ] << 8 ) + state->buffer->text[ *position+1 ];
            }
            *position += 2;
          }
        }
        else
        {
          item->item_type = IT_LINKEND;

          if( processing )
          {
            attributes->has_link = FALSE;
            attributes->font_def.lfUnderline = FALSE;
            attributes->text_color = state->info->defaults.text_color;
          }
        }
        if( processing ) static_load_font( state );
        break;

      case PLKR_FC_CODE( 0x11 ): /* set font */
        if( processing )
        {
          static_configure_font( state, state->buffer->text[ *position ] );
          static_load_font( state );
        }
        *position += 1;
        break;

      case PLKR_FC_CODE( 0x1A ): /* embedded image */
        i = ( (unsigned short)state->buffer->text[ *position ] << 8 ) + state->buffer->text[ *position+1 ];
        *position += 2;

        item->item_type = IT_IMAGE;
        if( processing )
        {
          /* make sure the current record remains in the cache */
          PluckerDBTouchItem( state->info->document, state->info->doc_record );

          /* get the image record */
          
          item->image = PluckerDBGetImage( state->info->document, i );

          if( state->large_insets && item->image->width > (unsigned)line_width )
          {
            LargeInset inset;
            inset.scrollpos = 0;
            inset.type = IT_IMAGE;
            inset.maxpos = item->image->width - state->info->page_width;
            inset.y_offs = state->info->height;
            LinkedList_insert( state->large_insets, &inset );
          }          
          
        }
        break;

      case PLKR_FC_CODE( 0x5C ): /* multiple embedded image */
        i = ( (unsigned short)state->buffer->text[ *position ] << 8 ) + state->buffer->text[ *position+1 ];
        j = ( (unsigned short)state->buffer->text[ *position+2 ] << 8 ) + state->buffer->text[ *position+3 ];

        *position += 4;
        item->item_type = IT_IMAGE;

        if( processing )
        {
          /* make sure the current record remains in the cache */
          PluckerDBTouchItem( state->info->document, state->info->doc_record );
          /* get the image record */          
          item->image = PluckerDBGetImage( state->info->document, j );

          if( state->large_insets && item->image->width > (unsigned)line_width )
          {
            LargeInset inset;
            inset.scrollpos = 0;
            inset.type = IT_IMAGE;
            inset.maxpos = item->image->width - state->info->page_width;
            inset.y_offs = state->info->height;
            LinkedList_insert( state->large_insets, &inset );
          }
          /*if( !attributes->has_link )
          {
            attributes->link_record = i;
            attributes->link_para = 0;
          }*/
          attributes->link_record = i;
          attributes->link_para = 0;
          attributes->has_link = TRUE;
        }
        break;

      case PLKR_FC_CODE( 0x33 ): /* horizontal rule */
        item->item_type = IT_RULE;
        if( processing )
        {
          item->rule.height = state->buffer->text[ (*position)++ ];
          i                 = state->buffer->text[ (*position)++ ];
          j                 = state->buffer->text[ (*position)++ ];

          if( j < 1 ) j = 100; /* the python version apparently makes j=0 for page-width 100% rules */

          if( i > 0 )
          {
            item->rule.width = i;
          }
          else
          {
            item->rule.width = line_width * j / 100;
          }
        }
        else
        {
          *position += 3;
        }
        break;

      case PLKR_FC_CODE( 0x38 ): /* new line */
        item->item_type = IT_NEWLINE;
        break;
      
      case PLKR_FC_CODE( 0x40 ): /* italic text begins */
        if( processing )
        { 
          attributes->font_def.lfItalic = TRUE;
          static_load_font( state );
        }
        break;

      case PLKR_FC_CODE( 0x48 ): /* italic text ends */
        item->item_type = IT_DONEITAL;
        if( processing ) 
        {
          if( state->last_c != 0 )
          {
            item->italic_correction = static_compute_italic_correction( state, state->last_c );
          }
          attributes->font_def.lfItalic = FALSE;
          static_load_font( state );
        }
        break;

      case PLKR_FC_CODE( 0x53 ): /* set text color */
        if( processing )
        {
          i = state->buffer->text[ (*position)++ ]; /* red */
          j = state->buffer->text[ (*position)++ ]; /* green */
          k = state->buffer->text[ (*position)++ ]; /* blue */
          /* windows COLORREF values are 0x00bbggrr */
          attributes->text_color = ( k << 16 ) + ( j << 8 ) + i;
        }
        else
        {
          *position += 3;
        }
        break;

      case PLKR_FC_CODE( 0x60 ): /* underline text begins */
        if( processing ) 
        {
          attributes->font_def.lfUnderline = TRUE;
          static_load_font( state );
        }
        break;

      case PLKR_FC_CODE( 0x68 ): /* underline text ends */
        if( processing ) 
        {
          attributes->font_def.lfUnderline = FALSE;
          static_load_font( state );
        }
        break;

      case PLKR_FC_CODE( 0x70 ): /* strike-through text begins */
        if( processing ) 
        {
          attributes->font_def.lfStrikeOut = TRUE;
          static_load_font( state );
        }
        break;

      case PLKR_FC_CODE( 0x78 ): /* strike-through text ends */
        if( processing ) 
        {
          attributes->font_def.lfStrikeOut = FALSE;
          static_load_font( state );
        }
        break;

      case PLKR_FC_CODE( 0x83 ): /* Unicode character */
        i = state->buffer->text[ *position ];
        item->item_type = IT_CHAR;

        *position += 1;
        if( size == 3 )
        {
          memcpy( &s, &state->buffer->text[ *position ], 2 );
          item->c = FIX_BYTE_ORDER_S( s );
          *position += 2;
        }
        else
        {
          memcpy( &item->c, &state->buffer->text[ *position ], 4 );
          item->c = FIX_BYTE_ORDER_L( item->c );
          *position += 4;
        }

        *position += i; /* skip over the alternate text */
        break;

      case PLKR_FC_CODE( 0x22 ): /* set margin --- should only be paragraph-level */
        item->item_type = IT_MARGINS;
        item->margins.left = (signed char)state->buffer->text[ (*position)++ ];
        item->margins.right = (signed char)state->buffer->text[ (*position)++ ];
        break;

      case PLKR_FC_CODE( 0x29 ): /* alignment of text -- should only be paragraph-level */
        item->item_type = IT_ALIGN;
        item->alignment = state->buffer->text[ (*position)++ ];
        break;

      default:
        *position += size;
        break;
    }
  }
}


static PaginatedRowRecord *static_paginate_row( PaginationState *state,
                                                int             *position,
                                                int              end_of_paragraph,
                                                int              max_line_width,
                                                int             *break_type )
{
  PaginatedRowRecord *row;
  int           line_width = 0;
  int           max_ascent = 0;
  int           max_descent = 0;
  int           spaces = 0;
  int           new_line = 0;
  int           tmp_height = 0;
  int           use_image_as_link=0;

  int           saved_line_width = -1;
  int           saved_max_ascent = -1;
  int           saved_max_descent = -1;
  int           saved_spaces = -1;
  int           saved_position = -1;
  int           last_c = 0;

  BOOL          seen_space = FALSE;
  BOOL          seen_non_space = FALSE;
  int           text_seen = 0;
  BOOL          eol = FALSE;

  ProcessedItem item;
  AttributeInfo saved_attributes;
  GlyphInfo    *glyph;
  TEXTMETRIC    metrics;


  *break_type = BREAK_ON_EOL;

  /* skip initial white-space */
  while( state->buffer->text[ *position ] == ' ' ) (*position)++;
  if( *position >= end_of_paragraph )
  {
    return NULL;
  }

  row = (PaginatedRowRecord*)malloc( sizeof( PaginatedRowRecord ) );
  memset( row, 0, sizeof( *row ) );

  row->attributes = static_get_attribute_info( state );
  row->xofs = 0;

  GetTextMetrics( state->hdc, &metrics );
  max_ascent = row->ascent = (unsigned short int)metrics.tmAscent;
  max_descent = row->descent = (unsigned short int)metrics.tmDescent;

  row->text_index = *position;
  eol = FALSE;

  while( !eol && ( *position < end_of_paragraph ) )
  {
    static_process_item( state, position, max_line_width, TRUE, &item );

    switch( item.item_type )
    {
      case IT_DONEITAL:
        if( line_width + item.italic_correction > max_line_width )
        {
          if( saved_position < 0 )
          {
            eol = 1;
            break;
          }

          *position        = saved_position;

          row->length      = saved_line_width;
          row->ascent      = saved_max_ascent;
          row->descent     = saved_max_descent;
          row->spaces      = saved_spaces;
          row->text_length = (unsigned short int)( *position - row->text_index );
          row->row_padding = ( max_line_width - row->length );

          memcpy( &state->current_attributes, &saved_attributes, sizeof( AttributeInfo ) );

          return row;
        }

        line_width += item.italic_correction;
        /* and then, fall through so we can update the text metrics */

      case IT_LINKSTART:
      case IT_LINKEND:
      case IT_NONE:
        /* font may have changed--check height */
        GetTextMetrics( state->hdc, &metrics );
        if( metrics.tmAscent > max_ascent ) 
          max_ascent = metrics.tmAscent;
        if( metrics.tmDescent > max_descent )
          max_descent = metrics.tmDescent;

        /* if a link just began or ended, update the "links" list. */
        switch( item.item_type )
        {
          case IT_LINKSTART:
            static_start_link( state, row, line_width, 0 );
            break;

          case IT_LINKEND:
            static_end_link( state, row, line_width, 0 );
            break;
        }
        break;

      case IT_IMAGE:        
        use_image_as_link = 0;

        /* does the image make the line too wide? */
        if( line_width + (signed)item.image->width > max_line_width )
        {
          row->text_length = (unsigned short int)( *position - row->text_index );

          /* if we HAVE seen some text on this line, break the line here.
           * Otherwise, break the line AFTER the image. */

          if( text_seen > 0 )
          {
           
            if( saved_position < 0 )
            {
              eol = 1;
              break;
            }

            //*position       = saved_position;
            /* Rewind the position to just before the image, depends on how long the instruction is.
               For now we treat an image as being preceded by a whitespace. */
            if( state->buffer->text[ *position - 4 ] == 0 )
            {
              *position -= 4;
            }
            else
            {
              *position -= 6;
            }

            row->length      = saved_line_width;
            row->ascent      = saved_max_ascent;
            row->descent     = saved_max_descent;
            row->spaces      = saved_spaces;
            row->text_length = (unsigned short int)( *position - row->text_index );

            memcpy( &state->current_attributes, &saved_attributes, sizeof( AttributeInfo ) );
            
          }
          else
          {
            /*if( !state->active_link )
            {
              static_start_link( state, row, line_width, 0 );
              use_image_as_link = 1;
            }*/

            if( item.image->width > (unsigned)max_line_width )
            {
              tmp_height = item.image->height + GetSystemMetrics( SM_CYHSCROLL ); 
              row->length = max_line_width;
              row->ascent = ( max_ascent < tmp_height ? tmp_height : max_ascent );
            }
            /*else
            {
              row->length = item.image->width;
              row->ascent = ( max_ascent < item.image->height ? item.image->height : max_ascent );
            }*/

            row->descent = max_descent;
            row->spaces = 0;
          }

          row->row_padding = ( max_line_width - row->length );

          /*if( use_image_as_link )
            static_end_link( state, row, max_line_width, 0 );*/
          return row;
        }
        else
        {
          /*if( !state->active_link )
          {
            static_start_link( state, row, line_width, 0 );
            use_image_as_link = 1;
          }*/
        }

        if( max_ascent < (signed)item.image->height )
          max_ascent = item.image->height;

        line_width += item.image->width;
        last_c = 0;

        /*if( use_image_as_link )
          static_end_link( state, row, line_width, 0 );*/

        /* consider the image a valid text element */
        text_seen++;

        /* consider a an image to be followed by a whitespace if we need to break */
        saved_line_width = line_width;
        saved_max_ascent = max_ascent;
        saved_max_descent = max_descent;
        saved_spaces     = spaces;
        saved_position   = *position;
        *break_type      = BREAK_ON_SPACE;

        memcpy( &saved_attributes, &state->current_attributes, sizeof( saved_attributes ) );

        break;

      case IT_RULE:
        text_seen++; /* horizontal rule counts as text */
        
        line_width += item.rule.width;
        last_c = 0;
        break;

      case IT_NEWLINE:
        spaces = 0;
        eol = TRUE;
        break;

      case IT_ALIGN:
      case IT_MARGINS:
        /* ignore these two cases, since they should have been handled at the
         * paragraph level. */
        break;

      case IT_CHAR:
        if( IS_SPACE( item.c ) )
        {
          if( seen_non_space || item.c == NBSP || state->current_attributes.fixed_width )
          {
            if( !seen_space && item.c != NBSP )
            {
              saved_line_width = line_width;
              saved_max_ascent = max_ascent;
              saved_max_descent = max_descent;
              saved_spaces     = spaces;
              saved_position   = *position-1;
              *break_type      = BREAK_ON_SPACE;

              memcpy( &saved_attributes, &state->current_attributes, sizeof( saved_attributes ) );

              seen_space = 1;
            }

            line_width += state->space_width;
            spaces++;
          }
        }
        else
        {
          text_seen++;

          if( last_c == UNICODE_M_DASH ) /* if the last character seen was an m-dash,
                                           insert an invisible space to the right of it. */
          {
            spaces++;
          }

          seen_space = 0;
          seen_non_space = 1;

          if( item.c == MINUS || item.c == UNICODE_N_DASH || item.c == UNICODE_M_DASH )
          {
            int dwid = 0;

            switch( item.c )
            {
              case MINUS:  dwid = state->minus_width;  break;
              case UNICODE_N_DASH: dwid = state->ndash_width; break;
              case UNICODE_M_DASH: dwid = state->mdash_width; break;
            }

            if( line_width + dwid > max_line_width )
            {
              if( saved_position < 0 )
              {
                eol = 1;
                break;
              }

              *position       = saved_position;

              row->length      = saved_line_width;
              row->ascent      = saved_max_ascent;
              row->descent     = saved_max_descent;
              row->spaces      = saved_spaces;
              row->text_length = (unsigned short int)( *position - row->text_index );
              row->row_padding = ( max_line_width - row->length );

              memcpy( &state->current_attributes, &saved_attributes, sizeof( AttributeInfo ) );

              return row;
            }

            line_width += dwid;

            if( last_c > ' ' && item.c == UNICODE_M_DASH ) /* allow space on the left of it if a line is full justified */
            {
              spaces++;
            }

            saved_line_width = line_width;
            saved_max_ascent = max_ascent;
            saved_max_descent = max_descent;
            saved_spaces     = spaces;
            saved_position   = *position; /* break AFTER the dash */
            *break_type      = BREAK_ON_DASH;

            memcpy( &saved_attributes, &state->current_attributes, sizeof( saved_attributes ) );
          }
          else if( IsSoftHyphen(item.c, state->info)/*item.c == SOFT_HYPHEN*/ )
          {
            /* if we had to break the line here, could we fit a minus at the end? */
            if( line_width + state->minus_width > max_line_width )
            {
              /* if not, then break the line at the last saved position */

              if( saved_position < 0 )
              {
                eol = 1;
                break;
              }

              *position       = saved_position;

              row->length      = saved_line_width;
              row->ascent      = saved_max_ascent;
              row->descent     = saved_max_descent;
              row->spaces      = saved_spaces;
              row->text_length = (unsigned short int)( *position - row->text_index );
              row->row_padding = ( max_line_width - row->length );

              memcpy( &state->current_attributes, &saved_attributes, sizeof( AttributeInfo ) );

              return row;
            }

            saved_line_width = line_width + state->minus_width; /* if we break here, we include the minus */
            saved_max_ascent = max_ascent;
            saved_max_descent = max_descent;
            saved_spaces     = spaces;
            saved_position   = *position; /* break AFTER the dash */
            *break_type      = BREAK_ON_SOFT_HYPHEN;

            memcpy( &saved_attributes, &state->current_attributes, sizeof( saved_attributes ) );
          }
          else
          {
            glyph = FontCache_getGlyph( state->info->font_cache,
                                        state->hdc,
                                        state->current_font,
                                        item.c );

            /* if adding this glyph would exceed the line width, break here */
            if( line_width + glyph->width > max_line_width )
            {
              if( saved_position < 0 )
              {
                eol = 1;
                break;
              }

              *position       = saved_position;

              row->length      = saved_line_width;
              row->ascent      = saved_max_ascent;
              row->descent     = saved_max_descent;
              row->spaces      = saved_spaces;
              row->text_length = (unsigned short int)( *position - row->text_index );
              row->row_padding = ( max_line_width - row->length );

              memcpy( &state->current_attributes, &saved_attributes, sizeof( AttributeInfo ) );
              return row;
            } /* if line_width + char_width > max_line_width */

            line_width += glyph->width;
          } /* else if not dash and not discretionary hyphen */
        } /* else if not space */
        last_c = item.c;
        break;
    }
  } /* while *position < end_of_paragraph */

  /* if the row ends and no significant text has been seen, ignore this row */
  if( text_seen < 1 )
  {
    free( row );
    return NULL;
  }

  /* break at end of line */

  row->length = line_width;
  row->ascent = max_ascent;
  row->descent = max_descent;
  row->spaces = spaces;
  row->text_length = (unsigned short int)( *position - row->text_index );
  row->row_padding = ( max_line_width - row->length );
  *break_type = BREAK_ON_EOL;

  return row;
}


static unsigned short int static_get_attribute_info( PaginationState *state )
{
  int i;
  int tmp;

  /* first, search the cache */
  for( i = 0; i < state->info->attributes_size; i++ )
  {
    if( memcmp( &state->info->attributes[ state->attribute_mru_cache[ i ] ],
                &state->current_attributes, sizeof( AttributeInfo ) ) == 0 )
    {
      /* move this attribute to the top of the attribute-mru-cache */
      tmp = state->attribute_mru_cache[ 0 ];
      state->attribute_mru_cache[ 0 ] = state->attribute_mru_cache[ i ];
      state->attribute_mru_cache[ i ] = tmp;
      /* return it */
      return state->attribute_mru_cache[ 0 ];
    }
  }

  /* if the cache misses, add this attribute to it */
  if( state->info->attributes_size >= state->info->attributes_capacity )
  {
    /* grow the cache */
    state->info->attributes_capacity *= 2;
    state->info->attributes = (AttributeInfo*)realloc( state->info->attributes,
                                  sizeof( AttributeInfo ) * state->info->attributes_capacity );
    state->attribute_mru_cache = (unsigned short*)realloc( state->attribute_mru_cache,
                                       sizeof( unsigned short ) * state->info->attributes_capacity );
  }

  /* put the item at the end of the list, and put its index at the front of the cache */
  memcpy( &state->info->attributes[ state->info->attributes_size ],
          &state->current_attributes, sizeof( AttributeInfo ) );
  state->attribute_mru_cache[ state->info->attributes_size ] = state->attribute_mru_cache[ 0 ];
  state->attribute_mru_cache[ 0 ] = state->info->attributes_size;

  state->info->attributes_size++;

  return state->info->attributes_size-1;
}


static void static_initialize_attributes( PaginationState *state,
                                          PaginatedRowRecord *row )
{
  memcpy( &state->current_attributes,
          &state->info->attributes[ row->attributes ],
          sizeof( state->current_attributes ) );
  state->original_font_size = state->info->attributes[ row->attributes ].font_def.lfHeight;
  static_load_font( state );
}


static void static_render_row( PaginationState *state,
                               PaginatedRowRecord *row,
                               int y )
{
  GlyphInfo *glyph;

  int i;
  int eol;
  int x;
  int extra_leeway = 0;
  int leeway_per_space = 0;
  int seen_non_space = 0;
  int line_width;
  int underline_from = -1;
  int last_c = 0;
  int width;
  int color_changed;
  TCHAR word[ 256 ];
  int start_x;
  int word_length;
  int biggest_height = 0;
  COLORREF current_color;
  LinkedListNode *oldest_highlight;
  LinkedListNode *active_highlight;
  LinkedListNode *temp_highlight;

  ProcessedItem item;

  if( row->paragraph )
  {
    static_configure_font( state, REGULAR_FONT );
    static_load_font( state );
  }

  if( row->spaces > 0 )
  {
    leeway_per_space = PADDING( row ) / row->spaces;
    extra_leeway     = PADDING( row ) % row->spaces;
  }

  oldest_highlight = state->highlights->list;
  active_highlight = NULL;

  current_color = state->info->attributes[ row->attributes ].text_color;
  SetTextColor( state->hdc, current_color );
 
  x = row->xofs /* - delta_x */;
  eol = row->text_index + row->text_length;

  line_width = state->info->page_width - ( state->info->defaults.left_margin + state->info->defaults.right_margin );

  word_length = 0;
  i = row->text_index;
  color_changed = 0;

  while( i < eol )
  {
    /* is the oldest_highlight obsolete? */
    while( oldest_highlight != NULL && NODEDATA(oldest_highlight,Highlight)->end_pos < i )
      oldest_highlight = oldest_highlight->next;

    temp_highlight = static_find_active_highlight( oldest_highlight, i );
    if( temp_highlight != active_highlight )
    {
      active_highlight = temp_highlight;
      color_changed = 1;
    }

    if( word_length > 0 && ( state->buffer->text[ i ] == '\0' || color_changed ) )
    {
      ExtTextOut( state->hdc, start_x, y+TO_BASELINE(row)-biggest_height-state->current_attributes.bline_ofs, 0, NULL, word, word_length, NULL );
      word_length = 0;
      biggest_height = 0;
    }

    if( color_changed )
    {
      if( active_highlight != NULL && NODEDATA(active_highlight,Highlight)->fg >= 0 )
      {
        SetTextColor( state->hdc, NODEDATA(active_highlight,Highlight)->fg );
      }
      else
      {
        SetTextColor( state->hdc, current_color );
      }
      color_changed = 0;
    }

    static_process_item( state, &i, line_width, TRUE, &item );

    switch( item.item_type )
    {
      case IT_DONEITAL:
        x += item.italic_correction;
        break;

      case IT_LINKSTART:
      case IT_LINKEND:
      case IT_NONE:
        if( state->current_attributes.text_color != current_color )
        {
          current_color = state->current_attributes.text_color;
          if( !active_highlight )
          {
            color_changed = 1;
          }
        }
        break;

      case IT_IMAGE:
        {
          HDC bmp_dc;
          RECT *rc;
          HWND scroll;
          SCROLLINFO si;

          static_draw_hilite( state, NODEDATA(active_highlight,Highlight), x, item.image->width, y, row );

          bmp_dc = CreateCompatibleDC( state->hdc );
          SelectObject( bmp_dc, item.image->image );
          if( (signed)item.image->width > line_width )
          {
            //tmpHeight = line_width * item.image->height / item.image->width;
            //StretchBlt( state->hdc, x, y+TO_BASELINE(row)-tmpHeight, line_width, tmpHeight,
            //            bmp_dc, 0, 0, item.image->width, item.image->height, SRCCOPY );
            //scroll = GetDlgItem( state->hWnd, ID_LARGE_INSET_SCROLL+state->current_large_inset );
            BitBlt( state->hdc, 
                    0, 
                    y + row->pre_padding, 
                    state->info->page_width, 
                    item.image->height, 
                    bmp_dc, 
                    state->info->large_insets[ state->current_large_inset ].scrollpos, 
                    0, 
                    SRCCOPY );

            scroll = GetDlgItem( state->hWnd, ID_LARGE_INSET_SCROLL+state->current_large_inset );
            if( scroll )
            {
              rc = (RECT*)GetWindowLong( scroll, GWL_USERDATA );
              rc->left = 0;
              rc->right = state->info->page_width;

              si.cbSize = sizeof( si );
              si.fMask = SIF_POS | SIF_RANGE | SIF_PAGE;
              si.nMin = 0;
              si.nMax = item.image->width;
              si.nPos = state->info->large_insets[ state->current_large_inset ].scrollpos;
              si.nPage = state->info->page_width;
              SetScrollInfo( scroll, SB_CTL, &si, TRUE );            
            }

            state->current_large_inset++;
          }
          else
          {
            BitBlt( state->hdc, x, y+TO_BASELINE(row)-item.image->height, item.image->width, item.image->height,
                    bmp_dc, 0, 0, SRCCOPY );
          }
          DeleteObject( bmp_dc );
          x += item.image->width;

          last_c = 0;
        }
        break;

      case IT_RULE:
        static_draw_hilite( state, NODEDATA(active_highlight,Highlight), x, item.rule.width, y, row );

        static_horizontal_rule( state, x, y, &item, row );
        x += item.rule.width;
        last_c = 0;
        break;

      case IT_NEWLINE:
        break;

      case IT_ALIGN:
      case IT_MARGINS:
        /* ignore these two cases, since they should have been handled at the
         * paragraph level. */
        break;

      case IT_CHAR:
        if( IS_SPACE( item.c ) ) 
        {
          if( seen_non_space || item.c == NBSP || state->current_attributes.fixed_width )
          {
            seen_non_space = 1; /* NBSP counts as a nonspace, too, in this case */

            width = state->space_width;
            width += leeway_per_space;
            if( extra_leeway > 0 )
            {
              width++;
              extra_leeway--;
            }

            /* draw the hilite BEFORE any text may be written, because otherwise the hilite may cut off
             * any text that overlapped the hilight's rectangle */
            static_draw_hilite( state, NODEDATA(active_highlight,Highlight), x, width, y, row );

            if( word_length > 0 )
            {
              if (state->current_attributes.font_def.lfUnderline)
              {
                word[word_length++] = ' ';
                word[word_length++] = ' ';
              }
              ExtTextOut( state->hdc, start_x, y+TO_BASELINE(row)-biggest_height-state->current_attributes.bline_ofs, 0, NULL, word, word_length, NULL );
              word_length = 0;
              biggest_height = 0;
            }

            x += width;
          }
        }
        else
        {
          seen_non_space = 1;

          if( IsSoftHyphen(item.c, state->info)/*item.c == SOFT_HYPHEN*/ && i == eol ) /* end of line, and a soft hyphen */
          {
            item.c = MINUS;
          }
          else if( IsSoftHyphen(item.c, state->info)/*item.c == SOFT_HYPHEN*/ )
          {
            continue;
          }

          if( item.c == UNICODE_M_DASH || last_c == UNICODE_M_DASH )
          {
            width = leeway_per_space;
            if( extra_leeway > 0 )
            {
              width++;
              extra_leeway--;
            }

            if( width > 0 )
            {
              static_draw_hilite( state, NODEDATA(active_highlight,Highlight), x, width, y, row );
              x += width;
            }

            if( word_length > 0 )
            {
              ExtTextOut( state->hdc, start_x, y+TO_BASELINE(row)-biggest_height-state->current_attributes.bline_ofs, 0, NULL, word, word_length, NULL );
              word_length = 0;
              biggest_height = 0;
            }
          }

          if( word_length < 1 )
          {
            start_x = x;
          }

          glyph = FontCache_getGlyph( state->info->font_cache, state->hdc, state->current_font, item.c );

          word[ word_length++ ] = item.c;

          if( biggest_height == 0 )
          {
            TEXTMETRIC metric;

            GetTextMetrics( state->hdc, &metric );
            biggest_height = metric.tmAscent;
          }

          static_draw_hilite( state, NODEDATA(active_highlight,Highlight), x, glyph->width, y, row );

          x += glyph->width;

          last_c = item.c;
        }
      }
  }

  if( word_length > 0 )
  {
    ExtTextOut( state->hdc, start_x, y+TO_BASELINE(row)-biggest_height-state->current_attributes.bline_ofs, 0, NULL, word, word_length, NULL );
  }
}


static void static_horizontal_rule( PaginationState *state,
                                    int x,
                                    int y,
                                    ProcessedItem *item,
                                    PaginatedRowRecord *row )
{
  HBRUSH hbr;
  RECT rect;

  hbr = CreateSolidBrush( state->current_attributes.text_color );

  rect.left = x;
  rect.top  = y + row->pre_padding + ( row->ascent - item->rule.height ) / 2;
  rect.right = rect.left + item->rule.width + 1;
  rect.bottom = rect.top + item->rule.height + 1;

  FillRect( state->hdc, &rect, hbr );

  DeleteObject( hbr );
}

static void static_dump_pagination( PaginationState *state )
{
  PaginatedRowRecord *row;
  unsigned int i;
  int code;
  int size;
  int v1;
  int v2;
  int v3;
  RangeRecord *link;
  int link_count;

  state->buffer = PluckerDBGetTextRecord( state->info->document, state->info->doc_record );

  WRITE_OUT(( L"Pagination information:\n" ));
  WRITE_OUT(( L"- row-count:   %d\n", state->info->row_count ));
  WRITE_OUT(( L"- height:      %d\n", state->info->height ));
  
  link_count = 0;
  link = state->info->links;
  while( link != NULL )
  {
    link_count++;
    link = link->next;
  }

  WRITE_OUT(( L"- links:     %d\n", link_count ));

  WRITE_OUT(( L"- memory used:\n" ));
  WRITE_OUT(( L"    pagination record: %-4d x %-4d = %-6d\n",
                    1,
                    sizeof( PaginationInfoRecord ),
                    sizeof( PaginationInfoRecord ) ));
  WRITE_OUT(( L"    row records:       %-4d x %-4d = %-6d\n",
                    state->info->row_count,
                    sizeof( PaginatedRowRecord ),
                    state->info->row_count * sizeof( PaginatedRowRecord ) ));
  WRITE_OUT(( L"    attribute records: %-4d x %-4d = %-6d\n",
                    state->info->attributes_capacity,
                    sizeof( AttributeInfo ),
                    state->info->attributes_capacity * sizeof( AttributeInfo ) ));
  WRITE_OUT(( L"    link records:      %-4d x %-4d = %-6d\n", link_count, sizeof( RangeRecord ),
                    link_count * sizeof( RangeRecord ) ));
  WRITE_OUT(( L"    total:                           %-6d\n",
                    sizeof( PaginationInfoRecord ) +
                    state->info->row_count * sizeof( PaginatedRowRecord ) +
                    state->info->attributes_capacity * sizeof( AttributeInfo ) +
                    link_count * sizeof( RangeRecord ) ));

  WRITE_OUT(( L"- rows:\n" ));

  row = state->info->rows;
  while( row != NULL )
  {
    WRITE_OUT(( L"  - %p\n", row ));
    WRITE_OUT(( L"    text: '" ));

    i = row->text_index;
    while( i < row->text_index + row->text_length )
    {
      if( state->buffer->text[ i ] < 32 || state->buffer->text[ i ] > 126 )
      {
        if( state->buffer->text[ i ] == 0 )
        {
          i++;

          code = PLKR_FC_CODE( state->buffer->text[ i ] );
          size = ( state->buffer->text[ i ] & 0x07 );

          switch( code )
          {
            case 0:
              i++;
              code = PLKR_FC_CODE( state->buffer->text[ i ] );
              size = ( state->buffer->text[ i ] & 0x07 );

              switch( code )
              {
                case PLKR_FC_CODE( 0x08 ): /* superscript start */
                  WRITE_OUT(( L"[sup]" ));
                  break;
                case PLKR_FC_CODE( 0x10 ): /* superscript end */
                  WRITE_OUT(( L"[/sup]" ));
                  break;
                case PLKR_FC_CODE( 0x18 ): /* subscript start */
                  WRITE_OUT(( L"[sub]" ));
                  break;
                case PLKR_FC_CODE( 0x20 ): /* subscript end */
                  WRITE_OUT(( L"[/sub]" ));
                  break;
              }
              break;

            case PLKR_FC_CODE( 0x0A ): /* page-link begins */
              v1 = v2 = -1;
              if( size > 0 )
              {
                i++; v1 = ( (unsigned short)state->buffer->text[ i ] << 8 ) + state->buffer->text[ i+1 ]; i++;
                if( size > 2 )
                {
                  i++; v2 = ( (unsigned short)state->buffer->text[ i ] << 8 ) + state->buffer->text[ i+1 ]; i++;
                }
              }
              if( v1 < 0 && v2 < 0 )
                WRITE_OUT(( L"[/a]" ));
              else if( v1 >= 0 && v2 >= 0 )
                WRITE_OUT(( L"[a.%d.%d]", v1, v2 ));
              else if( v1 > 0 )
                WRITE_OUT(( L"[a.%d.*]", v1 ));
              else
                WRITE_OUT(( L"[a.*.%d]", v2 ));
              break;

            case PLKR_FC_CODE( 0x11 ): /* set font */
              WRITE_OUT(( L"[f.%d]", state->buffer->text[ ++i ] ));
              break;

            case PLKR_FC_CODE( 0x1A ): /* embedded image */
              i++; v1 = ( (unsigned short)state->buffer->text[ i ] << 8 ) + state->buffer->text[ i+1 ]; i++;
              WRITE_OUT(( L"[i.%d]", v1 ));
              break;

            case PLKR_FC_CODE( 0x33 ): /* horizontal rule */
              v1 = state->buffer->text[ ++i ];
              v2 = state->buffer->text[ ++i ];
              v3 = state->buffer->text[ ++i ];
              WRITE_OUT(( L"[r.%d.%d.%d]", v1, v2, v3 ));
              break;

            case PLKR_FC_CODE( 0x38 ): /* new line */
              WRITE_OUT(( L"[\\n]" ));
              break;
              
            case PLKR_FC_CODE( 0x40 ): /* italic text begins */
              WRITE_OUT(( L"[i]" ));
              break;

            case PLKR_FC_CODE( 0x48 ): /* italic text ends */
              WRITE_OUT(( L"[/i]" ));
              break;

            case PLKR_FC_CODE( 0x53 ): /* set text color */
              v1 = state->buffer->text[ ++i ];
              v2 = state->buffer->text[ ++i ];
              v3 = state->buffer->text[ ++i ];
              WRITE_OUT(( L"[c.%d.%d.%d]", v1, v2, v3 ));
              break;

            case PLKR_FC_CODE( 0x5C ): /* multiple embedded image */
              i++; v1 = ( (unsigned short)state->buffer->text[ i ] << 8 ) + state->buffer->text[ i+1 ]; i++;
              i++; v2 = ( (unsigned short)state->buffer->text[ i ] << 8 ) + state->buffer->text[ i+1 ]; i++;
              WRITE_OUT(( L"[e.%d.%d]", v1, v2 ));
              break;

            case PLKR_FC_CODE( 0x60 ): /* underline text begins */
              WRITE_OUT(( L"[u]" ));
              break;

            case PLKR_FC_CODE( 0x68 ): /* underline text ends */
              WRITE_OUT(( L"[/u]" ));
              break;

            case PLKR_FC_CODE( 0x70 ): /* strike-through text begins */
              WRITE_OUT(( L"[s]" ));
              break;

            case PLKR_FC_CODE( 0x78 ): /* strike-through text ends */
              WRITE_OUT(( L"[/s]" ));
              break;

            case PLKR_FC_CODE( 0x83 ): /* Unicode character */
              v1 = state->buffer->text[ ++i ];
              if( size == 3 )
              {
                i++; v2 = ( (unsigned short)state->buffer->text[ i ] << 8 ) + state->buffer->text[ i+1 ]; i++;
              }
              else
              {
                i++;
                memcpy( &v2, &state->buffer->text[ i ], 4 );
                i += 3;
              }
              WRITE_OUT(( L"[u.%d.%04X]", v1, v2 ));
              break;

            case PLKR_FC_CODE( 0x22 ): /* set margin --- should only be paragraph-level */
              v1 = state->buffer->text[ ++i ];
              v2 = state->buffer->text[ ++i ];
              WRITE_OUT(( L"[m.%d.%d]", v1, v2 ));
              break;

            case PLKR_FC_CODE( 0x29 ): /* alignment of text -- should only be paragraph-level */
              v1 = state->buffer->text[ ++i ];
              switch( v1 )
              {
                case ALIGN_LEFT:     WRITE_OUT(( L"[left]" )); break;
                case ALIGN_RIGHT:    WRITE_OUT(( L"[right]" )); break;
                case ALIGN_CENTER:   WRITE_OUT(( L"[center]" )); break;
                case ALIGN_JUSTIFY:  WRITE_OUT(( L"[justify]" )); break;
              }
              break;

          }
        }
        else
          WRITE_OUT(( L"[\\x%02X]", state->buffer->text[ i ] ));
      }
      else
        WRITE_OUT(( L"%c", state->buffer->text[ i ] ));

      i++;
    }
    WRITE_OUT(( L"'\n" ));

    WRITE_OUT(( L"    ascent:      %d\n", row->ascent ));
    WRITE_OUT(( L"    descent:     %d\n", row->descent ));
    WRITE_OUT(( L"    pre-padding: %d\n", row->pre_padding ));
    WRITE_OUT(( L"    xofs:        %d\n", row->xofs ));
    WRITE_OUT(( L"    text_index:  %d\n", row->text_index ));
    WRITE_OUT(( L"    text_length: %d\n", row->text_length ));
    WRITE_OUT(( L"    length:      %d\n", row->length ));
    WRITE_OUT(( L"    row-padding: %d\n", row->row_padding ));
    WRITE_OUT(( L"    spaces:      %d\n", row->spaces ));
    WRITE_OUT(( L"    paragraph:   %d\n", row->paragraph ));
    WRITE_OUT(( L"    justified:   %d\n", row->justified ));
    WRITE_OUT(( L"    attributes:\n" ));
    WRITE_OUT(( L"       text-color:  %06X\n", state->info->attributes[ row->attributes ].text_color ));
    WRITE_OUT(( L"       has-link:    %d\n", state->info->attributes[ row->attributes ].has_link ));
    WRITE_OUT(( L"       link-record: %d\n", state->info->attributes[ row->attributes ].link_record ));
    WRITE_OUT(( L"       link-para:   %d\n", state->info->attributes[ row->attributes ].link_para ));
    WRITE_OUT(( L"       italic:      %d\n", state->info->attributes[ row->attributes ].font_def.lfItalic ));
    WRITE_OUT(( L"       weight:      %d\n", state->info->attributes[ row->attributes ].font_def.lfWeight ));
    WRITE_OUT(( L"       height:      %d\n", state->info->attributes[ row->attributes ].font_def.lfHeight ));
    WRITE_OUT(( L"       strikethru:  %d\n", state->info->attributes[ row->attributes ].font_def.lfStrikeOut ));
    WRITE_OUT(( L"       underline:   %d\n", state->info->attributes[ row->attributes ].font_def.lfUnderline ));
    WRITE_OUT(( L"       face:        '%s'\n", state->info->attributes[ row->attributes ].font_def.lfFaceName ));

    row = row->next;
  }
}


static void static_clear_links( PaginationInfoRecord *record )
{
  RangeRecord *link;
  RangeRecord *next;

  link = record->links;
  while( link != NULL )
  {
    next = link->next;
    free( link );
    link = next;
  }

  record->links = record->links_tail = NULL;
}


static void static_initialize_progress_meter( PaginationState *state,
                                              int max_size )
{
  RECT rect;

  GetClientRect( state->hWnd, &rect );
  state->hProgressMeter = CreateWindowEx( 0, PROGRESS_CLASS, NULL,
                               WS_CHILD | WS_VISIBLE | WS_BORDER | PBS_SMOOTH,
                               rect.left+20, ( rect.bottom - rect.top - 20 ) / 2,
                               rect.right-rect.left-40, 20,
                               state->hWnd, (HMENU) 0, GetModuleHandle( NULL ), NULL );

  SendMessage( state->hProgressMeter, PBM_SETRANGE, 0, MAKELPARAM( 0, max_size ) );
  SendMessage( state->hProgressMeter, PBM_SETPOS, 0, 0 );
}


static void static_update_progress_meter( PaginationState *state,
                                          int pos )
{
  SendMessage( state->hProgressMeter, PBM_SETPOS, pos, 0 );
}


static void static_destroy_progress_meter( PaginationState *state )
{
  DestroyWindow( state->hProgressMeter );
}


static void static_start_link( PaginationState *state,
                               PaginatedRowRecord *row,
                               int start_x, int start_y )
{
  RangeRecord *link;

  if( state->active_link != NULL )
  {
    /* this will happen if, during pagination, a link is started, and then
     * later found to not fit on the current row... In this case, simply
     * set the active link to be at the given row, start-x, and start-y */

    state->active_link->start_row = (PaginatedRow)row;
    state->active_link->start_x = start_x;
    state->active_link->start_y = start_y;

    return;
  }

  link = (RangeRecord*)malloc( sizeof( *link ) );
  link->start_row = (PaginatedRow)row;
  link->start_x = start_x;
  link->start_y = start_y;
  link->next = NULL;

  if( state->info->links_tail != NULL )
  {
    state->info->links_tail->next = link;
  }
  else
  {
    state->info->links = link;
  }

  state->info->links_tail = link;
  state->active_link = link;
}


static void static_end_link( PaginationState *state,
                             PaginatedRowRecord *row,
                             int end_x, int end_y )
{
  RangeRecord *r = state->active_link;

  if( state->active_link == NULL )
  {
    /* if this is the case, then the link was terminated on the previous row, but later
     * it was found that the line needed to be broken BEFORE where the link was terminated...
     * in that case, we simply alter the end-link info for the tail-most link record, 
     * and continue as if nothing happened... */

    r = state->info->links_tail;
  }

  r->end_row = (PaginatedRow)row;
  r->end_x = end_x;
  r->end_y = end_y + ROW_HEIGHT( row );

  state->active_link = NULL;
}


static void static_correct_link_offsets( PaginationState *state,
                                         RangeRecord *last_processed,
                                         PaginatedRowRecord *row,
                                         int dy )
{
  int idx;
  RangeRecord *r;
  int leeway_per_space = 0;
  int extra_leeway = 0;
  int accumulated_width;
  int seen_nonblank = 0;
  TCHAR last_c = 0;

  /* compute how much "room" we have per space/mdash */
  if( row->spaces > 0 )
  {
    leeway_per_space = PADDING( row ) / row->spaces;
    extra_leeway = PADDING( row ) % row->spaces;
  }

  /* here's the algorithm:
   * 
   *   1) if last_processed
   *       a) if last_processed->start_row == row then fix start_row values
   *       b) if last_processed->end_row == row then fix end_row values
   *       c) r = last_processed->next
   *   2) else if !last_processed
   *       a) r = state->info->links
   *   3) while r != NULL
   *       a) fix start_row values
   *       b) if r->end_row then fix end_row values
   *       c) r = r->next
   */

  idx = row->text_index;
  accumulated_width = row->xofs;

  if( last_processed != NULL )
  {
    if( last_processed->start_row == (PaginatedRow)row )
    {
      accumulated_width += static_compute_extra_width( state, &idx, leeway_per_space, &extra_leeway, &seen_nonblank, &last_c );
      last_processed->start_x += accumulated_width;
      last_processed->start_y  = dy;
    }

    if( last_processed->end_row == (PaginatedRow)row )
    {
      accumulated_width += static_compute_extra_width( state, &idx, leeway_per_space, &extra_leeway, &seen_nonblank, &last_c );
      last_processed->end_x += accumulated_width;
      last_processed->end_y  = dy + ROW_HEIGHT( row );
    }

    r = last_processed->next;
  }
  else
  {
    r = state->info->links;
  }

  while( r != NULL )
  {
    accumulated_width += static_compute_extra_width( state, &idx, leeway_per_space, &extra_leeway, &seen_nonblank, &last_c );
    r->start_x += accumulated_width;
    r->start_y  = dy;

    if( r->end_row == (PaginatedRow)row )
    {
      accumulated_width += static_compute_extra_width( state, &idx, leeway_per_space, &extra_leeway, &seen_nonblank, &last_c );
      r->end_x += accumulated_width;
      r->end_y  = dy + ROW_HEIGHT( row );
    }

    r = r->next;
  }
}


static int static_compute_extra_width( PaginationState *state,
                                       int *idx,
                                       int  leeway_per_space,
                                       int *extra_leeway,
                                       int *seen_nonblank,
                                       TCHAR *last_c )
{
  ProcessedItem item;
  BOOL looking = TRUE;
  int extra = 0;

  /* consider all spaces from the current index until the next beginning or ending of
   * a hyperlink.  Ignore leading whitespace if seen_nonblank is false.
   *
   * We only need to consider spaces and the padding arounnd m-dashes, because everything
   * else was counted correctly during the pagination stage. */

  do
  {
    static_process_item( state, idx, 0, FALSE, &item );

    switch( item.item_type )
    {
      case IT_LINKEND:
      case IT_LINKSTART:
        looking = FALSE;
        break;

      case IT_CHAR:
        if( IS_SPACE( item.c ) )
        {
          if( *seen_nonblank || item.c == NBSP )
          {
            *seen_nonblank = TRUE;

            extra += leeway_per_space;
            if( *extra_leeway > 0 )
            {
              extra += 1;
              *extra_leeway -= 1;
            }
          }
        }
        else
        {
          *seen_nonblank = TRUE;

          if( item.c == UNICODE_M_DASH || *last_c == UNICODE_M_DASH )
          {
            extra += leeway_per_space;
            if( *extra_leeway > 0 )
            {
              extra += 1;
              *extra_leeway -= 1;
            }
          }
        }

        *last_c = item.c;
        break;
    }
  }
  while( looking );

  return extra;
}


static LinkedListNode *static_find_active_highlight( LinkedListNode *list,
                                                     int idx )
{
  LinkedListNode *i;
  LinkedListNode *active = NULL;
  Highlight *h;

  i = list;
  while( i != NULL )
  {
    h = NODEDATA(i,Highlight);
    if( h->start_pos <= idx && h->end_pos >= idx ) active = i;
    else if( h->start_pos > idx ) break;

    i = i->next;
  }

  return active;
}


static void static_draw_hilite( PaginationState *state,
                                Highlight *h,
                                int x,
                                int w,
                                int y,
                                PaginatedRowRecord *row )
{
  HBRUSH hbr;
  RECT rect;

  if( h != NULL && h->bg >= 0 )
  {
    hbr = CreateSolidBrush( h->bg );

    rect.left = x;
    rect.top = y;
    rect.right = x + w + 1;
    rect.bottom = y + ROW_HEIGHT( row ) + 1;

    FillRect( state->hdc, &rect, hbr );

    DeleteObject( hbr );
  }
}


static int static_compute_italic_correction( PaginationState *state,
                                             TCHAR c )
{
  GlyphInfo *info;

  info = FontCache_getGlyph( state->info->font_cache, state->hdc, state->current_font, 'm' );

  return info->width/6;
}
