/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: pagination module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * This forms the 'core' of the reader.  This module defines the routines
 * and data structures that paginate and render a plucker document, as well
 * providing an API for identifying positions within the document both by
 * (x,y) coordinates, and by character position.  A simple "selection" API
 * is also provided to allow applications to interact with the document by
 * selecting portions of it with a pointing device.
 *
 * Every link on a page is preprocessed and assigned a "range".  This makes
 * it easier for applications to tell where a clicked link starts and stops,
 * meaning that a multi-line link may be correctly highlighted, as needed.
 *
 * Serialization routines are also provided, to facilitate saving the pagination
 * info for a document record and reloading it later, so that the pagination
 * process does not have to be repeated for every request.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __PAGINATION_H__
#define __PAGINATION_H__

#include "plkfmt.h"       /* defines the plucker document interface */
#include "highlights.h"   /* defines the highlight structure */
#include "font_cache.h"   /* the FontCache, used in paginating and rendering */
#include "linked-list.h"  /* the LinkedList object, used with Highlights */

/* constants used for determining where on a rendered page a click occurred */
#define CLICK_AT_EOF          0   /* the click occurred after the end of the text */
#define CLICK_IN_LEFT_MARGIN  1   /* the click occurred in the left margin of the text */
#define CLICK_IN_RIGHT_MARGIN 2   /* the click occurred in the right margin of the text */
#define CLICK_ON_TEXT         3   /* the click occurred in the text itself */

/* full-justification rule constants */
#define JUSTIFY_NORMAL        0  /* full-justify when requested */
#define JUSTIFY_ALWAYS        1  /* treat left-justification as full-justification */
#define JUSTIFY_NEVER         2  /* treat full-justification as left-justification */

/* base id of the large inset scrolls */
#define ID_LARGE_INSET_SCROLL ( 45000 )

typedef struct _range_record_  RangeRecord;

typedef void *PaginationInfo;
typedef void *PaginatedRow;
typedef RangeRecord *RangeInfo;


/* The range record defines a range text in a document, from a given row and x-offset to
 * another given row and x-offset.  The y-offsets are treated as coming from the top of the
 * document, and not the screen. */
struct _range_record_
{
  RangeRecord *next;               /* for linked lists of ranges (like visible anchors) */

  PaginatedRow *start_row;         /* the row on which the range starts */
  unsigned short start_x;          /* the x-ofs from the left side of the screen */

  PaginatedRow *end_row;           /* the row on which the range ends */
  unsigned short end_x;            /* the x-ofs from the left side of the screen */

  unsigned short start_y;          /* the y-ofs from the top of the document to where the range starts (top of row) */
  unsigned short end_y;            /* the y-ofs from the top of the document to where the range ends (bottom of row) */
};


/* The definition for the default formatting options.  These are used to alter how the paginator and
 * renderer behave.
 *
 * If the interparagraph or interline spaces are negative, the summed space may not be less than zero.  If it is,
 * it is adjusted to zero internally.  This prevents lines from overlapping (I can't imagine why anyone would want
 * that.)
 *
 * The first lines of paragraphs are only indented if the paragraph is left or full justified.  Centered and
 * right justified paragraphs are not indented.
 */
typedef struct
{
  TCHAR normal_face[ 30 ];        /* the font name for the regular font */
  TCHAR monotype_face[ 30 ];      /* the font name for the monospace font */

  int   font_size;                /* the height of the font (pixels) */

  unsigned long int text_color;   /* the color of the text */

  BYTE text_quality;              /* the quality of the text (antialiased, cleartype, etc.) */

  unsigned int left_margin;       /* size (in pixels) of the left margin */
  unsigned int right_margin;      /* size (in pixels) of the right margin */

  int  interparagraph_space;      /* amount of extra space to add between paragraphs (may be negative) */
  int  interline_space;           /* amount of extra space to add between lines (may be negative) */
  BYTE indent_size;               /* size (in pixels) of the extra indentation for the first line of a paragraph */
  BYTE justification_rule;        /* the justfication rule to use */

  BOOL enable_soft_hyphens;       /* uses soft hyphens, disable it to allow other charsets work properly (e.g. thai) */
} FormattingDefaults;


/* Defines a specific location in a document, and encapsulates part of the state of the document at that
 * point.
 */
typedef struct
{
  PaginatedRow row;                 /* the row object at that point */
  unsigned int xofs;                /* the x-offset from the left of the document */
  unsigned int yofs;                /* the y-offset from the top of the document */
  Highlight *highlight;             /* the highlight object that is active at that point (or NULL if none is) */

  unsigned int text_index;          /* the index into the plucker document buffer at this point */

  unsigned int character;           /* the character at this position, if any */
  unsigned int char_width;          /* the width (in pixels) of the character at this position */

  BOOL has_link;                    /* whether or not a hyperlink is active at this point */
  unsigned short link_record;       /* the record that the active hyperlink links to */
  unsigned short link_paragraph;    /* the paragraph # within the record that the active hyperlink links to */
  RangeInfo link_range;             /* the range record defining where the active link starts and stops */
} DocumentPoint;


/* Defines a document selection. Either start or end (or both) will always equal anchor. */
typedef struct
{
  DocumentPoint anchor;  /* where the selection first began. */
  DocumentPoint start;   /* the earliest point in the selection.  (start < end) */
  DocumentPoint end;     /* the latest point in the selection. (end > start) */

  RangeRecord   range;   /* the record that defines the document range between start and end. */
} SelectionState;

typedef struct
{
  unsigned char type;
  int maxpos;
  int scrollpos;
  int y_offs;
} LargeInset;


#ifdef __cplusplus
extern "C" {
#endif

/* Defines a new (and empty) PaginationInfo record, which uses the given font cache to store fonts. */
PaginationInfo PaginationInfo_new( FontCache font_cache );

/* Deletes the given record, with all associated rows, links, and so forth.  The font cache for the
 * record is *not* deleted, facilitating the reuse of that structure. */
void PaginationInfo_delete( PaginationInfo info );

/* Invalidates the pagination info by returning the record to its "empty" state. */
void PaginationInfo_invalidate( PaginationInfo info );

/* Returns the total height of the document, in pixels. This is basically the sum of the heights of all of
 * the paginated rows, which means that until a document has been paginated, this routine will not return
 * a meaningful number. */
unsigned int PaginationInfo_getHeight( PaginationInfo info );

/* Returns the widest width of the document, in pixels.  A document may be wider than the screen if there is
 * a particularly large image in the document. */
unsigned int PaginationInfo_getWidth( PaginationInfo info );

/**
 * Paginates the given record of the plucker document, using the given window and device context handles as
 * needed.  The routine attempts to paginate the record into a page width of 'width', but the result may be
 * wider if it cannot fit the page into that width.  The given formatting defaults are used to determine the
 * behavior of the paginator.
 */
void PaginationInfo_paginate( PaginationInfo info,
                              PluckerDB document,
                              int record,
                              HWND hWnd,
                              HDC hdc,
                              int width,
                              FormattingDefaults *defaults );

/**
 * Renders a previously paginated document onto the given device context.  The other parameters are
 * interpreted as follows:
 * 
 * dy: the y-offset from top of the document to the top of the visible area.
 * page_ofs: the y-offset from the top of the device context to the top visible area.
 * canvas_top: the y-offset from the top of the device context to the start of the area to render.
 * height: the height of the area to render.
 * highlights: a linked list of highlights to render with the page.
 *
 * canvas_top and height allow horizontal slices of the page to rerendered, without having to redraw
 * the entire page.  This helps speed things like scrolling.
 */
void PaginationInfo_render( PaginationInfo info,
                            HDC hdc,
                            int dy,
                            int page_ofs,
                            int canvas_top,
                            int height,
                            LinkedList *highlights,
                            HWND hWnd );

/**
 * Translates the given (delta_x+x,delta_y+y) coordinate pair into a point on the document (see
 * DocumentPoint, above).  The paramters are interpreted as follows:
 *
 * page_ofs: the y-offset from the top of the device context to the top of the visible area.
 * delta_y: the y-offset from the top of the document to the top of the visible area.
 * delta_x: the x-offset from the left of the document to the left of the visible area.
 * x: the x coordinate on the device context to query.
 * y: the y coordinate on the device context to query.
 * highlights: the list of highlights that are active on the page.
 * point: the document point to recieve the queried position.
 *
 * This routines returns one of the CLICKED_xxx constants, defining which area of the document
 * was derived from the coordinates.
 */
int PaginationInfo_lookupPoint( PaginationInfo info,
                                HDC hdc,
                                int page_ofs,
                                int delta_y,
                                int delta_x,
                                int x, int y,
                                LinkedList *highlights,
                                DocumentPoint *point );

/**
 * Inverts the colors of the given range of text. Parameters are interpreted as follows:
 *
 *  page_ofs: the y-offset from the top of the device context to the top of the visible area.
 *  yofs: the y-offset from the top of the document to the top of the device context.
 *  range: the range of text to invert.
 *
 * This is used for highlighting selections and clicked links.
 */
void PaginationInfo_invertRange( PaginationInfo info,
                                 HDC hdc,
                                 int page_ofs,
                                 int yofs,
                                 RangeInfo range );

/**
 * Takes a paragraph number and returns the y-offset from the top of the document to the top
 * of that paragraph.
 */
long int PaginationInfo_findParagraphY( PaginationInfo info,
                                        unsigned short int paragraph );

/**
 * Takes a delta_y (y-offset from top of document to the top of a visible area) and returns
 * the character index of the first character of the first row that completely fits (vertically)
 * on the visible area.  The y-offset from the top of the visible area to the top of the row
 * is stored in y_offset.
 */
long int PaginationInfo_deltaYToCharacterIndex( PaginationInfo info,
                                                int delta_y,
                                                int *y_offset );

/**
 * Takes a character index and returns the y-offset from the top of the document to the top of
 * the character's row.
 */
long int PaginationInfo_characterIndexToDeltaY( PaginationInfo info,
                                                int index );

/**
 * Initializes the given range record to the range defined by the given start and end points.
 * If start > end, the points are reordered (in place, so the start and end parameters are swapped).
 */
void PaginationInfo_createRange( PaginationInfo info,
                                 DocumentPoint *start,
                                 DocumentPoint *end,
                                 RangeRecord *range );

/**
 * Swaps the values of start and end if start > end, otherwise this function does nothing.
 */
void PaginationInfo_reorderPoints( PaginationInfo info,
                                   DocumentPoint *start,
                                   DocumentPoint *end );

/**
 * Compares two points and returns -1 if p1 < p2, 1 if p1 > p2, and 0 if p1 == p2.  Comparisons are
 * based only on the positional information in the document point records.
 */
int PaginationInfo_comparePoints( PaginationInfo info,
                                  DocumentPoint *p1,
                                  DocumentPoint *p2 );

/**
 * Initializes the given selection to be anchored on the given point.  The start and end of the
 * selection are also initialized to the anchor value.
 */
void PaginationInfo_selectionInit( PaginationInfo  info,
                                   SelectionState* selection,
                                   DocumentPoint*  anchor );

/**
 * Extends the given selection to include the given point.  This may cause the selection to rotate
 * about its anchor, as needed.  The yofs is the offset from the top of the document to the top of
 * the viewable area.  The selected range is inverted and updated as needed.
 */
void PaginationInfo_extendSelection( PaginationInfo  info,
                                     HDC             hdc,
                                     int             yofs,
                                     SelectionState* selection,
                                     DocumentPoint*  point );

/**
 * Unhighlights the given selection and resets its values so that it can be seen to be empty.
 */
void PaginationInfo_selectionDone( PaginationInfo  info,
                                   HDC             hdc,
                                   int             yofs,
                                   SelectionState *selection );

/**
 * Returns a boolean.  TRUE if the selection is empty, and FALSE otherwise.
 */
int PaginationInfo_emptySelection( PaginationInfo info,
                                   SelectionState *selection );

/**
 * Copies the text in the given selection into the given buffer.  All formatting is stripped from
 * the text, although newlines, paragraph breaks, and (rudimentarily) horizontal rules are
 * preserved.  The buffer must be large enough to hold the selection, or memory leaks will occur.
 *
 * Returns 0.
 */
int PaginationInfo_copySelectionTo( PaginationInfo info,
                                    SelectionState *selection,
                                    TCHAR* buffer );

/**
 * Serializes the given pagination info to the given file.  Returns 0 if successful, and -1
 * if there was an error.
 */
int PaginationInfo_serialize( PaginationInfo info,
                              FILE*          f );

/**
 * Unserializes the given pagination info for the given document and record, from the given
 * FILE.  The formatting defaults for the pagination info are given to be 'defaults'.  The
 * pagination info must have previously been created.  If the record contains prior pagination
 * info, it will be invalidated.
 *
 * Returns 0 if successful, or -1 if there was an error.
 */
int PaginationInfo_unserialize( PaginationInfo  info,
                                PluckerDB       document,
                                int             record,
                                FormattingDefaults *defaults,
                                FILE*           f );

/**
 * Returns TRUE if the given page begins with any amount of whitespace followed by a link.
 * If so, the 'record' and 'paragraph' parameters are populated with the link's relevant
 * information.
 */
int PaginationInfo_startsWithLink( PaginationInfo  info,
                                   int            *record,
                                   int            *paragraph );

/**
 * Returns TRUE if the given page ends with a link followed by any amount of whitespace.
 * If so, the 'record' and 'paragraph' parameters are populated with the link's relevant
 * information.
 */
int PaginationInfo_endsWithLink( PaginationInfo  info,
                                 int            *record,
                                 int            *paragraph );

int PaginationInfo_getNrLargeInsets( PaginationInfo info );

void PaginationInfo_updateLargeInsetScrolls( PaginationInfo info,
                                             HWND hWnd,
                                             int y_offset,
                                             const RECT *reader_bounds );
void PaginationInfo_updateLargeInset( PaginationInfo info, int id, int delta, BOOL rel );
#ifdef __cplusplus
}
#endif

#endif /* __PAGINATION_H__ */
