/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: pagination cache module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <string.h>
#include "pagination_cache.h"
#include "library.h"



PAGINATION_CACHE *PaginationCache_new( PluckerDB document,
                                       Library *library,
                                       int capacity )
{
  PAGINATION_CACHE *cache;

  cache = (PAGINATION_CACHE*)malloc( sizeof( PAGINATION_CACHE ) );
  if( cache == NULL ) return NULL;

  memset( cache, 0, sizeof( *cache ) );
  cache->cache_capacity = capacity;
  cache->document = document;
  cache->library = library;

  return cache;
}


void PaginationCache_delete( PAGINATION_CACHE *cache )
{
  PaginationCache_clear( cache );
  memset( cache, 0, sizeof( *cache ) );
  free( cache );
}


void PaginationCache_serialize( PAGINATION_CACHE *cache,
                                FILE *f )
{
  PAGINATION_CACHE_ITEM *item;
  int size;

  /* write the current set of formatting options */
  size = sizeof( cache->library->options );
  fwrite( &size, sizeof( size ), 1, f );
  fwrite( &cache->library->options, sizeof( cache->library->options ), 1, f );

  fwrite( &cache->cache_size, sizeof( cache->cache_size ), 1, f );

  item = cache->cache;
  while( item != NULL )
  {
    fwrite( &item->record_id, sizeof( item->record_id ), 1, f );
    PaginationInfo_serialize( item->pagination, f );
    item = item->next;
  }
}


PAGINATION_CACHE *PaginationCache_unserialize( PluckerDB document,
                                               Library *library,
                                               int capacity,
                                               FILE* f )
{
  PAGINATION_CACHE *cache;
  PAGINATION_CACHE_ITEM *item;
  PAGINATION_CACHE_ITEM *last_item;
  int i;
  int size;
  FormattingDefaults saved_opts;

  cache = PaginationCache_new( document, library, capacity );

  fread( &size, sizeof( size ), 1, f );
  if( size < sizeof( saved_opts ) )
  {
    Library_initializeDefaults( &saved_opts );
    fread( &saved_opts, sizeof( size ), 1, f );
  }
  else
  {
    fread( &saved_opts, sizeof( saved_opts ), 1, f );
    fseek( f, size-sizeof(saved_opts), SEEK_CUR );
  }

  if( fread( &cache->cache_size, sizeof( cache->cache_size ), 1, f ) != 1 )
  {
    free( cache );
    return NULL;
  }

  /* if the number of cached records is greater than the requested cache capacity,
   * ignore any records after the capacity is reached.  This should be okay anyway,
   * since items are stored in MRU order. */

  if( cache->cache_size > cache->cache_capacity )
  {
    cache->cache_size = cache->cache_capacity;
  }

  if( ( size != sizeof( saved_opts ) ) || ( memcmp( &saved_opts, &library->options, sizeof( saved_opts ) ) != 0 ) )
  {
    /* if the formatting options have changed, don't use these cached pages */
    cache->cache_size = 0;
    return cache;
  }

  last_item = NULL;
  for( i = 0; i < cache->cache_size; i++ )
  {
    item = (PAGINATION_CACHE_ITEM*)malloc( sizeof( PAGINATION_CACHE_ITEM ) );
    item->next = NULL;
    item->locked = 0;

    fread( &item->record_id, sizeof( item->record_id ), 1, f );
    item->pagination = PaginationInfo_new( library->font_cache );
    PaginationInfo_unserialize( item->pagination, document, item->record_id, &library->options, f );

    if( last_item == NULL )
    {
      cache->cache = item;
    }
    else
    {
      last_item->next = item;
    }

    last_item = item;
  }

  return cache;
}


PaginationInfo PaginationCache_lock( PAGINATION_CACHE *cache,
                                     int record_id )
{
  PAGINATION_CACHE_ITEM *item;
  PAGINATION_CACHE_ITEM *prev;

  /* first, check the cache to see if the requested record_id is there */
  prev = NULL;
  for( item = cache->cache; item != NULL; item = item->next )
  {
    if( item->record_id == record_id )
    {
      item->locked++;
      
      /* move item to the head of the list, as the MRU item */
      if( prev != NULL )
      {
        prev->next = item->next;
        item->next = cache->cache;
        cache->cache = item;
      }

      return item->pagination;
    }

    prev = item;
  }

  /* if we get here, then the cache missed... */
  return NULL;
}


void PaginationCache_release( PAGINATION_CACHE *cache,
                              int record_id )
{
  PAGINATION_CACHE_ITEM *item;

  /* decrement the requested record_id's lock count by one */
  for( item = cache->cache; item != NULL; item = item->next )
  {
    if( item->record_id == record_id )
    {
      if( item->locked > 0 ) item->locked--;
      break;
    }
  }
}


int PaginationCache_add( PAGINATION_CACHE *cache,
                         int record_id,
                         PaginationInfo info )
{
  PAGINATION_CACHE_ITEM *item;
  PAGINATION_CACHE_ITEM *prev;
  PAGINATION_CACHE_ITEM *not_locked;
  PAGINATION_CACHE_ITEM *prev_not_locked;

  /* first, ensure there is room in the cache for another item */
  if( cache->cache_size >= cache->cache_capacity )
  {
    /* remove the last item in the cache that is not locked */
    item = cache->cache;
    not_locked = NULL;
    prev_not_locked = NULL;
    prev = NULL;

    while( item != NULL )
    {
      if( !item->locked )
      {
        not_locked = item;
        prev_not_locked = prev;
      }

      prev = item;
      item = item->next;
    }

    if( not_locked == NULL )
    {
      /* the cache is full and all items are locked! */
      return 0;
    }

    /* delete the unlocked item */
    PaginationInfo_delete( not_locked->pagination );

    if( prev_not_locked == NULL )
    {
      /* we're deleting the head of the list */
      cache->cache = not_locked->next;
    }
    else
    {
      prev_not_locked->next = not_locked->next;
    }
    
    free( not_locked );

    cache->cache_size--;
  }

  /* here, we know there is room in the cache for one more... */
  item = (PAGINATION_CACHE_ITEM*)malloc( sizeof( PAGINATION_CACHE_ITEM ) );
  item->locked = 0;
  item->next = cache->cache;
  item->record_id = record_id;
  item->pagination = info;

  cache->cache = item;
  cache->cache_size++;

  return 1;
}


void PaginationCache_clear( PAGINATION_CACHE* cache )
{
  PAGINATION_CACHE_ITEM *item;
  PAGINATION_CACHE_ITEM *next;

  item = cache->cache;
  while( item != NULL )
  {
    next = item->next;
    PaginationInfo_delete( item->pagination );
    free( item );
    item = next;
  }

  cache->cache = NULL;
  cache->cache_size = 0;
}
