/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: pagination cache module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * These routines allow previously generated pagination records to be cached
 * for future reuse.  The cache itself may be serialized and unserialized.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __PAGINATION_CACHE_H__
#define __PAGINATION_CACHE_H__

#include <stdio.h>
#include "pagination.h"
#include "vade_mecum.h"


typedef struct pagination_cache_item_struct PAGINATION_CACHE_ITEM;

/* a single item in the pagination cache */
struct pagination_cache_item_struct
{
  PAGINATION_CACHE_ITEM *next;    /* the next-most-recently-used item in the cache */

  int record_id;                  /* the document record-id for this pagination record */
  int locked;                     /* whether or not the record is 'locked' (in use) */
  PaginationInfo pagination;      /* the (previously generated) pagination record */
};

typedef struct
{
  PAGINATION_CACHE_ITEM *cache;   /* the linked-list of cache items */

  int cache_capacity;             /* the maximum capacity of the cache */
  int cache_size;                 /* the # of elements in the cache */

  PluckerDB document;             /* the plucker document object that "owns" the pagination records in the cache */
  Library *library;               /* the library object representing all documents in the library */
} PAGINATION_CACHE;



/**
 * Create a new pagination cache for the given document and library.  The cache will be able to
 * hold 'capacity' active pagination information records.  The capacity must not be less than 1..
 */
PAGINATION_CACHE *PaginationCache_new( PluckerDB document,
                                       Library* library,
                                       int capacity );

/**
 * Delete the given cache and all component pagination data.
 */
void PaginationCache_delete( PAGINATION_CACHE *cache );

/**
 * Serializes the given cache (composed of all of its pagination data) to the given file.  The
 * default formatting options of the library are also serialized, so that when the cache is
 * unserialized the application can determine if the pagination records are obsolute (needing
 * reprocessing due to formatting changes.)
 */
void PaginationCache_serialize( PAGINATION_CACHE *cache,
                                FILE *f );

/**
 * Unserializes a pagination cache from teh given file, for the given document, library, and
 * capacity.  Returns NULL if the file does not exist, or is not a valid serialized cache.
 * Returns an empty pagination cache if the formatting options in the serialized cache differ
 * from the formatting options in the library.  If the given capacity is less than the number
 * of items in the serialized cache, the serialized cache will be truncated.
 */
PAGINATION_CACHE *PaginationCache_unserialize( PluckerDB document,
                                               Library *library,
                                               int  capacity,
                                               FILE* f );

/**
 * Returns NULL if the given record_id does not exist in the cache.  Otherwise, returns the
 * corresponding cache item and increments its lock count.
 */
PaginationInfo PaginationCache_lock( PAGINATION_CACHE *cache,
                                     int record_id );

/**
 * If the given record ID does not exist in the cache, this function does nothing.  Otherwise
 * it decrements the lock count of the given record_id.
 */
void PaginationCache_release( PAGINATION_CACHE *cache,
                              int record_id );

/**
 * Adds the given pagination info to the cache, associated with the given record-id, as the
 * most recently used item.  If doing so would exceed the cache's capacity, the least-recently-used
 * item that is not locked is deleted.  Returns 1 if it succeeds, or 0 if room cannot be made in
 * the cache for the new item (ie, the cache is full and all items are locked).
 */
int PaginationCache_add( PAGINATION_CACHE *cache,
                         int record_id,
                         PaginationInfo info );

/**
 * Clears the cache by deleting all of its component items.
 */
void PaginationCache_clear( PAGINATION_CACHE* cache );

#endif /* __PAGINATION_CACHE_H__ */
