/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: plucker database module
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2002 by Bill Janssen, Mark Ian Lillywhite and Michael
 *   Nordstrm, also under the GPL
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <stdio.h>
#include <windows.h>
#include <string.h>
#include <ctype.h>

#include "plkfmt.h"
#include "Bitmap.h"

typedef unsigned short int __uint16_t;
typedef unsigned long int  __uint32_t;
typedef unsigned char      __uint8_t;
typedef signed short int   __int16_t;
typedef signed long int    __int32_t;


#define DBPFX_LEN              ( 72 )
#define DBPFX_DOCNAME_LEN      ( 32 )

#define DBPFX_VERSION_OFFSET   ( 34 )
#define DBPFX_APPINFO_OFFSET   ( 52 )

#ifdef PDAV_SHOWGUIDE
#define DBPFX_MAGIC            "DataShwg"
#else
#define DBPFX_MAGIC            "DataPlkr"
#endif
#define DBPFX_MAGIC_LEN        (  8 )
#define DBPFX_MAGIC_OFFSET     ( 60 )

#define RIDL_OFFSET            ( DBPFX_LEN )
#define RIDL_LEN               (  6 )
#define RIDL_NUMREC_OFFSET     (  4 )
#define RIDL_FIRST_OFFSET      ( RIDL_OFFSET + RIDL_LEN )

#define DATARECHDR_LEN         (  8 )

#define LSQUO   ( 8216 )
#define RSQUO   ( 8217 )
#define LDQUO   ( 8220 )
#define RDQUO   ( 8221 )

#define RECORD_IS_CONTINUED( hdr )    ( (hdr)[7] & 0x1 ? TRUE : FALSE )
#define RECORD_HAS_METADATA( hdr )    ( ((hdr)[7] >> 1) & 0x1 ? TRUE : FALSE )
#define NR_PARAGRAPHS( hdr )          ( PLKR_INT16( hdr+2 ) )


typedef struct __pluckerDBCacheItem PluckerDBCacheItem;

struct __pluckerDBCacheItem
{
  unsigned char *buffer;       /* buffer containing uncompressed text */
  __uint16_t     uid;          /* the unique ID for the record */
  __uint16_t     type;

  PluckerDBCacheItem *next;
};


typedef struct
{
  PluckerDBOperations operations;
  void *stream;

  char docName[ DBPFX_DOCNAME_LEN ];
  __uint16_t version;

  __uint16_t numRecords;
  __uint32_t appInfoOffset;
  __uint32_t version_ofs;
  __uint32_t version_size;
  __uint32_t icon_ofs;
  __uint16_t icon_size;
  __uint32_t sm_icon_ofs;
  __uint16_t sm_icon_size;

  __uint16_t compression_type;
  __uint16_t reserved_count;
  __uint32_t reserved_ofs;

  PluckerDBCacheItem *cache; /* head of the list (most recently used item) */

  __uint8_t  cache_size;
  __uint8_t  cache_capacity;

  BOOL owned;
}
PluckerDBImpl;


typedef struct
{
  FILE *handle;
  int   active;
  char *filename;
}
PluckerDBFileStream;


#define IMAGE_IS_COMPRESSED_FLAG        0x8000
#define IMAGE_HAS_COLORMAP_FLAG        0x4000
#define IMAGE_HAS_TRANSPARENCY_FLAG    0x2000
#define IMAGE_DIRECT_COLOR_FLAG        0x0400
#define IMAGE_4_BYTE_FIELD_FLAG        0x0200

#define IMAGE_COMPRESSION_SCANLINE    0x00
#define IMAGE_COMPRESSION_RLE        0x01
#define IMAGE_COMPRESSION_PACKBITS    0x02
#define IMAGE_COMPRESSION_NONE        0xFF


typedef struct {
  unsigned char red;
  unsigned char green;
  unsigned char blue;
} ColorMapEntry;

static ColorMapEntry Palm8BitColormap[] = {
  { 255, 255, 255 }, { 255, 204, 255 }, { 255, 153, 255 }, { 255, 102, 255 }, 
  { 255,  51, 255 }, { 255,   0, 255 }, { 255, 255, 204 }, { 255, 204, 204 }, 
  { 255, 153, 204 }, { 255, 102, 204 }, { 255,  51, 204 }, { 255,   0, 204 }, 
  { 255, 255, 153 }, { 255, 204, 153 }, { 255, 153, 153 }, { 255, 102, 153 }, 
  { 255,  51, 153 }, { 255,   0, 153 }, { 204, 255, 255 }, { 204, 204, 255 },
  { 204, 153, 255 }, { 204, 102, 255 }, { 204,  51, 255 }, { 204,   0, 255 },
  { 204, 255, 204 }, { 204, 204, 204 }, { 204, 153, 204 }, { 204, 102, 204 },
  { 204,  51, 204 }, { 204,   0, 204 }, { 204, 255, 153 }, { 204, 204, 153 },
  { 204, 153, 153 }, { 204, 102, 153 }, { 204,  51, 153 }, { 204,   0, 153 },
  { 153, 255, 255 }, { 153, 204, 255 }, { 153, 153, 255 }, { 153, 102, 255 },
  { 153,  51, 255 }, { 153,   0, 255 }, { 153, 255, 204 }, { 153, 204, 204 },
  { 153, 153, 204 }, { 153, 102, 204 }, { 153,  51, 204 }, { 153,   0, 204 },
  { 153, 255, 153 }, { 153, 204, 153 }, { 153, 153, 153 }, { 153, 102, 153 },
  { 153,  51, 153 }, { 153,   0, 153 }, { 102, 255, 255 }, { 102, 204, 255 },
  { 102, 153, 255 }, { 102, 102, 255 }, { 102,  51, 255 }, { 102,   0, 255 },
  { 102, 255, 204 }, { 102, 204, 204 }, { 102, 153, 204 }, { 102, 102, 204 },
  { 102,  51, 204 }, { 102,   0, 204 }, { 102, 255, 153 }, { 102, 204, 153 },
  { 102, 153, 153 }, { 102, 102, 153 }, { 102,  51, 153 }, { 102,   0, 153 },
  {  51, 255, 255 }, {  51, 204, 255 }, {  51, 153, 255 }, {  51, 102, 255 },
  {  51,  51, 255 }, {  51,   0, 255 }, {  51, 255, 204 }, {  51, 204, 204 },
  {  51, 153, 204 }, {  51, 102, 204 }, {  51,  51, 204 }, {  51,   0, 204 },
  {  51, 255, 153 }, {  51, 204, 153 }, {  51, 153, 153 }, {  51, 102, 153 },
  {  51,  51, 153 }, {  51,   0, 153 }, {   0, 255, 255 }, {   0, 204, 255 },
  {   0, 153, 255 }, {   0, 102, 255 }, {   0,  51, 255 }, {   0,   0, 255 },
  {   0, 255, 204 }, {   0, 204, 204 }, {   0, 153, 204 }, {   0, 102, 204 },
  {   0,  51, 204 }, {   0,   0, 204 }, {   0, 255, 153 }, {   0, 204, 153 },
  {   0, 153, 153 }, {   0, 102, 153 }, {   0,  51, 153 }, {   0,   0, 153 },
  { 255, 255, 102 }, { 255, 204, 102 }, { 255, 153, 102 }, { 255, 102, 102 },
  { 255,  51, 102 }, { 255,   0, 102 }, { 255, 255,  51 }, { 255, 204,  51 },
  { 255, 153,  51 }, { 255, 102,  51 }, { 255,  51,  51 }, { 255,   0,  51 },
  { 255, 255,   0 }, { 255, 204,   0 }, { 255, 153,   0 }, { 255, 102,   0 },
  { 255,  51,   0 }, { 255,   0,   0 }, { 204, 255, 102 }, { 204, 204, 102 },
  { 204, 153, 102 }, { 204, 102, 102 }, { 204,  51, 102 }, { 204,   0, 102 },
  { 204, 255,  51 }, { 204, 204,  51 }, { 204, 153,  51 }, { 204, 102,  51 },
  { 204,  51,  51 }, { 204,   0,  51 }, { 204, 255,   0 }, { 204, 204,   0 },
  { 204, 153,   0 }, { 204, 102,   0 }, { 204,  51,   0 }, { 204,   0,   0 },
  { 153, 255, 102 }, { 153, 204, 102 }, { 153, 153, 102 }, { 153, 102, 102 },
  { 153,  51, 102 }, { 153,   0, 102 }, { 153, 255,  51 }, { 153, 204,  51 },
  { 153, 153,  51 }, { 153, 102,  51 }, { 153,  51,  51 }, { 153,   0,  51 },
  { 153, 255,   0 }, { 153, 204,   0 }, { 153, 153,   0 }, { 153, 102,   0 },
  { 153,  51,   0 }, { 153,   0,   0 }, { 102, 255, 102 }, { 102, 204, 102 },
  { 102, 153, 102 }, { 102, 102, 102 }, { 102,  51, 102 }, { 102,   0, 102 },
  { 102, 255,  51 }, { 102, 204,  51 }, { 102, 153,  51 }, { 102, 102,  51 },
  { 102,  51,  51 }, { 102,   0,  51 }, { 102, 255,   0 }, { 102, 204,   0 },
  { 102, 153,   0 }, { 102, 102,   0 }, { 102,  51,   0 }, { 102,   0,   0 },
  {  51, 255, 102 }, {  51, 204, 102 }, {  51, 153, 102 }, {  51, 102, 102 },
  {  51,  51, 102 }, {  51,   0, 102 }, {  51, 255,  51 }, {  51, 204,  51 },
  {  51, 153,  51 }, {  51, 102,  51 }, {  51,  51,  51 }, {  51,   0,  51 },
  {  51, 255,   0 }, {  51, 204,   0 }, {  51, 153,   0 }, {  51, 102,   0 },
  {  51,  51,   0 }, {  51,   0,   0 }, {   0, 255, 102 }, {   0, 204, 102 },
  {   0, 153, 102 }, {   0, 102, 102 }, {   0,  51, 102 }, {   0,   0, 102 },
  {   0, 255,  51 }, {   0, 204,  51 }, {   0, 153,  51 }, {   0, 102,  51 },
  {   0,  51,  51 }, {   0,   0,  51 }, {   0, 255,   0 }, {   0, 204,   0 },
  {   0, 153,   0 }, {   0, 102,   0 }, {   0,  51,   0 }, {  17,  17,  17 },
  {  34,  34,  34 }, {  68,  68,  68 }, {  85,  85,  85 }, { 119, 119, 119 },
  { 136, 136, 136 }, { 170, 170, 170 }, { 187, 187, 187 }, { 221, 221, 221 },
  { 238, 238, 238 }, { 192, 192, 192 }, { 128,   0,   0 }, { 128,   0, 128 },
  {   0, 128,   0 }, {   0, 128, 128 }, {   0,   0,   0 }, {   0,   0,   0 },
  {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 },
  {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 },
  {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 },
  {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 },
  {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 },
  {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }, {   0,   0,   0 }};

static ColorMapEntry Palm1BitColormap[] = {{ 255, 255, 255 }, { 0, 0, 0 }};

static ColorMapEntry Palm2BitColormap[] = {
  { 255, 255, 255 }, { 192, 192, 192 }, { 128, 128, 128 }, { 0, 0, 0 }};

static ColorMapEntry Palm4BitColormap[] = {
  { 255, 255, 255 }, { 238, 238, 238 }, { 221, 221, 221 }, { 204, 204, 204 },
  { 187, 187, 187 }, { 170, 170, 170 }, { 153, 153, 153 }, { 136, 136, 136 },
  { 119, 119, 119 }, { 102, 102, 102 }, {  85,  85,  85 }, {  68,  68,  68 },
  {  51,  51,  51 }, {  34,  34,  34 }, {  17,  17,  17 }, {   0,   0,   0 }};


/* default FILE-based plucker file operations */
static int plucker_FILE_seek( void* stream, off_t offset, int whence );
static int plucker_FILE_close( void* stream );
static int plucker_FILE_read( void* stream, void *buffer, size_t size );
static int plucker_FILE_deactivate( void* stream );
static int plucker_FILE_reactivate( void* stream );
static int plucker_FILE_isactive( void* stream );


/* after calling this function, if the record is found, the stream will
 * be positioned immediately after the header for the record.  The 'size'
 * parameter will be the size of the (possibly compressed) record,
 * INCLUDING the header */
static int static_find_record_with_uid( PluckerDBImpl *impl,
                                        __uint16_t     uid,
                                        unsigned char *hdr,
                                        __uint16_t    *size );

/* assumes the stream points to the data-portion of the record to load
 * (ie, static_find_record_with_uid was called immediately prior to
 * this), and that hdr is the header record for that data record. 
 * The 'size' is the size of the (possibly compressed) record, INCLUDING
 * the size of the header record. */
static unsigned char* static_load_record( PluckerDBImpl *impl,
                                          unsigned char *hdr,
                                          __uint16_t     size );

static unsigned char* static_get_record( PluckerDBImpl *impl,
                                         __uint16_t     uid,
                                         __int16_t     *record_type );

/* Like static_get_record, but we only get text records and everything is 
 * put in a custom buffer. Works with continuous records. */

static PluckerTextRecord *static_get_text_record( PluckerDBImpl *impl,
                                                  __uint16_t     uid );

static PluckerImage *static_get_image_record( PluckerDBImpl *impl,
                                              __uint16_t     uid,
                                              int            type);


static void static_get_navigation_metadata( PluckerDBImpl *impl,
                                            __uint16_t uid,
                                           NavigationMetadata **meta );

/* returns non-zero if the given record type represents a compressed record */
static int static_type_is_compressed( int type );

static unsigned char *static_uncompress_record( PluckerDBImpl *impl,
                                                unsigned char *data,
                                                __uint16_t     size,
                                                __uint16_t     uncompressed_size );

static void static_doc_decompression( unsigned char *data,
                                      __uint32_t     size,
                                      unsigned char *uncompressed,
                                      __uint32_t     uncompressed_size );

static void static_zlib_decompression( unsigned char *data,
                                       __uint32_t     size,
                                       unsigned char *uncompressed,
                                       __uint32_t     uncompressed_size );

static void static_delete_cache_item( PluckerDBCacheItem *item );

static int static_find_next_html_record( PluckerDBSearchMark *mark );

static int static_search_paragraph( PluckerTextRecord *record,
                                    PluckerParagraphDescriptor *paragraph,
                                    int paragraph_end,
                                    PluckerDBSearchMatch *match,
                                    PluckerDBSearchMark *mark );

/* These make sure that there are no conflicts with the search thread.
 * It's a very simple mutex like function */
static void static_wait_for_owned( PluckerDBImpl *impl );
static void static_own( PluckerDBImpl *impl );
static void static_unown( PluckerDBImpl *impl );

/* translating from palm pixmap to bitmap */
static void static_transcode_image( unsigned char *source,
                                    unsigned char *dest,
                                    unsigned int  destWidth,
                                    unsigned int  destHeight,
                                    unsigned int  startCol,
                                    unsigned int  startRow );

/* convenience macros for accessing a plucker stream */
#define PLKR_SEEK( impl, ofs, wh ) (impl)->operations.plkr_seek( (impl)->stream, (ofs), (wh) )
#define PLKR_CLOSE( impl )         (impl)->operations.plkr_close( (impl)->stream )
#define PLKR_READ( impl, buf, sz ) (impl)->operations.plkr_read( (impl)->stream, buf, sz )
#define PLKR_DEACTIVATE( impl )    (impl)->operations.plkr_deactivate( (impl)->stream )
#define PLKR_REACTIVATE( impl )    (impl)->operations.plkr_reactivate( (impl)->stream )
#define PLKR_ISACTIVE( impl )      (impl)->operations.plkr_isactive( (impl)->stream )

PluckerDB OpenPluckerDBFile( const char *filename )
{
  PluckerDBOperations ops;
  PluckerDBFileStream *stream;

  /* set up the operations */
  ops.plkr_seek       = plucker_FILE_seek;
  ops.plkr_close      = plucker_FILE_close;
  ops.plkr_read       = plucker_FILE_read;
  ops.plkr_reactivate = plucker_FILE_reactivate;
  ops.plkr_deactivate = plucker_FILE_deactivate;
  ops.plkr_isactive   = plucker_FILE_isactive;

  stream = (PluckerDBFileStream*)malloc( sizeof( PluckerDBFileStream ) );

  /* open the requested file */
  stream->handle = fopen( filename, "rb" );
  if( stream->handle == NULL )
  {
    fprintf( stderr, "[fopen failed]\n" );
    free( stream );
    return NULL;
  }

  stream->filename = _strdup( filename );
  stream->active = TRUE;

  return ConnectPluckerDBStream( (void*)stream, &ops );
}

/*
    @@@chrisb
    Best way to deal with reading constructed plucker data from RAM
    may be to just make an OpenPluckerRAMFile or similar. Need to look
    at how to do that for Palm.
*/


PluckerDB ConnectPluckerDBStream( void *stream, PluckerDBOperations *ops )
{
  PluckerDBImpl *impl;
  unsigned char  buffer[ DBPFX_LEN ];
  __int16_t      word;
  __int32_t      ofs;

  /* set up the stream data and operations */
  impl = (PluckerDBImpl*)malloc( sizeof( PluckerDBImpl ) );
  memset( impl, 0, sizeof( PluckerDBImpl ) );

  /* cache the last six records */
  impl->cache_capacity = 6;

  memcpy( &impl->operations, ops, sizeof( PluckerDBOperations ) );
  impl->stream = stream;

  static_own( impl );

  if( ( word = PLKR_READ( impl, buffer, DBPFX_LEN ) ) != DBPFX_LEN )
  {
    PLKR_CLOSE( impl );
    free( impl );
    fprintf( stderr, "[read of db prefix failed (%d!=%d)]\n", word, DBPFX_LEN );
    return NULL;
  }

  /* examine the magic to see if this is a valid plucker document */
  if( strncmp( buffer+DBPFX_MAGIC_OFFSET, DBPFX_MAGIC, DBPFX_MAGIC_LEN ) != 0 )
  {
		/* It has come to my attention that eyenotes.net uses the plucker format for some of their
		   texts. To read them in VadeMecum we need to look for a magic word of "DataiTxt" here.
			 I will add this if it is deemed necessary, but omit it for now. 
		*/
    PLKR_CLOSE( impl );
    free( impl );
    fprintf( stderr, "[bad magic value]\n" );
    return NULL;
  }

  /* the first 32 bytes (0-31) are a NUL-terminated string containing the name of
   * the document. */
  strcpy( impl->docName, buffer );

  /* get the version number */
  impl->version = PLKR_INT16( buffer+DBPFX_VERSION_OFFSET );

  /* get the offset to the AppInfo record */
  impl->appInfoOffset = PLKR_INT32( buffer+DBPFX_APPINFO_OFFSET );

  /* read the Record-ID List header */
  if( PLKR_READ( impl, buffer, RIDL_LEN ) != RIDL_LEN )
  {
    PLKR_CLOSE( impl );
    free( impl );
    fprintf( stderr, "[eof reading record-id list header]\n" );
    return NULL;
  }

  /* get the number of records in the file */
  impl->numRecords = PLKR_INT16( buffer+RIDL_NUMREC_OFFSET );

  /* get the offset for the first index record */
  PLKR_READ( impl, buffer, 4 );
  ofs = PLKR_INT32( buffer );

  if( impl->appInfoOffset > 0 )
  {
    /* seek to the start of the AppInfo block */
    PLKR_SEEK( impl, impl->appInfoOffset, SEEK_SET );

    /* get the version information */
    impl->version_ofs = PLKR_SEEK( impl, 4+2+2, SEEK_CUR )+2;
    PLKR_READ( impl, buffer, 2 );
    impl->version_size = PLKR_INT16( buffer ) * 2;

    /* move to the title information (which we'll ignore) */
    PLKR_SEEK( impl, impl->version_size, SEEK_CUR );
    PLKR_READ( impl, buffer, 2 );
    word = PLKR_INT16( buffer ) * 2;

    /* get the icon information */
    impl->icon_ofs = PLKR_SEEK( impl, word, SEEK_CUR )+2;
    PLKR_READ( impl, buffer, 2 );
    impl->icon_size = PLKR_INT16( buffer ) * 2;

    /* get the small icon information */
    impl->sm_icon_ofs = PLKR_SEEK( impl, impl->icon_size, SEEK_CUR )+2;
    PLKR_READ( impl, buffer, 2 );
    impl->sm_icon_size = PLKR_INT16( buffer ) * 2;
  }

  /* seek to the index record */
  PLKR_SEEK( impl, ofs, SEEK_SET );

  /* we're now at the first byte of the index record */
  PLKR_READ( impl, buffer, 6 );
  impl->compression_type = PLKR_INT16( buffer+2 );
  impl->reserved_count = PLKR_INT16( buffer+4 );
  impl->reserved_ofs = ofs+6;

  static_unown( impl );

  return (PluckerDB)impl;
}


void ClosePluckerDB( PluckerDB db )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  PluckerDBCacheItem *item;
  PluckerDBCacheItem *next;

  static_wait_for_owned( impl );
  static_own( impl );

  PLKR_CLOSE( impl );

  item = impl->cache;
  while( item != NULL )
  {
    next = item->next;
    static_delete_cache_item( item );
    item = next;
  }

  memset( impl, 0, sizeof( *impl ) );
  free( impl );
}


int PluckerDBVersion( PluckerDB db )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  return impl->version;
}


char* PluckerDBName( PluckerDB db )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  return impl->docName;
}


unsigned char *PluckerDBGetRecord( PluckerDB db,
                                   int       uid,
                                   int      *record_type )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  unsigned char *buf;
  __int16_t rectyp;

  static_wait_for_owned( impl );
  static_own( impl );

  buf = static_get_record( impl, (__uint16_t)uid, &rectyp );
  *record_type = rectyp;

  static_unown( impl );
  return buf;
}

PluckerTextRecord *PluckerDBGetTextRecord( PluckerDB db,
                                           int       uid )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  PluckerTextRecord *buffer;

  static_wait_for_owned( impl );
  static_own( impl );

  buffer = static_get_text_record( impl, (unsigned short)uid );

  static_unown( impl );
  return buffer;
}

int PluckerUIDOfReservedRecord( PluckerDB db,
                                int index_id )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  unsigned char buffer[ 4 ];
  int i;
  __uint16_t name;

  static_wait_for_owned( impl );
  static_own( impl );

  PLKR_REACTIVATE( impl );

  /* seek to the start of the reserved record list */
  PLKR_SEEK( impl, impl->reserved_ofs, SEEK_SET );

  /* search the reserved record list for the one with the indicated name */
  for( i = 0; i < impl->reserved_count; i++ )
  {
    PLKR_READ( impl, buffer, 4 );
    name = PLKR_INT16( buffer );

    if( name == index_id )
    {
      /* return the uid */
      PLKR_DEACTIVATE( impl );
      static_unown( impl );
      return PLKR_INT16( buffer+2 );
    }
  }

  /* if not found, return -1 */
  PLKR_DEACTIVATE( impl );

  static_unown( impl );
  return -1;
}


int PluckerDBGetHomeID( PluckerDB db )
{
  return PluckerUIDOfReservedRecord( db, INDEX_ID_HOME );
}


unsigned char *PluckerDBGetMetadataItem( PluckerDB db,
                                         int item )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  int uid;
  __int16_t record_type;
  __uint16_t record_size;
  __uint16_t type_code;
  __uint16_t length;
  __uint16_t idx;
  unsigned char *buffer;
  unsigned char *ptr;

  uid = PluckerUIDOfReservedRecord( db, INDEX_ID_METADATA );
  if( uid < 0 ) return NULL;

  static_wait_for_owned( impl );
  static_own( impl );

  buffer = static_get_record( impl, (__uint16_t)uid, &record_type );
  if( buffer == NULL )
  {
    static_unown( impl );
    return NULL;
  }

  ptr = buffer+4;
  record_size = PLKR_INT16( ptr );
  ptr += 4;

  idx = 0;
  while( idx < record_size )
  {
    type_code = PLKR_INT16( ptr ); ptr += 2;
    length = PLKR_INT16( ptr ) * 2; ptr += 2;

    if( type_code == item )
    {
      static_unown( impl );
      return ptr;
    }

    ptr += length;
    idx += length + 4;
  }

  static_unown( impl );
  return NULL;
}


int PluckerGetRecordSize( unsigned char *record )
{
  /* 8 byte record header
   * 4 bytes per paragraph record
   * # of bytes of record data */
  return ( 8 + PLKR_INT16( record+2 )*4 + PLKR_INT16( record+4 ) );
}

int PluckerDBGetRecordType( PluckerDB db,
                            int        uid )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  unsigned char hdr[8];
  __uint16_t size;

  static_wait_for_owned( impl );
  static_own( impl );

  PLKR_REACTIVATE( impl );
  if ( static_find_record_with_uid( impl, (unsigned short)uid, hdr, &size ) == -1 )
  {
    PLKR_DEACTIVATE( impl );
    static_unown( impl );
    return -1;
  }
  PLKR_DEACTIVATE( impl );

  static_unown( impl );
  return hdr[6];
}

int PluckerDBTouchItem( PluckerDB db,
                        int uid )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  PluckerDBCacheItem *item;
  PluckerDBCacheItem *prev = NULL;

  static_wait_for_owned( impl );
  static_own( impl );

  /* examine the cache, looking for the given uid */
  item = impl->cache;
  while( item != NULL )
  {
    if( item->uid == uid )
    {
      /* move the item to the front of the cache */
      if( prev != NULL )
      {
        prev->next = item->next;
        item->next = impl->cache;
        impl->cache = item;
      }

      static_unown( impl );

      return 0;
    }

    prev = item;
    item = item->next;
  }

  static_unown( impl );
  /* if the item was not in the cache, return 1 */
  return 1;
}


int PluckerDBFindFirst( PluckerDB db,
                        int record_id,
                        char *pattern,
                        int flags,
                        PluckerDBSearchMatch *match,
                        PluckerDBSearchMark *mark )
{
  /* initialize the mark record */
  mark->abort = 0;
  mark->db = db;
  mark->pattern = pattern;
  mark->restrict_to_record = record_id;
  mark->record_id = record_id;
  mark->record_index = -1;
  mark->text_index = 0;
  mark->paragraph_index = 0;
  mark->flags = flags;

  if( record_id >= 0 )
  {    
    return PluckerDBFindNext( match, mark );
  }

  return PluckerDBFindNextRecord( match, mark );
}

int PluckerDBFindNextRecord( PluckerDBSearchMatch *match,
                             PluckerDBSearchMark  *mark )
{
  PluckerTextRecord *record;
  int position;
  PluckerParagraphDescriptor paragraph_descriptor;
  PluckerDBImpl *impl = (PluckerDBImpl*)mark->db;

  static_wait_for_owned( impl );
  static_own( impl );
  PLKR_REACTIVATE( impl );

  mark->paragraph_index = 0;
  mark->text_index = 0;
  static_find_next_html_record( mark );

  if ( mark->record_index >= 0 )
  {
    record = static_get_text_record( impl, (__uint16_t)mark->record_id );

    if ( record != NULL )
    {
      position = 0;

      while ( mark->paragraph_index < record->nr_paragraphs && !mark->abort )
      {
        GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, record->paragraphs,
          mark->paragraph_index * sizeof( PluckerParagraphDescriptor ) );

        if ( static_search_paragraph( record, &paragraph_descriptor,
                                      position + paragraph_descriptor.size,
                                      match, mark ) )
        {
          /* found one */
          match->record_id = (__uint16_t)mark->record_id;
          static_unown( impl );
          return 1;
        }
        mark->paragraph_index += 1;
        position += paragraph_descriptor.size;
        mark->text_index = position;
      }
      /* found nothing in this record */
      static_unown( impl );
      return -1;
    }
  }

  PLKR_DEACTIVATE( impl );
  static_unown( impl );

  return 0;
}

int PluckerDBFindNext( PluckerDBSearchMatch *match,
                       PluckerDBSearchMark  *mark )
{
  __uint16_t record_id;
  int position;
  int paragraph;
  PluckerTextRecord *record;
  PluckerParagraphDescriptor paragraph_descriptor;
  PluckerDBImpl *impl = (PluckerDBImpl*)mark->db;

  static_wait_for_owned( impl );
  static_own( impl );
  PLKR_REACTIVATE( impl );

  record_id = (__uint16_t)mark->record_id;
  record = static_get_text_record( impl, record_id );

  paragraph = 0;
  position = 0;

  while( paragraph < mark->paragraph_index && !mark->abort )
  {
    GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, record->paragraphs, 
      paragraph * sizeof( PluckerParagraphDescriptor ) );

    position += paragraph_descriptor.size;

    paragraph += 1;
  }

  /* skip through the rest of the paragraphs, searching each one for the text in question */

  while( mark->paragraph_index < record->nr_paragraphs )
  {
    GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, record->paragraphs, 
      mark->paragraph_index * sizeof( PluckerParagraphDescriptor ) );

    if( static_search_paragraph( record, &paragraph_descriptor,                
                                 position + paragraph_descriptor.size,
                                 match, mark ) )
    {
      match->record_id = mark->record_id;
      static_unown( impl );
      PLKR_DEACTIVATE( impl );
      return 1;
    }

    position += paragraph_descriptor.size;

    mark->paragraph_index += 1;
    mark->text_index = position; /* move the text index to the start of the next paragraph */
  }

  static_unown( impl );
  PLKR_DEACTIVATE( impl );
  return 0;
}


int PluckerDBAbortSearch( PluckerDBSearchMark *mark )
{
  mark->abort = 1;
  return 0;
}


int PluckerDBGetText( PluckerDB db,
                      int record_id,
                      int from_index,
                      int to_index,
                      char *buffer,
                      int buffer_size )
{
  PluckerTextRecord *record;
  int index;
  int position;
  int size;
  unsigned char c;
  PluckerParagraphDescriptor paragraph_descriptor;
  int paragraph;

  record = PluckerDBGetTextRecord( db, record_id );

  /* FIXME: what needs to happen, is that the paragraph containing from_index should be searched forward,
   * until it encounters from_index, and if from_index is in the middle of a formatting code, it should
   * be moved forward to the end of the formatting code.  That's what needs to happen.  Yah. */

  /* Stop getting text if we reach the end of the paragraph */
  paragraph = 0;
  position = 0;
  while( position <= from_index && paragraph < record->nr_paragraphs )
  {
    GET_PARAGRAPH_DESCRIPTOR( &paragraph_descriptor, record->paragraphs, 
      paragraph * sizeof( PluckerParagraphDescriptor ) );

    position += paragraph_descriptor.size;

    paragraph += 1;
  }
  
  if ( position < to_index )
    to_index = position;

  /* Now roll back to the beginning of the paragraph and make sure that we start at a proper index and not 
   * in the middle of formatting code */
  position -= paragraph_descriptor.size;

  while( position < from_index )
  {
    if( record->text[ position ] == 0 )
    {      
      position += 1;
      size = ( record->text[ position ] & 0x07 );
      position += size;
      
      /* If we're in the middle of a formatting code, we move the starting point forward */
      if( from_index < position )
        from_index = position;
    }
    else
    {
      position += 1;
    }
  }

  index = from_index;
  position = 0;
  while( index < to_index && position < buffer_size-1 )
  {
    c = record->text[ index ];
    
    if( c == 0 )
    {
      index += 1;

      /* skip over the formatting */
      size = ( record->text[ index ] & 0x07 );

      index += size;
    }
    else if( c == (unsigned char)'\xAD' ) /* soft hyphen */
    {
      /* do nothing */
    }
    else
    {      
      if( c == LDQUO || c == RDQUO )
      {
        c = '"';
      }
      else if( c == LSQUO || c == RSQUO )
      {
        c = '\'';
      }
      else if( c == '\x96' ) /* n-dash */
      {
        c = '-';
      }
      else if( c == '\x97' ) /* m-dash */
      {
        c = '-';
      }

      buffer[ position++ ] = c;
    }

    index += 1;
  }

  buffer[ position ] = 0;

  return 0;
}

int PluckerDBGetShortTitle( PluckerDB db,
                            int record_id,
                            char *buffer,
                            int buffer_size,
                            int use_alt )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  PluckerTextRecord *record;

  static_wait_for_owned( impl );
  static_own( impl );

  record = static_get_text_record( impl, (__uint16_t)record_id );

  static_unown( impl );
  if ( record->navigation )
  {
    strncpy( buffer, record->navigation->short_record_title, buffer_size-1 );
    buffer[buffer_size-1] = 0;
  }
  else if( use_alt )
  {
    PluckerDBGetText( db, record_id, 0, buffer_size, buffer, buffer_size );
  }
  else
  {
    strncpy( buffer, "\0", 1 );
  }

  return 0;
}

int PluckerDBGetLongTitle( PluckerDB db,
                            int record_id,
                            char *buffer,
                            int buffer_size,
                            int use_alt )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  PluckerTextRecord *record;

  static_wait_for_owned( impl );
  static_own( impl );

  record = static_get_text_record( impl, (__uint16_t)record_id );

  static_unown( impl );

  if ( record->navigation )
  {
    strncpy( buffer, record->navigation->long_record_title, buffer_size-1 );
    buffer[buffer_size-1] = 0;
  }
  else if( use_alt )
  {
    PluckerDBGetText( db, record_id, 0, buffer_size, buffer, buffer_size );
  }
  else
  {
    strncpy( buffer, "\0", 1 );
  }
  
  return 0;
}

int PluckerDBGetRecordCount( PluckerDB db )
{
  PluckerDBImpl *impl = (PluckerDBImpl*)db;
  return (int)impl->numRecords;
}

PluckerImage *PluckerDBGetImage( PluckerDB db, 
                                 int record )
{
  PluckerImage *image;
  int type;
  PluckerDBImpl *impl = (PluckerDBImpl*)db;

  type = PluckerDBGetRecordType( db, record );

  static_wait_for_owned( impl );
  static_own( impl );
  
  image = static_get_image_record( impl, (__uint16_t)record, type );

  static_unown( impl );
  return image;
}

static int plucker_FILE_seek( void* stream, off_t offset, int whence )
{
  PluckerDBFileStream *fp = (PluckerDBFileStream*)stream;
  if( !fp->active ) return -1;
  if (fseek( fp->handle, offset, whence ) == 0)
  {
    return (int)ftell(fp->handle);
  }
  return -1;
}


static int plucker_FILE_close( void* stream )
{
  PluckerDBFileStream *fp = (PluckerDBFileStream*)stream;
  if( fp->active )
  {
    fclose( fp->handle );
  }
  free( fp->filename );
  free( fp );
  return 0;
}


static int plucker_FILE_read( void* stream, void *buffer, size_t size )
{
  PluckerDBFileStream *fp = (PluckerDBFileStream*)stream;
  if( !fp->active ) return -1;
  return fread( buffer, 1, size, fp->handle );
}

static int plucker_FILE_deactivate( void* stream )
{
  PluckerDBFileStream *fp = (PluckerDBFileStream*)stream;
  if( !fp->active ) return -1;
  fclose( fp->handle );
  fp->handle = NULL;
  fp->active = FALSE;
  return 0;
}

static int plucker_FILE_reactivate( void* stream )
{
  PluckerDBFileStream *fp = (PluckerDBFileStream*)stream;
  if( fp->active ) return -1;

  fp->handle = fopen( fp->filename, "rb" );
  if( fp->handle == NULL )
    return -1;
    
  fp->active = TRUE;
  return 0;
}

static int plucker_FILE_isactive( void* stream )
{
  PluckerDBFileStream *fp = (PluckerDBFileStream*)stream;
  return fp->active;
}


/* this assumes that the records in the file are in ascending order
 * by uid. */

static int static_find_record_with_uid( PluckerDBImpl *impl,
                                        __uint16_t     uid,
                                        unsigned char *hdr,
                                        __uint16_t    *size )
{
  __uint32_t rec_ofs;
  __uint32_t temp_rec_ofs;
  __uint16_t uid_tmp;
  unsigned char buffer[ 4 ];

  __uint32_t left;
  __uint32_t right;
  __uint32_t mid;

  left = 0;
  right = impl->numRecords - 1;

  while( left <= right )
  {
    mid = ( left + right ) / 2;

    /* seek to the current record-ID list entry */
    PLKR_SEEK( impl, RIDL_FIRST_OFFSET + mid*8, SEEK_SET );
    PLKR_READ( impl, buffer, 4 ); /* read the recordOffset */
    rec_ofs = PLKR_INT32( buffer );

    /* seek to that position */
    PLKR_SEEK( impl, rec_ofs, SEEK_SET );
    PLKR_READ( impl, hdr, DATARECHDR_LEN );
    uid_tmp = PLKR_INT16( hdr );

    if( uid_tmp == uid )
    {
      if( mid+1 < impl->numRecords )
      {
        /* get the beginning of the next record */
        PLKR_SEEK( impl, RIDL_FIRST_OFFSET + (mid+1)*8, SEEK_SET );
        PLKR_READ( impl, buffer, 4 );
        temp_rec_ofs = PLKR_INT32( buffer );
      }
      else
      {
        /* if this is the last record, then we use the end of the file as the begining
         * of the "next record". */
        temp_rec_ofs = PLKR_SEEK( impl, 0, SEEK_END );
      }
      *size = (__uint16_t)( temp_rec_ofs - rec_ofs );
      PLKR_SEEK( impl, rec_ofs+8, SEEK_SET );
      return mid;
    }
    else if( uid  < uid_tmp )
    {
      right = mid-1;
    }
    else
    {
      left = mid+1;
    }
  }

  return -1; /* no such record */
}


static int static_find_record_with_index( PluckerDBImpl *impl,
                                          int             index,
                                          unsigned char *hdr,
                                          __uint16_t    *size )
{
  unsigned char buffer[4];
  __uint32_t ofs,
             ofs2;

  if ( index >= impl->numRecords )
    return -1;

  PLKR_SEEK( impl, RIDL_FIRST_OFFSET + index*8, SEEK_SET );
  PLKR_READ( impl, buffer, 4 );
  ofs = PLKR_INT32( buffer );

  if ( index+1 < impl->numRecords )
  {
    PLKR_SEEK( impl, RIDL_FIRST_OFFSET + (index+1)*8, SEEK_SET );
    PLKR_READ( impl, buffer, 4 );
    ofs2 = PLKR_INT32( buffer );
  }
  else
  {
    ofs2 = PLKR_SEEK( impl, 0, SEEK_END );
  }

  PLKR_SEEK( impl, ofs, SEEK_SET );
  PLKR_READ( impl, hdr, 8 );

  *size = (__uint16_t)(ofs2 - ofs);

  return index;
}

static unsigned char* static_load_record( PluckerDBImpl *impl,
                                          unsigned char *hdr,
                                          __uint16_t     size )
{
  __uint8_t type;
  __uint16_t uncompressed_size;
  unsigned char *buffer;

  uncompressed_size = 8 + PLKR_INT16( hdr+2 )*4 + PLKR_INT16( hdr+4 );
  type = *(hdr+6);

  buffer = (char*)malloc( size );

  memcpy( buffer, hdr, 8 );
  PLKR_READ( impl, buffer+8, size-8 );

  if( static_type_is_compressed( type ) )
  {
    unsigned char *uncompressed;

    uncompressed = static_uncompress_record( impl, buffer, size, uncompressed_size );
    free( buffer );

    buffer = uncompressed;
  }

  return buffer;
}


static unsigned char* static_get_record( PluckerDBImpl *impl,
                                         __uint16_t     uid,
                                         __int16_t     *record_type )
{
  PluckerDBCacheItem *item;
  PluckerDBCacheItem *prev = NULL;
  char hdr[8];
  __uint16_t size;

  *record_type = DATATYPE_INVALID;

  /* first, examine the cache */
  item = impl->cache;
  while( item != NULL )
  {
    if( item->uid == uid )
    {
      /* move the item to the front of the cache */
      if( prev != NULL )
      {
        prev->next = item->next;
        item->next = impl->cache;
        impl->cache = item;
      }

      *record_type = *( item->buffer+6 );
      return item->buffer;
    }

    prev = item;
    item = item->next;
  }

  /* cache missed, make sure there is room in the cache for a new item */
  if( impl->cache_size+1 == impl->cache_capacity )
  {
    /* delete the least recently used item from the cache (the tailmost item) */
    item = impl->cache;
    while( item->next->next != NULL ) item = item->next;
    static_delete_cache_item( item->next );
    item->next = NULL;

    impl->cache_size--;
  }

  PLKR_REACTIVATE( impl );

  /* create a new item and add it to the front of the cache */
  item = (PluckerDBCacheItem*)malloc( sizeof( PluckerDBCacheItem ) );
  item->uid = uid;
  if( static_find_record_with_uid( impl, uid, hdr, &size ) < 0 )
  {
    free( item );
    PLKR_DEACTIVATE( impl );
    return NULL;
  }
  item->buffer = static_load_record( impl, hdr, size );

  if( item->buffer == NULL )
  {
    free( item );
    PLKR_DEACTIVATE( impl );
    return NULL;
  }

  PLKR_DEACTIVATE( impl );

  item->next = impl->cache;
  impl->cache = item;
  impl->cache_size++;

  *record_type = hdr[6];
  item->type = *record_type;

  return item->buffer;
}

static PluckerTextRecord *static_get_text_record( PluckerDBImpl *impl,
                                                  __uint16_t     uid )
{
  PluckerDBCacheItem *item;
  PluckerDBCacheItem *prev = NULL;
  unsigned char hdr[8];
  __uint16_t size;
  __int32_t index,
             nr_paragraphs,
             text_length,
             i,
             tmp_par;
  unsigned char *buffer;
  unsigned char par[4];
              
  PluckerTextRecord *record;

  /* first, examine the cache */
  item = impl->cache;
  while( item != NULL )
  {
    if( item->uid == uid )
    {
      /* move the item to the front of the cache */
      if( prev != NULL )
      {
        prev->next = item->next;
        item->next = impl->cache;
        impl->cache = item;
      }

      if ( item->type == DATATYPE_PHTML || item->type == DATATYPE_PHTML_COMPRESSED )
        return (PluckerTextRecord*)item->buffer;
      else
        return NULL;
    }

    prev = item;
    item = item->next;
  }

  /* cache missed, make sure there is room in the cache for a new item */
  if( impl->cache_size+1 == impl->cache_capacity )
  {
    /* delete the least recently used item from the cache (the tailmost item) */
    item = impl->cache;
    while( item->next->next != NULL ) item = item->next;
    static_delete_cache_item( item->next );
    item->next = NULL;

    impl->cache_size--;
  }

  PLKR_REACTIVATE( impl );
  
  /* create a new item and add it to the front of the cache */
  item = (PluckerDBCacheItem*)malloc( sizeof( PluckerDBCacheItem ) );
  item->uid = uid;
  if( (index = static_find_record_with_uid( impl, uid, hdr, &size )) < 0 )
  {
    free( item );
    PLKR_DEACTIVATE( impl );
    return NULL;
  }

  item->type = hdr[6];

  /* Make sure that this really is a text record */
  if ( item->type != DATATYPE_PHTML && item->type != DATATYPE_PHTML_COMPRESSED )
  {
    free( item );
    PLKR_DEACTIVATE( impl );
    return NULL;
  }
  
  record = (PluckerTextRecord*)malloc( sizeof( PluckerTextRecord ) );
  record->navigation = NULL;
  record->uid = uid;

  /* Is this record continued? */
  if ( RECORD_IS_CONTINUED( hdr ) )
  {
    /* If so we first calculate the total number of paragraphs and the total length of text */
    nr_paragraphs = NR_PARAGRAPHS( hdr );
    text_length = 0;
    for ( i = 0; i < nr_paragraphs; i++ )
    {
      PLKR_READ( impl, par, 4 );
      text_length += PLKR_INT16( par );
    }
    
    do
    {
      /* Go to next record that contains text */
      do
      {
        index++;
        static_find_record_with_index( impl, index, hdr, &size );
      } while ( hdr[6] != DATATYPE_PHTML && hdr[6] != DATATYPE_PHTML_COMPRESSED );

      /* Calculate the text length */
      for ( i = 0; i < NR_PARAGRAPHS( hdr ); i++)
      {
        PLKR_READ( impl, par, 4 );
        text_length += PLKR_INT16( par );
      }

      /* Calculate nr of paragraphs */
      nr_paragraphs += NR_PARAGRAPHS( hdr );

    } while ( RECORD_IS_CONTINUED( hdr ) );

    /* Then we create two buffers of appropriate length */
    record->nr_paragraphs = nr_paragraphs;
    record->text_length = text_length;
    record->paragraphs = (unsigned char*)malloc( nr_paragraphs*4 );
    record->text = (unsigned char*)malloc( text_length );

    /* Now we wind back and copy everything */
    index = static_find_record_with_uid( impl, uid, hdr, &size );
    buffer = static_load_record( impl, hdr, size );

    /* Copy paragraph data */
    memcpy( record->paragraphs, buffer+DATARECHDR_LEN, NR_PARAGRAPHS( hdr )*4 );

    tmp_par = 0;
    for (i = 0; i < NR_PARAGRAPHS( hdr ); i++ )
      tmp_par += PLKR_INT16( buffer+DATARECHDR_LEN+i*4 );

    text_length = tmp_par;
    nr_paragraphs = NR_PARAGRAPHS( hdr );

    /* Copy text data */
    memcpy( record->text, buffer+DATARECHDR_LEN+nr_paragraphs*4, text_length );

    /* Make sure that we don't leak any memory */
    free( buffer );

    /* Get the navigation metadata. */
    if ( RECORD_HAS_METADATA( hdr ) )
    {
      static_get_navigation_metadata( impl, PLKR_INT16( hdr ), &record->navigation );
    }
    

    /* This is all like above except we also copy the data into the appropriate buffers */
    do
    {
      do
      {
        index++;
        static_find_record_with_index( impl, index, hdr, &size );
      } while ( hdr[6] != DATATYPE_PHTML && hdr[6] != DATATYPE_PHTML_COMPRESSED );

      buffer = static_load_record( impl, hdr, size );

      /* Copy the paragraph data */
      memcpy( record->paragraphs + nr_paragraphs*4, buffer+DATARECHDR_LEN, NR_PARAGRAPHS( hdr )*4 );

      tmp_par = 0;
      for (i = 0; i < NR_PARAGRAPHS( hdr ); i++ )
        tmp_par += PLKR_INT16( buffer+DATARECHDR_LEN+i*4 );

      /* Copy the actual text data */
      memcpy( record->text + text_length, buffer+DATARECHDR_LEN+NR_PARAGRAPHS( hdr )*4, tmp_par );

      text_length += tmp_par;
      nr_paragraphs += NR_PARAGRAPHS( hdr );

      /* No leaks here */
      free( buffer );

      /* Get the navigation metadata. */
      if ( RECORD_HAS_METADATA( hdr ) )
      {        
        static_get_navigation_metadata( impl, PLKR_INT16( hdr ), &record->navigation );
      }

      record->last_uid = PLKR_INT16( hdr );

    } while ( RECORD_IS_CONTINUED( hdr ));

  }
  else
  {
    /* If we get here this is just a standard text record */
    buffer = static_load_record( impl, hdr, size );

    if( buffer == NULL )
    {
      free( item );
      free( record );
      PLKR_DEACTIVATE( impl );
      return NULL;
    }

    /* Create a PluckerTextRecord and fill it */
    record->nr_paragraphs = NR_PARAGRAPHS( hdr );
    record->text_length = 0;

    for ( i = 0; i < record->nr_paragraphs; i++ )
    {
      record->text_length += PLKR_INT16( buffer+DATARECHDR_LEN+i*4 );
    }

    record->paragraphs = (unsigned char *)malloc( i*4 );
    memcpy( record->paragraphs, buffer+DATARECHDR_LEN, i*4 );

    record->text = (unsigned char *)malloc( record->text_length );
    memcpy( record->text, buffer+DATARECHDR_LEN+i*4, record->text_length );

    free( buffer );

    /* Get the navigation metadata. */
    if ( RECORD_HAS_METADATA( hdr ) )
    {
      static_get_navigation_metadata( impl, PLKR_INT16( hdr ), &record->navigation );
    }

    record->last_uid = record->uid;
  }

  PLKR_DEACTIVATE( impl );

  item->buffer = (unsigned char*)record;

  item->next = impl->cache;
  impl->cache = item;
  impl->cache_size++;

  return record;
}

static PluckerImage *static_get_image_record( PluckerDBImpl *impl,
                                              __uint16_t     uid,
                                              int            type )
{
  PluckerDBCacheItem *item;
  PluckerDBCacheItem *prev = NULL;
  int image_cols, image_rows;
  unsigned char hdr[8];
  __uint16_t size;
  unsigned char *multiImage=0;
  __uint32_t miPos=0;
  unsigned int x, y;
  PluckerImage *image;
  unsigned char *imageBuffer;
  unsigned char *dib;
  BITMAPINFO bmi;
  unsigned int currentCol, currentRow, lastRow;

  /* first, examine the cache */
  item = impl->cache;
  while( item != NULL )
  {
    if( item->uid == uid )
    {
      /* move the item to the front of the cache */
      if( prev != NULL )
      {
        prev->next = item->next;
        item->next = impl->cache;
        impl->cache = item;
      }

      if ( item->type == DATATYPE_TBMP || item->type == DATATYPE_TBMP_COMPRESSED || item->type == DATATYPE_COMPOSITE_IMAGE )
        return (PluckerImage*)item->buffer;
      else
        return NULL;
    }

    prev = item;
    item = item->next;
  }

  /* cache missed, make sure there is room in the cache for a new item */
  if( impl->cache_size+1 == impl->cache_capacity )
  {
    /* delete the least recently used item from the cache (the tailmost item) */
    item = impl->cache;
    while( item->next->next != NULL ) item = item->next;
    static_delete_cache_item( item->next );
    item->next = NULL;

    impl->cache_size--;
  }

  PLKR_REACTIVATE( impl );

  if( type == DATATYPE_TBMP || type == DATATYPE_TBMP_COMPRESSED )
  {
    image_cols = image_rows = 1;
  }
  else if( type == DATATYPE_COMPOSITE_IMAGE )
  {
    static_find_record_with_uid( impl, uid, hdr, &size );
    multiImage = static_load_record( impl, hdr, size );
    image_cols = PLKR_INT16( multiImage + 8 );
    image_rows = PLKR_INT16( multiImage + 10 );
    miPos = 12;
  }
  else
  {
    /* wrong type of record */
    PLKR_DEACTIVATE( impl );
    return NULL;
  }

  item = (PluckerDBCacheItem*)malloc( sizeof( PluckerDBCacheItem ) );
  item->uid = uid;
  item->type = type;

  image = (PluckerImage*)malloc( sizeof( PluckerImage ) );
  item->buffer = (unsigned char*)image;

  /* now find the actual dimensions */

  if( type == DATATYPE_TBMP || type == DATATYPE_TBMP_COMPRESSED )
  {    
    BitmapPtr tmpBitmap;
    BitmapPtr bitmap;
    UInt16 idealDensity = kDensityOneAndAHalf;
    UInt8 idealBpp = 8;

    static_find_record_with_uid( impl, uid, hdr, &size );
    imageBuffer = static_load_record( impl, hdr, size );

    tmpBitmap = imageBuffer+8;
    bitmap = tmpBitmap;
    
    do
    {
      if (BmpGetDensity(tmpBitmap) == idealDensity)
      {
        if (BmpGetBitDepth(tmpBitmap) == idealBpp)
        {
          bitmap = tmpBitmap;
          break;
        }
      }

      tmpBitmap = BmpGetNextBitmap(tmpBitmap);
    } while (tmpBitmap != NULL);

    image->width = PLKR_INT16( bitmap + 0 );
    image->height = PLKR_INT16( bitmap + 2 );
    free( imageBuffer );
  }
  else
  {
    image->width = 0;
    for( x = 0; x < (unsigned int)image_cols; x++ )
    {
      static_find_record_with_uid( impl, PLKR_INT16( multiImage + miPos + x*2 ), hdr, &size );
      imageBuffer = static_load_record( impl, hdr, size );
      image->width += PLKR_INT16( imageBuffer + 8 );
      free( imageBuffer );
    }

    image->height = 0;

    for( y = 0; y < (unsigned int)image_rows; y++ )
    {
      static_find_record_with_uid( impl, PLKR_INT16( multiImage + miPos + y*image_cols*2 ), hdr, &size );
      imageBuffer = static_load_record( impl, hdr, size );
      image->height += PLKR_INT16( imageBuffer + 10 );
      free( imageBuffer );
    }
  }

  /* create the bitmap */
  bmi.bmiHeader.biSize = sizeof( bmi );
  bmi.bmiHeader.biWidth = image->width;
  bmi.bmiHeader.biHeight = -(signed int)image->height;
  bmi.bmiHeader.biPlanes = 1;
  bmi.bmiHeader.biSizeImage = 0;
  bmi.bmiHeader.biXPelsPerMeter = 0;
  bmi.bmiHeader.biYPelsPerMeter = 0;
  bmi.bmiHeader.biClrUsed = 0;
  bmi.bmiHeader.biClrImportant = 0;
  bmi.bmiHeader.biBitCount = 24;
  bmi.bmiHeader.biCompression = BI_RGB;

  image->image = CreateDIBSection( 0, &bmi, DIB_RGB_COLORS, &dib, NULL, 0 );
  memset( dib, 0, image->width*image->height*3 );

  /* now add the data to the bitmap */
  currentRow = 0;
  for( y = 0; y < (unsigned int)image_rows; y++ )
  {
    currentCol = 0;
    for( x = 0; x < (unsigned int)image_cols; x++ )
    {
      if( multiImage )
      {
        static_find_record_with_uid( impl, PLKR_INT16( multiImage + miPos ), hdr, &size );
        miPos += 2;
      }
      else
      {
        static_find_record_with_uid( impl, uid, hdr, &size );
      }
      imageBuffer = static_load_record( impl, hdr, size );
      static_transcode_image( imageBuffer, dib, image->width, image->height, currentCol, currentRow );
      currentCol += PLKR_INT16( imageBuffer + 8 );
      lastRow = PLKR_INT16( imageBuffer + 10 );
      free( imageBuffer );      
    }
    currentRow += lastRow;
  }

  if( multiImage )
    free( multiImage );

  item->next = impl->cache;
  impl->cache = item;
  impl->cache_size++;

  PLKR_DEACTIVATE( impl );

  return image;
}


static void static_get_navigation_metadata( PluckerDBImpl *impl,
                                            __uint16_t uid,
                                           NavigationMetadata **meta )
{
  __uint16_t position;
  __uint16_t short_title_len;
  __uint16_t long_title_len;
  __uint16_t n;
  __uint16_t size;
  __uint16_t metadata;
  unsigned char hdr[8];
  unsigned char *buffer;
  long f;
  
  /* We assume that the caller has checked for the occurrence of metadata.
   */

  /* If meta is NULL we allocate the memory for it */
  if ( *meta == NULL )
  {
    *meta = (NavigationMetadata*)malloc( sizeof( NavigationMetadata ) );
    memset( *meta, 0, sizeof( NavigationMetadata ) );
  }
  /*Find the metadata*/
  n = static_find_record_with_uid( impl, uid, hdr, &size );  
  f = ftell(((PluckerDBFileStream*)(impl->stream))->handle);
  PLKR_SEEK( impl, size-10, SEEK_CUR );
  f = ftell(((PluckerDBFileStream*)(impl->stream))->handle);
  PLKR_READ( impl, hdr, 2 );
  f = ftell(((PluckerDBFileStream*)(impl->stream))->handle);
  metadata = PLKR_INT16( hdr );

  static_find_record_with_index( impl, n, hdr, &size );
  PLKR_SEEK( impl, metadata-8, SEEK_CUR );
  f = ftell(((PluckerDBFileStream*)(impl->stream))->handle);
  metadata = size - metadata - 2;
  buffer = (unsigned char*)malloc( metadata );
  PLKR_READ( impl, buffer, metadata );
  f = ftell(((PluckerDBFileStream*)(impl->stream))->handle);


  /* First we get the titles */
  long_title_len = strlen( buffer+8 )+1;
  short_title_len = strlen( buffer+8+long_title_len )+1;

  if ( (*meta)->long_record_title == NULL )
  {
    (*meta)->long_record_title = (char*)malloc( long_title_len );
    strcpy( (*meta)->long_record_title, buffer+8 );
  }

  if ( (*meta)->short_record_title == NULL )
  {
    (*meta)->short_record_title = (char*)malloc( short_title_len );
    strcpy( (*meta)->short_record_title, buffer+8+long_title_len );
  }

  /* Then we do the page lists */

  position = PLKR_INT16( buffer+2 );
  
  if ( position != 0xffff )
  {
    if ( (*meta)->pagelists == NULL )
    {
      (*meta)->nr_pagelists = PLKR_INT16( buffer+position );
      (*meta)->pagelists = (PagelistData*)malloc( sizeof ( PagelistData )*(*meta)->nr_pagelists );
      position += 2;

      for ( n = 0; n < (*meta)->nr_pagelists; n++ )
      {
        (*meta)->pagelists[ n ].list_id  = PLKR_INT16( buffer+position );
        (*meta)->pagelists[ n ].prev_uid = PLKR_INT16( buffer+position+2 );
        (*meta)->pagelists[ n ].next_uid = PLKR_INT16( buffer+position+4 );
        position += 6;
      }
    }
    else
    {
      /* If here then add the new unique pagelist elements to the existing */
      /* Will this ever be necessary? */
    }
  }

  /* Next we must load all the other metadata that hasn't been implemented yet :) */


  free( buffer );
}

static int static_type_is_compressed( int type )
{
  return ( ( type == DATATYPE_PHTML_COMPRESSED ) ||
           ( type == DATATYPE_TBMP_COMPRESSED ) ||
           ( type == DATATYPE_LINKS_COMPRESSED ) ||
           ( type == DATATYPE_TABLE_COMPRESSED ) );
}


static unsigned char *static_uncompress_record( PluckerDBImpl *impl,
                                                unsigned char *data,
                                                __uint16_t     size,
                                                __uint16_t     uncompressed_size )
{
  unsigned char *uncompressed;
  int ofs;

  uncompressed = (unsigned char*)malloc( uncompressed_size );

  /* copy the record header into the new buffer */
  ofs = 8 + PLKR_INT16( data+2 ) * 4;
  memcpy( uncompressed, data, ofs );

  if( impl->compression_type == DOC_COMPRESSION )
  {
    static_doc_decompression( data+ofs, size-ofs, uncompressed+ofs, uncompressed_size-ofs );
  }
  else if( impl->compression_type == ZLIB_COMPRESSION )
  {
    static_zlib_decompression( data+ofs, size-ofs, uncompressed+ofs, uncompressed_size-ofs );
  }
  else
  {
    /* unknown compression type! */
    free( uncompressed );
    return NULL;
  }

  return uncompressed;
}


static void static_doc_decompression( unsigned char *data,
                                      __uint32_t     size,
                                      unsigned char *uncompressed,
                                      __uint32_t     uncompressed_size )
{
  /* I confess, this code is taken from Bill Janssen's 'unpluck' library, since I couldn't figure
   * out why my version wouldn't work...  Thanks, Bill! */

  unsigned int  offset;
  unsigned int  src_index;
  unsigned int  dest_index;

  offset      = 0;
  src_index   = 0;
  dest_index  = 0;
  memset( uncompressed, 0, uncompressed_size );

  while ( src_index < size ) {
      unsigned int token;

      token = (unsigned int) data[ src_index++ ];
      if ( 0 < token && token < 9 ) {
          while ( token != 0 ) {
              uncompressed[ dest_index++ ] = data[ src_index++ ];
              token--;
          }
      }
      else if ( token < 0x80 ) {
          uncompressed[ dest_index++ ] = token;
      }
      else if ( 0xc0 <= token ) {
          uncompressed[ dest_index++ ] = ' ';
          uncompressed[ dest_index++ ] = token ^ 0x80;
      } 
      else {
          int m;
          int n;

          token *= 256;
          token += data[ src_index++ ];

          m  = ( token & 0x3fff ) / 8;
          n  = token & 7;
          n += 3;
          while ( n != 0 ) {
              uncompressed[ dest_index ] = uncompressed[ dest_index - m ];
              dest_index++;
              n--;
          }
      }
  }

  return;
}


static void static_zlib_decompression( unsigned char *data,
                                       __uint32_t     size,
                                       unsigned char *uncompressed,
                                       __uint32_t     uncompressed_size )
{
#ifdef PDAV_SHOWGUIDE
  MessageBox(NULL, TEXT("ZLib compression is not supported. If you're seeing this, something went wrong."), TEXT("Unexpected Zlib compression"),MB_OK);
  exit(0);
#else
  int rc;
  uLongf ucsz = uncompressed_size;

  /* this is easy -- just call the zlib uncompress routine */
  rc = uncompress( uncompressed, &ucsz, data, size );
  if( rc == Z_MEM_ERROR )
  {
    fprintf( stderr, "[out of memory doing uncompress]\n" );
  }
  else if( rc == Z_BUF_ERROR )
  {
    fprintf( stderr, "[output buffer too small doing uncompress]\n" );
  }
  else if( rc == Z_DATA_ERROR )
  {
    fprintf( stderr, "[input data corrupted doing uncompress]\n" );
  }
  else if( rc != Z_OK )
  {
    fprintf( stderr, "[error while uncompressing (%d)]\n", rc );
  }
  else if( ucsz != uncompressed_size )
  {
    fprintf( stderr, "[wrong number of bytes was uncompressed (%d instead of %d)]\n", ucsz, uncompressed_size );
  }
#endif
}

static void static_delete_text_cache_item( PluckerDBCacheItem *item )
{
  PluckerTextRecord *tmp = (PluckerTextRecord*)item->buffer;

  if ( tmp->paragraphs != NULL )
    free( tmp->paragraphs );
  if ( tmp->text != NULL )
    free( tmp->text );

  if ( tmp->navigation != NULL )
  {
    if ( tmp->navigation->long_record_title )
      free (tmp->navigation->long_record_title );
    if ( tmp->navigation->short_record_title )
      free (tmp->navigation->short_record_title );
    if ( tmp->navigation->pagelists )
      free (tmp->navigation->pagelists );

    free ( tmp->navigation );
  }
  
  free( tmp );
  memset( item, 0, sizeof( *item ) );
  free( item );
}

static void static_delete_image_cache_item( PluckerDBCacheItem *item )
{
  PluckerImage *image = (PluckerImage*)item->buffer;
  if( image->image )
    DeleteObject( image->image );

  free( image );
  memset( item, 0, sizeof( *item ) );
  free( item );
}

static void static_delete_cache_item( PluckerDBCacheItem *item )
{
  if ( item->type == DATATYPE_PHTML || item->type == DATATYPE_PHTML_COMPRESSED )
  {
    static_delete_text_cache_item( item );
  }
  else if( item->type == DATATYPE_TBMP || item->type == DATATYPE_TBMP_COMPRESSED || item->type == DATATYPE_COMPOSITE_IMAGE )
  {
    static_delete_image_cache_item( item );
  }
  else
  {
    free( item->buffer );
    memset( item, 0, sizeof( *item ) );
    free( item );
  }
}


static int static_find_next_html_record( PluckerDBSearchMark *mark )
{
  __uint32_t rec_ofs;
  __uint16_t uid;
  __uint8_t  type;
  unsigned char buffer[ 4 ];
  unsigned char hdr[ DATARECHDR_LEN ];
  PluckerDBImpl *impl = (PluckerDBImpl*)mark->db;
  BOOL continued;

  PLKR_REACTIVATE( impl );

  continued = FALSE;

  if ( mark->record_index >= 0 )
  {
    PLKR_SEEK( impl, RIDL_FIRST_OFFSET + mark->record_index*8, SEEK_SET );
    PLKR_READ( impl, buffer, 4 );
    rec_ofs = PLKR_INT32( buffer );

    /* seek to that position */
    PLKR_SEEK( impl, rec_ofs, SEEK_SET );
    PLKR_READ( impl, hdr, DATARECHDR_LEN );

    type = hdr[ 6 ];
    continued = RECORD_IS_CONTINUED( hdr );
  }
  else
   mark->record_index = 0; /* to make sure we skip the index record */

  mark->record_id = -1;
  mark->record_index += 1;

  while( mark->record_index < impl->numRecords && !mark->abort )
  {
    PLKR_SEEK( impl, RIDL_FIRST_OFFSET + mark->record_index*8, SEEK_SET );
    PLKR_READ( impl, buffer, 4 );
    rec_ofs = PLKR_INT32( buffer );

    /* seek to that position */
    PLKR_SEEK( impl, rec_ofs, SEEK_SET );
    PLKR_READ( impl, hdr, DATARECHDR_LEN );

    type = hdr[ 6 ];
    uid = PLKR_INT16( hdr );

    if( type == DATATYPE_PHTML || type == DATATYPE_PHTML_COMPRESSED && !continued )
    {
      mark->record_id = uid;
      PLKR_DEACTIVATE( impl );
      return 1;
    }

    mark->record_index += 1;
    continued = RECORD_IS_CONTINUED( hdr );
  }

  PLKR_DEACTIVATE( impl );
  return 0; /* no more records */
}

static int static_search_paragraph( PluckerTextRecord *record,
                                    PluckerParagraphDescriptor *paragraph,
                                    int paragraph_end,
                                    PluckerDBSearchMatch *match,
                                    PluckerDBSearchMark *mark )
{
  unsigned char c;
  unsigned char pattern_c;
  int pattern_position = 0;
  int size;


  match->text_index = mark->text_index;
  match->earliest_look_behind = paragraph_end - paragraph->size;
  match->latest_look_ahead = paragraph_end;

  while( mark->text_index < paragraph_end && !mark->abort )
  {
    c = record->text[ mark->text_index ];

    if( c == 0 )
    {
      mark->text_index += 1;

      /* skip over the formatting */
      size = ( record->text[ mark->text_index ] & 0x07 );

      mark->text_index += size;
    }
    else if( c == (unsigned char)'\xAD' ) /* soft hyphen */
    {
      /* do nothing */
    }
    else
    {
      pattern_c = mark->pattern[ pattern_position ];

      if( ( mark->flags & SEARCH_CASE_INSENSITIVE ) != 0 )
      {
        c = tolower( c );
        pattern_c = tolower( pattern_c );
      }

      if( ( mark->flags & SEARCH_NORMALIZE_QUOTES ) != 0 )
      {
        if( c == LDQUO || c == RDQUO )
        {
          c = '"';
        }
        else if( c == LSQUO || c == RSQUO )
        {
          c = '\'';
        }
      }

      if( ( mark->flags & SEARCH_NORMALIZE_DASH ) != 0 )
      {
        if( c == '\x96' ) /* n-dash */
        {
          c = '-';
        }
        else if( c == '\x97' ) /* m-dash */
        {
          c = '-';
        }
      }

      if( pattern_c != c )
      {
        pattern_position = 0;
      }
      else
      {
        /* if the first character of the pattern matches, then mark this as the start of a tentative
         * match. */
        if( pattern_position == 0 )
        {
          match->text_index = mark->text_index;
        }
        pattern_position += 1;

        /* we found a match! */
        if( mark->pattern[ pattern_position ] == '\0' )
        {
          return 1;
        }
      }
    }

    mark->text_index += 1;
  }

  return 0;
}

static void static_wait_for_owned( PluckerDBImpl *impl )
{
  while ( impl->owned )
  {
    Sleep( 100 );
  }
}

static void static_own( PluckerDBImpl *impl )
{
  impl->owned = 1;
}

static void static_unown( PluckerDBImpl *impl )
{
  impl->owned = 0;
}

/* adapted from the explode utility provided with Plucker */
static void static_transcode_image( unsigned char *source,
                                    unsigned char *dest,
                                    unsigned int  destWidth,
                                    unsigned int  destHeight,
                                    unsigned int  startCol,
                                    unsigned int  startRow )
{
  UInt16   width;
  UInt16   height;
  UInt16   bytesPerRow;
  unsigned int   flags;
  unsigned char  bitsPerPixel;
  unsigned char  version;
  unsigned char  compressionType;
  ColorMapEntry *colorMap;
  unsigned char *imageDataStart, *ptr, *outPtr, *inbyte;
  unsigned char *rowBuf, *lastRow;
  unsigned char  redBits, greenBits, blueBits;
  unsigned long  redMask, greenMask, blueMask;
  unsigned long  redStretch, greenStretch, blueStretch;
  unsigned int   i, j;
  unsigned int   incount, inval, inbit;
  unsigned int   mask;
  unsigned int   imgBytesPerRow;

  UInt16 idealDensity = kDensityOneAndAHalf;
  UInt16 idealBpp = 8;
  BitmapPtr tmpBitmap;
  BitmapPtr bitmap;

  if( source[6] != DATATYPE_TBMP && source[6] != DATATYPE_TBMP_COMPRESSED && source[6] != DATATYPE_COMPOSITE_IMAGE )
  {
    return;
  }

  source += 8;

  tmpBitmap = source;
  bitmap = tmpBitmap;
  
  do
  {
    if (BmpGetDensity(tmpBitmap) == idealDensity)
    {
      if (BmpGetBitDepth(tmpBitmap) == idealBpp)
      {
        bitmap = tmpBitmap;
        break;
      }
    }

    tmpBitmap = BmpGetNextBitmap(tmpBitmap);
  } while (tmpBitmap != NULL);

  source = bitmap;
  BmpGetDimensions(bitmap, &width, &height, &bytesPerRow);
  version = BmpGetVersion(bitmap);
  compressionType   = BmpGetCompressionType(bitmap);
  bitsPerPixel      = BmpGetBitDepth(bitmap);

  flags             = PLKR_INT16( bitmap + 6);

  if( flags & IMAGE_HAS_COLORMAP_FLAG )
  {
    fprintf( stderr, "Palm images with custom colormaps are not currently supported.\n" );
    return;
  }
  else if( bitsPerPixel == 1 )
  {
    colorMap = Palm1BitColormap;
  }
  else if( bitsPerPixel == 2 )
  {
    colorMap = Palm2BitColormap;
  }
  else if( bitsPerPixel == 4 )
  {
    colorMap = Palm4BitColormap;
  }
  else if( bitsPerPixel == 8 )
  {
    colorMap = Palm8BitColormap;
  }
  else if( bitsPerPixel == 16 && ( flags & IMAGE_DIRECT_COLOR_FLAG ) )
  {
    colorMap = NULL;
    redBits   = source[16];
    greenBits = source[17];
    blueBits  = source[18];

    if (blueBits > 8 || greenBits > 8 || redBits > 8) 
    {
      fprintf( stderr, "Can't handle this format DirectColor image -- too wide in some color (%d:%d:%d)\n",
               redBits, greenBits, blueBits);
      return;
    }
    redMask   = (1 << redBits) - 1;
    blueMask  = (1 << blueBits) - 1;
    greenMask = (1 << greenBits) - 1;

    redStretch    = (255 << 24) / redMask;
    greenStretch  = (255 << 24) / greenMask;
    blueStretch   = (255 << 24) / blueMask;

    if( bitsPerPixel > ( 8 * sizeof( unsigned long ) ) )
    {
      fprintf( stderr, "Can't handle this format DirectColor image -- too many bits per pixel (%d)\n",
               bitsPerPixel );
      return;
    }
  }
  else
  {
    fprintf( stderr, "Unknown bits-per-pixel of %d encountered.\n", bitsPerPixel );
    return;
  }

  if ((version == 3) || ( bitsPerPixel == 16 && ( flags & IMAGE_DIRECT_COLOR_FLAG ) ))
  {
    imageDataStart = &bitmap[24];
  }
  else
  {
    imageDataStart = &bitmap[16];
  }


  imgBytesPerRow = destWidth * 3;
  if( imgBytesPerRow % 4 > 0 ) imgBytesPerRow += ( 4 - ( imgBytesPerRow % 4 ) );

  /* row by row, uncompress the Palm image and copy it to the image_bits buffer */
  rowBuf = (unsigned char*)malloc( bytesPerRow * width );
  lastRow = (unsigned char*)malloc( bytesPerRow * width );

  if ( compressionType != BitmapCompressionTypeNone )
  {
    /*
      Skip the compressed length
    */
    if (version >= 3)
    {
      imageDataStart += 4;
    }
    else
    {
      imageDataStart += 2;
    }
  }


  for( i=0, ptr=imageDataStart; i < height; i++ )
  {
    /* first, uncompress the Palm image */
    if( ( flags & IMAGE_IS_COMPRESSED_FLAG ) && ( compressionType == IMAGE_COMPRESSION_RLE ) )
    {
      for( j = 0; j < bytesPerRow; )
      {
        incount = *ptr++;
        inval = *ptr++;
        memset( rowBuf + j, inval, incount );
        j += incount;
      }
    }
    else if( ( flags & IMAGE_IS_COMPRESSED_FLAG ) && ( compressionType == IMAGE_COMPRESSION_SCANLINE ) )
    {

      for (j = 0; j < bytesPerRow;  j += 8)
      {
        incount = *ptr++;
        inval = ( ( bytesPerRow - j ) < 8 ) ? ( bytesPerRow - j ) : 8;
        for( inbit = 0; inbit < inval; inbit += 1 )
        {
          if( incount & ( 1 << ( 7 - inbit ) ) )
            rowBuf[j + inbit] = *ptr++;
          else
            rowBuf[j + inbit] = lastRow[j + inbit];
        }
      }
      memcpy( lastRow, rowBuf, bytesPerRow );
    }
    else if( ( ( flags & IMAGE_IS_COMPRESSED_FLAG ) && ( compressionType == IMAGE_COMPRESSION_NONE ) ) ||
             ( flags & IMAGE_IS_COMPRESSED_FLAG ) == 0 )
    {
      memcpy( rowBuf, ptr, bytesPerRow );
      ptr += bytesPerRow;
    }

    outPtr = dest + imgBytesPerRow * (startRow + i) + startCol*3;
    /* next, write it to the GDK bitmap */
    if( colorMap )
    {
      mask = ( 1 << bitsPerPixel ) - 1;
      for( inbit = 8 - bitsPerPixel, inbyte = rowBuf, j = 0; j < width; j++ )
      {
        inval = ( (*inbyte) & ( mask << inbit ) ) >> inbit;

        /* correct for oddity of the 8-bit color Palm pixmap... */
        if( ( bitsPerPixel == 8 ) && ( inval == 0xFF ) ) inval = 231;

        outPtr[ j*3 + 0 ] = colorMap[ inval ].blue;
        outPtr[ j*3 + 1 ] = colorMap[ inval ].green;
        outPtr[ j*3 + 2 ] = colorMap[ inval ].red;

        if( !inbit )
        {
          inbyte++;
          inbit = 8 - bitsPerPixel;
        }
        else
        {
          inbit -= bitsPerPixel;
        }
      }
    }
    else if( bitsPerPixel == 16 )
    {
      for( inbyte = rowBuf, j = 0; j < width; ++j )
      {
        inval = ( inbyte[0] << 8 ) | inbyte[1];

        outPtr[ j*3 + 0 ] = (unsigned char)((((inval >> 0) & blueMask) * blueStretch) >> 24);
        outPtr[ j*3 + 1 ] = (unsigned char)((((inval >> blueBits) & greenMask) * greenStretch) >> 24);
        outPtr[ j*3 + 2 ] = (unsigned char)((((inval >> (bitsPerPixel - redBits)) & redMask) * redStretch) >> 24);
        inbyte += 2;
      }
    }
  }

  free( rowBuf );
  free( lastRow );
}