/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: plucker database module
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2002 by Bill Janssen, Mark Ian Lillywhite and Michael
 *   Nordstrm, also under the GPL
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * These routines encapsulate the Plucker file format and provide means for
 * an application to query records in a plucker database.  Unlike the
 * 'unpluck' library by Bill Janssen et. al. (which is most appropriate for
 * a PC environment with fewer resource constraints), this library's goal
 * is minimal resource consumption, which means each lookup requires
 * file I/O reads simply to locate the record.  A binary search is used to
 * speed the search.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __PLKFMT_H__
#define __PLKFMT_H__

#include <stdio.h>

typedef long off_t;

/* the type of compression used in the document */
#define DOC_COMPRESSION    ( 0x0001 )
#define ZLIB_COMPRESSION   ( 0x0002 )


/* Datatype constants (mostly borrowed from the 'unpluck' library). */
#define DATATYPE_INVALID               (     -1 )

#define DATATYPE_PHTML                 (      0 )
#define DATATYPE_PHTML_COMPRESSED      (      1 )
#define DATATYPE_TBMP                  (      2 )
#define DATATYPE_TBMP_COMPRESSED       (      3 )
#define DATATYPE_MAILTO                (      4 )
#define DATATYPE_LINK_INDEX            (      5 )
#define DATATYPE_LINKS                 (      6 )
#define DATATYPE_LINKS_COMPRESSED      (      7 )
#define DATATYPE_BOOKMARKS             (      8 )
#define DATATYPE_CATEGORY              (      9 )
#define DATATYPE_METADATA              (     10 )
#define DATATYPE_STYLE_SHEET           (     11 )
#define DATATYPE_FONT_PAGE             (     12 )
#define DATATYPE_TABLE                 (     13 )
#define DATATYPE_TABLE_COMPRESSED      (     14 )
#define DATATYPE_COMPOSITE_IMAGE       (     15 )
#define DATATYPE_PAGELIST_METADATA 	   (     16 )
#define DATATYPE_SORTED_URL_INDEX 	   (     17 )
#define DATATYPE_SORTED_URL 	         (     18 )
#define DATATYPE_SORTED_URL_COMPRESSED (     19 )
#define DATATYPE_EXT_ANCHOR_INDEX 	   (     20 )
#define DATATYPE_EXT_ANCHOR 	         (     21 )
#define DATATYPE_EXT_ANCHOR_COMPRESSED (     22 )


/* Index constants, used to identify which type of reserved record to retrieve */
#define INDEX_ID_HOME                  (      0 )
#define INDEX_ID_BOOKMARKS             (      1 )
#define INDEX_ID_URL_HANDLING          (      2 )
#define INDEX_ID_DEFAULT_CATEGORY      (      3 )
#define INDEX_ID_METADATA              (      4 )

/* Metadata constants, used to identify which metadata item to retrieve */
#define METADATA_ID_CHARSET            (      0 )
#define METADATA_ID_EXCEPTCHARSET      (      1 )
#define METADATA_ID_OWNERID            (      2 )
#define METADATA_ID_AUTHOR             (      3 )
#define METADATA_ID_TITLE              (      4 )
#define METADATA_ID_PUBDATE            (      5 )

/* "borrowed" this from unpluck.h: decodes a plucker instruction */
#define PLKR_FC_CODE(v)         (((v)>>3) & 0x1F)

/* These macros fix the byte order of the integer at the address given by 's'.
 * 's' must be an unsigned character pointer */
#define PLKR_INT16( s )       ( ( (s)[0] << 8 ) + (s)[1] )
#define PLKR_INT24( s )       ( ( (s)[0] << 16 ) + ( (s)[1] << 8 ) + (s)[2] )
#define PLKR_INT32( s )       ( ( (s)[0] << 24 ) + ( (s)[1] << 16 ) + ( (s)[2] << 8 ) + (s)[3] )

/* these are flags that define the behavior of the search */
#define SEARCH_CASE_INSENSITIVE   ( 0x0001 )
#define SEARCH_NORMALIZE_QUOTES   ( 0x0002 )
#define SEARCH_NORMALIZE_DASH     ( 0x0004 )

/* define the set of valid operations that may be performed on a plucker stream */
typedef struct
{
  /* returns byte position in the stream */
  int (*plkr_seek)( void* stream, off_t offset, int whence );

  /* closes the stream and disposes of any allocated memory */
  int (*plkr_close)( void* stream );

  /* reads up to 'size' bytes into buffer, and returns the number of bytes read */
  int (*plkr_read)( void* stream, void* buffer, size_t size );

  /* "deactivates" the stream, so that it is disconnected from any external resources */
  int (*plkr_deactivate)( void* stream );

  /* "activates" the stream, so that it is reassociated with any external resources */
  int (*plkr_reactivate)( void* stream );

  /* indicates whether the stream is active or not */
  int (*plkr_isactive)( void* stream );
}
PluckerDBOperations;

/* define some opaque datatypes, for encapsulation */
typedef void* PluckerDB;

/* define the structure to be used to return matches while searching */
typedef struct
{
  int record_id;            /* the record-id containing the match */
  int text_index;           /* the index in the text record that matches */
  int earliest_look_behind; /* the furthest we can look backwards in displaying context */
  int latest_look_ahead;    /* the furthest we can look forwards in displaying context */
}
PluckerDBSearchMatch;

/* define the structure to be used while searching a plucker database for text */
typedef struct
{
  PluckerDB db;           /* the database being searched */
  int record_id;          /* the record-id currently being searched */
  int record_index;       /* the index in the record list of the current record */
  int paragraph_index;    /* the index of the current paragraph being searched */
  int text_index;         /* the index in the text record to inspect next */
  char *pattern;          /* the pattern to search for */
  int restrict_to_record; /* restrict the search to the given record only */
  int flags;              /* the flags applicable to this search */
  int abort;              /* used internally to detect whether the search has been aborted by another thread */
}
PluckerDBSearchMark;

/* the format of a plucker database record header */
typedef struct 
{
  int uid        : 16;
  int paragraphs : 16;
  int size       : 16;
  int type       : 8;
  int reserved   : 8;
} PluckerDataRecordHeader;

/* the format of a plucker paragraph descriptor */
typedef struct
{
  unsigned int size       : 16;
  unsigned int attributes : 16;
} PluckerParagraphDescriptor;

typedef struct
{
  unsigned short list_id;
  unsigned short prev_uid;
  unsigned short next_uid;
} PagelistData;

typedef struct
{
  char *long_record_title;
  char *short_record_title;

  unsigned short nr_pagelists;
  PagelistData *pagelists;

} NavigationMetadata;

typedef struct
{
  unsigned short uid;
  unsigned short last_uid;
  long nr_paragraphs;
  unsigned long text_length;

  unsigned char *paragraphs;
  unsigned char *text;

  NavigationMetadata *navigation;
} PluckerTextRecord;

#ifdef __WINDOWS__
typedef struct
{
  unsigned short width;
  unsigned short height;
  HBITMAP image;
} PluckerImage;
#endif

#define GET_DATA_RECORD_HEADER( hdr, data, pos ) \
  memcpy( hdr, &data[ pos ], sizeof( PluckerDataRecordHeader ) ); \
  (hdr)->uid = FIX_BYTE_ORDER_S( (hdr)->uid ); \
  (hdr)->paragraphs = FIX_BYTE_ORDER_S( (hdr)->paragraphs ); \
  (hdr)->size = FIX_BYTE_ORDER_S( (hdr)->size );

#define GET_PARAGRAPH_DESCRIPTOR( para, data, pos ) \
  memcpy( para, &data[ pos ], sizeof( PluckerParagraphDescriptor ) ); \
  (para)->size = FIX_BYTE_ORDER_S( (para)->size ); \
  (para)->attributes = FIX_BYTE_ORDER_S( (para)->attributes );

#define FIX_BYTE_ORDER_S( s )  ( ( s >> 8 ) & 0x00FF ) + ( ( s & 0x00FF ) << 8 )

#define FIX_BYTE_ORDER_L( l )  ( ( l >> 24 ) & 0x000000FF ) + \
                               ( ( ( l >> 16 ) & 0x000000FF ) << 8 ) + \
                               ( ( ( l >> 8 ) & 0x000000FF ) << 16 ) + \
                               ( ( l & 0x000000FF ) << 24 )

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Opens the given plucker database identified by the given file.  Returns NULL if it
 * is unsuccessful.
  */
PluckerDB OpenPluckerDBFile( const char *filename );

/**
 * this function assumes that 'stream' is pointing to the first byte of the
 * data in a plucker database.  The seek operation should always treat this position as
 * the beginning of the file.
 */
PluckerDB ConnectPluckerDBStream( void *stream, PluckerDBOperations *ops );

/**
 * Closes (and deallocates) the given plucker database object.
 */
void ClosePluckerDB( PluckerDB db );

/**
 * Returns the minimum version of plucker needed to read this database.
 */
int PluckerDBVersion( PluckerDB db );

/**
 * Returns the name of the plucker database.  The return value should be considered read-only, and
 * should never be freed.
 */
char* PluckerDBName( PluckerDB db );

/* A macro for retrieving the title of the plucker database. */
#define PluckerDBTitle( db )   PluckerDBGetMetadataItem( db, METADATA_ID_TITLE )

/* A macro for retrieving the author of the plucker database */
#define PluckerDBAuthor( db )  PluckerDBGetMetadataItem( db, METADATA_ID_AUTHOR )

/**
 * Returns a pointer to the record identified by uid, with the record's type being returned in
 * the 'record_type' variable.  If the given uid does not exist in the database, NULL is
 * returned and 'record_type' is DATATYPE_INVALID.
 */
unsigned char *PluckerDBGetRecord( PluckerDB db,
                                   int       uid,
                                   int      *record_type );

/**
 * Like PluckerDBGetRecord but only works with text records and returns a custom buffer.
 * Supports continuous records. 
 */

PluckerTextRecord *PluckerDBGetTextRecord( PluckerDB db,
                                           int       uid );

/**
 * If the given record is cached, promote to the front of the cache and return 0.  If the
 * record is not in the cache, do nothing and return 1.
 */
int PluckerDBTouchItem( PluckerDB db,
                        int uid );

/**
 * Returns the record id of the record for the given INDEX_ID_xxx constant, or -1 if the record
 * does not exist.
 */
int PluckerUIDOfReservedRecord( PluckerDB db,
                                int index_id );

/**
 * Returns the record id of the 'home' record for this database.  This function should always
 * return a valid record ID; if it does not, the database itself should be considered corrupt.
 */
int PluckerDBGetHomeID( PluckerDB db );

/**
 * Returns the requested metadata item (identified by a METADATA_xxx constant).  Returns NULL
 * if the requested item does not exist in the database.
 */
unsigned char *PluckerDBGetMetadataItem( PluckerDB db,
                                         int item );

/**
 * Returns the size of the record whose buffer is given as the parameter.  This is the
 * complete size of the record, including all header information.  This function assumes
 * that a valid plucker database record (as returned from PluckerDBGetRecord) is passed
 * as the parameter.
 */
int PluckerGetRecordSize( unsigned char *record );


/**
 * Returns the type of the record without reading the whole record.
 */
int PluckerDBGetRecordType( PluckerDB db,
                            int        uid );

/**
 * Search the given database for the first instance of the given text string.  If the record-id
 * is >= 0, then search only the given record.  If a match is found, the function returns non-zero
 * and populates the 'match' parameter.  Otherwise it returns zero.
 *
 * Subsequent matches may be obtained by calling the PluckerDBFindNext function.
 */
int PluckerDBFindFirst( PluckerDB db,
                        int record_id,
                        char *pattern,
                        int flags,
                        PluckerDBSearchMatch *match,
                        PluckerDBSearchMark *mark );

/**
 * Search for the next match, using the same parameters as the PluckerDBFindFirst function that
 * must have been called prior to this.  If a match is found, returns non-zero and populates the
 * 'match' parameter, otherwise returns zero.
 */

int PluckerDBFindNextRecord( PluckerDBSearchMatch *match,
                             PluckerDBSearchMark  *mark );

int PluckerDBFindNext( PluckerDBSearchMatch *match,
                       PluckerDBSearchMark *mark );

/**
 * Causes the current search to abort.
 */
int PluckerDBAbortSearch( PluckerDBSearchMark *mark );

/**
 * Extract the text in the given record between the given indices.  The text is copied into buffer,
 * and all formatting is stripped from the text.
 */
int PluckerDBGetText( PluckerDB db,
                      int record_id,
                      int from_index,
                      int to_index,
                      char *buffer,
                      int buffer_size );

int PluckerDBGetShortTitle( PluckerDB db,
                            int record_id,
                            char *buffer,
                            int buffer_size,
                            int use_alt );

int PluckerDBGetLongTitle(  PluckerDB db,
                            int record_id,
                            char *buffer,
                            int buffer_size,
                            int use_alt );

int PluckerDBGetRecordCount( PluckerDB db );

#ifdef __WINDOWS__
PluckerImage *PluckerDBGetImage( PluckerDB db, 
                                 int record );
#endif

#ifdef __cplusplus
}
#endif

#endif /* __PLKFMT_H__ */
