/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: reader interface
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

/* TODO:
 *   Handle WM_HIBERNATE and WM_ACTIVATE properly
 */

#define CINTERFACE
#define COBJMACROS
#define INITGUID


#include <windows.h>
#include <dbgapi.h>
#include <aygshell.h>
#include <commdlg.h>
#include <commctrl.h>

#include "resrc1.h"
#include "reader-ui.h"

#include "plkfmt.h"

#include "pagination.h"
#include "pagination_cache.h"
#include "reader-settings.h"
#include "library.h"
#include "library-ui.h"
#include "reader-toolbar.h"
#include "highlights.h"
#include "bookmarks.h"
#include "get-text.h"
#include "colorpicker.h"
#include "notes.h"
#include "about.h"
#include "search-dlg.h"
#include "image-viewer.h"

#ifdef PDAV_SHOWGUIDE
#include "MapViewer.h"
#include "SessionViewer.h"
#include "SessionSearch.h"
#include "SessionSearchScreen.h"
#include "ExhibitorViewer.h"
#include "ExhibitorListViewer.h"
#include "PluckerWriter.h"
#include "PDAVShowGuide.h"

#include <Pimstore.h>

#endif


#define SEL_START                  (    0 )
#define SEL_EXTEND                 (    1 )
#define SEL_END                    (    2 )
#define SEL_CLEARSEL               (    3 )


#define RS_HIDE_BARS               (    0 )
#define RS_REPAGINATE              (    1 )
#define RS_SHOW_BARS               (    2 )
#define RS_RENDER_TEXT             (    3 )

#define RS_START_STATE             RS_HIDE_BARS
#define RS_END_STATE               RS_RENDER_TEXT


#define WM_READER_CTX_MENU         ( WM_USER+1 )
#define WM_READER_ENDSELECTION     ( WM_USER+2 )
#define WM_READER_OPENRECORD       ( WM_USER+3 )


#define ID_OPT_ADD_BOOKMARK        (    1 )
#define ID_GOTO_BOOKMARK           (   10 )
#define ID_REPOSITION_BOOKMARK     (  500 )
#define ID_RENAME_BOOKMARK         ( 1000 )
#define ID_DELETE_BOOKMARK         ( 1500 )

#define ID_CHANGE_HIGHLIGHT_COLOR  ( 40001 )
#define ID_EDIT_ANNOTATION         ( 40002 )
#define ID_DELETE_HIGHLIGHT        ( 40003 )

#define HOTSPOT_SIZE        ( 20 )

#define MAX_READER_WINDOWS  ( 1 )


typedef struct
{
  HWND hWnd;
  
  Library* library;
  Document* document;
  
  PluckerDB plkr;
  int current_record;

  PaginationInfo info;
  int redraw_state;
  RECT original_size;

  int delta_y;
  int delta_x;

  BOOL cpos_valid;
  long int cpos;
  int y_offset;

  BOOL add_to_history;

  BOOL change_para;
  unsigned short int new_para;

  HWND hscroll;
  HWND vscroll;
  HWND toolbar;

  UINT autoscroll_timer_id;

  SelectionState selection;
  DocumentPoint clickPt;
  BOOL mouse_down;

  PAGINATION_CACHE *cache;
  HighlightManager  highlight_manager;
  LinkedList       *highlights;

  BookmarkManager   bookmark_manager;
  NoteManager       note_manager;
#ifndef PDAV_SHOWGUIDE
  HISTORY_LIST     *history;
#endif
  HISTORY_LIST     *current_search_list;
  SearchReturn      current_search;

  int               nr_inset_scrolls;

} ReaderWindowData;


/* this makes no sense now, but eventually I'd like to make this so you can have multiple
 * document windows open simultaneously... this facilitates that. */

static ReaderWindowData g_windows[ MAX_READER_WINDOWS ] = { 0 };


static ReaderWindowData *FindReaderWindow( HWND hWnd );
static void CreateReaderWindow( HWND hWnd, LPARAM lParam );
static void DestroyReaderWindow( HWND hWnd );
static void PaintReaderWindow( HWND hWnd );
static void GetReaderBounds( HWND hWnd, RECT *bounds );
static BOOL RecognizeContextMenuGesture( HWND hWnd, LPARAM lParam );
static void DoPopupMenu( HWND hWnd, LPARAM lParam );
static void DoSelectionPopupMenu( HWND hWnd, LPARAM lParam );
static void HandleKeyUp( HWND hWnd, WPARAM wParam, LPARAM lParam );
static BOOL DoButtonDown( HWND hWnd, LPARAM lParam, int mode, BOOL check_tap_action );
static void RedrawReaderWindow( HWND hWnd, BOOL repaginate );
static void NextRedrawState( HWND hWnd );
static void DoToggleFullscreen( HWND hWnd );
static void DoResizeScreen( HWND hWnd );
static void UpdateScrollPos( HWND hWnd );
static void DoReaderSettings( HWND hWnd );
static void DoCopySelection( HWND hWnd );
static int  DoVerticalScroll( HWND hWnd, WPARAM wParam, LPARAM lParam );
static int  DoHorizontalScroll( HWND hWnd, WPARAM wParam, LPARAM lParam );
#ifndef PDAV_SHOWGUIDE
static void GoHome( HWND hWnd );
#else
void GoHome( HWND hWnd );
#endif

static void ComputeScrollbars( HWND hWnd, ReaderWindowData *data, int cx, int cy );
static void RefreshToolbar( HWND hWnd, ReaderWindowData *data );
static UINT DoToolbarMenu( HWND hWnd, WPARAM wParam, LPARAM lParam );
static void InterpretTapAction( HWND hWnd, LPARAM lParam, ReaderWindowData *data );
static int  ScrollDocumentBy( HWND hWnd, int dy, ReaderWindowData *data );
static HMENU BuildPopupMenu( HWND hWnd );
static void GoTop( HWND hWnd );
static void GoBottom( HWND hWnd );
static void StartAutoscroll( HWND hWnd );
static void StopAutoscroll( HWND hWnd );
static void DoHiliteSelection( HWND hWnd, BOOL quick, BOOL annotate );
static void DoBookmarksMenu( HWND hWnd );
static void DoHighlightPopupMenu( HWND hWnd, Highlight *highlight, LPARAM lParam );
static BOOL GetHighlightColor( HWND hWnd, Highlight *highlight );
static BOOL GetAnnotationText( HWND hWnd, int *id );
static void UpdateCurrentHistory( HWND hWnd );
static void AddCurrentHistory( HWND hWnd );
static void GoPrevious( HWND hWnd );
static void GoNext( HWND hWnd );
#ifdef PDAV_SHOWGUIDE
void GoCurrent(Library *library, HWND hWnd);
#else
void GoCurrent( HWND hWnd );
#endif
static void DoExitVadeMecum( HWND hWnd );
static void ToggleToolbar( HWND hWnd );
static void DoSearchDocument( HWND hWnd, BOOL forceNewSearch );
static void DoRecordInfo( HWND hWnd );
static void GoNextSearch( HWND hWnd );
static void GoPreviousSearch( HWND hWnd );
static void SetInSearch( HWND hWnd, BOOL in );

static void OpenRecord( HWND hWnd, int record, int para, BOOL use_cpos, int cpos, int yofs, BOOL add_history );
static void OpenTextRecord( HWND hWnd, int record, int para, BOOL use_cpos, int cpos, int yofs, BOOL add_history );
static void OpenMailToRecord( HWND hWnd, int record );
static void OpenInvalidRecord( HWND hWnd, int record );
static void OpenImageRecord( HWND hWnd, int record );

static int  ComputeVerticalPageSize( HWND hWnd, ReaderWindowData *data );
static int  ComputeVerticalLineSize( HWND hWnd, ReaderWindowData *data );

static UINT HandleNotification( HWND hWnd, WPARAM wParam, LPARAM lParam );
static void HandleExternalLink( HWND hWnd, int record );
static void CreateInsetScrolls( HWND hWnd, int nr );
static void DeleteInsetScrolls( HWND hWnd );


LRESULT CALLBACK ReaderWindowProc( HWND hWnd,
                                   UINT uMsg,
                                   WPARAM wParam,
                                   LPARAM lParam )
{
  ReaderWindowData *data;
  MSG msg;

  switch( uMsg )
  {
    case WM_CREATE:
      CreateReaderWindow( hWnd, lParam );
      break;
    case WM_DESTROY:
      DestroyReaderWindow( hWnd );
      break;
    case WM_LBUTTONDOWN:
      StopAutoscroll( hWnd );

      data = FindReaderWindow( hWnd );

      SetCapture( hWnd );
      DoButtonDown( hWnd, lParam, SEL_START, TRUE );
      if( RecognizeContextMenuGesture( hWnd, lParam ) )
      {
        DoButtonDown( hWnd, lParam, SEL_END, FALSE );
        PostMessage( hWnd, WM_READER_CTX_MENU, wParam, lParam );
      }

      if (data)
      {
        data->mouse_down = TRUE;
      }

      break;
    case WM_MOUSEMOVE:
      /* if there is a repaint message in the queue, don't process mouse-move */
      PeekMessage( &msg, hWnd, WM_PAINT, WM_PAINT, PM_NOREMOVE );
      if( msg.message == WM_PAINT ) break;

      data = FindReaderWindow( hWnd );
      if (data)
      {
        if( data->mouse_down )
        {
          DoButtonDown( hWnd, lParam, SEL_EXTEND, TRUE );
        }
      }
      break;
    case WM_LBUTTONUP:
      data = FindReaderWindow( hWnd );
      ReleaseCapture();
      DoButtonDown( hWnd, lParam, SEL_END, TRUE );
      if (data)
      {
        data->mouse_down = FALSE;
      }
      break;
    case WM_KEYUP:
      StopAutoscroll( hWnd );
      HandleKeyUp( hWnd, wParam, lParam );
      break;
    case WM_ACTIVATE:
      if( LOWORD( wParam ) != WA_INACTIVE )
      {
        DoResizeScreen( hWnd );
      }
      break;
    case WM_READER_CTX_MENU:
      DoPopupMenu( hWnd, lParam );
      break;
    case WM_READER_ENDSELECTION:
      DoButtonDown( hWnd, lParam, SEL_CLEARSEL, TRUE );
      break;
    case WM_PAINT:
      PaintReaderWindow( hWnd );
      break;
    case WM_VSCROLL:
      return DoVerticalScroll( hWnd, wParam, lParam );
    case WM_HSCROLL:
      return DoHorizontalScroll( hWnd, wParam, lParam );
    case WM_SIZE:
      data = FindReaderWindow( hWnd );
      if (data)
      {
        RefreshToolbar( hWnd, data );
      }
      break;
    case WM_TIMER:
      if( wParam == ID_AUTOSCROLL_INCREMENT )
      {
        data = FindReaderWindow( hWnd );
        if (data)
        {
          data->autoscroll_timer_id = 0;

          ScrollDocumentBy( hWnd, data->library->autoscroll_amount, data );
          StartAutoscroll( hWnd );
        }
      }
      break;
    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
#ifndef PDAV_SHOWGUIDE
        case ID_GO_TO_LIBRARY:
          DestroyWindow( hWnd );
          break;
#endif
        case ID_TOGGLE_FULLSCREEN:
          DoToggleFullscreen( hWnd );
          break;
#ifndef PDAV_SHOWGUIDE
        case ID_READER_SETTINGS:
          DoReaderSettings( hWnd );
          break;
        case ID_COPY_TEXT:
          DoCopySelection( hWnd );
          break;
        case ID_HILITE_TEXT:
          DoHiliteSelection( hWnd, FALSE, FALSE );
          break;
        case ID_QUICK_HILITE_TEXT:
          DoHiliteSelection( hWnd, TRUE, FALSE );
          break;
        case ID_ANNOTATE_TEXT:
          DoHiliteSelection( hWnd, FALSE, TRUE );
          break;
        case ID_GO_HOME:
          GoHome( hWnd );
          break;
        case ID_GO_TOP:
          GoTop( hWnd );
          break;
        case ID_GO_BOTTOM:
          GoBottom( hWnd );
          break;
        case ID_START_AUTOSCROLL:
          StartAutoscroll( hWnd );
          break;
        case ID_STOP_AUTOSCROLL:
          StopAutoscroll( hWnd );
          break;
        case ID_BOOKMARKS_MENU:
          DoBookmarksMenu( hWnd );
          break;
        case ID_DISPLAY_ABOUT:
          ShowAboutDialog( hWnd );
          break;
#else
        case ID_GO_HOME:
        {
            Document *document;
            Library *library;
            data = FindReaderWindow( hWnd );
            if (data)
            {
              library = data->library;
              if (_tcscmp(library->current_document, library->home_document))
              {
                document = Library_findDocument( library, library->home_document );
                if( document != NULL )
                {
                  DestroyWindow(hWnd);
                  OpenDocument( library->hWnd, document );
                }
              }
              GoHome( hWnd );
            }
          break;
        }
#endif
        case ID_PREVIOUS_ITEM:
          GoPrevious( hWnd );
          break;
        case ID_NEXT_ITEM:
          GoNext( hWnd );
          break;
        case ID_EXIT_VADE_MECUM:
          DoExitVadeMecum( hWnd );
          break;
#ifndef PDAV_SHOWGUIDE
        case ID_SEARCH_DOCUMENT:
        case ID_SEARCH_BUTTON:
          DoSearchDocument( hWnd, LOWORD( wParam ) == ID_SEARCH_DOCUMENT );
          break;
        case ID_RECORD_INFO:
          DoRecordInfo( hWnd );
          break;
#endif
      }
      break;
    case WM_NOTIFY:
      return HandleNotification( hWnd, wParam, lParam );
    case WM_READER_OPENRECORD:
      OpenRecord( hWnd, wParam, 0, TRUE, lParam, 0, TRUE );
      break;
    default:
      return DefWindowProc( hWnd, uMsg, wParam, lParam );
  }

  return 0;
}


HWND OpenReaderWindow( Library *library,
                       Document *document,
                       RECT *area )
{
  ReaderWindowData *data;
  TCHAR window_class[ 64 ];
  char file_name[ MAX_PATH ];
  POINT orig;
  int w;
  int h;
  FILE *file;

  /* find the first blank slot available */
  data = FindReaderWindow( NULL );

  if( data == NULL )
  {
    /* no available slots open */
    return NULL;
  }

  memset( data, 0, sizeof( ReaderWindowData ) );

  orig.x = area->left;
  orig.y = area->top;

  ClientToScreen( library->hWnd, &orig );

  w = area->right - area->left;
  h = area->bottom - area->top;

  area->left = orig.x;
  area->top = orig.y;
  area->right = area->left + w;
  area->bottom = area->top + h;

  memcpy( &data->original_size, area, sizeof( *area ) );

  data->document = document;

  wcstombs( file_name, document->filename, MAX_PATH );
  data->plkr = OpenPluckerDBFile( file_name );
  if( data->plkr == NULL )
  {
    return NULL;
  }

  data->current_record = PluckerDBGetHomeID( data->plkr );
  data->library = library;
  data->cache = NULL;
  data->autoscroll_timer_id = 0;
  data->cpos_valid = FALSE;

#ifndef PDAV_SHOWGUIDE
  /* first, try to load the pagination cache from a file, if it exists */
  file = Library_openMetadataFile( data->library, document, MF_CACHE, MF_READ );
  if( file )
  {
    data->cache = PaginationCache_unserialize( data->plkr, data->library, data->library->pagination_cache_size, file );
    fclose( file );
  }
#endif
  /* if the file doesn't exist, or the load failed, just create the pagination cache from scratch */
  if( data->cache == NULL )
  {
    data->cache = PaginationCache_new( data->plkr, data->library, data->library->pagination_cache_size );
  }



  data->current_search_list = NULL;

  data->nr_inset_scrolls = 0;

#ifndef PDAV_SHOWGUIDE
  /* open/create the reader history */
  file = Library_openMetadataFile( data->library, document, MF_HISTORY, MF_READ );
  if( file )
  {
    data->history = HistoryList_unserialize( file, data->library->history_size );
    fclose( file );
  }
  else
  {
    data->history = HistoryList_new( data->library->history_size );
  }
#endif

  Library_getMetadataFilename( data->library, document, MF_HIGHLIGHTS, file_name );
  data->highlight_manager = HighlightManager_create( file_name );
  if( data->highlight_manager == NULL )
  {
    MessageBox( library->hWnd, L"Could not open highlight file", L"Error", MB_OK | MB_ICONEXCLAMATION );
    return NULL;
  }

  Library_getMetadataFilename( data->library, document, MF_BOOKMARKS, file_name );
  data->bookmark_manager = BookmarkManager_create( file_name );
  if( data->bookmark_manager == NULL )
  {
    MessageBox( library->hWnd, L"Could not open bookmarks file", L"Error", MB_OK | MB_ICONEXCLAMATION );
    return NULL;
  }

  Library_getMetadataFilename( data->library, document, MF_NOTES, file_name );
  data->note_manager = NoteManager_create( file_name );
  if( data->note_manager == NULL )
  {
    MessageBox( library->hWnd, L"Could not open notes file", L"Error", MB_OK | MB_ICONEXCLAMATION );
    return NULL;
  }

  LoadString( library->hInstance, IDS_READER_CLASS, window_class, 64 );
  data->hWnd = CreateWindow( window_class, L"", WS_BORDER | WS_POPUP,
                             orig.x, orig.y, w, h,
                             library->hWnd, (HMENU)NULL, library->hInstance, (LPVOID)data );

  if( data->hWnd == NULL )
  {
    PaginationCache_delete( data->cache );
    return NULL;
  }

  if( library->full_screen )
  {
    library->full_screen = FALSE; /* force it to false, so the toggle really does do the full screen */
    PostMessage( data->hWnd, WM_COMMAND, ID_TOGGLE_FULLSCREEN, (LPARAM)NULL );
  }

  return data->hWnd;
}


void RepaginateReader( HWND hWnd )
{
  RedrawReaderWindow( hWnd, TRUE );
}


static ReaderWindowData *FindReaderWindow( HWND hWnd )
{
  int i;

  for( i = 0; i < MAX_READER_WINDOWS; i++ )
  {
    if( g_windows[ i ].hWnd == hWnd )
    {
      return &g_windows[ i ];
    }
  }

  return NULL;
}


static void CreateReaderWindow( HWND hWnd, LPARAM lParam )
{
  CREATESTRUCT *cs = (CREATESTRUCT*)lParam;
  ReaderWindowData *data = (ReaderWindowData*)cs->lpCreateParams;

  data->info = NULL;
  data->delta_x = data->delta_y = 0;

  /* if the history was restored from disk and it remembers the last record viewed
   * for this document, restore that record and position */

  if( HistoryList_current( data->library->history ) != NULL )
  {
    HISTORY_ITEM *hi;

#ifdef PDAV_SHOWGUIDE
    hi = HistoryList_current( data->library->history );

    if ((hi->type == HISTORY_ITEM_TYPE_PLUCKER) && (_tcscmp(hi->current_document,data->library->current_document)==0))
    {
      data->cpos_valid = TRUE;
      data->cpos = hi->item.plucker.char_index;
      data->y_offset = hi->item.plucker.y_offset;
      data->current_record = hi->item.plucker.record;
    }
    else
    {
      data->cpos_valid = FALSE;
    }
#else
    data->cpos_valid = TRUE;

    hi = HistoryList_current( data->history );
    data->cpos = hi->char_index;
    data->y_offset = hi->y_offset;
    data->current_record = hi->record;
#endif
    data->add_to_history = FALSE;
  }
  else
  {
    data->add_to_history = TRUE;
  }

  data->redraw_state = RS_START_STATE;
  data->mouse_down = FALSE;
  data->hscroll = data->vscroll = NULL;
  data->toolbar = NULL;

  RefreshToolbar( hWnd, data );
  ComputeScrollbars( hWnd, data, cs->cx, cs->cy );
}


static void DestroyReaderWindow( HWND hWnd )
{
  ReaderWindowData *data;
  FILE *file;

  DeleteInsetScrolls( hWnd );

  data = FindReaderWindow( hWnd );
#ifndef PDAV_SHOWGUIDE
  /* try to save the pagination cache */
  file = Library_openMetadataFile( data->library, data->document, MF_CACHE, MF_WRITE );
  if( file )
  {
    PaginationCache_serialize( data->cache, file );
    fclose( file );
  }
#endif
  UpdateCurrentHistory( hWnd );

#ifndef PDAV_SHOWGUIDE
  /* try to save the history */
  file = Library_openMetadataFile( data->library, data->document, MF_HISTORY, MF_WRITE );
  if( file )
  {
    HistoryList_serialize( data->history, file );
    fclose( file );
  }
  HistoryList_delete( data->history );
#endif

  if (data)
  {
    PaginationCache_delete( data->cache );
    if( data->highlights != NULL ) LinkedList_delete( data->highlights );
    HighlightManager_destroy( data->highlight_manager );
    BookmarkManager_destroy( data->bookmark_manager );
    NoteManager_destroy( data->note_manager );

    ClosePluckerDB( data->plkr );

    /* free the window slot */
    data->hWnd = NULL;
  }
}


static void PaintReaderWindow( HWND hWnd )
{
  ReaderWindowData *data;
  RECT bounds;
  PAINTSTRUCT ps;
  SCROLLINFO si;
  int step_size;
  int text_size;
  int window_size;
  BOOL enable;
  int nr_large_insets;

  data = FindReaderWindow( hWnd );

  BeginPaint( hWnd, &ps );
  GetReaderBounds( hWnd, &bounds );

  switch( data->redraw_state )
  {
    case RS_HIDE_BARS:
      if( data->library->use_hscroll ) ShowWindow( data->hscroll, SW_HIDE );
      if( data->library->use_vscroll ) ShowWindow( data->vscroll, SW_HIDE );
      NextRedrawState( hWnd );
      break;

    case RS_REPAGINATE:
      data->info = PaginationCache_lock( data->cache, data->current_record );
      if( data->info == NULL )
      {
        data->info = PaginationInfo_new( data->library->font_cache );
        PaginationInfo_paginate( data->info,
                                 data->plkr,
                                 data->current_record,
                                 hWnd,
                                 ps.hdc,
                                 bounds.right - bounds.left,
                                 &data->library->options );
        if( PaginationCache_add( data->cache, data->current_record, data->info ) != 1 )
        {
          /* there was no room in the cache! */
          MessageBox( hWnd, L"The pagination cache has run out of space.  Please exit the program!",
                      L"Error (bug?)", MB_OK | MB_ICONEXCLAMATION );
        }
      }

      nr_large_insets = PaginationInfo_getNrLargeInsets( data->info );
      CreateInsetScrolls( hWnd, nr_large_insets );


      /* lookup the highlight records for the current record */
      if( data->highlights ) LinkedList_delete( data->highlights );
      data->highlights = HighlightManager_getHighlights( data->highlight_manager, data->current_record );

      NextRedrawState( hWnd );
      break;

    case RS_SHOW_BARS:
      if( data->library->use_hscroll )
      {
        text_size = PaginationInfo_getWidth( data->info );
        window_size = bounds.right - bounds.left;
        enable = ( text_size > window_size );

        step_size = text_size / window_size;
        if( text_size % window_size > 0 ) step_size++;

        si.cbSize = sizeof( si );
        si.fMask = SIF_PAGE | SIF_POS | SIF_RANGE;
        si.nMin = 0;
        si.nMax = ( enable ? step_size * window_size : 0 );
        si.nPage = window_size;
        si.nPos = data->delta_x;

        SetScrollInfo( data->hscroll, SB_CTL, &si, TRUE );

        EnableWindow( data->hscroll, enable );
        ShowWindow( data->hscroll, SW_SHOW );
      }

      if( data->library->use_vscroll ) 
      {
        int extra = 0;

        text_size = PaginationInfo_getHeight( data->info ) + ComputeVerticalPageSize( hWnd, data );
        window_size = bounds.bottom - bounds.top;
        enable = ( text_size > window_size );

        step_size = text_size / window_size;
        extra = text_size % window_size;

        si.cbSize = sizeof( si );
        si.fMask = SIF_PAGE | SIF_POS | SIF_RANGE;
        si.nMin = 0;
        si.nMax = ( enable ? step_size * window_size + extra : 0 );
        si.nPage = window_size;
        si.nPos = data->delta_y;

        SetScrollInfo( data->vscroll, SB_CTL, &si, TRUE );

        EnableWindow( data->vscroll, enable );
        ShowWindow( data->vscroll, SW_SHOW );
      }      

      NextRedrawState( hWnd );
      break;

    case RS_RENDER_TEXT:      

      if( data->change_para )
      {
        data->change_para = FALSE;
        data->delta_y = PaginationInfo_findParagraphY( data->info, data->new_para );
      }
      else if( data->cpos_valid )
      {
        /* if the delta-y needs to be recomputed from a character index and y-offset, do so */
        data->delta_y = PaginationInfo_characterIndexToDeltaY( data->info, data->cpos );
        if( data->delta_y > data->y_offset ) data->delta_y -= data->y_offset;
        data->cpos_valid = FALSE;
      }

      if( data->add_to_history )
      {
        AddCurrentHistory( hWnd );
        data->add_to_history = FALSE;
      }

      if ( data->current_search_list )
      {
        EnableToolbarButton( data->toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( data->current_search_list ) );
        EnableToolbarButton( data->toolbar, ID_NEXT_ITEM, TRUE );
      }
      else
      {
#ifdef PDAV_SHOWGUIDE
        EnableToolbarButton( data->toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( data->library->history ) );
        EnableToolbarButton( data->toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( data->library->history ) );
#else
        EnableToolbarButton( data->toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( data->history ) );
        EnableToolbarButton( data->toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( data->history ) );
#endif
      }

      /* have to do this here, since this is where the delta_y will have been modified (above) */
      UpdateScrollPos( hWnd );

      PaginationInfo_updateLargeInsetScrolls( data->info, hWnd, data->delta_y, &bounds );

      IntersectClipRect( ps.hdc, bounds.left, bounds.top, bounds.right, bounds.bottom );
      PaginationInfo_render( data->info,
                             ps.hdc,
                             data->delta_y,
#ifndef PDAV_SHOWGUIDE
                             ( data->library->toolbar_pos == TOOLBAR_TOP ? GetReaderToolbarHeight( data->toolbar ) : 0 ),
#else
                             0,
#endif
                             ps.rcPaint.top,
                             ps.rcPaint.bottom - ps.rcPaint.top,
                             data->highlights,
                             hWnd );

      SetFocus( hWnd );
      NextRedrawState( hWnd );
      break;
  }

  EndPaint( hWnd, &ps );
}


static void GetReaderBounds( HWND hWnd, RECT *bounds )
{
  ReaderWindowData *data;
  RECT tmp;

  data = FindReaderWindow( hWnd );

  GetClientRect( hWnd, bounds );

#ifndef PDAV_SHOWGUIDE
  if( data->library->toolbar_pos == TOOLBAR_TOP )
  {
    bounds->top += GetReaderToolbarHeight( data->toolbar );
  }
  else if( data->library->toolbar_pos == TOOLBAR_BOTTOM )
#endif
  {
    bounds->bottom -= GetReaderToolbarHeight( data->toolbar );
  }

  if( data->library->use_hscroll )
  {
    GetClientRect( data->hscroll, &tmp );
    bounds->bottom -= ( tmp.bottom - tmp.top );
  }
  if( data->library->use_vscroll )
  {
    GetClientRect( data->vscroll, &tmp );
    bounds->right -= ( tmp.right - tmp.left );
  }
}


static BOOL RecognizeContextMenuGesture( HWND hWnd, LPARAM lParam )
{
  SHRGINFO shrg;
  DWORD rc;

  shrg.cbSize = sizeof( shrg );
  shrg.hwndClient = hWnd;
  shrg.ptDown.x = LOWORD( lParam );
  shrg.ptDown.y = HIWORD( lParam );
  shrg.dwFlags = SHRG_RETURNCMD;

  rc = SHRecognizeGesture( &shrg );

#ifndef PDAV_SHOWGUIDE
  if( rc == GN_CONTEXTMENU )
  {
    return TRUE;
  }
#endif
  return FALSE;
}


static void DoPopupMenu( HWND hWnd, LPARAM lParam )
{
  HMENU menu;
  POINT pt;

  menu = BuildPopupMenu( hWnd );

  pt.x = LOWORD( lParam );
  pt.y = HIWORD( lParam );
  ClientToScreen( hWnd, &pt );

  TrackPopupMenu( menu, TPM_LEFTALIGN | TPM_TOPALIGN, pt.x, pt.y,
                  0, hWnd, NULL );

  DestroyMenu( menu );
}


static void HandleKeyUp( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );


  switch( wParam )
  {
    case 0x25: /* left */
      /*ScrollDocumentBy( hWnd, -(int)( ComputeVerticalPageSize( hWnd, data ) * 0.5 ), data );*/
      GoPrevious( hWnd );
      break;
    case 0x26: /* up */
      ScrollDocumentBy( hWnd, -ComputeVerticalPageSize( hWnd, data ), data );
      break;
    case 0x27: /* right */
      /*ScrollDocumentBy( hWnd, (int)( ComputeVerticalPageSize( hWnd, data ) * 0.5 ), data );*/
      GoNext( hWnd );
      break;
    case 0x28: /* down */
      ScrollDocumentBy( hWnd, ComputeVerticalPageSize( hWnd, data ), data );
      break;
    default:
      DefWindowProc( hWnd, WM_KEYUP, wParam, lParam );
  }
}


static BOOL DoButtonDown( HWND hWnd, LPARAM lParam, int mode, BOOL check_tap_action )
{
  ReaderWindowData *data;
  DocumentPoint pt;
  int rc;
  int xPos, yPos;
  BOOL check_selection = TRUE;
  HDC dc;

  xPos = LOWORD( lParam );
  yPos = HIWORD( lParam );

  data = FindReaderWindow( hWnd );

#ifndef PDAV_SHOWGUIDE
  if( check_tap_action && mode == SEL_END && data->library->toolbar_hotspot != HOTSPOT_NONE )
  {
    int x;
    int y;
    RECT bounds;

    GetReaderBounds( hWnd, &bounds );

    switch( data->library->toolbar_hotspot )
    {
      case HOTSPOT_TOPLEFT:
        x = bounds.left; y = bounds.top;
        break;
      case HOTSPOT_TOPRIGHT:
        x = bounds.right - HOTSPOT_SIZE; y = bounds.top;
        break;
      case HOTSPOT_BOTTOMLEFT:
        x = bounds.left; y = bounds.bottom - HOTSPOT_SIZE;
        break;
      case HOTSPOT_BOTTOMRIGHT:
        x = bounds.right - HOTSPOT_SIZE; y = bounds.bottom - HOTSPOT_SIZE;
        break;
    }

    if( ( xPos >= x ) && ( yPos >= y ) && ( xPos <= x+HOTSPOT_SIZE ) && ( yPos <= y+HOTSPOT_SIZE ) )
    {
      ToggleToolbar( hWnd );
      return TRUE;
    }
  }
#endif
  dc = GetDC( hWnd );
  rc = PaginationInfo_lookupPoint( data->info, dc,
#ifndef PDAV_SHOWGUIDE
                                   ( data->library->toolbar_pos == TOOLBAR_TOP ? GetReaderToolbarHeight( data->toolbar ) : 0 ),
#else
                                   0,
#endif
                                   data->delta_y, data->delta_y,
                                   xPos, yPos, data->highlights,
                                   &pt );

  switch( mode )
  {
    case SEL_START:
      data->clickPt = pt;
      PaginationInfo_selectionInit( data->info, &data->selection, &pt );
      if( pt.has_link )
      {
        PaginationInfo_invertRange( data->info, dc,
#ifndef PDAV_SHOWGUIDE
                                    ( data->library->toolbar_pos == TOOLBAR_TOP ? GetReaderToolbarHeight( data->toolbar ) : 0 ),
#else
                                   0,
#endif
                                    data->delta_y, pt.link_range );
      }
      break;

    case SEL_EXTEND:
      if( data->clickPt.has_link )
      {
        data->clickPt.has_link = FALSE;
        PaginationInfo_invertRange( data->info, dc,
#ifndef PDAV_SHOWGUIDE
                                    ( data->library->toolbar_pos == TOOLBAR_TOP ? GetReaderToolbarHeight( data->toolbar ) : 0 ),
#else
                                   0,
#endif
                                    data->delta_y, data->clickPt.link_range );
      }
      PaginationInfo_extendSelection( data->info, dc, data->delta_y, &data->selection, &pt );
      break;

    case SEL_END:
      if( data->clickPt.has_link )
      {
        PaginationInfo_invertRange( data->info, dc, 
#ifndef PDAV_SHOWGUIDE
                                    ( data->library->toolbar_pos == TOOLBAR_TOP ? GetReaderToolbarHeight( data->toolbar ) : 0 ),
#else
                                   0,
#endif
                                    data->delta_y, data->clickPt.link_range );
        if( data->clickPt.link_range == pt.link_range )
        {
          SetInSearch( hWnd, FALSE );
          OpenRecord( hWnd, pt.link_record, pt.link_paragraph, FALSE, 0, 0, TRUE );
 
          check_selection = FALSE;
        }
      }

      if( check_selection )
      {
        if( !PaginationInfo_emptySelection( data->info, &data->selection ) )
        {
          DoSelectionPopupMenu( hWnd, lParam );
        }
        else if( pt.highlight && data->clickPt.highlight == pt.highlight )
        {
          PostMessage( hWnd, WM_READER_ENDSELECTION, 0, 0 );
          DoHighlightPopupMenu( hWnd, pt.highlight, lParam );
          check_selection = FALSE;
        }
        else if( check_tap_action )
        {
          InterpretTapAction( hWnd, lParam, data );
        }
        PostMessage( hWnd, WM_READER_ENDSELECTION, 0, 0 );
      }
      break;

    case SEL_CLEARSEL:
      PaginationInfo_selectionDone( data->info, dc, data->delta_y, &data->selection );
      break;
  }

  ReleaseDC( hWnd, dc );

  return TRUE;
}


static void RedrawReaderWindow( HWND hWnd, BOOL repaginate )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if( repaginate )
  {
    data->redraw_state = RS_START_STATE;
  }
  else
  {
    data->redraw_state = RS_RENDER_TEXT;
  }

  InvalidateRect( hWnd, NULL, TRUE );
}


static void NextRedrawState( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if( data->redraw_state == RS_END_STATE )
  {
    return;
  }

  data->redraw_state += 1;
  InvalidateRect( hWnd, NULL, TRUE );
}


static void DoResizeScreen( HWND hWnd )
{
  ReaderWindowData *data;
  RECT rect;
  HDC dc;

  data = FindReaderWindow( hWnd );

  SHFullScreen( hWnd, SHFS_HIDESIPBUTTON );

  if (data)
  {
    if( data->library->full_screen )
    {
      SHFullScreen( hWnd, SHFS_HIDETASKBAR );

      dc = GetDC( hWnd );
      rect.left = 0;
      rect.top = 0;
      rect.right = GetDeviceCaps( dc, HORZRES );
      rect.bottom = GetDeviceCaps( dc, VERTRES );
      ReleaseDC( hWnd, dc );
    }
    else
    {
      SHFullScreen( hWnd, SHFS_SHOWTASKBAR );

      memcpy( &rect, &data->original_size, sizeof( RECT ) );
    }
  }
  MoveWindow( hWnd, rect.left, rect.top,
                    rect.right - rect.left, rect.bottom - rect.top,
                    TRUE );

  GetClientRect( hWnd, &rect );

  ComputeScrollbars( hWnd, data, rect.right-rect.left, rect.bottom-rect.top );

  UpdateScrollPos( hWnd );
  SetFocus( hWnd );
}


static void DoToggleFullscreen( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  data->library->full_screen = !data->library->full_screen;

  DoResizeScreen( hWnd );
}


static void UpdateScrollPos( HWND hWnd )
{
  SCROLLINFO si;
  ReaderWindowData *data;
  int pct;

  data = FindReaderWindow( hWnd );
  if (!data) return;

  if( data->library->use_hscroll )
  {
    si.cbSize = sizeof( si );
    si.fMask = SIF_POS;
    si.nPos = data->delta_x;
    SetScrollInfo( data->hscroll, SB_CTL, &si, TRUE );
  }

  if( data->library->use_vscroll )
  {
    si.cbSize = sizeof( si );
    si.fMask = SIF_POS;
    si.nPos = data->delta_y;
    SetScrollInfo( data->vscroll, SB_CTL, &si, TRUE );
  }

  if( data->info != NULL )
  {
    pct = (int)( (unsigned long)( data->delta_y + ComputeVerticalPageSize( hWnd, data ) ) * 100 / PaginationInfo_getHeight( data->info ) );
  }
  else
  {
    pct = 0;
  }

  SetViewedPercentage( data->toolbar, pct );

  if( data->library->progress_meter != PROGRESS_METER_NONE )
  {
    RECT bounds;
    int  x;
    int  y;
    HDC  dc;
    HPEN pen;
    HPEN oldpen;

    GetReaderBounds( hWnd, &bounds );
    y = bounds.top + pct * ( bounds.bottom - bounds.top ) / 100;
    x = ( data->library->progress_meter == PROGRESS_METER_LEFT ? bounds.left : bounds.right-2 );

    dc = GetDC( hWnd );
    pen = CreatePen( PS_SOLID, 1, RGB( 255, 0, 0 ) );
    oldpen = SelectObject( dc, pen );

    Rectangle( dc, x, bounds.top, x+2, y+1 );

    pen = CreatePen( PS_SOLID, 1, RGB( 255, 255, 255 ) );
    DeleteObject( SelectObject( dc, pen ) );

    /* draw the entire line, in case we are scrolling up */
    Rectangle( dc, x, y+1, x+2, bounds.bottom+1 );

    SelectObject( dc, oldpen );
    DeleteObject( pen );
    ReleaseDC( hWnd, dc );
  }
}


static void DoReaderSettings( HWND hWnd )
{
  UINT32 rc;
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );
  if (!data) return;

  /* save the character offset and y-offset of the first fully visible row, in case a repagination occurs */
  data->cpos = PaginationInfo_deltaYToCharacterIndex( data->info, data->delta_y, &data->y_offset );

  if( data->library->full_screen )
  {
    SHFullScreen( hWnd, SHFS_SHOWTASKBAR );
  }

  rc = OpenReaderSettings( hWnd, data->library );

  if( ( rc & RS_BOOKPATH_CHANGED ) != 0 )
  {
    Library_scanBooks( data->library );
  }
  if( ( rc & RS_CHANGE_NEEDS_REFRESH ) != 0 )
  {
    RefreshToolbar( hWnd, data );
  }
  if( ( rc & RS_DISPLAY_PARMS_CHANGED ) != 0 )
  {
    RECT rect;

    GetClientRect( hWnd, &rect );

    ComputeScrollbars( hWnd, data, rect.right - rect.left, rect.bottom - rect.top );

    data->cpos_valid = TRUE;
    data->add_to_history = FALSE;

    PaginationCache_clear( data->cache );
  }
  if( ( rc & ( RS_CHANGE_NEEDS_REFRESH | RS_DISPLAY_PARMS_CHANGED ) ) != 0 )
  {
    RedrawReaderWindow( hWnd, ( rc & RS_DISPLAY_PARMS_CHANGED ) != 0 );
  }

  if( data->library->full_screen )
  {
    SHFullScreen( hWnd, SHFS_HIDETASKBAR );
  }
}


static void DoSelectionPopupMenu( HWND hWnd, LPARAM lParam )
{
#ifndef PDAV_SHOWGUIDE
  HMENU menu;
  ReaderWindowData *data;
  POINT pt;

  data = FindReaderWindow( hWnd );
  menu = CreatePopupMenu();

  AppendMenu( menu, MF_STRING, ID_COPY_TEXT, L"&Copy" );
  AppendMenu( menu, MF_STRING, ID_QUICK_HILITE_TEXT, L"&Quick Highlight" );
  AppendMenu( menu, MF_STRING, ID_HILITE_TEXT, L"&Highlight" );
  AppendMenu( menu, MF_STRING, ID_ANNOTATE_TEXT, L"&Annotate" );
  AppendMenu( menu, MF_SEPARATOR, 0, NULL );
  AppendMenu( menu, MF_STRING, ID_CANCEL_SELECTION, L"Cancel" );

  pt.x = LOWORD( lParam );
  pt.y = HIWORD( lParam );
  ClientToScreen( hWnd, &pt );

  TrackPopupMenu( menu, TPM_LEFTALIGN | TPM_TOPALIGN, pt.x, pt.y,
                  0, hWnd, NULL );

  DestroyMenu( menu );
#endif
}


static void DoCopySelection( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if( OpenClipboard( hWnd ) )
  {
    HLOCAL hMem;
    UINT   size;

    size = data->selection.end.text_index - data->selection.start.text_index + 2;

    hMem = LocalAlloc( LMEM_FIXED, size * sizeof( TCHAR ) );

    PaginationInfo_copySelectionTo( data->info, &data->selection, (TCHAR*)hMem );

    EmptyClipboard();
    SetClipboardData( CF_UNICODETEXT, hMem );

    CloseClipboard();
  }
}


static int DoVerticalScroll( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
  short int scroll_code;
  int       pos;
  int       dy;
  MSG       msg;
  SCROLLINFO si;

  ReaderWindowData *data;

  /* don't process a scroll message if there is a paint message waiting in the
   * buffer. */

  if( PeekMessage( &msg, hWnd, WM_PAINT, WM_PAINT, PM_NOREMOVE ) )
  {
    return 1;
  }

  data = FindReaderWindow( hWnd );

  scroll_code = LOWORD( wParam );

  ZeroMemory( &si, sizeof( SCROLLINFO ) );
  si.cbSize = sizeof( SCROLLINFO );

  switch( scroll_code )
  {
    case SB_PAGEUP:
      dy = -ComputeVerticalPageSize( hWnd, data );
      break;

    case SB_LINEUP:
      dy = -ComputeVerticalLineSize( hWnd, data );
      break;

    case SB_LINEDOWN:
      dy = ComputeVerticalLineSize( hWnd, data );
      break;

    case SB_PAGEDOWN:
      dy = ComputeVerticalPageSize( hWnd, data );
      break;

    case SB_THUMBTRACK:
      si.fMask = SIF_TRACKPOS;
      if ( !GetScrollInfo( data->vscroll, SB_CTL, &si ) )
        return 1;
      pos = si.nTrackPos;
      dy = pos - data->delta_y;
      if( data->delta_y + dy < 0 ) dy = -data->delta_y;
      break;

    case SB_THUMBPOSITION:
      si.fMask = SIF_POS;
      if ( !GetScrollInfo( data->vscroll, SB_CTL, &si ) )
        return 1;
      pos = si.nPos;
      dy = pos - data->delta_y;
      if( data->delta_y + dy < 0 ) dy = -data->delta_y;
      break;

    default:
      return 1;
  }

  return ScrollDocumentBy( hWnd, dy, data );
}

static int DoHorizontalScroll( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
  ReaderWindowData *data = FindReaderWindow( hWnd );
  int scroll_code = LOWORD(wParam);
  int pos = HIWORD(wParam);
  HWND scroll_hwnd = (HWND)lParam;
  int delta;
  RECT *rc;
  SCROLLINFO si;
  BOOL rel;

  if( scroll_hwnd == data->hscroll )
    return 0;

  rc = (RECT*)GetWindowLong( scroll_hwnd, GWL_USERDATA );
  si.cbSize = sizeof(si);
  si.fMask = SIF_PAGE | SIF_POS | SIF_TRACKPOS;
  GetScrollInfo( scroll_hwnd, SB_CTL, &si );
  rel = TRUE;

  switch( scroll_code )
  {
  case SB_LINELEFT:
    delta = -(signed)(si.nPage/5);
    break;
  case SB_LINERIGHT:
    delta = si.nPage/5;
    break;
  case SB_PAGELEFT:  
    delta = -(signed)si.nPage;
    break;
  case SB_PAGERIGHT:
    delta = si.nPage;
    break;
  case SB_THUMBTRACK:
    delta = si.nTrackPos;
    rel = FALSE;
    break;
  case SB_THUMBPOSITION:
    delta = si.nPos;
    rel = FALSE;
    break;

  default:
    return 1;
  }

  PaginationInfo_updateLargeInset( data->info, GetWindowLong( scroll_hwnd, GWL_ID)-ID_LARGE_INSET_SCROLL, delta, rel );

  data->redraw_state = RS_RENDER_TEXT;
  InvalidateRect( hWnd, rc, FALSE );

  return 0;
}

static int ComputeVerticalPageSize( HWND hWnd, ReaderWindowData *data )
{
  RECT rect;
  GetReaderBounds( hWnd, &rect );

  /* the size of the reader area, minus the height of a single line */
  return rect.bottom - rect.top - ComputeVerticalLineSize( hWnd, data );
}


static int ComputeVerticalLineSize( HWND hWnd, ReaderWindowData *data )
{
  return (int)( data->library->options.font_size * 1.2 ) + data->library->options.interline_space;
}


static UINT HandleNotification( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
  LPNMHDR notify = (LPNMHDR)lParam;
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );
  if (data)
  {
    if( notify->hwndFrom == data->toolbar )
    {
      switch( notify->code )
      {
        case NM_CUSTOMDRAW:
          return DrawToolbar( data->toolbar, wParam, lParam );

        case TBN_DROPDOWN:
          return DoToolbarMenu( hWnd, wParam, lParam );
      }
    }
  }
  return 1;
}


#ifndef PDAV_SHOWGUIDE
static 
#endif
void GoHome( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if (data)
  {
    SetInSearch( hWnd, FALSE );
    OpenRecord( hWnd, PluckerDBGetHomeID( data->plkr ), -1, FALSE, 0, 0, TRUE );
  }
}


static void ComputeScrollbars( HWND hWnd, ReaderWindowData *data, int cx, int cy )
{
  int width;
  int height;
  int vscroll_wid;
  int hscroll_hei;
  int top;
  int y;

  if (!data)
  {
    return;
  }

  vscroll_wid = GetSystemMetrics( SM_CXVSCROLL );
  hscroll_hei = GetSystemMetrics( SM_CYHSCROLL );

  top = 0;

#ifndef PDAV_SHOWGUIDE
  if( data->library->toolbar_pos == TOOLBAR_TOP )
  {
    top += GetReaderToolbarHeight( data->toolbar );
  }
  
  if( data->library->toolbar_pos != TOOLBAR_NONE )
#endif
  {
    cy -= GetReaderToolbarHeight( data->toolbar );
  }
  width = cx - ( data->library->use_vscroll ? vscroll_wid : 0 ) + 1;
  height = cy - ( data->library->use_hscroll ? hscroll_hei : 0 ) + 1;

  y = cy + 1;
#ifndef PDAV_SHOWGUIDE
  if( data->library->toolbar_pos == TOOLBAR_TOP ) y += GetReaderToolbarHeight( data->toolbar );
#endif
  if( !data->hscroll && data->library->use_hscroll )
  {

    data->hscroll = CreateWindow( L"scrollbar", L"", SBS_HORZ | WS_CHILD,
                                  0, y - hscroll_hei - 1,
                                  width, hscroll_hei, hWnd,
                                  (HMENU)NULL, data->library->hInstance, NULL );
  }
  else if( data->hscroll && data->library->use_hscroll )
  {
    MoveWindow( data->hscroll, 0, y - hscroll_hei - 1, width, hscroll_hei, TRUE );
  }
  else /* data->hscroll && !data->library->use_hscroll */
  {
    DestroyWindow( data->hscroll );
    data->hscroll = NULL;
  }

  if( !data->vscroll && data->library->use_vscroll )
  {
    data->vscroll = CreateWindow( L"scrollbar", L"", SBS_VERT | WS_CHILD,
                                  cx - vscroll_wid, top,
                                  vscroll_wid, height, hWnd,
                                  (HMENU)NULL, data->library->hInstance, NULL );
  }
  else if( data->vscroll && data->library->use_vscroll )
  {
    MoveWindow( data->vscroll, cx - vscroll_wid, top, vscroll_wid, height, TRUE );
  }
  else /* data->vscroll && !data->library->use_vscroll */
  {
    DestroyWindow( data->vscroll );
    data->vscroll = NULL;
  }
}


static void RefreshToolbar( HWND hWnd, ReaderWindowData *data )
{
  if( !data->toolbar
#ifndef PDAV_SHOWGUIDE
    && data->library->toolbar_pos != TOOLBAR_NONE && !data->library->toolbar_hidden 
#endif
    )
  {
    data->toolbar = CreateReaderToolbar( hWnd, data->library, false );
  }
#ifndef PDAV_SHOWGUIDE
  else if( data->toolbar && data->library->toolbar_pos != TOOLBAR_NONE && !data->library->toolbar_hidden )
  {
    MoveToolbarPosition( data->toolbar, data->library->toolbar_pos );
  }
  else if( data->toolbar ) /* data->library->toolbar_pos == TOOLBAR_NONE || data->library->toolbar_hidden */
  {
    DestroyWindow( data->toolbar );
    data->toolbar = NULL;
  }
#endif

  if( data->toolbar != NULL )
  {
    if ( data->current_search_list )
    {
      EnableToolbarButton( data->toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( data->current_search_list ) );
      EnableToolbarButton( data->toolbar, ID_NEXT_ITEM, TRUE );
    }
    else
    {
      
#ifdef PDAV_SHOWGUIDE
      EnableToolbarButton( data->toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( data->library->history ) );
      EnableToolbarButton( data->toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( data->library->history ) );
#else
      EnableToolbarButton( data->toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( data->history ) );
      EnableToolbarButton( data->toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( data->history ) );
#endif

    }
  }
}


static UINT DoToolbarMenu( HWND hWnd, WPARAM wParam, LPARAM lParam )
{
  LPNMTOOLBAR nm = (LPNMTOOLBAR)lParam;
  ReaderWindowData *data;
  TPMPARAMS tpm;
  HMENU menu;
  RECT rect;

  data = FindReaderWindow( hWnd );
  menu = BuildPopupMenu( hWnd );

  SendMessage( data->toolbar, TB_GETRECT, nm->iItem, (LPARAM)&rect );
  MapWindowPoints( data->toolbar, HWND_DESKTOP, (LPPOINT)&rect, 2 );

  tpm.cbSize = sizeof( tpm );
  CopyRect( &tpm.rcExclude, &rect );

  TrackPopupMenuEx( menu, TPM_LEFTALIGN | TPM_VERTICAL, rect.left, rect.bottom, hWnd, &tpm );

  DestroyMenu( menu );

  return FALSE;
}


static void InterpretTapAction( HWND hWnd, LPARAM lParam, ReaderWindowData *data )
{
  int x;
  int y;
  int area = -1;
  int page_size;
  int line_size;
  RECT bounds;

  x = LOWORD( lParam );
  y = HIWORD( lParam );

  GetReaderBounds( hWnd, &bounds );

  switch( data->library->screen_division )
  {
    case SCREEN_DIV_NONE:
      break;

    case SCREEN_DIV_QUADRANTS:
      /* +---+---+
       * | 0 | 1 |
       * +---+---+
       * | 2 | 3 |
       * +---+---+
       */

      area = ( ( y - bounds.top ) * 2 / ( bounds.bottom - bounds.top  + 1 ) ) * 2 + 
             ( ( x - bounds.left ) * 2 / ( bounds.right - bounds.left + 1 ) );
      break;

    case SCREEN_DIV_VERTICAL:
      /* +---+
       * | 0 |
       * +---+
       * | 1 |
       * +---+
       * | 2 |
       * +---+
       * | 3 |
       * +---+
       */

      area = ( ( y - bounds.top ) * 4 / ( bounds.bottom - bounds.top + 1 ) );
      break;

    case SCREEN_DIV_HORIZONTAL:
      /* +---+---+---+---+
       * | 0 | 1 | 2 | 3 |
       * +---+---+---+---+
       */

      area = ( ( x - bounds.left ) * 4 / ( bounds.right - bounds.left + 1 ) );
      break;
  }

  if( area < 0 ) return;

  page_size = ComputeVerticalPageSize( hWnd, data );
  line_size = ComputeVerticalLineSize( hWnd, data );

  switch( data->library->tap_actions[ area ] )
  {
    case TAP_ACTION_NONE:
      break;

    case TAP_ACTION_UP_ONELINE:
      ScrollDocumentBy( hWnd, -line_size, data );
      break;

    case TAP_ACTION_UP_QUARTERPAGE:
      ScrollDocumentBy( hWnd, -(int)( page_size * 0.25 ), data );
      break;

    case TAP_ACTION_UP_HALFPAGE:
      ScrollDocumentBy( hWnd, -(int)( page_size * 0.5 ), data );
      break;

    case TAP_ACTION_UP_FULLPAGE:
      ScrollDocumentBy( hWnd, -page_size, data );
      break;

    case TAP_ACTION_DOWN_FULLPAGE:
      ScrollDocumentBy( hWnd, page_size, data );
      break;

    case TAP_ACTION_DOWN_HALFPAGE:
      ScrollDocumentBy( hWnd, (int)( page_size * 0.5 ), data );
      break;

    case TAP_ACTION_DOWN_QUARTERPAGE:
      ScrollDocumentBy( hWnd, (int)( page_size * 0.25 ), data );
      break;

    case TAP_ACTION_DOWN_ONELINE:
      ScrollDocumentBy( hWnd, line_size, data );
      break;

    case TAP_ACTION_TOP:
      GoTop( hWnd );
      break;

    case TAP_ACTION_BOTTOM:
      GoBottom( hWnd );
      break;

    case TAP_ACTION_HOME:
      GoHome( hWnd );
      break;
  }
}


static int ScrollDocumentBy( HWND hWnd, int dy, ReaderWindowData *data )
{
  RECT rect;
  int height;
  int new_record;
  int new_paragraph;
  int follow_link = 0;
  int top;

  if( data->delta_y + dy < 0 )
  {
    dy = -data->delta_y;

    if( dy == 0 && data->library->auto_follow_first_link )
    {
      follow_link = PaginationInfo_startsWithLink( data->info, &new_record, &new_paragraph );
    }
  }
  else
  {
    height = PaginationInfo_getHeight( data->info ) - ComputeVerticalLineSize( hWnd, data );
    if( data->delta_y + dy > height )
    {
      dy = height - data->delta_y;

      if( dy == 0 && data->library->auto_follow_last_link )
      {
        follow_link = PaginationInfo_endsWithLink( data->info, &new_record, &new_paragraph );
      }
    }
  }

  if( dy == 0 ) 
  {
    if( follow_link )
    {
      SetInSearch( hWnd, FALSE );
      OpenRecord( hWnd, new_record, new_paragraph, FALSE, 0, 0, TRUE );
    }
    return 1;
  }

  data->delta_y += dy;

  GetReaderBounds( hWnd, &rect );

  UpdateScrollPos( hWnd );

  ScrollWindowEx( hWnd, 0, -dy, &rect, &rect, NULL, NULL, 0 );

  /* compute the area that needs to be redrawn */
  if( dy < 0 ) /* scroll up... the top needs to be redrawn */
  {
    rect.bottom = rect.top - dy;
  }
  else /* scroll down... the bottom needs to be redrawn */
  {
    top = rect.bottom - dy - GetSystemMetrics( SM_CYHSCROLL );
    if( top > rect.top ) rect.top = top;
    /*rect.top = rect.bottom - dy;*/
    
  }

  data->redraw_state = RS_RENDER_TEXT; /* we only want to redraw the text, nothing else */
  InvalidateRect( hWnd, &rect, TRUE );

  return 0;
}


static HMENU BuildPopupMenu( HWND hWnd )
{
  HMENU menu;
  HMENU submenu;
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );
  menu = CreatePopupMenu();
  submenu = CreatePopupMenu();

  AppendMenu( submenu, MF_STRING, ID_GO_HOME, L"&Home" );
  AppendMenu( submenu, MF_SEPARATOR, 0, NULL );
  AppendMenu( submenu, MF_STRING, ID_GO_TOP, L"&Top" );
  AppendMenu( submenu, MF_STRING, ID_GO_BOTTOM, L"&Bottom" );
  AppendMenu( submenu, MF_SEPARATOR, 0, NULL );
  AppendMenu( submenu, MF_STRING, ID_PREVIOUS_ITEM, L"B&ack" );
  AppendMenu( submenu, MF_STRING, ID_NEXT_ITEM, L"&Forward" );

#ifdef PDAV_SHOWGUIDE
  
  EnableMenuItem( submenu, ID_PREVIOUS_ITEM,
    ( MF_BYCOMMAND | ( HistoryList_priorElementsExist( data->library->history ) ? MF_ENABLED : MF_GRAYED ) ) );
  EnableMenuItem( submenu, ID_NEXT_ITEM, 
    ( MF_BYCOMMAND | ( HistoryList_subsequentElementsExist( data->library->history ) ? MF_ENABLED : MF_GRAYED ) ) );
#else
  EnableMenuItem( submenu, ID_PREVIOUS_ITEM,
    ( MF_BYCOMMAND | ( HistoryList_priorElementsExist( data->history ) ? MF_ENABLED : MF_GRAYED ) ) );
  EnableMenuItem( submenu, ID_NEXT_ITEM, 
    ( MF_BYCOMMAND | ( HistoryList_subsequentElementsExist( data->history ) ? MF_ENABLED : MF_GRAYED ) ) );
#endif
  AppendMenu( menu, MF_POPUP, (unsigned int)submenu, L"&Go" );
  AppendMenu( menu, MF_SEPARATOR, 0, NULL );
  AppendMenu( menu, MF_STRING, ID_RECORD_INFO, L"&Record info" );
  AppendMenu( menu, MF_STRING, ID_SEARCH_DOCUMENT, L"S&earch" );
  AppendMenu( menu, MF_STRING, ID_GO_TO_LIBRARY, L"&Library" );
  AppendMenu( menu, MF_STRING | ( data->library->full_screen ? MF_CHECKED : 0 ), ID_TOGGLE_FULLSCREEN, L"&Full Screen" );
  AppendMenu( menu, MF_SEPARATOR, 0, NULL );
  AppendMenu( menu, MF_STRING, ID_START_AUTOSCROLL, L"&Autoscroll" );
  AppendMenu( menu, MF_SEPARATOR, 0, NULL );
  AppendMenu( menu, MF_STRING, ID_READER_SETTINGS, L"&Settings" );
  AppendMenu( menu, MF_STRING, ID_DISPLAY_ABOUT, L"A&bout Vade Mecum" );
  AppendMenu( menu, MF_STRING, ID_EXIT_VADE_MECUM, L"E&xit" );

  return menu;
}


static void GoTop( HWND hWnd )
{
  ReaderWindowData *data;

  UpdateCurrentHistory( hWnd );

  data = FindReaderWindow( hWnd );
  ScrollDocumentBy( hWnd, -data->delta_y, data );

  AddCurrentHistory( hWnd );
}


static void GoBottom( HWND hWnd )
{
  ReaderWindowData *data;
  int page_size;

  UpdateCurrentHistory( hWnd );

  data = FindReaderWindow( hWnd );

  page_size = ComputeVerticalPageSize( hWnd, data );
  ScrollDocumentBy( hWnd, PaginationInfo_getHeight( data->info ) - data->delta_y - page_size, data );

  AddCurrentHistory( hWnd );
}


static void StartAutoscroll( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if( data->autoscroll_timer_id < 1 )
  {
    data->autoscroll_timer_id = SetTimer( hWnd, ID_AUTOSCROLL_INCREMENT, data->library->autoscroll_delay * 10, NULL );
  }
}


static void StopAutoscroll( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if( data->autoscroll_timer_id > 0 )
  {
    KillTimer( hWnd, data->autoscroll_timer_id );
    data->autoscroll_timer_id = 0;
  }
}


static void DoHiliteSelection( HWND hWnd, BOOL quick, BOOL annotate )
{
  ReaderWindowData *data;
  Highlight h;
  int start_sel;
  int end_sel;

  data = FindReaderWindow( hWnd );

  start_sel = data->selection.start.text_index;
  end_sel = data->selection.end.text_index - 1;

  h.record_id = data->current_record;
  h.fg = ( data->library->foreground_selected ? data->library->foreground : -1 );
  h.bg = ( data->library->background_selected ? data->library->background : -1 );
  h.start_pos = start_sel;
  h.end_pos = end_sel;
  h.annotation = -1;

  if( annotate )
  {
    if( !GetAnnotationText( hWnd, &h.annotation ) )
    {
      return;
    }
  }

  if( quick || GetHighlightColor( hWnd, &h ) )
  {
    HighlightManager_addHighlight( data->highlight_manager, data->highlights, &h );
    RedrawReaderWindow( hWnd, FALSE );
  }
  else
  {
    /* if the highlight was cancelled, and an annotation was associated with the highlight,
     * we need to delete the annotation. */
    if( h.annotation >= 0 )
    {
      Note *note;
      note = NoteManager_getNote( data->note_manager, h.annotation );
      NoteManager_deleteNote( data->note_manager, note );
    }
  }
}


static void DoBookmarksMenu( HWND hWnd )
{
  ReaderWindowData *data;
  LinkedList *list;
  LinkedListNode *node;
  HMENU menu;
  HMENU submenu;
  int id;
  RECT rect;
  TPMPARAMS tpm;
  Bookmark *mptr;
  Bookmark m;
  GETUSERTEXTOPTS opts;

  data = FindReaderWindow( hWnd );

  list = BookmarkManager_getBookmarks( data->bookmark_manager );

  menu = CreatePopupMenu();

  AppendMenu( menu, MF_STRING, ID_OPT_ADD_BOOKMARK, L"Add Bookmark" );

  node = list->list;

  if( node != NULL )
  {
    AppendMenu( menu, MF_SEPARATOR, 0, NULL );

    /* Go to Bookmark */
    submenu = CreatePopupMenu();
    id = 1;
    while( node != NULL )
    {
      AppendMenu( submenu, MF_STRING, ID_GOTO_BOOKMARK + id++, NODEDATA(node,Bookmark)->name );
      node = node->next;
    }

    AppendMenu( menu, MF_STRING | MF_POPUP, (UINT)submenu, L"Go to..." );

    /* Reposition Bookmark */
    submenu = CreatePopupMenu();
    node = list->list;
    id = 1;
    while( node != NULL )
    {
      AppendMenu( submenu, MF_STRING, ID_REPOSITION_BOOKMARK + id++, NODEDATA(node,Bookmark)->name );
      node = node->next;
    }

    AppendMenu( menu, MF_STRING | MF_POPUP, (UINT)submenu, L"Reposition..." );

    /* Rename Bookmark */
    submenu = CreatePopupMenu();
    node = list->list;
    id = 1;
    while( node != NULL )
    {
      AppendMenu( submenu, MF_STRING, ID_RENAME_BOOKMARK + id++, NODEDATA(node,Bookmark)->name );
      node = node->next;
    }

    AppendMenu( menu, MF_STRING | MF_POPUP, (UINT)submenu, L"Rename..." );

    /* Delete Bookmark */
    submenu = CreatePopupMenu();
    node = list->list;
    id = 1;
    while( node != NULL )
    {
      AppendMenu( submenu, MF_STRING, ID_DELETE_BOOKMARK + id++, NODEDATA(node,Bookmark)->name );
      node = node->next;
    }

    AppendMenu( menu, MF_STRING | MF_POPUP, (UINT)submenu, L"Delete..." );
  }

  SendMessage( data->toolbar, TB_GETRECT, ID_BOOKMARKS_MENU, (LPARAM)&rect );
  MapWindowPoints( data->toolbar, HWND_DESKTOP, (LPPOINT)&rect, 2 );

  tpm.cbSize = sizeof( tpm );
  CopyRect( &tpm.rcExclude, &rect );

  id = TrackPopupMenuEx( menu, TPM_RETURNCMD | TPM_LEFTALIGN | TPM_VERTICAL, rect.left, rect.bottom, hWnd, &tpm );

  if( id == 0 ) /* cancelled */
  {
  }
  else if( id == ID_OPT_ADD_BOOKMARK )
  {
    m.char_pos = PaginationInfo_deltaYToCharacterIndex( data->info, data->delta_y, &m.delta_y );
    m.record_id = data->current_record;
    
    wcscpy( m.name, L"" );

    opts.flags = ES_LEFT | ES_AUTOHSCROLL;
    opts.hInstance = data->library->hInstance;
    opts.hWnd = hWnd;
    opts.prompt = L"Enter this bookmark's name:";
    opts.title = L"New Bookmark";
    opts.width = 180;
    opts.height = 20;
    opts.text = m.name;
    opts.text_len = BOOKMARK_TITLE_LEN;

    if( GetUserText( &opts ) && wcscmp( m.name, L"" ) != 0 )
    {
      BookmarkManager_addBookmark( data->bookmark_manager, list, &m );
    }
  }
  else if( id > ID_DELETE_BOOKMARK )
  {
    id -= ID_DELETE_BOOKMARK;

    node = list->list;
    while( id > 1 )
    {
      node = node->next;
      id--;
    }

    BookmarkManager_deleteBookmark( data->bookmark_manager, list, NODEDATA(node,Bookmark) );
  }
  else if( id > ID_RENAME_BOOKMARK )
  {
    id -= ID_RENAME_BOOKMARK;

    node = list->list;
    while( id > 1 )
    {
      node = node->next;
      id--;
    }

    mptr = NODEDATA(node,Bookmark);

    opts.flags = ES_LEFT | ES_AUTOHSCROLL;
    opts.hInstance = data->library->hInstance;
    opts.hWnd = hWnd;
    opts.prompt = L"Edit this bookmark's name:";
    opts.title = L"Rename Bookmark";
    opts.width = 180;
    opts.height = 20;
    opts.text = mptr->name;
    opts.text_len = BOOKMARK_TITLE_LEN;

    if( GetUserText( &opts ) )
    {
      BookmarkManager_modifyBookmark( data->bookmark_manager,
                                      mptr,
                                      mptr->record_id,
                                      mptr->char_pos,
                                      mptr->delta_y,
                                      mptr->name );
    }
  }
  else if( id > ID_REPOSITION_BOOKMARK )
  {
    id -= ID_REPOSITION_BOOKMARK;

    node = list->list;
    while( id > 1 )
    {
      node = node->next;
      id--;
    }

    mptr = NODEDATA(node,Bookmark);

    mptr->char_pos = PaginationInfo_deltaYToCharacterIndex( data->info, data->delta_y, &mptr->delta_y );
    mptr->record_id = data->current_record;

    BookmarkManager_modifyBookmark( data->bookmark_manager,
                                    mptr,
                                    mptr->record_id,
                                    mptr->char_pos,
                                    mptr->delta_y,
                                    mptr->name );
  }
  else if( id > ID_GOTO_BOOKMARK ) /* goto bookmark */
  {
    id -= ID_GOTO_BOOKMARK;

    node = list->list;
    while( id > 1 )
    {
      node = node->next;
      id--;
    }

    SetInSearch( hWnd, FALSE );
    OpenRecord( hWnd, NODEDATA( node, Bookmark )->record_id, -1, TRUE,
                NODEDATA( node, Bookmark )->char_pos,
                NODEDATA( node, Bookmark )->delta_y,
                TRUE );
  }

  LinkedList_delete( list );
  DestroyMenu( menu );
}


static void DoHighlightPopupMenu( HWND hWnd, Highlight *highlight, LPARAM lParam )
{
  ReaderWindowData *data;
  HMENU menu;
  UINT id;

  data = FindReaderWindow( hWnd );

  menu = CreatePopupMenu();

  AppendMenu( menu, MF_STRING, ID_CHANGE_HIGHLIGHT_COLOR, L"Change Color" );

  if( highlight->annotation >= 0 )
    AppendMenu( menu, MF_STRING, ID_EDIT_ANNOTATION, L"Edit Annotation" );
  else
    AppendMenu( menu, MF_STRING, ID_EDIT_ANNOTATION, L"Add Annotation" );

  AppendMenu( menu, MF_STRING, ID_DELETE_HIGHLIGHT, L"Delete" );
  AppendMenu( menu, MF_SEPARATOR, 0, NULL );
  AppendMenu( menu, MF_STRING, IDCANCEL, L"Cancel" );

  id = TrackPopupMenu( menu, TPM_RETURNCMD | TPM_LEFTALIGN | TPM_VERTICAL, LOWORD(lParam), HIWORD(lParam), 0, hWnd, NULL );

  DestroyMenu( menu );

  switch( id )
  {
    case ID_CHANGE_HIGHLIGHT_COLOR:
      if( GetHighlightColor( hWnd, highlight ) )
      {
        HighlightManager_modifyHighlight( data->highlight_manager, highlight, highlight->fg, highlight->bg, highlight->annotation );
        RedrawReaderWindow( hWnd, FALSE );
      }
      break;

    case ID_EDIT_ANNOTATION:
      if( GetAnnotationText( hWnd, &highlight->annotation ) )
      {
        HighlightManager_modifyHighlight( data->highlight_manager, highlight, highlight->fg, highlight->bg, highlight->annotation );
      }
      break;

    case ID_DELETE_HIGHLIGHT:
      if( highlight->annotation >= 0 )
      {
        NoteManager_deleteNoteByID( data->note_manager, highlight->annotation );
      }
      HighlightManager_deleteHighlight( data->highlight_manager, data->highlights, highlight );
      RedrawReaderWindow( hWnd, FALSE );
      break;
  }
}


static BOOL GetHighlightColor( HWND hWnd, Highlight *highlight )
{
  COLORPICKERDATA clr;
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  memset( &clr, 0, sizeof( clr ) );

  clr.title = L"Select Highlight Color";
  clr.hWnd = hWnd;
  clr.selected_background = ( highlight->bg >= 0 ? TRUE : FALSE );
  clr.selected_foreground = ( highlight->fg >= 0 ? TRUE : FALSE );
  clr.foreground = highlight->fg;
  clr.background = highlight->bg;
  clr.fgcolors = data->library->fg_colors;
  clr.bgcolors = data->library->bg_colors;

  if( ColorPicker( &clr ) && ( clr.selected_foreground || clr.selected_background ) )
  {
    data->library->foreground_selected = clr.selected_foreground;
    data->library->background_selected = clr.selected_background;

    if( data->library->foreground_selected ) data->library->foreground = clr.foreground;
    if( data->library->background_selected ) data->library->background = clr.background;

    highlight->bg = ( clr.selected_background ? clr.background : -1 );
    highlight->fg = ( clr.selected_foreground ? clr.foreground : -1 );
    return TRUE;
  }

  return FALSE;
}


static void OpenRecord( HWND hWnd, int record, int para, BOOL use_cpos, int cpos, int yofs, BOOL add_history )
{
  ReaderWindowData *data;
  BOOL repaginate = FALSE;
  int record_type = DATATYPE_PHTML;


  data = FindReaderWindow( hWnd );
  if (!data) return;

  record_type = PluckerDBGetRecordType( data->plkr,
                                        record );

  switch( record_type )
  {
    case DATATYPE_PHTML:
    case DATATYPE_PHTML_COMPRESSED:
      OpenTextRecord( hWnd, record, para, use_cpos, cpos, yofs, add_history );
      break;

    case DATATYPE_MAILTO:
      OpenMailToRecord( hWnd, record );
      break;

    case DATATYPE_TBMP:
    case DATATYPE_TBMP_COMPRESSED:
    case DATATYPE_COMPOSITE_IMAGE:
      OpenImageRecord( hWnd, record );
      break;

    case DATATYPE_INVALID:
      OpenInvalidRecord( hWnd, record );
      break;

    default:
      {
        TCHAR msg[130];
        wsprintf( msg, L"I'm not sure how you did it, but you shouldn't be able to open a record of that type (%d)!", record_type );
        MessageBox( hWnd, msg, L"DEBUG", MB_OK | MB_ICONEXCLAMATION );
      }
  }
}


static void OpenTextRecord( HWND hWnd, int record, int para, BOOL use_cpos, int cpos, int yofs, BOOL add_history )
{
  ReaderWindowData *data;
  BOOL repaginate = FALSE;
  PluckerTextRecord *text;

  data = FindReaderWindow( hWnd );

  text = PluckerDBGetTextRecord( data->plkr,
                                 record );

  /* save the current position in the history */
  if( add_history )
  {
    UpdateCurrentHistory( hWnd );
  }

  if( data->current_record != record )
  {
    repaginate = TRUE;
    PaginationCache_release( data->cache, data->current_record );
    data->current_record = record;
  }

  data->add_to_history = add_history;
  data->delta_x = data->delta_y = 0;
  data->change_para = ( para >= 0 && !use_cpos );
  data->new_para = para;
  data->cpos_valid = use_cpos;
  data->cpos = cpos;
  data->y_offset = yofs;

  RedrawReaderWindow( hWnd, repaginate );

}


static void OpenMailToRecord( HWND hWnd, int record )
{
  int to_offset;
  int cc_offset;
  int subject_offset;
  int body_offset;
  unsigned char *text;
  int record_type;
  ReaderWindowData *data;
  size_t count;
  SHELLEXECUTEINFO sei;

  TCHAR *to = NULL;
  TCHAR *cc = NULL;
  TCHAR *subject = NULL;
  TCHAR *body = NULL;

  TCHAR  parameter_string[ 1024 ];


  data = FindReaderWindow( hWnd );

  text = PluckerDBGetRecord( data->plkr, record, &record_type );
  text += 8; /* skip the header */

  to_offset = PLKR_INT16( text );
  cc_offset = PLKR_INT16( text+2 );
  subject_offset = PLKR_INT16( text+4 );
  body_offset = PLKR_INT16( text+6 );

  if( to_offset >= 0 )
  {
    count = strlen( text+to_offset ) + 1;
    to = (TCHAR*)malloc( sizeof( TCHAR ) * count );
    mbstowcs( to, text+to_offset, count );
  }

  if( cc_offset >= 0 )
  {
    count = strlen( text+cc_offset ) + 1;
    cc = (TCHAR*)malloc( sizeof( TCHAR ) * count );
    mbstowcs( cc, text+cc_offset, count );
  }

  if( subject_offset >= 0 )
  {
    count = strlen( text+subject_offset ) + 1;
    subject = (TCHAR*)malloc( sizeof( TCHAR ) * count );
    mbstowcs( subject, text+subject_offset, count );
  }

  if( body_offset >= 0 )
  {
    count = strlen( text+body_offset ) + 1;
    body = (TCHAR*)malloc( sizeof( TCHAR ) * count );
    mbstowcs( body, text+body_offset, count );
  }

  wcscpy( parameter_string, L"mailto:" );
  if( to ) wcscat( parameter_string, to );
  if( cc || subject || body )
  {
    wcscat( parameter_string, L"?" );
    if( cc )
    {
      wcscat( parameter_string, L"cc=" );
      wcscat( parameter_string, cc );
      if( subject || body ) wcscat( parameter_string, L"&" );
    }
    if( subject )
    {
      wcscat( parameter_string, L"subject=" );
      wcscat( parameter_string, subject );
      if( body ) wcscat( parameter_string, L"&" );
    }
    if( body )
    {
      wcscat( parameter_string, L"body=" );
      wcscat( parameter_string, body );
    }
  }

  memset( &sei, 0, sizeof( sei ) );
  sei.cbSize = sizeof( sei );
  sei.lpFile = L"\\windows\\tmail.exe";
  sei.nShow = SW_SHOW;
  sei.lpParameters = parameter_string;

  if( ShellExecuteEx( &sei ) == 0 )
  {
    MessageBox( hWnd, L"The mail program could not be launched.", L"Error", MB_OK | MB_ICONEXCLAMATION );
  }

  if( to != NULL ) free( to );
  if( cc != NULL ) free( cc );
  if( subject != NULL ) free( subject );
  if( body != NULL ) free( body );
}


static void OpenInvalidRecord( HWND hWnd, int record )
{
  ReaderWindowData *data;
  
  data = FindReaderWindow( hWnd );

  if( PluckerUIDOfReservedRecord( data->plkr, INDEX_ID_URL_HANDLING ) == -1 )
  {
    MessageBox( hWnd, L"That link was not plucked with the rest of this document, and cannot be displayed.", L"Information",
                MB_OK | MB_ICONINFORMATION );
  }
  else
  {
    HandleExternalLink( hWnd, record );
  }
}


static void OpenImageRecord( HWND hWnd, int record )
{
  ReaderWindowData *data;
  RECT rect;

  data = FindReaderWindow( hWnd );

  if( data->library->full_screen )
  {
    SHFullScreen( hWnd, SHFS_SHOWTASKBAR );
  }

  GetClientRect( hWnd, &rect );

  OpenImageViewer( hWnd, data->library, data->plkr, record );

  if( data->library->full_screen )
  {
    SHFullScreen( hWnd, SHFS_HIDETASKBAR );
  }

}

static BOOL GetAnnotationText( HWND hWnd, int *id )
{
  ReaderWindowData *data;
  GETUSERTEXTOPTS opts;
  TCHAR annotation_text[ 4096 ];
  Note *note = NULL;
  BOOL rc = FALSE;

  data = FindReaderWindow( hWnd );

  if( *id >= 0 )
  {
    note = NoteManager_getNote( data->note_manager, *id );
    wcscpy( annotation_text, note->note );
  }
  else
  {
    wcscpy( annotation_text, L"" );
  }

  opts.flags = ES_LEFT | ES_MULTILINE | WS_VSCROLL | ES_WANTRETURN | ES_AUTOVSCROLL;
  opts.hInstance = data->library->hInstance;
  opts.hWnd = hWnd;
  opts.prompt = L"Annotation text:";
  opts.title = ( *id < 0 ? L"New Annotation" : L"Edit Annotation" );
  opts.width = 220;
  opts.height = 130;
  opts.text = annotation_text;
  opts.text_len = 4096;

  if( GetUserText( &opts ) )
  {
    /* if a blank buffer was entered, then don't add or modify the associated note... delete the
     * note if one already existed, though. */
    if( wcslen( opts.text ) < 1 )
    {
      if( *id >= 0 )
      {
        NoteManager_deleteNote( data->note_manager, note );
        note = NULL;
        *id = -1;
      }
    }
    else
    {
      if( *id < 0 )
      {
        *id = NoteManager_addNote( data->note_manager, opts.text );
      }
      else
      {
        NoteManager_modifyNote( data->note_manager, note, opts.text );
      }
    }
    rc = TRUE;
  }

  if( note != NULL ) NoteManager_releaseNote( note );
  return rc;
}


static void UpdateCurrentHistory( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

#ifdef PDAV_SHOWGUIDE
  if( HistoryList_current( data->library->history ) == NULL )
#else
  if( HistoryList_current( data->history ) == NULL )
#endif
  {
    AddCurrentHistory( hWnd );
  }
  else
  {
    int cpos;
    int yofs;

    cpos = PaginationInfo_deltaYToCharacterIndex( data->info, data->delta_y, &yofs );

#ifdef PDAV_SHOWGUIDE
    if (data->library->history->current->type == HISTORY_ITEM_TYPE_PLUCKER)
    {
      data->library->history->current->item.plucker.char_index = cpos;
      data->library->history->current->item.plucker.y_offset = yofs;
      data->library->history->current->item.plucker.record = data->current_record;
    }
#else
    data->history->current->char_index = cpos;
    data->history->current->y_offset = yofs;
    data->history->current->record = data->current_record;
#endif
  }
}


static void AddCurrentHistory( HWND hWnd )
{
  int cpos;
  int yofs;
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  cpos = PaginationInfo_deltaYToCharacterIndex( data->info, data->delta_y, &yofs );

#ifndef PDAV_SHOWGUIDE
  HistoryList_add( data->history, data->current_record, cpos, yofs );
#else
  HistoryList_setHistoryItemPlucker(HistoryList_add( data->library->history),data->library->current_document, data->current_record, cpos, yofs);
#endif
}


static void GoPrevious( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if ( data->current_search_list )
  {
    GoPreviousSearch( hWnd );
  }
  else
  {
    UpdateCurrentHistory( hWnd );

#ifdef PDAV_SHOWGUIDE
    if( HistoryList_goBack( data->library->history ) )
    {
      GoCurrent (data->library, hWnd);
    }
#else
    if( HistoryList_goBack( data->history ) )
    {
      GoCurrent( hWnd );
    }
#endif
  }
}


static void GoNext( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if ( data->current_search_list )
  {
    GoNextSearch( hWnd );
  }
  else
  {
    UpdateCurrentHistory( hWnd );

#ifdef PDAV_SHOWGUIDE
    if( HistoryList_goForward( data->library->history ) )
    {
      GoCurrent (data->library, hWnd);
    }
#else
    if( HistoryList_goForward( data->history ) )
    {
      GoCurrent( hWnd );
    }
#endif
  }
}


#ifdef PDAV_SHOWGUIDE

void GoCurrent( Library *library, HWND hWnd)
{
  HISTORY_ITEM *item = HistoryList_current( library->history );

  switch(item->type)
  {
    case HISTORY_ITEM_TYPE_PLUCKER:
    {
      Document *document;
      int record, char_index, y_offset;
      ReaderWindowData *data;

      record = item->item.plucker.record;
      char_index = item->item.plucker.char_index;
      y_offset = item->item.plucker.y_offset;

      if (_tcscmp(item->current_document, library->current_document))
      {
        document = Library_findDocument( library, item->current_document);
        if( document != NULL )
        {
          if (hWnd)
          {
            DestroyWindow(hWnd);
          }
          OpenDocument( library->hWnd, document );

          hWnd = g_windows[0].hWnd;
          data = FindReaderWindow( hWnd );
          data->current_record = record + 1;
        }
      }


      OpenRecord( hWnd, record, 0, TRUE, char_index, y_offset, FALSE );
      break;
    }
    case HISTORY_ITEM_TYPE_MAP:
    {
      RECT windowRect;

      GetWindowRect(hWnd,&windowRect);
      DestroyWindow(hWnd);

      library->current_document[0] = '\0';
      theMapViewer = MapViewer_new(MAP_DB_NAME, library);
      MapViewer_show(theMapViewer, library, &windowRect);

      /*
        Set the highlight if there is one.
      */
      if ((item->item.map.highlightRect.left < item->item.map.highlightRect.right) && (item->item.map.highlightRect.top < item->item.map.highlightRect.bottom))
      {
        MapViewer_setHighlight(theMapViewer, &item->item.map.highlightRect);
      }

      EnableToolbarButton( theMapViewer->m_toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( library->history ) );
      EnableToolbarButton( theMapViewer->m_toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( library->history ) );

      break;
    }
    case HISTORY_ITEM_TYPE_EXHIBITOR:
    {
      Document *document;

      ExhibitorViewer_new(item->item.exhibitor.id);
      document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
      if( document != NULL )
      {
        if (hWnd)
        {
          DestroyWindow(hWnd);
        }
        OpenDocument( library->hWnd, document );
      }

      break;
    }
    case HISTORY_ITEM_TYPE_EXHIBITOR_LIST:
    {
      Document *document;

      ExhibitorListViewer_new(item->item.exhibitorList.startId,item->item.exhibitorList.limitChar);
      document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
      if( document != NULL )
      {
        if (hWnd)
        {
          DestroyWindow(hWnd);
        }
        OpenDocument( library->hWnd, document );
      }
      break;
    }
    case HISTORY_ITEM_TYPE_SESSION:
    {
      Document *document;

      SessionViewer_new(item->item.session.id);
      document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
      if( document != NULL )
      {
        if (hWnd)
        {
          DestroyWindow(hWnd);
        }
        OpenDocument( library->hWnd, document );
      }
      break;
    }
    case HISTORY_ITEM_TYPE_SESSION_LIST:
    {
      Document *document;
      
      SessionSearch *ss = SessionSearch_new(item->item.sessionList.year,
                                            item->item.sessionList.month, 
                                            item->item.sessionList.day, 
                                            item->item.sessionList.keyword, 
                                            item->item.sessionList.track, 
                                            item->item.sessionList.cme);

      SessionSearch_generatePlucker(ss);
      SessionSearch_delete(ss);

      document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
      if( document != NULL )
      {
        if (hWnd)
        {
          DestroyWindow(hWnd);
        }
        OpenDocument( library->hWnd, document );
      }
      break;
    }
    case HISTORY_ITEM_TYPE_SESSION_SEARCH:
    {
      RECT windowRect;

      GetWindowRect(hWnd,&windowRect);
      DestroyWindow(hWnd);

      library->current_document[0] = '\0';
      theSessionSearchScreen = SessionSearchScreen_new(library,
                                                       &windowRect,
                                                       item->item.sessionSearch.year,
                                                       item->item.sessionSearch.month, 
                                                       item->item.sessionSearch.day, 
                                                       item->item.sessionSearch.keyword, 
                                                       item->item.sessionSearch.track, 
                                                       item->item.sessionSearch.cme);

      EnableToolbarButton( theSessionSearchScreen->m_toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( library->history ) );
      EnableToolbarButton( theSessionSearchScreen->m_toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( library->history ) );
      break;
    }
  }
}

#else

static void GoCurrent( HWND hWnd )
{
  ReaderWindowData *data;
  HISTORY_ITEM *item;

  data = FindReaderWindow( hWnd );

  item = HistoryList_current( data->history );

  OpenRecord( hWnd, item->record, 0, TRUE, item->char_index, item->y_offset, FALSE );

}

#endif


static void DoExitVadeMecum( HWND hWnd )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  PostMessage( hWnd, WM_COMMAND, ID_GO_TO_LIBRARY, (LPARAM)hWnd );
  PostMessage( data->library->hWnd, WM_EXIT_VADE_MECUM, 0, (LPARAM)( data->document->filename ) );
}


static void ToggleToolbar( HWND hWnd )
{
#ifndef PDAV_SHOWGUIDE
  ReaderWindowData *data;
  RECT rect;

  data = FindReaderWindow( hWnd );

  data->library->toolbar_hidden = !data->library->toolbar_hidden;
  RefreshToolbar( hWnd, data );

  GetClientRect( hWnd, &rect );
  ComputeScrollbars( hWnd, data, rect.right - rect.left, rect.bottom - rect.top );
  InvalidateRect( hWnd, &rect, TRUE );
#endif
}

static void DoSearchDocument( HWND hWnd, BOOL forceNewSearch )
{
#ifndef PDAV_SHOWGUIDE

  ReaderWindowData *data;
  BOOL result;

  data = FindReaderWindow( hWnd );

  if ( data->current_search_list || forceNewSearch )
  {
    SetInSearch( hWnd, FALSE );

    if ( !forceNewSearch ) return;
  }

  
  if ( data->library->full_screen )
  {
    SHFullScreen( hWnd, SHFS_SHOWTASKBAR );
  }

  result = SearchDialog( hWnd, data->library, data->plkr, data->current_record, &data->current_search );

  if ( result )
  {  
    if ( data->current_search.just_this_record )
    {
      OpenRecord( hWnd, data->current_search.record, 0, TRUE, data->current_search.text_offset, 0, FALSE );
    }
    else
    {
      PostMessage( hWnd, WM_READER_OPENRECORD, data->current_search.record, data->current_search.text_offset );
    }

    SetInSearch( hWnd, TRUE );
    HistoryList_add( data->current_search_list, data->current_search.record, data->current_search.text_offset, 0 );
  }

  DoResizeScreen( hWnd );
#endif
}

static void DoRecordInfo( HWND hWnd )
{
  ReaderWindowData *data;
  TCHAR info[512];
  PluckerTextRecord *record;

  data = FindReaderWindow( hWnd );

  record = PluckerDBGetTextRecord( data->plkr, data->current_record );

  if ( record->navigation == NULL )
  {
  wsprintf( info, L"UID: %d\nParagraphs: %d\nSize: %d",
    record->uid, record->paragraphs, record->text_length );
  }
  else  
  {
    wsprintf( info, L"UID: %d\nParagraphs: %d\nSize: %d\n%S\n%S",
      record->uid, record->paragraphs, record->text_length, record->navigation->short_record_title,
      record->navigation->long_record_title );

  }
    
  MessageBox( hWnd, info, L"Record info", MB_OK );  
}

static void GoNextSearch( HWND hWnd )
{
#ifndef PDAV_SHOWGUIDE
  ReaderWindowData *data;
  HISTORY_ITEM *hi;
  BOOL found;
  PluckerDBSearchMatch match;
  PluckerDBSearchMark mark;

  data = FindReaderWindow( hWnd );

  if ( HistoryList_subsequentElementsExist( data->current_search_list) )
  {
    hi = HistoryList_goForward( data->current_search_list );
    OpenRecord( hWnd, hi->record, 0, TRUE, hi->char_index, 0, FALSE );

  }
  else
  {
    mark.abort = 0;
    mark.db = data->plkr;
    mark.flags = data->current_search.flags;
    mark.paragraph_index = data->current_search.paragraph;
    mark.pattern = data->current_search.pattern;
    mark.record_id = data->current_search.record;
    mark.text_index = data->current_search.text_offset;

    found = PluckerDBFindNext( &match, &mark );

    if ( found )
    {
      data->current_search.paragraph = mark.paragraph_index;
      data->current_search.text_offset = mark.text_index;

      HistoryList_add( data->current_search_list, mark.record_id, mark.text_index, 0 );

      OpenRecord( hWnd, match.record_id, 0, TRUE, match.text_index, 0, FALSE );
    }
    else
    {
      MessageBox( hWnd, L"No more occurrences found", L"Search", MB_OK );
    }
  }
#endif
}

static void GoPreviousSearch( HWND hWnd )
{
  ReaderWindowData *data;
  HISTORY_ITEM *hi;

  data = FindReaderWindow( hWnd );

  if ( HistoryList_priorElementsExist( data->current_search_list ) )
  {
    hi = HistoryList_goBack( data->current_search_list );
#ifdef PDAV_SHOWGUIDE
    OpenRecord( hWnd, hi->item.plucker.record, 0, TRUE, hi->item.plucker.char_index, 0, FALSE );
#else
    OpenRecord( hWnd, hi->record, 0, TRUE, hi->char_index, 0, FALSE );
#endif
  }

}

static void SetInSearch( HWND hWnd, BOOL in )
{
  ReaderWindowData *data;

  data = FindReaderWindow( hWnd );

  if ( in )
  {
    data->current_search_list = HistoryList_new( 100 );
  }
  else
  {
    if ( data->current_search_list )
    {
      HistoryList_delete( data->current_search_list );
      data->current_search_list = NULL;
    }
  }

  CheckToolbarButton( data->toolbar, ID_SEARCH_BUTTON, in );
  PostMessage( hWnd, WM_SIZE, 0, 0 );
}

static void HandleExternalLink( HWND hWnd, int record )
{
  ReaderWindowData *data;
  int uid;
#ifndef PDAV_SHOWGUIDE
  int result;
#endif
  unsigned char *buffer;
  int type;
  int offset;
  int length;
  int lastURL;
  int count;
#ifndef PDAV_SHOWGUIDE
  HLOCAL hMem;
#endif  

  data = FindReaderWindow( hWnd );

  uid = PluckerUIDOfReservedRecord( data->plkr, INDEX_ID_URL_HANDLING );

  if( uid != -1 )
  {
#ifndef PDAV_SHOWGUIDE
    result = MessageBox( hWnd, L"That link was not plucked with the rest of this document. Copy URL to clipboard?", L"Information",
                         MB_YESNO | MB_ICONINFORMATION );
    
    if( result == IDNO )
      return;

    if( OpenClipboard( hWnd ) )
    {
#endif

      buffer = PluckerDBGetRecord( data->plkr, uid, &type );    
    
      length = PLKR_INT16( buffer+4 );
      buffer += 8;
      offset = 0;
      lastURL = 0;
      
      while( PLKR_INT16( buffer + offset) < record && offset < length )
      {
        lastURL = PLKR_INT16( buffer + offset );
        offset += 4;
      }

      if( offset < length)
      {
        /* Just to make sure that we're not out of bounds */

        uid = PLKR_INT16( buffer + offset + 2 );
        count = record - lastURL;

        buffer = PluckerDBGetRecord( data->plkr, uid, &type );
        if( type == DATATYPE_LINKS || type == DATATYPE_LINKS_COMPRESSED )
        {
          buffer += 8;
          while( --count )
          {
            buffer += strlen( buffer ) + 1;
          }

#ifdef PDAV_SHOWGUIDE
          if (strncmp(buffer, "map:",4) == 0)
          {
            RECT windowRect;
            RECT highlightRect={99999,99999,0,0};
            Library *library = data->library;

            UpdateCurrentHistory(hWnd);

            GetWindowRect(hWnd,&windowRect);

            DestroyWindow(hWnd);
            library->current_document[0] = '\0';

            theMapViewer = MapViewer_new(MAP_DB_NAME, library);
            HistoryList_setHistoryItemMap(HistoryList_add( library->history),theMapViewer->m_scale, theMapViewer->m_xCentre, theMapViewer->m_yCentre, &highlightRect);

            MapViewer_show(theMapViewer, data->library, &windowRect);

            /*
              Set the highlight if there is one.
            */
            sscanf(&buffer[4],"%u,%u,%u,%u",&highlightRect.left, &highlightRect.top, &highlightRect.right, &highlightRect.bottom);
            if ((highlightRect.left < highlightRect.right) && (highlightRect.top < highlightRect.bottom))
            {
              MapViewer_setHighlight(theMapViewer, &highlightRect);
            }

            EnableToolbarButton( theMapViewer->m_toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( library->history ) );
            EnableToolbarButton( theMapViewer->m_toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( library->history ) );
          }
          else if (strncmp(buffer, "exhibitor:",10) == 0)
          {
            UInt16 id;
            Library *library = data->library;
            Document *document;

            UpdateCurrentHistory(hWnd);

            sscanf(&buffer[10],"%u",&id);
            ExhibitorViewer_new(id);

            document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
            if( document != NULL )
            {
              DestroyWindow(hWnd);
              OpenDocument( library->hWnd, document );
              HistoryList_setHistoryItemExhibitor(HistoryList_add( data->library->history), TEMPORARY_PLUCKER_DB_NAME, id);
            }
          }
          else if (strncmp(buffer, "exhibitorlist:",14) == 0)
          {
            UInt16 startId;
            char limitChar;
            Library *library = data->library;
            Document *document;

            UpdateCurrentHistory(hWnd);

            sscanf(&buffer[14],"%u,%c",&startId, &limitChar);
            ExhibitorListViewer_new(startId, limitChar);

            document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
            if( document != NULL )
            {
              extern void OpenDocument( HWND hWnd, Document *document );
              DestroyWindow(hWnd);
              OpenDocument( library->hWnd, document );
              HistoryList_setHistoryItemExhibitorList(HistoryList_add( data->library->history), TEMPORARY_PLUCKER_DB_NAME, startId, limitChar);
            }
          }
          else if (strncmp(buffer, "session:",8) == 0)
          {
            UInt16 id;
            Library *library = data->library;
            Document *document;

            UpdateCurrentHistory(hWnd);

            sscanf(&buffer[8],"%u",&id);
            SessionViewer_new(id);

            document = Library_findDocument( library, TEMPORARY_PLUCKER_DB_NAME );
            if( document != NULL )
            {
              DestroyWindow(hWnd);
              OpenDocument( library->hWnd, document );
              HistoryList_setHistoryItemSession(HistoryList_add( data->library->history), TEMPORARY_PLUCKER_DB_NAME, id);
            }
          }
          else if (strncmp(buffer, "sessionsearch:",14) == 0)
          {
            RECT windowRect;
            Library *library = data->library;

            UpdateCurrentHistory(hWnd);

            GetWindowRect(hWnd,&windowRect);
            DestroyWindow(hWnd);
            library->current_document[0] = '\0';

            theSessionSearchScreen = SessionSearchScreen_new(library, &windowRect, 0, 0, 0, "", "", "");
            HistoryList_setHistoryItemSessionSearch(HistoryList_add( library->history), 0, 0, 0, "", "", "");

            EnableToolbarButton( theSessionSearchScreen->m_toolbar, ID_PREVIOUS_ITEM, HistoryList_priorElementsExist( library->history ) );
            EnableToolbarButton( theSessionSearchScreen->m_toolbar, ID_NEXT_ITEM, HistoryList_subsequentElementsExist( library->history ) );

          }
          else if (strncmp(buffer, "event:",6) == 0)
          {
            TCHAR title[256+1];
            TCHAR message[256+1];
            Library *library = data->library;

            UpdateCurrentHistory(hWnd);

            /*
              Pop up a box asking the user if they want to add the event
            */
            LoadString(library->hInstance, IDS_MBOX_EVENT_TITLE, title, 256);
            LoadString(library->hInstance, IDS_MBOX_EVENT_MESSAGE, message, 256);

            if (MessageBox(hWnd, message, title, MB_YESNO) == IDYES)
            {
              int i=0;
              UInt32 yr,mo,da,sh,sm,eh,em;

              const char *p = buffer;
              const char *end = buffer+strlen(buffer);
              const char *facility="", *room="", *description="";

              while (p != end)
              {
                if (*(p++) == '~')
                {
                  i++;
                  if (i == 5)
                  {
                    facility = p;
                  }
                  else if (i == 6)
                  {
                    room = p;
                  }
                  else if (i == 7)
                  {
                    description = p;
                    break;
                  }
                }
              }

              if ((i == 7) && (sscanf(&buffer[6], "%u~%u~%u~%u:%u~%u:%u", &yr,&mo,&da,&sh,&sm,&eh,&em) == 7))
              {
                SYSTEMTIME st;
                st.wDayOfWeek = st.wMilliseconds = st.wSecond = 0;
                st.wYear = yr;
                st.wMonth = mo;
                st.wDay = da;
                st.wHour = sh;
                st.wMinute = sm;

                {

                  HRESULT hr;
                  IPOutlookApp *polApp = NULL;
                      
                  CoInitializeEx (NULL, 0);

                  hr = CoCreateInstance (&CLSID_Application, NULL, CLSCTX_INPROC_SERVER, &IID_IPOutlookApp, (LPVOID*)&polApp);
                  if (SUCCEEDED (hr))
                  {
                    hr = IPOutlookApp_Logon(polApp,0);
                    if (SUCCEEDED (hr))
                    {
                      IAppointment *pAppt = NULL;

                      hr = IPOutlookApp_CreateItem (polApp, olAppointmentItem, (IDispatch**)&pAppt);
                      if (SUCCEEDED(hr))
                      {
                        DATE date;
                        SystemTimeToVariantTime(&st, &date);
                        hr = IAppointment_put_Start(pAppt, date);
                        if (SUCCEEDED(hr))
                        {
                          st.wHour = eh;
                          st.wMinute = em;
                          SystemTimeToVariantTime(&st, &date);
                          hr = IAppointment_put_End(pAppt, date);
                          if (SUCCEEDED(hr))
                          {
                            LPTSTR descriptionString;
                            LPTSTR locationString;

                            descriptionString = malloc(sizeof(TCHAR)*(strlen(description)+1));
                            locationString = malloc(sizeof(TCHAR)*((description-facility)+10));
                            if (descriptionString && locationString)
                            {
                              LPTSTR wp;

                              wp = locationString;
                              while (facility != (room - 1))
                              {
                                *(wp++) = *((unsigned char *)(facility++));                                
                              }

                              if (room != (description - 1))
                              {
                                *(wp++) = ',';
                                *(wp++) = ' ';

                                while (room != (description - 1))
                                {
                                  *(wp++) = *((unsigned char *)(room++));
                                }
                              }
                              *wp = '\0';


                              wp = descriptionString;
                              while (description != end)
                              {
                                *(wp++) = *((unsigned char *)(description++));
                              }
                              *wp = '\0';

                              hr = IAppointment_put_Subject(pAppt, descriptionString);
                              if (SUCCEEDED(hr))
                              {
                                hr = IAppointment_put_Location(pAppt, locationString);
                                if (SUCCEEDED(hr))
                                {
                                  hr = IAppointment_Save(pAppt);
                                  if (SUCCEEDED(hr))
                                  {
                                    /*
                                      If we get here, then we've made the appointment.
                                    */
                                  }
                                }
                              }
                            }
                            free(descriptionString);
                            free(locationString);
                          }
                        }
                      }

                      IPOutlookApp_Logoff (polApp);
                      IPOutlookApp_Release (polApp);
                    }
                  }
                }
              }
            }
          }
          else if (strncmp(buffer, "plkr:",5) == 0)
          {
            /*
              A plkr: link is a link to another plucker document.
              We can only jump to the "home page" of that document
              at this point.

              Try and find the database first - if it's not there,
              or it exists and its version number is 0, we assume it's
              an optional download. We ask the user if they want to
              download it later, and if they say yes, we create it with
              a version number of 0, which will cause the updater/kiosk
              to download it later. If the user says no, we delete any
              existing one, so it won't get downloaded unnecessarily.

              If the database is actually there, and kosher, we open it.
            */

            const Char *dbName = &buffer[5];
            int len = strlen(dbName);
            int len2 = _tcslen(SHOWGUIDE_DATA_DIR);
            LPTSTR filename;
            LPTSTR p;
            DmOpenRef db;
            Boolean foundValid=false;

            int i;

            UpdateCurrentHistory(hWnd);


            filename = malloc(sizeof(TCHAR) * (len2 + len + 5));

            _tcscpy(filename, SHOWGUIDE_DATA_DIR);
            p = &filename[len2];

            for (i = 0; i < len; i++)
            {
              *(p++) = (UInt8)(dbName[i]);
            }
            _tcscpy(p, TEXT(".pdb"));

            db = DmOpenDatabase(filename);
            if (db)
            {
              if (db->m_header.version != 0)
              {
                foundValid = true;
              }
              DmCloseDatabase(db);
            }

            if (foundValid)
            {
              Library *library = data->library;
              Document *document;
              document = Library_findDocument( library, filename );
              if( document != NULL )
              {
                DestroyWindow(hWnd);
                OpenDocument( library->hWnd, document );
                HistoryList_setHistoryItemPlucker(HistoryList_add(library->history), filename,g_windows[0].current_record, 0, 0);
              }
            }
            else
            {
              if (MessageBox(data->hWnd,
                             TEXT("Would you like to download this document next time you visit a beaming station?"),
                             TEXT("Document Not Available"),
                             MB_YESNO) == IDYES)
              {
                UInt8 *dbPtr = malloc(PLUCKER_WRITER_FILE_HEADER_LENGTH+4);
                UInt8 *tmp = dbPtr;
                tmp = PluckerWriter_fileHeader(tmp, 0, dbName);
                tmp = PluckerWriter_uInt16(tmp, 0); /* num record entries */
                tmp = PluckerWriter_uInt16(tmp, 0); /* placeholder bytes */
                DmCreateDatabaseFromImage(filename, dbPtr,(PLUCKER_WRITER_FILE_HEADER_LENGTH+4));
                free(dbPtr);
              }
              else
              {
                if (db != NULL)
                {
                  DeleteFile(filename);
                }
              }
            }

            free(filename);
          }
          else
          {
            int q = 1;
          }
#else
          length = strlen( buffer ) + 1;
          hMem = LocalAlloc( LMEM_FIXED, length * sizeof( TCHAR ) );
          mbstowcs( (TCHAR*)hMem, buffer, length );        

          EmptyClipboard();
          SetClipboardData( CF_UNICODETEXT, hMem );          
#endif
        }
      }

#ifndef PDAV_SHOWGUIDE
      CloseClipboard();
    }
#endif
  }
}

static void CreateInsetScrolls( HWND hWnd, int nr )
{
  ReaderWindowData *data;
  HWND scroll;
  RECT *rc;

  data = FindReaderWindow( hWnd );

  DeleteInsetScrolls( hWnd );

  while( data->nr_inset_scrolls < nr )
  {
    scroll = CreateWindow( L"scrollbar",
                  L"",
                  SBS_HORZ | WS_CHILD,
                  0,
                  100,
                  CW_USEDEFAULT,
                  CW_USEDEFAULT,
                  hWnd,
                  (HMENU)(ID_LARGE_INSET_SCROLL+data->nr_inset_scrolls),
                  data->library->hInstance,
                  NULL );
    rc = (RECT*)malloc( sizeof( RECT ) );
    SetWindowLong( scroll, GWL_USERDATA, (LONG)rc );
    data->nr_inset_scrolls++;
  }
}

static void DeleteInsetScrolls( HWND hWnd )
{
  ReaderWindowData *data;
  HWND scroll;
  RECT *rc;
  
  data = FindReaderWindow( hWnd );

  while( data->nr_inset_scrolls )
  {
    data->nr_inset_scrolls--;
    scroll = GetDlgItem( hWnd, ID_LARGE_INSET_SCROLL+data->nr_inset_scrolls );
    rc = (RECT *)GetWindowLong( scroll, GWL_USERDATA );
    free( rc );
    DestroyWindow( scroll );
  }
}