/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: SearchDialog module definition
 * Copyright 2003 by Jamis Buck
 * Portions copyright 2004 by Henrik Kjoelhede
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include "search-dlg.h"
#include "utilities.h"

#include <aygshell.h>
#include <stdlib.h>
#include <string.h>

#define ID_SEARCH_PATTERN   ( 40001 )
#define ID_THIS_RECORD      ( 40002 )
#define ID_ABORT_SEARCH     ( 40003 )
#define ID_MATCH_LIST       ( 40004 )
#define ID_START_SEARCH     ( 40005 )
#define ID_CASE_SENSITIVE   ( 40006 )
#define ID_PROGRESS         ( 40007 )

#define SEARCH_DIALOG_CLASS L"VadeMecumSearchDialog"
#define INTERNAL_BORDER     ( 5 )
#define MAX_TEXT_SIZE ( MAX_SEARCH_PATTERN )


typedef struct
{
  HWND hWnd;
  char pattern[ MAX_SEARCH_PATTERN ];
  PluckerDBSearchMark mark;
  PluckerDB db;
  int record;
  int flags;
  BOOL use_first_paragraph;
} SearchInitData;

typedef struct
{
  Library *library;
  PluckerDB db;
  int record;
  SearchReturn* match;
  SearchInitData *searchInitData;
  char pattern[ MAX_TEXT_SIZE ];
} SearchData;


static BOOL  CALLBACK SearchDialogWindowProc( HWND, UINT, WPARAM, LPARAM );
static BOOL  CreateSearchDialog( HWND, LPCREATESTRUCT );
static void  doStartSearch( HWND );
static void  doAbortSearch( HWND );
static void  toggleButtonStates( HWND );
static DWORD searchThreadProc( LPVOID );
static void  setOkEnabled( HWND );
static void  captureSelected( HWND );
static void  releaseListResources( HWND );
static void  addToRecent( HWND, LPSTR );

ATOM RegisterSearchDialog( HINSTANCE hInst )
{
  WNDCLASS wc;

  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = SearchDialogWindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = sizeof( SearchData* );
  wc.hInstance = hInst;
  wc.hIcon = (HICON)NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject( WHITE_BRUSH );
  wc.hCursor = (HCURSOR)NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = SEARCH_DIALOG_CLASS;

  return RegisterClass( &wc );
}


BOOL SearchDialog( HWND parent, Library *library, PluckerDB db, int record, SearchReturn *selected )
{
  MSG msg;
  HWND hSearch;
  SHMENUBARINFO mbi;
  SearchData data;

  data.library = library;
  data.db = db;
  data.record = record;
  data.match = selected;
  data.searchInitData = NULL;

  hSearch = CreateWindow( SEARCH_DIALOG_CLASS,
                          L"Search", WS_VISIBLE | WS_POPUP,
                          CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
                          parent, (HMENU)NULL, library->hInstance, &data );

  if( hSearch == NULL )
  {
    return FALSE;
  }

  EnableWindow( parent, FALSE );

  mbi.cbSize = sizeof( mbi );
  mbi.hwndParent = hSearch;
  mbi.dwFlags = SHCMBF_EMPTYBAR;
  mbi.hInstRes = library->hInstance;

  SHFullScreen( hSearch, SHFS_SHOWSIPBUTTON );
  SHCreateMenuBar( &mbi );

  ShowWindow( hSearch, SW_SHOW );
  UpdateWindow( hSearch );

  while( GetMessage( &msg, NULL, 0, 0 ) ) 
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  EnableWindow( parent, TRUE );  

  return msg.wParam;
}


static BOOL CALLBACK SearchDialogWindowProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{

  switch( uMsg )
  {
    case WM_CREATE:
      return CreateSearchDialog( hWnd, (LPCREATESTRUCT)lParam );

    case WM_DESTROY:
      releaseListResources( hWnd );
      return 0;


    case WM_COMMAND:
      switch( LOWORD( wParam ) )
      {
        case IDOK:
          captureSelected( hWnd );

        case IDCANCEL:
          doAbortSearch( hWnd );
          SHFullScreen( hWnd, SHFS_HIDESIPBUTTON );
          PostQuitMessage( LOWORD( wParam ) == IDOK );
          DestroyWindow( hWnd );
          break;

        case ID_START_SEARCH:
          doStartSearch( hWnd );
          break;

        case ID_ABORT_SEARCH:
          doAbortSearch( hWnd );
          break;

        
        case ID_SEARCH_PATTERN:
          switch( HIWORD( wParam) )
          {
          case CBN_SETFOCUS:
            SHSipPreference( hWnd, SIP_UP );
            break;
          case CBN_KILLFOCUS:
            SHSipPreference( hWnd, SIP_FORCEDOWN );
            break;
          }
          break;
      }
      break;

    case WM_NOTIFY:
      switch( wParam )
      {
        case ID_MATCH_LIST:
          setOkEnabled( hWnd );
          break;
      }
      return DefWindowProc( hWnd, uMsg, wParam, lParam );

    default:
      return DefWindowProc( hWnd, uMsg, wParam, lParam );
  }

  return TRUE;
}

static BOOL CreateSearchDialog( HWND hWnd, LPCREATESTRUCT create )
{
  RECT rect;
  HWND ctl;
  int xofs;
  int yofs;
  int width;
  int height;
  SearchData *data;
  SHINITDLGINFO shidi;
  int i;
  TCHAR textW[ MAX_SEARCH_PATTERN ];
  LV_COLUMN column;
  unsigned long lstyle;
  
  shidi.dwMask = SHIDIM_FLAGS;
  shidi.hDlg = hWnd;
  shidi.dwFlags = SHIDIF_SIZEDLG;
  SHInitDialog( &shidi );

  data = (SearchData*)create->lpCreateParams;
  SetWindowLong( hWnd, 0, (LONG)data );

  GetClientRect( hWnd, &rect );

#define SEARCH_LABEL_WIDTH  ( 50 )
#define SEARCH_LABEL_HEIGHT ( 20 )
#define CONTROL_PADDING     ( 5 )

  yofs = INTERNAL_BORDER;
  height = SEARCH_LABEL_HEIGHT;
  xofs = INTERNAL_BORDER;
  width = SEARCH_LABEL_WIDTH;

  CreateStaticControl( hWnd, GetModuleHandle( NULL ), 0, L"Search:",
                       xofs, yofs+4, width, height-4, SS_LEFT | WS_VISIBLE );

  xofs += width + CONTROL_PADDING;
  width = rect.right - SEARCH_LABEL_WIDTH - CONTROL_PADDING - INTERNAL_BORDER*2;

  ctl = CreateComboBox( hWnd, GetModuleHandle( NULL ), ID_SEARCH_PATTERN, xofs, yofs, width, height*(NUMBER_RECENT_SEARCHES+1),
                        CBS_DROPDOWN | WS_VISIBLE );

  for ( i = 0; i < NUMBER_RECENT_SEARCHES; i++ )
  {
    if ( *data->library->recent_searches[ i ].pattern == 0 )
      break;

    mbstowcs( textW, data->library->recent_searches[ i ].pattern, MAX_SEARCH_PATTERN );
    data->library->recent_searches[ i ].index = SendMessage( ctl, CB_ADDSTRING, 0, (LPARAM)textW ); 
  }


#define CHECKBOX_WIDTH      ( 100 )

  yofs += height + CONTROL_PADDING;
  xofs = INTERNAL_BORDER;
  width = CHECKBOX_WIDTH;

  ctl = CreateCheckBox( hWnd, GetModuleHandle( NULL ), ID_THIS_RECORD, L"Just this page", xofs, yofs, width, height, 
                        WS_VISIBLE );

  if( data->record < 0 )
  {
    EnableWindow( ctl, FALSE );
  }

  xofs += width + CONTROL_PADDING;
  width = ( rect.right - xofs - CONTROL_PADDING - INTERNAL_BORDER ) / 2;

  CreateButtonControl( hWnd, GetModuleHandle( NULL ), ID_START_SEARCH, L"Go", xofs, yofs, width, height,
                       WS_VISIBLE, FALSE );

  xofs += width + CONTROL_PADDING;

  ctl = CreateButtonControl( hWnd, GetModuleHandle( NULL ), ID_ABORT_SEARCH, L"Abort", xofs, yofs, width, height, WS_VISIBLE, FALSE );
  EnableWindow( ctl, FALSE );

  yofs += height + CONTROL_PADDING;
 
  

  xofs = INTERNAL_BORDER;
  width = CHECKBOX_WIDTH;

  ctl = CreateCheckBox( hWnd, GetModuleHandle( NULL ), ID_CASE_SENSITIVE, L"Case sensitive", xofs, yofs, width, height, WS_VISIBLE );


  xofs += width + CONTROL_PADDING;
  width = ( rect.right - xofs - INTERNAL_BORDER );

  ctl = CreateWindow( PROGRESS_CLASS, NULL, WS_CHILD | WS_BORDER| PBS_SMOOTH, 
                      xofs, yofs, width, height,
                      hWnd, (HMENU)ID_PROGRESS, GetModuleHandle( NULL ), NULL );

  SendMessage( ctl, PBM_SETRANGE32, 0, PluckerDBGetRecordCount( data->db ) );
  
  ShowWindow( ctl, SW_SHOW );

  xofs = INTERNAL_BORDER;
  yofs += height + CONTROL_PADDING;

#define BUTTON_HEIGHT ( 25 )
#define BUTTON_WIDTH  ( 60 )

  height = rect.bottom - yofs - CONTROL_PADDING - BUTTON_HEIGHT - INTERNAL_BORDER;
  width = rect.right - rect.left - INTERNAL_BORDER*2;
  

  ctl = CreateWindowEx( 0, WC_LISTVIEW, NULL,
                        WS_VISIBLE | WS_BORDER | WS_CHILD |
                          LVS_REPORT | LVS_SINGLESEL | LVS_SHOWSELALWAYS | LVS_NOLABELWRAP | LVS_NOSORTHEADER,
                          xofs, yofs, width, height,
                          hWnd, (HMENU)ID_MATCH_LIST, GetModuleHandle( NULL ), NULL );

  column.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT;

  column.fmt = LVCFMT_LEFT;
  column.cx = width / 2;
  column.pszText = L"Title";
  ListView_InsertColumn( ctl, 0, &column );

  column.pszText = L"Context";
  ListView_InsertColumn( ctl, 1, &column );

  /* set full-row selection */
  lstyle = ListView_GetExtendedListViewStyle( ctl );
  lstyle |= LVS_EX_FULLROWSELECT;
  ListView_SetExtendedListViewStyle( ctl, lstyle );

  yofs += height + CONTROL_PADDING;
  xofs = INTERNAL_BORDER;
  width = BUTTON_WIDTH;
  height = BUTTON_HEIGHT;

  ctl = CreateButtonControl( hWnd, GetModuleHandle( NULL ), IDOK, L"OK", xofs, yofs, width, height, WS_VISIBLE, TRUE );
  EnableWindow( ctl, FALSE );

  xofs = rect.right - INTERNAL_BORDER - BUTTON_WIDTH;

  CreateButtonControl( hWnd, GetModuleHandle( NULL ), IDCANCEL, L"Cancel", xofs, yofs, width, height, WS_VISIBLE, FALSE );

  return TRUE;
}


static void doStartSearch( HWND hWnd )
{
  HWND ctl;
  TCHAR textW[ MAX_TEXT_SIZE ];
  SearchData *windowData;
  int flags;

  ctl = GetDlgItem( hWnd, ID_SEARCH_PATTERN );
  GetWindowText( ctl, textW, MAX_TEXT_SIZE );

  if( wcslen( textW ) < 1 )
  {
    MessageBox( hWnd, L"You must specify a pattern to search for.", L"Oops", MB_OK | MB_ICONWARNING );
    return;
  }

  toggleButtonStates( hWnd );
  EnableWindow( GetDlgItem( hWnd, IDCANCEL ), FALSE );

  windowData = (SearchData*)GetWindowLong( hWnd, 0 );

  wcstombs( windowData->pattern, textW, MAX_TEXT_SIZE );

  flags = SEARCH_NORMALIZE_DASH | SEARCH_NORMALIZE_QUOTES;
  if( SendMessage( GetDlgItem( hWnd, ID_CASE_SENSITIVE ), BM_GETCHECK, 0, 0 ) == BST_UNCHECKED )
  {
    flags |= SEARCH_CASE_INSENSITIVE;
  }

  addToRecent( hWnd, windowData->pattern );

  if( SendMessage( GetDlgItem( hWnd, ID_THIS_RECORD ), BM_GETCHECK, 0, 0 ) == BST_CHECKED )
  {
    BOOL found;
    PluckerDBSearchMatch match;
    PluckerDBSearchMark mark;

    found = PluckerDBFindFirst( windowData->db, windowData->record, windowData->pattern, flags, &match, &mark );

    if ( found )
    {
      windowData->match->record = match.record_id;
      windowData->match->text_offset = mark.text_index;
      windowData->match->paragraph = mark.paragraph_index;
      windowData->match->flags = flags;
      windowData->match->just_this_record = TRUE;
      strncpy( windowData->match->pattern, windowData->pattern, MAX_SEARCH_PATTERN );

      SHFullScreen( hWnd, SHFS_HIDESIPBUTTON );
      PostQuitMessage( found != 0 );
      DestroyWindow( hWnd );
    }
    else
    {
      MessageBox( hWnd, L"No occurrences found", L"Search", MB_OK );
      toggleButtonStates( hWnd );
      EnableWindow( GetDlgItem( hWnd, IDCANCEL ), TRUE );
    }
  }
  else
  {
    SearchInitData *data;

    data = (SearchInitData*)malloc( sizeof( SearchInitData ) );
    data->db = windowData->db;
    data->hWnd = hWnd;
    data->record = windowData->record;
    strncpy( data->pattern, windowData->pattern, MAX_SEARCH_PATTERN );
    data->flags = flags;
    data->use_first_paragraph = windowData->library->first_paragraph_is_title;

    windowData->searchInitData = data;

    CreateThread( NULL, 0, searchThreadProc, data, 0, NULL );
  }
}


static void doAbortSearch( HWND hWnd )
{
  SearchData *windowData;

  windowData = (SearchData*)GetWindowLong( hWnd, 0 );
  if ( windowData->searchInitData )
    PluckerDBAbortSearch( &windowData->searchInitData->mark );
}


static void toggleButtonStates( HWND hWnd )
{
  HWND startBtn;
  HWND endBtn;

  startBtn = GetDlgItem( hWnd, ID_START_SEARCH );
  endBtn = GetDlgItem( hWnd, ID_ABORT_SEARCH );

  EnableWindow( startBtn, !IsWindowEnabled( startBtn ) );
  EnableWindow( endBtn, !IsWindowEnabled( endBtn ) );
}

static DWORD searchThreadProc( LPVOID parameter )
{
  SearchInitData *data = (SearchInitData*)parameter;
  SearchMatch *itemMatch;
  PluckerDBSearchMatch match;
  HWND list, progress;
  int found;
  TCHAR textW[ MAX_TEXT_SIZE ];
  char text[ MAX_TEXT_SIZE ];
  int pos;
  LVITEM item;

  list = GetDlgItem( data->hWnd, ID_MATCH_LIST );
  progress = GetDlgItem( data->hWnd, ID_PROGRESS );
  releaseListResources( data->hWnd );

  SendMessage( progress, PBM_SETPOS, 0, 0 );

  found = PluckerDBFindFirst( data->db, -1, data->pattern, data->flags, &match, &data->mark );

  while( 1 )
  {
    if( found > 0 )
    {
      int from_idx;
      int to_idx;
      /* found a record */

      itemMatch = (SearchMatch*)malloc( sizeof( SearchMatch ) );
      itemMatch->record = match.record_id;
      itemMatch->text_offset = data->mark.text_index;
      itemMatch->paragraph = data->mark.paragraph_index;

      PluckerDBGetLongTitle( data->db, match.record_id, text, 40, data->use_first_paragraph );
      mbstowcs( textW, text, MAX_TEXT_SIZE );

      item.mask = LVIF_TEXT | LVIF_PARAM;
      item.iItem = 0x7ffffff;
      item.iSubItem = 0;
      item.lParam = (LPARAM)itemMatch;
      item.pszText = textW;

      pos = ListView_InsertItem( list, &item );
      
      from_idx = match.text_index - 20;
      to_idx = match.text_index + 20;
      if( from_idx < match.earliest_look_behind ) from_idx = match.earliest_look_behind;
      if( to_idx > match.latest_look_ahead ) to_idx = match.latest_look_ahead;

      PluckerDBGetText( data->db, match.record_id, from_idx, to_idx, text, MAX_TEXT_SIZE );
      mbstowcs( textW, text, MAX_TEXT_SIZE );
      ListView_SetItemText( list, pos, 1, textW );
      
      ListView_SetColumnWidth( list, 0, LVSCW_AUTOSIZE );
      ListView_SetColumnWidth( list, 1, LVSCW_AUTOSIZE );

    }
    else if( found == 0 )
    {
      /* done searching */
      break;
    }
    /* else found < 0, means that we have finished checking a record. Update progress bar and search next */

    PostMessage( progress, PBM_SETPOS, data->mark.record_index + 1, 0 );

    found = PluckerDBFindNextRecord( &match, &data->mark );
  }

  SendMessage( progress, PBM_SETPOS, SendMessage( progress, PBM_GETRANGE, 0, 0 ), 0 );

  toggleButtonStates( data->hWnd );
  EnableWindow( GetDlgItem( data->hWnd, IDCANCEL ), TRUE );

  free( data );

  return 0;
}

static void setOkEnabled( HWND hWnd )
{
  HWND ctl;
  HWND list;

  ctl = GetDlgItem( hWnd, IDOK );
  list = GetDlgItem( hWnd, ID_MATCH_LIST );

  EnableWindow( ctl, ( ListView_GetSelectedCount( list ) != 0 ) );
}


static void captureSelected( HWND hWnd )
{
  HWND list;
  SearchMatch *match;
  SearchData *data;
  LVITEM item;

  
  list = GetDlgItem( hWnd, ID_MATCH_LIST );

  item.mask = LVIF_PARAM;
  item.iSubItem = 0;
  item.iItem = ListView_GetSelectionMark( list );
  ListView_GetItem( list, &item );
  match = (SearchMatch*)item.lParam;

  data = (SearchData*)GetWindowLong( hWnd, 0 );

  data->match->record = match->record;
  data->match->text_offset = match->text_offset;
  data->match->paragraph = match->paragraph;
  data->match->flags = data->searchInitData->mark.flags;
  data->match->just_this_record = FALSE;
  strncpy( data->match->pattern, data->searchInitData->pattern, MAX_SEARCH_PATTERN );
  
}


static void releaseListResources( HWND hWnd )
{
  HWND list;
  LVITEM item;
  SearchMatch *match;

  list = GetDlgItem( hWnd, ID_MATCH_LIST );
  item.mask = LVIF_PARAM;
  item.iItem = -1;
  item.iSubItem = 0;

  while ( (item.iItem = ListView_GetNextItem( list, item.iItem, LVNI_ALL ) ) != -1 )
  {
    ListView_GetItem( list, &item );
    match = (SearchMatch*)item.lParam;
    free( match );
  }

  ListView_DeleteAllItems( list );
}

static void  addToRecent( HWND hWnd, LPSTR str )
{
  SearchData *data;
  int i, j;
  HWND ctl;
  TCHAR textW[ MAX_SEARCH_PATTERN ];

  data = (SearchData*)GetWindowLong( hWnd, 0 );

  for ( i = 0; i < NUMBER_RECENT_SEARCHES-1; i++ )
  {
    if ( strncmp( data->library->recent_searches[ i ].pattern, str, MAX_SEARCH_PATTERN ) == 0 )
      break;
  }

  if ( i == 0 )
    return;

  for ( j = i-1; i; j--, i-- )
  {
    strncpy( data->library->recent_searches[ i ].pattern, data->library->recent_searches[ j ].pattern, MAX_SEARCH_PATTERN );
  }

  strncpy( data->library->recent_searches[ 0 ].pattern, str, MAX_SEARCH_PATTERN );

  ctl = GetDlgItem( hWnd, ID_SEARCH_PATTERN );
  SendMessage( ctl, CB_RESETCONTENT, 0, 0 );

  for ( i = 0; i < NUMBER_RECENT_SEARCHES; i++ )
  {
    mbstowcs( textW, data->library->recent_searches[ i ].pattern, MAX_SEARCH_PATTERN );
    data->library->recent_searches[ i ].index = SendMessage( ctl, CB_ADDSTRING, 0, (LPARAM)textW );
  }

  SendMessage( ctl, CB_SETCURSEL, data->library->recent_searches[ 0 ].index, 0 );
}
