/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: simple database module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *>=-----------------------------------------------------------------------=<*/

#include <stdio.h>
#include <string.h>

#include "simple-db.h"

#define DBHANDLE(var,parm)   SimpleDBRecord *var = (SimpleDBRecord*)parm


typedef struct
{
  BOOL valid;
} SimpleDBRecordHeader;

typedef struct
{
  SimpleDBRecordHeader header;
  BOOL is_first_chunk : 1;
  INT  next_chunk : 31;
} SimpleDBVariableRecordHeader;


typedef struct
{
  FILE *handle;
  char *filename;
  UINT  record_size;
  BOOL  variable_sized_records;
} SimpleDBRecord;



static int static_get_db_full_record( SimpleDBRecord *db,
                                      UINT  id,
                                      void *header,
                                      int   header_size,
                                      void *data,
                                      int   data_size );

static int static_get_db_record( SimpleDBRecord *db,
                                 void *header,
                                 int   header_size,
                                 void *data,
                                 int   data_size );

static int static_find_free_records( SimpleDBRecord *db,
                                     int header_size,
                                     int num_to_find,
                                     UINT *locations );

static int static_single_insert( SimpleDBRecord *db,
                                 void *buffer,
                                 int buffer_len );

static int static_variable_insert( SimpleDBRecord *db,
                                   void *buffer,
                                   int buffer_len );

static int static_single_update( SimpleDBRecord *db,
                                 UINT id,
                                 void *buffer,
                                 int buffer_len );

static int static_variable_update( SimpleDBRecord *db,
                                   UINT id,
                                   void *buffer,
                                   int buffer_len );

static int static_get_locations( SimpleDBRecord *db,
                                 UINT id,
                                 UINT *locations );

static void static_activate_db( SimpleDBRecord *db );

static void static_passivate_db( SimpleDBRecord *db );


SimpleDB SimpleDB_open( SimpleDBDefinition *def )
{
  SimpleDBRecord *rec;

  rec = (SimpleDBRecord*)malloc( sizeof( SimpleDBRecord ) );
  memset( rec, 0, sizeof( *rec ) );

  rec->filename = _strdup( def->filename );
  rec->handle = NULL;
  rec->record_size = def->record_size;
  rec->variable_sized_records = def->variable_sized_records;

  return (SimpleDB)rec;
}


void SimpleDB_close( SimpleDB db )
{
  DBHANDLE( rec, db );

  free( rec->filename );
  if( rec->handle ) fclose( rec->handle );
  memset( rec, 0, sizeof( *rec ) );
  free( rec );
}


int SimpleDB_get( SimpleDB db,
                  UINT     id,
                  void*    buffer,
                  UINT     len )
{
  DBHANDLE( rec, db );
  int rc;

  static_activate_db( rec );
  rc = static_get_db_full_record( db, id, NULL, 0, buffer, len );
  static_passivate_db( rec );

  return rc;
}


int SimpleDB_recordLength( SimpleDB db,
                           UINT     id,
                           UINT*    len )
{
  DBHANDLE( rec, db );
  SimpleDBVariableRecordHeader header;

  *len = rec->record_size;
  if( !rec->variable_sized_records ) return rec->record_size;

  *len = 0;

  static_activate_db( rec );

  do
  {
    fseek( rec->handle, id, SEEK_SET );
    if( static_get_db_record( rec, &header, sizeof( header ), NULL, 0 ) != 0 )
    {
      static_passivate_db( rec );
      return -1;
    }

    *len += rec->record_size;
    id = header.next_chunk;
  }
  while( header.next_chunk >= 0 );

  static_passivate_db( rec );
  return *len;
}


int SimpleDB_insert( SimpleDB db,
                     void*    buffer,
                     UINT     len )
{
  DBHANDLE( rec, db );
  int rc;

  static_activate_db( rec );

  if( rec->variable_sized_records )
  {
    rc = static_variable_insert( rec, buffer, len );
  }
  else
  {
    rc = static_single_insert( rec, buffer, len );
  }

  static_passivate_db( rec );

  return rc;
}


int SimpleDB_update( SimpleDB db,
                     UINT     id,
                     void*    buffer,
                     UINT     len )
{
  DBHANDLE( rec, db );
  int rc;

  static_activate_db( rec );

  if( rec->variable_sized_records )
  {
    rc = static_variable_update( rec, id, buffer, len );
  }
  else
  {
    rc = static_single_update( rec, id, buffer, len );
  }

  static_passivate_db( rec );

  return rc;
}


int SimpleDB_delete( SimpleDB db,
                     UINT     id )
{
  DBHANDLE( rec, db );

  static_activate_db( rec );

  if( rec->variable_sized_records )
  {
    SimpleDBVariableRecordHeader hdr;

    do
    {
      fseek( rec->handle, id, SEEK_SET );
      fread( &hdr, sizeof( hdr ), 1, rec->handle );
      fseek( rec->handle, id, SEEK_SET );
      hdr.header.valid = 0;
      fwrite( &hdr, sizeof( hdr.header ), 1, rec->handle );

      id = hdr.next_chunk;
    }
    while( hdr.next_chunk >= 0 );
  }
  else
  {
    SimpleDBRecordHeader hdr;

    fseek( rec->handle, id, SEEK_SET );
    hdr.valid = 0;
    fwrite( &hdr, sizeof( hdr ), 1, rec->handle );
  }

  static_passivate_db( rec );

  return 0;
}


int SimpleDB_enumerate( SimpleDB db,
                        void*    buffer,
                        UINT     len,
                        SimpleDBEnumerator* iterator )
{
  SimpleDBRecord *rec;
  UINT id;
  UINT record_and_header_size;
  int rc = -1;

  if( db != NULL )
  {
    rec = (SimpleDBRecord*)db;
    id = 0;
    iterator->db = db;

    static_activate_db( rec );

    fseek( rec->handle, 0, SEEK_END );
    iterator->len = ftell( rec->handle );
  }
  else
  {
    rec = (SimpleDBRecord*)iterator->db;
    id = iterator->id;
    static_activate_db( rec );
  }

  record_and_header_size = ( rec->variable_sized_records ? sizeof( SimpleDBVariableRecordHeader ) : sizeof( SimpleDBRecordHeader ) ) +
                           rec->record_size;

  while( id < iterator->len )
  {
    fseek( rec->handle, id, SEEK_SET );

    if( rec->variable_sized_records )
    {
      SimpleDBVariableRecordHeader hdr;

      fread( &hdr, sizeof( hdr ), 1, rec->handle );
      if( hdr.header.valid && hdr.is_first_chunk ) break;
    }
    else
    {
      SimpleDBRecordHeader hdr;

      fread( &hdr, sizeof( hdr ), 1, rec->handle );
      if( hdr.valid ) break;
    }

    id += record_and_header_size;
  }

  if( id < iterator->len )
  {
    iterator->id = id + record_and_header_size;

    rc = static_get_db_full_record( rec, id, NULL, 0, buffer, len );
  }
  else
  {
    iterator->id = id;
  }

  static_passivate_db( rec );

  return rc;
}


static int static_get_db_full_record( SimpleDBRecord *db,
                                      UINT  id,
                                      void *header,
                                      int   header_size,
                                      void *data,
                                      int   data_size )
{
  int buflen = data_size;
  SimpleDBVariableRecordHeader variable_header;
  SimpleDBRecordHeader fixed_header;

  if( header == NULL )
  {
    if( db->variable_sized_records )
    {
      header = &variable_header;
      header_size = sizeof( variable_header );
    }
    else
    {
      header = &fixed_header;
      header_size = sizeof( fixed_header );
    }
  }

  fseek( db->handle, id, SEEK_SET );

  if( db->variable_sized_records )
  {
    unsigned char *bufptr = (unsigned char*)data;

    if( static_get_db_record( db, header, header_size, (void*)bufptr, buflen ) != 0 )
    {
      return -1;
    }

    bufptr += db->record_size;
    buflen -= db->record_size;

    variable_header.next_chunk = ((SimpleDBVariableRecordHeader*)header)->next_chunk;
    
    while( buflen > 0 && variable_header.next_chunk >= 0 )
    {
      fseek( db->handle, variable_header.next_chunk, SEEK_SET );

      /* only return the FIRST header record; use a temp variable for the subsequent ones */
      if( static_get_db_record( db, &variable_header, sizeof( variable_header ), (void*)bufptr, buflen ) != 0 )
      {
        return -1;
      }

      bufptr += db->record_size;
      buflen -= db->record_size;
    }
  }
  else
  {
    if( static_get_db_record( db, header, header_size, data, data_size ) != 0 )
    {
      return -1;
    }
  }

  return id;
}


static int static_get_db_record( SimpleDBRecord *db,
                                 void *header,
                                 int   header_size,
                                 void *data,
                                 int   data_size )
{
  int size;

  if( fread( header, header_size, 1, db->handle ) != 1 )
  {
    return -1;
  }

  /* set data_size to 0 if all you want is the header of the record */
  if( data_size == 0 ) return 0;

  size = data_size;
  if( size > (int)db->record_size ) size = db->record_size;

  if( fread( data, size, 1, db->handle ) != 1 )
  {
    return -1;
  }

  return 0;
}


static int static_find_free_records( SimpleDBRecord *db,
                                     int header_size,
                                     int num_to_find,
                                     UINT *locations )
{
  int position = 0;
  int found = 0;
  int db_len = 0;
  SimpleDBRecordHeader hdr;

  fseek( db->handle, 0, SEEK_END );
  db_len = ftell( db->handle );

  while( found < num_to_find )
  {
    if( position >= db_len )
    {
      locations[ found ] = position;
      found++;
    }
    else
    {
      fseek( db->handle, position, SEEK_SET );
      static_get_db_record( db, &hdr, sizeof( hdr ), NULL, 0 );

      if( !hdr.valid )
      {
        locations[ found ] = position;
        found++;
      }
    }

    position += header_size + db->record_size;
  }

  return locations[0];
}


static int static_single_insert( SimpleDBRecord *db,
                                 void *buffer,
                                 int buffer_len )
{
  UINT location;

  if( buffer_len > (int)db->record_size ) return -1;

  static_find_free_records( db, sizeof( SimpleDBRecordHeader ), 1, &location );

  return static_single_update( db, location, buffer, buffer_len );
}


static int static_variable_insert( SimpleDBRecord *db,
                                   void *buffer,
                                   int buffer_len )
{
  SimpleDBVariableRecordHeader hdr;
  UINT *locations;
  int locations_needed;
  char *bufptr = (char*)buffer;
  int location = 0;
  int data_size;
  int id;

  locations_needed = buffer_len / db->record_size;
  if( buffer_len % db->record_size > 0 ) locations_needed++;

  locations = (UINT*)malloc( sizeof( UINT ) * locations_needed );
  static_find_free_records( db, sizeof( hdr ), locations_needed, locations );

  hdr.header.valid = 1;
  hdr.is_first_chunk = 1;

  id = locations[ 0 ];

  while( buffer_len > 0 )
  {
    fseek( db->handle, locations[ location ], SEEK_SET );

    hdr.next_chunk = ( location+1 < locations_needed ? locations[location+1] : -1 );

    fwrite( &hdr, sizeof( hdr ), 1, db->handle );

    data_size = db->record_size;
    if( data_size > buffer_len ) data_size = buffer_len;

    fwrite( bufptr, data_size, 1, db->handle );

    hdr.is_first_chunk = 0;
    bufptr += data_size;
    buffer_len -= data_size;
    location++;
  }

  free( locations );
  return id;
}


static int static_single_update( SimpleDBRecord *db,
                                 UINT id,
                                 void *buffer,
                                 int buffer_len )
{
  SimpleDBRecordHeader hdr;

  fseek( db->handle, id, SEEK_SET );
  
  hdr.valid = 1;
  fwrite( &hdr, sizeof( hdr ), 1, db->handle );
  fwrite( buffer, buffer_len, 1, db->handle );

  return id;
}


static int static_variable_update( SimpleDBRecord *db,
                                   UINT id,
                                   void *buffer,
                                   int buffer_len )
{
  UINT length;
  UINT *locations;
  UINT locations_needed;
  UINT locations_used;
  UINT location_count;
  int location;
  SimpleDBVariableRecordHeader hdr;
  char *bufptr;
  int data_size;

  SimpleDB_recordLength( (SimpleDB)db, id, &length );

  locations_needed = buffer_len / db->record_size;
  if( buffer_len % db->record_size > 0 ) locations_needed++;

  locations_used = length / db->record_size;
  if( length % db->record_size > 0 ) locations_used++;

  location_count = ( locations_used > locations_needed ? locations_used : locations_needed );
  locations = (UINT*)malloc( sizeof( UINT ) * location_count );

  static_get_locations( db, id, locations );
  if( location_count > locations_used )
  {
    static_find_free_records( db, sizeof( hdr ),
                              ( location_count - locations_used ),
                              locations+locations_used );
  }

  bufptr = buffer;
  location = 0;
  hdr.header.valid = 1;
  hdr.is_first_chunk = 1;

  do
  {
    fseek( db->handle, locations[location], SEEK_SET );
    hdr.next_chunk = ( location+1 < (int)locations_needed ? locations[location+1] : -1 );

    fwrite( &hdr, sizeof( hdr ), 1, db->handle );

    data_size = db->record_size;
    if( data_size > buffer_len ) data_size = buffer_len;

    fwrite( bufptr, data_size, 1, db->handle );

    bufptr += data_size;
    buffer_len -= data_size;
    location++;
  }
  while( buffer_len > 0 );

  /* delete remaining records */
  while( location < (int)location_count )
  {
    fseek( db->handle, locations[location], SEEK_SET );
    hdr.header.valid = 0;
    fwrite( &hdr, sizeof( hdr.header ), 1, db->handle );
    location++;
  }

  return id;
}


static int static_get_locations( SimpleDBRecord *db,
                                 UINT id,
                                 UINT *locations )
{
  SimpleDBVariableRecordHeader header;
  int location = 0;

  do
  {
    locations[ location ] = id;

    fseek( db->handle, id, SEEK_SET );
    fread( &header, sizeof( header ), 1, db->handle );

    id = header.next_chunk;
    location++;
  }
  while( header.next_chunk >= 0 );

  return location;
}


static void static_activate_db( SimpleDBRecord *db )
{
  if( db->handle == NULL )
  {
    db->handle = fopen( db->filename, "r+b" );
    if( db->handle == NULL ) /* try to open with the intent to create the file... */
    {
      db->handle = fopen( db->filename, "w+b" );
    }
  }
}


static void static_passivate_db( SimpleDBRecord *db )
{
  if( db->handle != NULL )
  {
    fclose( db->handle );
    db->handle = NULL;
  }
}
