/*>=-----------------------------------------------------------------------=<*
 * Vade Mecum: simple database module
 * Copyright 2003 by Jamis Buck
 * ------------------------------------------------------------------------
 * This file is part of Vade Mecum.
 *
 * Vade Mecum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Vade Mecum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Vade Mecum; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 * ------------------------------------------------------------------------
 * A (very) simple database API for storing either fixed or variable length
 * records.
 *
 * Author: Jamis Buck (jgb3@email.byu.edu)
 *>=-----------------------------------------------------------------------=<*/

#ifndef __SIMPLE_DB_H__
#define __SIMPLE_DB_H__

#include <windows.h>

typedef void *SimpleDB;

/* Defines a database file. */
typedef struct
{
  char *filename;               /* the name of the file to use for the persistant storage */
  BOOL  variable_sized_records; /* whether or not this database supports variable sized records */
  UINT  record_size;            /* the size of a single record (or record "chunk") */
} SimpleDBDefinition;


/* An enumerator object, used to traverse all records in the database */
typedef struct
{
  SimpleDB db;          /* the database handle */
  UINT     id;          /* the current record id that the cursor is pointing to */
  UINT     len;         /* the size of the database file */
} SimpleDBEnumerator;


/**
 * Opens a database using the given database definition.  If the database already exists, the
 * definition record must match the definition record that was given when the database was
 * created, or the results will be undefined (and likely unfortunate).  The database handle
 * is returned.  The FILE* field of the definition should not be closed---it is now the
 * property and responsibility of the SimpleDB object.
 */
SimpleDB SimpleDB_open( SimpleDBDefinition *def );

/**
 * Closes the given database and releases all associated memory (including the FILE handle
 * originally passed to open it).
 */
void SimpleDB_close( SimpleDB db );

/**
 * Retrieves a single record from the database, identified by 'id'.  The contents of the record
 * are stored in 'buffer'.  No more than 'len' bytes are copied to the buffer, which means that
 * if the buffer is not large enough to hold the entire record, the data will be truncated.
 * If successful, the 'id' is returned; otherwise -1 is returned.
 */
int SimpleDB_get( SimpleDB db,
                  UINT     id,
                  void*    buffer,
                  UINT     len );

/**
 * Computes and returns the length of the given record.  The length is stored in *len, and is
 * also returned if the operation was successful.  If there was an error, this routine returns
 * -1.
 */
int SimpleDB_recordLength( SimpleDB db,
                           UINT     id,
                           UINT*    len );

/**
 * Inserts the given buffer into the database, returning the new record id.  If the database
 * is of fixed-length records and the length of the buffer (len) is greater than the record
 * size, this operation fails.  The buffer must be valid, and len must be greater than 0.
 */
int SimpleDB_insert( SimpleDB db,
                     void*    buffer,
                     UINT     len );

/**
 * Updates the given record (id) with the given data.  For variable length records, this means
 * that the original chain of records needed to hold the data may be shortened or lengthened,
 * as needed, to hold all of the data.  Returns 'id' if successful, or -1 if there was an error.
 */
int SimpleDB_update( SimpleDB db,
                     UINT     id,
                     void*    buffer,
                     UINT     len );

/**
 * Deletes the given record from the database. Returns 0 if successful.
 */
int SimpleDB_delete( SimpleDB db,
                     UINT     id );

/**
 * Enumerates the records in the database.  The first time this function is called, a valid
 * pointer to an enumerator record and a valid db handle should be passed.  On subsequent
 * calls, the db parameter should be NULL, indicating that subsequent records should be returned.
 * Returns -1 when there are no more records; otherwise, returns the 'id' of the record that
 * is returned.  The contents of the record will be stored in 'buffer' (but no more than 'len'
 * bytes of each record).
 */
int SimpleDB_enumerate( SimpleDB db,
                        void*    buffer,
                        UINT     len,
                        SimpleDBEnumerator* iterator );

#endif /* __SIMPLE_DB_H__ */
